define([ "jquery", "base1/cockpit" ], function($, cockpit) {
    "use strict";
    var kubernetes = {};
    function debug() {
        if (window.debugging == "all" || window.debugging == "kubernetes") console.debug.apply(console, arguments);
    }
    function failure(ex) {
        console.warn(ex);
    }
    function hash(str) {
        var h, i, chr, len;
        if (str.length === 0) return 0;
        for (h = 0, i = 0, len = str.length; i < len; i++) {
            chr = str.charCodeAt(i);
            h = (h << 5) - h + chr;
            h |= 0;
        }
        return Math.abs(h);
    }
    function HashIndex(size) {
        var self = this;
        var array = [];
        self.add = function add(keys, value) {
            var i, j, p, length = keys.length;
            for (j = 0; j < length; j++) {
                i = hash("" + keys[j]) % size;
                p = array[i];
                if (p === undefined) p = array[i] = [];
                p.push(value);
            }
        };
        self.select = function select(keys) {
            var i, j, interim = [], length = keys.length;
            for (j = 0; j < length; j++) {
                i = hash("" + keys[j]) % size;
                interim.push.apply(interim, array[i] || []);
            }
            var result = [];
            interim.sort();
            length = interim.length;
            for (j = 0; j < length; j++) {
                if (interim[j - 1] !== interim[j]) result.push(interim[j]);
            }
            return result;
        };
    }
    function KubernetesWatch(api, type, update, remove) {
        var self = this;
        var lastResource;
        var stopping = false;
        var requested = false;
        var req = null;
        var buffer;
        function handle_watch(data) {
            if (buffer) data = buffer + data;
            var lines = data.split("\n");
            var i, length = lines.length - 1;
            buffer = lines[length];
            var action, object;
            for (i = 0; i < length; i++) {
                try {
                    action = JSON.parse(lines[i]);
                } catch (ex) {
                    failure(ex);
                    req.close();
                    continue;
                }
                object = action.object;
                if (!object) {
                    console.warn("invalid watch without object");
                    continue;
                }
                if (object.metadata) lastResource = object.metadata.resourceVersion; else lastResource = null;
                if (action.type == "ADDED") {
                    update(object);
                } else if (action.type == "MODIFIED") {
                    update(object);
                } else if (action.type == "DELETED") {
                    remove(object);
                } else if (action.type == "ERROR") {
                    if (lastResource) {
                        lastResource = null;
                        start_watch();
                    }
                } else {
                    console.warn("invalid watch action type: " + action.type);
                    continue;
                }
            }
        }
        function start_watch() {
            var uri = "/api/v1beta3/watch/" + type;
            if (requested && lastResource) uri += "?resourceVersion=" + encodeURIComponent(lastResource); else if (!requested) remove(null);
            var waited = false;
            window.setTimeout(function() {
                waited = true;
            }, 1e3);
            if (req) {
                req.cancelled = true;
                req.close();
            }
            req = api.get(uri).stream(handle_watch).fail(function(ex) {
                req = null;
                if (!stopping) console.warn("watching kubernetes " + type + " failed: " + ex);
            }).done(function(data) {
                var cancelled = req && req.cancelled;
                req = null;
                if (!stopping && !cancelled) {
                    if (!waited) console.warn("watching kubernetes " + type + " didn't block"); else start_watch();
                }
            });
            requested = true;
        }
        start_watch();
        self.stop = function stop() {
            stopping = true;
            if (req) req.close();
        };
    }
    function KubernetesClient() {
        var self = this;
        var api = cockpit.http(8080);
        var watches = [];
        self.objects = {};
        var index = new HashIndex(262139);
        self.resourceVersion = null;
        function bind(type) {
            var flat = null;
            var items = {};
            var timeout;
            function trigger() {
                flat = null;
                if (!timeout) {
                    timeout = window.setTimeout(function() {
                        timeout = null;
                        $(self).triggerHandler(type);
                    }, 100);
                }
            }
            function updated(item) {
                var meta = item.metadata;
                if (!meta || !meta.uid) {
                    console.warn("kubernetes item without uid");
                    return;
                }
                if (meta.resourceVersion && meta.resourceVersion > self.resourceVersion) self.resourceVersion = meta.resourceVersion;
                items[meta.uid] = item;
                self.objects[meta.uid] = item;
                var i, keys, length;
                if (meta.labels) {
                    keys = [];
                    for (i in meta.labels) keys.push(i + meta.labels[i]);
                    index.add(keys, meta.uid);
                }
                var spec = item.spec;
                if (spec && spec.selector) {
                    keys = [];
                    for (i in spec.selector) keys.push(i + spec.selector[i]);
                    index.add(keys, meta.uid);
                }
                var status = item.status;
                if (status && status.host) index.add([ status.host ], meta.uid);
                trigger();
            }
            function removed(item) {
                var key;
                if (!item) {
                    for (key in items) {
                        delete items[key];
                        delete self.objects[key];
                    }
                } else {
                    key = item.metadata ? item.metadata.uid : null;
                    if (!key) {
                        console.warn("kubernetes item without uid");
                        return;
                    }
                    delete items[key];
                    delete self.objects[key];
                }
                trigger();
            }
            Object.defineProperty(self, type, {
                enumerable: true,
                get: function get() {
                    if (!flat) {
                        flat = [];
                        for (var key in items) flat.push(items[key]);
                        flat.sort(function(a1, a2) {
                            return (a1.metadata.name || "").localeCompare(a2.metadata.name || "");
                        });
                    }
                    return flat;
                }
            });
            var wc = new KubernetesWatch(api, type, updated, removed);
            watches.push(wc);
        }
        bind("nodes");
        bind("pods");
        bind("services");
        bind("replicationcontrollers");
        this.select = function select(selector, kind) {
            var i, keys;
            var possible, match, results = [];
            if (selector) {
                keys = [];
                for (i in selector) keys.push(i + selector[i]);
                possible = index.select(keys);
            } else {
                possible = Object.keys(self.objects);
            }
            var obj, uid, j, length = possible.length;
            for (j = 0; j < length; j++) {
                uid = possible[j];
                obj = self.objects[uid];
                if (!obj || !obj.metadata) continue;
                if (selector && !obj.metadata.labels) continue;
                if (kind && obj.kind !== kind) continue;
                match = true;
                for (i in selector) {
                    if (obj.metadata.labels[i] !== selector[i]) {
                        match = false;
                        break;
                    }
                }
                if (match) results.push(obj);
            }
            return results;
        };
        this.infer = function infer(labels, kind) {
            var i, keys;
            var possible, match, results = [];
            if (labels) {
                keys = [];
                for (i in labels) keys.push(i + labels[i]);
                possible = index.select(keys);
            } else {
                possible = Object.keys(self.objects);
            }
            var obj, uid, j, length = possible.length;
            for (j = 0; j < length; j++) {
                uid = possible[j];
                obj = self.objects[uid];
                if (!obj || !obj.spec || !obj.spec.selector) continue;
                if (kind && obj.kind !== kind) continue;
                match = true;
                if (labels) {
                    for (i in obj.spec.selector) {
                        if (labels[i] !== obj.spec.selector[i]) {
                            match = false;
                            break;
                        }
                    }
                }
                if (match) results.push(obj);
            }
            return results;
        };
        this.hosting = function hosting(host) {
            var possible = index.select([ host ]);
            var obj, j, length = possible.length;
            var results = [];
            for (j = 0; j < length; j++) {
                obj = self.objects[possible[j]];
                if (obj && obj.status && obj.status.host === host) results.push(obj);
            }
            return results;
        };
        this.delete_pod = function delete_pod(pod_name) {
            api.request({
                method: "DELETE",
                body: "",
                path: "/api/v1beta3/namespaces/default/pods/" + encodeURIComponent(pod_name)
            }).fail(failure);
        };
        this.delete_node = function delete_node(nodes_name) {
            api.request({
                method: "DELETE",
                body: "",
                path: "/api/v1beta3/namespaces/default/nodes/" + encodeURIComponent(nodes_name)
            }).fail(failure);
        };
        this.delete_replicationcontroller = function delete_replicationcontroller(rc_name) {
            api.request({
                method: "DELETE",
                body: "",
                path: "/api/v1beta3/namespaces/default/replicationcontrollers/" + encodeURIComponent(rc_name)
            }).fail(failure);
        };
        this.delete_service = function delete_service(service_name) {
            api.request({
                method: "DELETE",
                body: "",
                path: "/api/v1beta3/namespaces/default/services/" + encodeURIComponent(service_name)
            }).fail(failure);
        };
        this.create_replicationcontroller = function create_replicationcontroller(replicationcontroller_json) {
            api.post("/api/v1beta3/namespaces/default/replicationcontrollers", replicationcontroller_json).fail(failure);
        };
        this.create_node = function create_node(node_json) {
            api.post("/api/v1beta3/namespaces/default/nodes", node_json).fail(failure);
        };
        this.create_pod = function create_pod(pod_json) {
            api.post("/api/v1beta3/namespaces/default/pods", pod_json).fail(failure);
        };
        this.create_service = function create_pod(service_json) {
            api.post("/api/v1beta3/namespaces/default/services", service_json).fail(failure);
        };
        this.update_replicationcontroller = function update_replicationcontroller(rc_json, rc_name) {
            api.request({
                method: "PUT",
                body: rc_json,
                path: "/api/v1beta3/namespaces/default/replicationcontrollers/" + encodeURIComponent(rc_name)
            }).fail(failure);
        };
        self.close = function close() {
            var w = watches;
            watches = [];
            $.each(w, function(i, wc) {
                wc.stop();
            });
        };
    }
    function EtcdClient() {
        var self = this;
        var etcd_api = cockpit.http(7001);
        var first = true;
        var later;
        function receive(data, what, kind) {
            var resp = JSON.parse(data);
            self[what] = resp;
            if (!first) $(self).triggerHandler(what, [ self[what] ]);
        }
        function update() {
            var reqs = [];
            reqs.push(etcd_api.get("/v2/admin/machines").fail(failure).done(function(data) {
                receive(data, "etcdHosts");
            }));
            reqs.push(etcd_api.get("/v2/keys/coreos.com/network/config").fail(failure).done(function(data) {
                receive(data, "flannelConfig");
            }));
            if (first) {
                $.when.apply($, reqs).always(function() {
                    first = false;
                    $(self).triggerHandler("etcdHosts", [ self.etcdHosts ]);
                });
            }
        }
        update();
    }
    kubernetes.k8client = function client() {
        return new KubernetesClient();
    };
    kubernetes.etcdclient = function client() {
        return new EtcdClient();
    };
    return kubernetes;
});