if (!String.prototype.startsWith) {
    Object.defineProperty(String.prototype, "startsWith", {
        enumerable: false,
        configurable: false,
        writable: false,
        value: function(searchString, position) {
            position = position || 0;
            return this.indexOf(searchString, position) === position;
        }
    });
}

if (!String.prototype.endsWith) {
    Object.defineProperty(String.prototype, "endsWith", {
        enumerable: false,
        configurable: false,
        writable: false,
        value: function(searchString, position) {
            position = position || this.length;
            position = position - searchString.length;
            var lastIndex = this.lastIndexOf(searchString);
            return lastIndex !== -1 && lastIndex === position;
        }
    });
}

if (!Array.prototype.find) {
    Object.defineProperty(Array.prototype, "find", {
        enumerable: false,
        configurable: true,
        writable: true,
        value: function(predicate) {
            if (this === null || this === undefined) {
                throw new TypeError("Array.prototype.find called on null or undefined");
            }
            if (typeof predicate !== "function") {
                throw new TypeError("predicate must be a function");
            }
            var list = Object(this);
            var length = list.length >>> 0;
            var thisArg = arguments[1];
            var value;
            for (var i = 0; i < length; i++) {
                if (i in list) {
                    value = list[i];
                    if (predicate.call(thisArg, value, i, list)) {
                        return value;
                    }
                }
            }
            return undefined;
        }
    });
}

var phantom_checkpoint = phantom_checkpoint || function() {};

var shell = shell || {};

(function($, cockpit, shell) {
    function dbus_debug() {
        if (window.debugging == "all" || window.debugging == "dbus") console.debug.apply(console, arguments);
    }
    function dbus_warning(str) {
        console.debug("WARNING: " + str);
    }
    function DBusVariant(signature, value) {
        this._init(signature, value);
    }
    DBusVariant.prototype = {
        _init: function(type, value) {
            this.sig = type;
            this.val = value;
        }
    };
    shell.variant = function variant(signature, value) {
        return new DBusVariant(signature, value);
    };
    function DBusError(name, message) {
        this._init(name, message);
    }
    DBusError.prototype = {
        _init: function(name, message) {
            this.name = name;
            this.message = message;
        }
    };
    function DBusInterface(json, iface_name, enclosing_object, client) {
        this._init(json, iface_name, enclosing_object, client);
    }
    DBusInterface.prototype = {
        _init: function(json, iface_name, enclosing_object, client) {
            this._iface_name = iface_name;
            this._enclosing_object = enclosing_object;
            this._client = client;
            this._set_props(json, "dbus_prop_", false);
        },
        _set_props: function(props, prefix, emit_signals) {
            var some_changed = false;
            for (var prop_name in props) {
                if (prop_name.startsWith(prefix)) {
                    var name = prop_name.substr(prefix.length);
                    if (!this.hasOwnProperty(name) || this[name] != props[prop_name]) {
                        this[name] = props[prop_name];
                        if (emit_signals) $(this).trigger("notify:" + name, this[name]);
                        some_changed = true;
                    }
                }
            }
            if (some_changed && emit_signals) {
                $(this).trigger("notify");
                $(this._client).trigger("propertiesChanged", [ this._enclosing_object, this ]);
            }
        },
        _reseed: function(json) {
            this._set_props(json, "dbus_prop_", true);
        },
        call: function() {
            var args = [];
            var dbus_method_name = arguments[0];
            for (var n = 1; n < arguments.length - 1; n++) {
                args.push(arguments[n]);
            }
            var callback = arguments[n];
            return this.call_with_args(dbus_method_name, args, callback);
        },
        call_with_args: function(dbus_method_name, args, callback) {
            return this._client.call_with_args(this._enclosing_object.objectPath, this._iface_name, dbus_method_name, args, callback);
        },
        propertiesChanged: function(dict) {
            this._set_props(dict, "", true);
        },
        getObject: function() {
            return this._enclosing_object;
        },
        toString: function() {
            return "[DBusInterface name=" + this._iface_name + " objectPath=" + this._enclosing_object.objectPath + "]";
        }
    };
    function DBusObject(json, client) {
        this._init(json, client);
    }
    DBusObject.prototype = {
        _init: function(json, client) {
            this.objectPath = json.objpath;
            this._ifaces = {};
            for (var iface_name in json.ifaces) {
                this._ifaces[iface_name] = new DBusInterface(json.ifaces[iface_name], iface_name, this, client);
            }
        },
        _reseed: function(json, client) {
            var iface_name;
            var iface;
            for (iface_name in json.ifaces) {
                if (iface_name in this._ifaces) {
                    this._ifaces[iface_name]._reseed(json.ifaces[iface_name]);
                } else {
                    iface = new DBusInterface(json.ifaces[iface_name], iface_name, this, client);
                    this._ifaces[iface_name] = iface;
                    $(this).trigger("interfaceAdded", iface);
                    $(client).trigger("interfaceAdded", [ this, iface ]);
                }
            }
            for (iface_name in this._ifaces) {
                if (!(iface_name in json.ifaces)) {
                    iface = this._ifaces[iface_name];
                    delete this._ifaces[iface_name];
                    $(this).trigger("interfaceRemoved", iface);
                    $(client).trigger("interfaceRemoved", [ this, iface ]);
                }
            }
        },
        lookup: function(iface_name) {
            return this._ifaces[iface_name];
        },
        getInterfaces: function() {
            var ret = [];
            for (var i in this._ifaces) ret.push(this._ifaces[i]);
            return ret;
        }
    };
    function DBusClient(target, options) {
        this._init(target, options);
    }
    DBusClient.prototype = {
        _init: function(target, options) {
            this.target = target;
            this.options = options;
            this.error = null;
            this.state = null;
            this._channel = null;
            this._objmap = {};
            this._cookie_counter = 0;
            this._call_reply_map = {};
            this._last_error = null;
            this._proxies = !options || options.proxies || options.proxies === undefined;
            this.error_details = {};
            this.byteorder = null;
            this.connect();
        },
        connect: function() {
            var channel_opts = {
                host: this.target,
                payload: "dbus-json1"
            };
            $.extend(channel_opts, this.options);
            if (!channel_opts["service"]) channel_opts["service"] = "com.redhat.Cockpit";
            if (channel_opts["object-manager"] === undefined && channel_opts["object-paths"] === undefined) channel_opts["object-manager"] = "/com/redhat/Cockpit";
            dbus_debug("Connecting DBusClient to " + channel_opts["service"] + " on " + this.target);
            var channel = cockpit.channel(channel_opts);
            var client = this;
            client._channel = channel;
            this._last_error = null;
            if (this.state !== null) {
                this.state = null;
                this.error = null;
                client._state_change();
            }
            $(client._channel).on("message", function(event, payload) {
                var decoded = JSON.parse(payload);
                dbus_debug("got message command=" + decoded.command);
                if (decoded.command == "seed") {
                    client._handle_seed(decoded.data, decoded.options);
                } else if (decoded.command == "interface-properties-changed") {
                    client._handle_properties_changed(decoded.data);
                } else if (decoded.command == "object-added") {
                    client._handle_object_added(decoded.data);
                } else if (decoded.command == "object-removed") {
                    client._handle_object_removed(decoded.data);
                } else if (decoded.command == "interface-added") {
                    client._handle_interface_added(decoded.data);
                } else if (decoded.command == "interface-removed") {
                    client._handle_interface_removed(decoded.data);
                } else if (decoded.command == "call-reply") {
                    client._handle_call_reply(decoded.data);
                } else if (decoded.command == "interface-signal") {
                    client._handle_interface_signal(decoded.data);
                } else if (decoded.command == "error") {
                    client._handle_error(decoded.data);
                } else {
                    dbus_warning("Unhandled command '" + decoded.command + "'");
                }
            });
            $(client._channel).on("close", function(event, options) {
                client.error = options.problem;
                client.error_details = options;
                client.state = "closed";
                client._state_change();
                for (var cookie in client._call_reply_map) {
                    var callback = client._call_reply_map[cookie];
                    if (callback) {
                        var error = new DBusError("Disconnected", "Disconnected from server.");
                        callback.apply(null, [ error ]);
                    }
                }
                client._call_reply_map = {};
                phantom_checkpoint();
            });
        },
        _state_change: function() {
            $(this).trigger("state-change");
            phantom_checkpoint();
        },
        _handle_error: function(data) {
            this._last_error = data;
        },
        _handle_seed: function(data, options) {
            if (options && options.byteorder) this.byteorder = options.byteorder;
            $(this).trigger("seed", data);
            if (!this._proxies) return;
            for (var objpath in data) {
                if (objpath in this._objmap) {
                    this._objmap[objpath]._reseed(data[objpath], this);
                } else {
                    this._objmap[objpath] = new DBusObject(data[objpath], this);
                    $(this).trigger("objectAdded", this._objmap[objpath]);
                }
            }
            for (objpath in this._objmap) {
                if (!(objpath in data)) {
                    var obj = this._objmap[objpath];
                    delete this._objmap[objpath];
                    $(this).trigger("objectRemoved", obj);
                }
            }
            this.state = "ready";
            this.error = null;
            this._state_change();
        },
        _handle_properties_changed: function(data) {
            $(this).trigger("properties-changed", [ data.objpath, data.iface_name, data.iface[iface_name] ]);
            if (!this._proxies) return;
            var objpath = data.objpath;
            var iface_name = data.iface_name;
            var existing_obj = this._objmap[objpath];
            if (!existing_obj) {
                dbus_warning("Received interface-properties-changed for non-existing object path " + objpath);
            } else {
                var existing_iface = existing_obj._ifaces[iface_name];
                if (!existing_iface) {
                    dbus_warning("Received interface-properties-changed for existing object path " + objpath + " but non-existant interface " + iface_name);
                } else {
                    var changed_properties = data.iface[iface_name];
                    for (var key in changed_properties) {
                        existing_iface[key] = changed_properties[key];
                        $(existing_iface).trigger("notify:" + key, changed_properties[key]);
                    }
                    $(existing_iface).trigger("notify");
                    $(this).trigger("propertiesChanged", [ existing_obj, existing_iface ]);
                }
            }
        },
        _handle_object_added: function(data) {
            $(this).trigger("object-added", [ data.object.objpath, data.object.ifaces ]);
            if (!this._proxies) return;
            var objpath = data.object.objpath;
            var existing_obj = this._objmap[objpath];
            if (existing_obj) {
                dbus_warning("Received object-added for already-existing object path " + objpath);
            }
            var obj = new DBusObject(data.object, this);
            this._objmap[objpath] = obj;
            $(this).trigger("objectAdded", obj);
        },
        _handle_object_removed: function(data) {
            $(this).trigger("object-removed", [ data[0] ]);
            if (!this._proxies) return;
            var objpath = data[0];
            var existing_obj = this._objmap[objpath];
            if (!existing_obj) {
                dbus_warning("Received object-added for non-existing object path " + objpath);
            } else {
                var obj = this._objmap[objpath];
                delete this._objmap[objpath];
                $(this).trigger("objectRemoved", obj);
            }
        },
        _handle_interface_added: function(data) {
            $(this).trigger("interface-added", [ data.objpath, data.iface_name, data.iface[data.iface_name] ]);
            if (!this._proxies) return;
            var objpath = data.objpath;
            var iface_name = data.iface_name;
            var existing_obj = this._objmap[objpath];
            if (!existing_obj) {
                dbus_warning("Received interface-added for non-existing object path " + objpath);
            } else {
                var existing_iface = existing_obj._ifaces[iface_name];
                if (existing_iface) {
                    dbus_warning("Received interface-added for existing object path " + objpath + " and existing interface " + iface_name);
                } else {
                    var iface = new DBusInterface(data.iface[iface_name], iface_name, existing_obj, this);
                    existing_obj._ifaces[iface_name] = iface;
                    $(existing_obj).trigger("interfaceAdded", iface);
                    $(this).trigger("interfaceAdded", [ existing_obj, iface ]);
                }
            }
        },
        _handle_interface_removed: function(data) {
            $(this).trigger("interface-removed", [ data.objpath, data.iface_name ]);
            if (!this._proxies) return;
            var objpath = data.objpath;
            var iface_name = data.iface_name;
            var existing_obj = this._objmap[objpath];
            if (!existing_obj) {
                dbus_warning("Received interface-removed for non-existing object path " + objpath);
            } else {
                var existing_iface = existing_obj._ifaces[iface_name];
                if (!existing_iface) {
                    dbus_warning("Received interface-removed for existing object path " + objpath + " and non-existing interface " + iface_name);
                } else {
                    var iface = existing_obj._ifaces[iface_name];
                    delete existing_obj._ifaces[iface_name];
                    $(existing_obj).trigger("interfaceRemoved", iface);
                    $(this).trigger("interfaceRemoved", [ existing_obj, iface ]);
                }
            }
        },
        _register_call_reply: function(callback) {
            var cookie = "cookie" + this._cookie_counter;
            this._cookie_counter++;
            this._call_reply_map[cookie] = callback;
            return cookie;
        },
        _handle_call_reply: function(data) {
            var cookie = data.cookie;
            var callback = this._call_reply_map[cookie];
            delete this._call_reply_map[cookie];
            if (!callback) {
                if (callback !== null) {
                    dbus_warning("Received call-reply for non-existing cookie " + cookie);
                }
            } else {
                var result = data.result;
                if (result) {
                    var na = [ null ];
                    callback.apply(null, na.concat(result));
                } else {
                    var error = new DBusError(data.error_name, data.error_message);
                    callback.apply(null, [ error ]);
                }
            }
        },
        _handle_interface_signal: function(data) {
            var objpath = data.objpath;
            var iface_name = data.iface_name;
            var signal_name = data.signal_name;
            var signal_args = data.args;
            $(this).trigger("signal", [ objpath, iface_name, signal_name, signal_args ]);
            if (!this._proxies) return;
            var existing_obj = this._objmap[objpath];
            if (!existing_obj) {
                dbus_debug("Received interface-signal for non-existing object path " + objpath);
            } else {
                var existing_iface = existing_obj._ifaces[iface_name];
                if (!existing_iface) {
                    dbus_debug("Received interface-signal for existing object path " + objpath + " and non-existing interface " + iface_name);
                } else {
                    $(existing_iface).trigger(signal_name, signal_args);
                    $(this).trigger("signalEmitted", [ existing_iface, signal_name, signal_args ]);
                }
            }
        },
        close: function(problem) {
            var chan = this._channel;
            this._channel = null;
            if (chan) chan.close();
        },
        lookup: function(objpath, iface_name) {
            if (iface_name) {
                var obj = this._objmap[objpath];
                if (obj) return obj._ifaces[iface_name]; else return null;
            } else {
                return this._objmap[objpath];
            }
        },
        get: function(objpath, iface_name) {
            var obj = this._objmap[objpath];
            if (!obj) {
                obj = new DBusObject({
                    objpath: objpath,
                    ifaces: {}
                }, this);
                this._objmap[objpath] = obj;
                $(this).trigger("objectAdded", obj);
            }
            if (iface_name) {
                var iface = obj._ifaces[iface_name];
                if (!iface) {
                    iface = new DBusInterface({}, iface_name, obj, this);
                    obj._ifaces[iface_name] = iface;
                    $(obj).trigger("interfaceAdded", iface);
                    $(this).trigger("interfaceAdded", [ obj, iface ]);
                }
                return iface;
            } else {
                return obj;
            }
        },
        getInterfacesFrom: function(path_prefix, iface_name) {
            var result = [];
            var obj, objpath, obj_iface;
            for (objpath in this._objmap) {
                if (objpath.indexOf(path_prefix) !== 0) continue;
                obj = this._objmap[objpath];
                obj_iface = obj._ifaces[iface_name];
                if (obj_iface) result.push(obj_iface);
            }
            return result;
        },
        getObjectsFrom: function(path_prefix) {
            var result = [];
            var obj, objpath;
            for (objpath in this._objmap) {
                if (objpath.indexOf(path_prefix) !== 0) continue;
                result.push(this._objmap[objpath]);
            }
            return result;
        },
        call: function() {
            var args = [];
            var path = arguments[0];
            var iface = arguments[1];
            var method = arguments[2];
            for (var n = 3; n < arguments.length - 1; n++) {
                args.push(arguments[n]);
            }
            var callback = arguments[n];
            return this.call_with_args(path, iface, method, args, callback);
        },
        call_with_args: function(path, iface, method, args, callback) {
            var error;
            if (this._channel && this._channel.valid) {
                var cookie = this._register_call_reply(callback);
                var call_obj = {
                    command: "call",
                    objpath: path,
                    iface: iface,
                    method: method,
                    cookie: cookie,
                    args: args
                };
                try {
                    this._channel.send(JSON.stringify(call_obj));
                } catch (e) {
                    delete this._call_reply_map[cookie];
                    error = new DBusError("NotConnected", "Can't send.");
                    callback.apply(null, [ error ]);
                }
            } else {
                error = new DBusError("NotConnected", "Not connected to server.");
                callback.apply(null, [ error ]);
            }
        },
        toString: function() {
            return "[DBusClient]";
        }
    };
    shell.dbus_client = function dbus_client(target, options) {
        return new DBusClient(target, options);
    };
})(jQuery, cockpit, shell);

var shell = shell || {};

(function($, cockpit, shell) {
    function reflow_bar_graph(graph, div) {
        var parts;
        if (graph) {
            var selector = "div.bar-row[graph='" + graph + "']";
            parts = $(selector);
        } else if (div) {
            parts = $(div);
        } else {
            parts = $([]);
        }
        function value_parts(el) {
            var value = $(el).attr("value");
            if (value === undefined) return [ NaN ];
            var values = value.split("/", 2);
            var portion = parseInt(values[0], 10);
            if (values.length == 1) return [ portion ];
            var limit = parseInt(values[1], 10);
            if (!isNaN(limit) && portion > limit) portion = limit;
            if (portion < 0) portion = 0;
            return [ portion, limit ];
        }
        var max = 0;
        parts.each(function() {
            var limit = value_parts(this).pop();
            if (!isNaN(limit) && limit > max) max = limit;
        });
        if (graph) {
            var bound = 100 * 1024 * 1024;
            max = max - max % bound + bound;
        }
        parts.each(function() {
            var bits = value_parts(this);
            var portion = bits.shift();
            var limit = bits.pop();
            if (isNaN(portion) || limit === 0) {
                $(this).css("visibility", "hidden");
            } else {
                var bar_progress = $(this).data("bar-progress");
                if (isNaN(limit)) {
                    bar_progress.addClass("progress-no-limit");
                    limit = portion;
                } else {
                    bar_progress.removeClass("progress-no-limit");
                }
                $(this).css("visibility", "visible");
                bar_progress.css("width", limit / max * 100 + "%");
                $(this).data("bar-progress-bar").css("width", portion / limit * 100 + "%").toggle(portion > 0);
            }
        });
    }
    var reflow_timeouts = {};
    function reflow_bar_graph_soon(graph, div) {
        if (graph === undefined) {
            if (div) graph = $(div).attr("graph");
        }
        if (!graph) {
            reflow_bar_graph(undefined, div);
            return;
        }
        if (reflow_timeouts[graph] !== "undefined") window.clearTimeout(reflow_timeouts[graph]);
        reflow_timeouts[graph] = window.setTimeout(function() {
            delete reflow_timeouts[graph];
            reflow_bar_graph(graph);
        }, 10);
    }
    function setup_bar_graph(div) {
        var progress_bar = $("<div>").addClass("progress-bar");
        var progress = $("<div>").addClass("progress").append(progress_bar);
        $(div).addClass("bar-row").append(progress).data("bar-progress", progress).data("bar-progress-bar", progress_bar);
        $(div).attrchange({
            trackValues: false,
            callback: function(event) {
                if (event.attributeName == "graph" || event.attributeName == "value") {
                    reflow_bar_graph_soon(this.getAttribute("graph"), this);
                }
            }
        });
        div.reflow = function() {
            reflow_bar_graph(this.getAttribute("graph"), this);
        };
        reflow_bar_graph_soon(undefined, div);
    }
    function setup_bar_graphs() {
        $("div.bar-row").each(function() {
            setup_bar_graph(this, false);
        });
    }
    shell.BarRow = function BarRow(graph) {
        var div = $("<div>").addClass("bar-row").attr("graph", graph);
        setup_bar_graph(div);
        return div;
    };
    $(document).ready(setup_bar_graphs);
})(jQuery, cockpit, shell);

(function($, cockpit, shell) {
    function resize_flex(slider, flex, total, part) {
        var value = 0;
        if (part > total) value = 1; else if (part < 0 || isNaN(part)) value = 0; else if (!isNaN(total) && total > 0 && part >= 0) value = part / total;
        $(flex).css("width", value * 100 + "%").next("div").css("margin-left", $(flex).css("width"));
        slider.value = value;
    }
    function update_value(slider) {
        resize_flex(slider, $(slider).children("div.slider-bar").first()[0], 1, slider.value);
    }
    function check_overflow(slider) {
        $(slider).toggleClass("slider-warning", slider.offsetWidth < slider.scrollWidth);
    }
    function setup_slider(slider) {
        $(slider).attr("unselectable", "on");
        Object.defineProperty(slider, "value", {
            get: function() {
                return parseFloat(this.getAttribute("value"));
            },
            set: function(v) {
                var s = String(v);
                if (s != this.getAttribute("value")) this.setAttribute("value", v);
            }
        });
        Object.defineProperty(slider, "disabled", {
            get: function() {
                if (!this.hasAttribute("disabled")) return false;
                return this.getAttribute("disabled").toLowerCase() != "false";
            },
            set: function(v) {
                this.setAttribute("disabled", v ? "true" : "false");
            }
        });
        update_value(slider);
        check_overflow(slider);
        $(slider).attrchange({
            trackValues: true,
            callback: function(event) {
                if (event.attributeName == "value" && event.oldValue !== event.newValue) update_value(slider);
                if (event.attributeName == "disabled") $(slider).toggleClass("slider-disabled", slider.disabled);
            }
        });
        if (slider.disabled) $(slider).addClass("slider-disabled");
        $(slider).on("mousedown", function(ev) {
            if (slider.disabled) return true;
            var flex;
            var offset = $(slider).offset().left;
            if ($(ev.target).hasClass("slider-thumb")) {
                var hitx = ev.offsetX || ev.clientX - $(ev.target).offset().left;
                offset += hitx - $(ev.target).outerWidth() / 2;
                flex = $(ev.target).parent()[0];
            } else {
                flex = $(slider).children("div.slider-bar").first()[0];
                resize_flex(slider, flex, $(slider).width(), ev.pageX - offset);
                $(slider).trigger("change", [ slider.value ]);
                check_overflow(slider);
            }
            $(document).on("mousemove.slider", function(ev) {
                resize_flex(slider, flex, $(slider).width(), ev.pageX - offset);
                $(slider).trigger("change", [ slider.value ]);
                check_overflow(slider);
                return false;
            }).on("mouseup.slider", function(ev) {
                $(document).off("mousemove.slider").off("mouseup.slider");
                return false;
            });
            return false;
        });
    }
    function setup_sliders() {
        $("div.slider").each(function() {
            setup_slider(this);
        });
    }
    shell.Slider = function Slider() {
        var div = $("<div class='slider'>").append($("<div class='slider-bar'>").append($("<div class='slider-thumb'>")));
        setup_slider(div);
        return div;
    };
    $(document).ready(setup_sliders);
    shell.OnOff = function OnOff(val, on, off, role_check, btn_classes) {
        function toggle(event) {
            if (role_check && !role_check()) return false;
            box.find(".btn").toggleClass("active");
            box.find(".btn").toggleClass("btn-primary");
            box.find(".btn").toggleClass("btn-default");
            if (on_btn.hasClass("active")) {
                if (off) on(); else on(true);
            } else {
                if (off) off(); else on(false);
            }
            return false;
        }
        var on_btn, off_btn;
        var box = $('<div class="btn-group btn-toggle">').append(on_btn = $('<button class="btn">').text("On").addClass(btn_classes).addClass(!val ? "btn-default" : "btn-primary active").click(toggle), off_btn = $('<button class="btn">').text("Off").addClass(btn_classes).addClass(val ? "btn-default" : "btn-primary active").click(toggle));
        box.set = function set(val) {
            (val ? on_btn : off_btn).addClass("btn-primary active").removeClass("btn-default");
            (val ? off_btn : on_btn).removeClass("btn-primary active").addClass("btn-default");
        };
        box.enable = function enable(val) {
            box.find("button").toggleClass("disabled", !val);
        };
        return box;
    };
    shell.update_privileged_ui = function update_privileged_ui(perm, selector, denied_message) {
        var allowed = perm.allowed !== false;
        $(selector).each(function() {
            var allowed_key = "allowed-title";
            if (typeof $(this).data(allowed_key) === "undefined" || $(this).data(allowed_key) === false) $(this).data(allowed_key, $(this).attr("title") || "");
            $(this).tooltip({
                html: true
            });
            if ($(this).hasClass("disabled") === allowed) {
                $(this).toggleClass("disabled", !allowed).attr("data-original-title", null);
                if (allowed) $(this).attr("title", $(this).data(allowed_key)); else $(this).attr("title", denied_message);
                $(this).tooltip("fixTitle");
            }
        });
    };
})(jQuery, cockpit, shell);

var shell = shell || {};

(function($, shell) {
    shell.plot = function plot(element, x_range_seconds, x_stop_seconds) {
        var options = {};
        var data = [];
        var flot = null;
        var walk_timer;
        var result = {};
        var now, x_range, x_offset, interval;
        function refresh() {
            if (flot === null) {
                if (element.height() === 0 || element.width() === 0) return;
                flot = $.plot(element, data, options);
            }
            flot.setData(data);
            var axes = flot.getAxes();
            axes.xaxis.options.min = now - x_range - x_offset;
            axes.xaxis.options.max = now - x_offset;
            if (options.setup_hook) options.setup_hook(flot);
            axes.xaxis.show = true;
            axes.xaxis.used = true;
            axes.yaxis.show = true;
            axes.yaxis.used = true;
            flot.setupGrid();
            flot.draw();
        }
        function start_walking() {
            if (!walk_timer) walk_timer = window.setInterval(function() {
                now += interval;
                for (var i = 0; i < data.length; i++) data[i].walk();
                refresh();
            }, interval);
        }
        function stop_walking() {
            if (walk_timer) window.clearInterval(walk_timer);
            walk_timer = null;
        }
        function reset(x_range_seconds, x_stop_seconds) {
            now = new Date().getTime();
            if (x_stop_seconds !== undefined) x_offset = now - x_stop_seconds * 1e3; else x_offset = 0;
            if (x_range_seconds !== undefined) {
                x_range = x_range_seconds * 1e3;
                interval = Math.ceil(x_range_seconds / 1e3) * 1e3;
            }
            for (var i = 0; i < data.length; i++) data[i].reset();
        }
        function destroy() {
            stop_walking();
            for (var i = 0; i < data.length; i++) data[i].stop();
            options = {};
            data = [];
            flot = null;
            $(element).empty();
        }
        function resize() {
            if (element.height() === 0 || element.width() === 0) return;
            if (flot) flot.resize();
            refresh();
        }
        function set_options(opts) {
            options = opts;
            flot = null;
        }
        function add_metrics_sum_series(desc, opts) {
            var series = opts;
            var series_data = null;
            var old_series_data = null;
            var timestamp;
            var archive_channel, real_time_channel;
            var cur_samples;
            var stopped = false;
            var refresh_on_samples = true;
            var real_time_samples = [];
            var max_real_time_samples = 1e3;
            var n_real_time_samples = 0;
            var real_time_samples_pos = -1;
            var real_time_samples_timestamp = null;
            var real_time_samples_interval = 2e3;
            var self = {
                options: series,
                move_to_front: move_to_front,
                remove: remove
            };
            series.stop = stop;
            series.reset = reset_series;
            series.hover = hover;
            series.walk = walk;
            function stop() {
                stopped = true;
                if (archive_channel) archive_channel.close();
                if (real_time_channel) real_time_channel.close();
            }
            function hover(val) {
                $(self).triggerHandler("hover", [ val ]);
            }
            function walk() {
                if (archive_channel === null) trim_series();
                var val = sample_real_time(now - interval, now);
                if (!isNaN(val)) series_data[series_data.length] = [ now, val ];
            }
            function add_series() {
                series.data = series_data;
                data.push(series);
            }
            function remove_series() {
                var pos = data.indexOf(series);
                if (pos >= 0) data.splice(pos, 1);
            }
            function move_to_front() {
                var pos = data.indexOf(series);
                if (pos >= 0) {
                    data.splice(pos, 1);
                    data.push(series);
                }
            }
            function remove() {
                stop();
                remove_series();
                refresh();
            }
            function trim_series() {
                for (var i = 0; i < series_data.length; i++) {
                    if (series_data[i] && series_data[i][0] >= now - x_range - x_offset) {
                        series_data.splice(0, i);
                        return;
                    }
                }
            }
            var metrics;
            function channel_sampler(options, callback) {
                var instances;
                var timestamp;
                var factor;
                factor = desc.factor || 1;
                function on_new_samples(msg) {
                    function compute_sample(samples) {
                        var i, j, sum = 0;
                        function count_sample(index, cur, samples) {
                            if (samples[index] || samples[index] === 0) cur[index] = samples[index];
                            sum += cur[index];
                        }
                        for (i = 0; i < metrics.length; i++) {
                            if (instances[i] !== undefined) {
                                for (j = 0; j < instances[i].length; j++) count_sample(j, cur_samples[i], samples[i]);
                            } else count_sample(i, cur_samples, samples);
                        }
                        return sum * factor;
                    }
                    var res = [];
                    var last = null;
                    for (var i = 0; i < msg.length; i++) {
                        last = compute_sample(msg[i]);
                        res[i] = [ timestamp, last ];
                        timestamp += options.interval;
                    }
                    callback(res);
                    if (last !== null) $(self).triggerHandler("value", [ last ]);
                }
                function handle_message(event, message) {
                    if (stopped) return;
                    var msg = JSON.parse(message);
                    var i, t;
                    if (msg.length) {
                        on_new_samples(msg);
                    } else {
                        instances = msg.metrics.map(function(m) {
                            return m.instances;
                        });
                        cur_samples = [];
                        for (i = 0; i < metrics.length; i++) {
                            if (instances[i] !== null) cur_samples[i] = [];
                        }
                        timestamp = new Date().getTime() + (msg.timestamp - msg.now);
                    }
                }
                var channel = cockpit.channel(options);
                $(channel).on("message", handle_message);
                return channel;
            }
            function build_metric(n) {
                return {
                    name: n,
                    units: desc.units,
                    derive: desc.derive
                };
            }
            metrics = desc.direct.map(build_metric);
            var fallback = null;
            if (desc.internal) fallback = desc.internal.map(build_metric);
            var chanopts = {
                payload: "metrics1",
                metrics: metrics,
                instances: desc.instances,
                "omit-instances": desc["omit-instances"],
                host: desc.host
            };
            function reset_series() {
                if (archive_channel) {
                    archive_channel.close();
                    archive_channel = null;
                }
                series_data = [];
                series.data = series_data;
                var plot_start = now - x_range - x_offset;
                var plot_end = now - x_offset;
                var real_time_start;
                var archive_start, archive_end, archive_limit, archive_index;
                var min_gap;
                if (real_time_samples_timestamp) real_time_start = real_time_samples_timestamp - (n_real_time_samples - 1) * real_time_samples_interval; else real_time_start = now;
                if (plot_start > real_time_start) real_time_start = plot_start;
                archive_start = plot_start;
                archive_end = Math.min(real_time_start, plot_end);
                archive_limit = Math.ceil((archive_end - archive_start) / interval);
                min_gap = Math.max(2 * 60 * 1e3, 2 * interval);
                if (archive_limit > 0) {
                    archive_index = 0;
                    archive_channel = channel_sampler($.extend({
                        source: "pcp-archive",
                        interval: interval,
                        timestamp: -x_range - x_offset,
                        limit: archive_limit
                    }, chanopts), function(vals) {
                        var i, gap, del;
                        if (!result.archives) {
                            result.archives = true;
                            $(result).triggerHandler("changed");
                        }
                        for (i = 0; i < vals.length; i++) {
                            if (vals[i][0] > archive_end) {
                                vals.length = i;
                                break;
                            }
                        }
                        if (vals.length > 0) {
                            del = 0;
                            if (archive_index > 0) {
                                gap = vals[0][0] - series_data[archive_index - 1][0];
                                if (gap < min_gap) del = 1;
                            }
                            Array.prototype.splice.apply(series_data, [ archive_index - del, del ].concat(vals));
                            archive_index += vals.length - del;
                            gap = real_time_start - vals[vals.length - 1][0];
                            if (gap > min_gap) {
                                series_data.splice(archive_index, 0, [ vals[vals.length - 1][0], null ]);
                                archive_index += 1;
                            } else if (archive_index == series_data.length) {
                                series_data.splice(archive_index, 0, [ real_time_start, vals[vals.length - 1][1] ]);
                                archive_index += 1;
                            }
                            refresh();
                        }
                    });
                    $(archive_channel).on("close", function(event, message) {
                        archive_channel = null;
                        trim_series();
                        refresh();
                        if (!result.archives) {
                            var lookup = cockpit.channel($.extend({
                                source: "pcp-archive",
                                limit: 1
                            }, chanopts));
                            $(lookup).on("message", function() {
                                result.archives = true;
                                lookup.close();
                            }).on("close", function() {
                                $(result).triggerHandler("changed");
                            });
                        }
                    });
                }
                for (var t = real_time_start; t <= plot_end; t += interval) {
                    var val = sample_real_time(t - interval, t);
                    if (val !== null) series_data[series_data.length] = [ t, val ];
                }
            }
            reset_series();
            add_series();
            function process_samples(vals) {
                real_time_samples_pos += 1;
                if (real_time_samples_pos >= max_real_time_samples) real_time_samples_pos = 0; else n_real_time_samples += 1;
                real_time_samples[real_time_samples_pos] = vals[0][1];
                real_time_samples_timestamp = vals[0][0];
            }
            function channel_closed(ev, options, desc) {
                real_time_channel = null;
                if (options.problem) console.log("problem retrieving " + desc + " metrics data: " + options.problem);
            }
            real_time_channel = channel_sampler($.extend({}, chanopts, {
                source: "direct",
                interval: real_time_samples_interval
            }), process_samples);
            $(real_time_channel).on("close", function(event, options) {
                if ((options.problem == "not-supported" || options.problem == "not-found") && fallback) {
                    real_time_channel = channel_sampler($.extend({}, chanopts, {
                        source: "internal",
                        interval: real_time_samples_interval,
                        metrics: fallback
                    }), process_samples);
                    $(real_time_channel).on("close", function(event, options) {
                        channel_closed(event, options, "internal");
                    });
                } else {
                    channel_closed(event, options, "real time");
                }
            });
            function sample_real_time(t1, t2) {
                var p1 = Math.min(Math.floor((t1 - real_time_samples_timestamp) / real_time_samples_interval), 0);
                var p2 = Math.min(Math.ceil((t2 - real_time_samples_timestamp) / real_time_samples_interval), 0);
                if (-p1 >= n_real_time_samples) p1 = -n_real_time_samples + 1;
                if (p1 > p2) return null;
                var sum = 0;
                for (var i = p1; i <= p2; i++) {
                    var p = real_time_samples_pos + i;
                    if (p > max_real_time_samples) p -= max_real_time_samples;
                    sum += real_time_samples[p];
                }
                return sum / (p2 - p1 + 1);
            }
            return self;
        }
        var cur_hover = null;
        function hover(series) {
            if (series !== cur_hover) {
                if (cur_hover && cur_hover.hover) cur_hover.hover(false);
                cur_hover = series;
                if (cur_hover && cur_hover.hover) cur_hover.hover(true);
            }
        }
        function hover_on(event, pos, item) {
            hover(item && item.series);
        }
        function hover_off(event) {
            hover(null);
        }
        $(element).on("plothover", hover_on);
        $(element).on("mouseleave", hover_off);
        reset(x_range_seconds, x_stop_seconds);
        $.extend(result, {
            archives: false,
            start_walking: start_walking,
            stop_walking: stop_walking,
            refresh: refresh,
            reset: reset,
            destroy: destroy,
            resize: resize,
            set_options: set_options,
            add_metrics_sum_series: add_metrics_sum_series
        });
        return result;
    };
    shell.plot_simple_template = function simple() {
        return {
            colors: [ "#0099d3" ],
            legend: {
                show: false
            },
            series: {
                shadowSize: 0,
                lines: {
                    lineWidth: 0,
                    fill: 1
                }
            },
            xaxis: {
                tickFormatter: function() {
                    return "";
                }
            },
            yaxis: {
                tickFormatter: function() {
                    return "";
                },
                ticks: 5
            },
            points: {
                radius: 0
            },
            grid: {
                borderWidth: 1,
                aboveData: true,
                color: "black",
                borderColor: $.color.parse("black").scale("a", .22).toString(),
                labelMargin: 0
            }
        };
    };
})(jQuery, shell);

var shell = shell || {};

(function($, shell) {
    shell.image_editor = function image_editor(element, width, height) {
        var self = {
            load_data: load_data,
            get_data: get_data,
            start_cropping: start_cropping,
            stop_cropping: stop_cropping,
            select_file: select_file,
            changed: false
        };
        var square_size = Math.min(width, height);
        var initial_crop_size = square_size;
        var crop_handle_width = 20;
        var $image_canvas, $overlay_canvas, $file_input;
        var image_canvas, overlay_canvas;
        var image_2d, overlay_2d;
        function setup() {
            element.empty().css("width", width).css("height", height).css("position", "relative").append($image_canvas = $("<canvas>"), $overlay_canvas = $("<canvas>").css("position", "absolute").css("top", 0).css("left", 0).css("z-index", 10));
            $("body").append($file_input = $('<input data-role="none" type="file">').hide());
            image_canvas = $image_canvas[0];
            image_2d = image_canvas.getContext("2d");
            overlay_canvas = $overlay_canvas[0];
            overlay_2d = overlay_canvas.getContext("2d");
            image_canvas.width = overlay_canvas.width = width;
            image_canvas.height = overlay_canvas.height = height;
            $file_input.on("change", load_file);
        }
        var cropping = false;
        var crop_x, crop_y, crop_s;
        function start_cropping() {
            cropping = true;
            set_crop((width - initial_crop_size) / 2, (height - initial_crop_size) / 2, initial_crop_size, true);
            $overlay_canvas.on("mousedown", mousedown);
        }
        function stop_cropping() {
            cropping = false;
            overlay_2d.clearRect(0, 0, width, height);
            $overlay_canvas.off("mousedown", mousedown);
        }
        function set_crop(x, y, s, fix) {
            function clamp(low, val, high) {
                if (val < low) return low;
                if (val > high) return high;
                return val;
            }
            x = Math.floor(x);
            y = Math.floor(y);
            s = Math.floor(s);
            var min_s = 2 * crop_handle_width;
            if (fix) {
                s = clamp(min_s, s, square_size);
                x = clamp(0, x, width - s);
                y = clamp(0, y, height - s);
            } else if (x < 0 || y < 0 || x + s > width || y + s > height || s < min_s) return;
            crop_x = x;
            crop_y = y;
            crop_s = s;
            draw_crop(x, y, x + s, y + s);
        }
        function draw_crop(x1, y1, x2, y2) {
            var ctxt = overlay_2d;
            function draw_box(x1, y1, x2, y2) {
                ctxt.strokeStyle = "black";
                ctxt.strokeRect(x1 + .5, y1 + .5, x2 - x1 - 1, y2 - y1 - 1);
                ctxt.strokeStyle = "white";
                ctxt.strokeRect(x1 + 1.5, y1 + 1.5, x2 - x1 - 3, y2 - y1 - 3);
            }
            ctxt.clearRect(0, 0, width, height);
            ctxt.fillStyle = "rgba(0,0,0,0.8)";
            ctxt.fillRect(0, 0, width, height);
            ctxt.clearRect(x1, y1, x2 - x1, y2 - y1);
            var h_w = crop_handle_width;
            draw_box(x1, y1, x1 + h_w, y1 + h_w);
            draw_box(x2 - h_w, y1, x2, y1 + h_w);
            draw_box(x1, y2 - h_w, x1 + h_w, y2);
            draw_box(x2 - h_w, y2 - h_w, x2, y2);
            draw_box(x1, y1, x2, y2);
        }
        function mousedown(ev) {
            var offset = $overlay_canvas.offset();
            var xoff = ev.pageX - offset.left - crop_x;
            var yoff = ev.pageY - offset.top - crop_y;
            var orig_x = crop_x;
            var orig_y = crop_y;
            var orig_s = crop_s;
            var proj_sign, dx_sign, dy_sign, ds_sign;
            var h_w = crop_handle_width;
            function mousemove(ev) {
                var x = ev.pageX - offset.left - xoff;
                var y = ev.pageY - offset.top - yoff;
                if (proj_sign === 0) set_crop(x, y, orig_s, true); else {
                    var d = Math.floor((x - orig_x + proj_sign * (y - orig_y)) / 2);
                    set_crop(orig_x + dx_sign * d, orig_y + dy_sign * d, orig_s + ds_sign * d, false);
                }
                self.changed = true;
            }
            function mouseup(ev) {
                $("body").off("mousemove", mousemove);
                $("body").off("mouseup", mouseup);
            }
            if (xoff > 0 && yoff > 0 && xoff < crop_s && yoff < crop_s) {
                if (xoff < h_w && yoff < h_w) {
                    proj_sign = 1;
                    dx_sign = 1;
                    dy_sign = 1;
                    ds_sign = -1;
                } else if (xoff > crop_s - h_w && yoff < h_w) {
                    proj_sign = -1;
                    dx_sign = 0;
                    dy_sign = -1;
                    ds_sign = 1;
                } else if (xoff < h_w && yoff > crop_s - h_w) {
                    proj_sign = -1;
                    dx_sign = 1;
                    dy_sign = 0;
                    ds_sign = -1;
                } else if (xoff > crop_s - h_w && yoff > crop_s - h_w) {
                    proj_sign = 1;
                    dx_sign = 0;
                    dy_sign = 0;
                    ds_sign = 1;
                } else {
                    proj_sign = 0;
                }
                $("body").on("mousemove", mousemove);
                $("body").on("mouseup", mouseup);
            }
        }
        function load_data(data) {
            var dfd = $.Deferred();
            var img = new window.Image();
            img.onerror = function() {
                dfd.reject();
            };
            img.onload = function() {
                var dest_w, dest_h;
                if (img.width > img.height) {
                    dest_w = width;
                    dest_h = dest_w * (img.height / img.width);
                } else {
                    dest_h = height;
                    dest_w = dest_h * (img.width / img.height);
                }
                image_2d.fillStyle = "rgb(255,255,255)";
                image_2d.fillRect(0, 0, width, height);
                image_2d.drawImage(img, (width - dest_w) / 2, (height - dest_h) / 2, dest_w, dest_h);
                initial_crop_size = Math.min(dest_h, dest_w);
                dfd.resolve();
            };
            img.src = data;
            return dfd.promise();
        }
        function get_data(width, height, format) {
            var dest = $("<canvas/>")[0];
            dest.width = width;
            dest.height = height;
            var ctxt = dest.getContext("2d");
            if (cropping) {
                ctxt.drawImage(image_canvas, crop_x, crop_y, crop_s, crop_s, 0, 0, width, height);
            } else {
                ctxt.drawImage(image_canvas, 0, 0, square_size, square_size, 0, 0, width, height);
            }
            return dest.toDataURL(format);
        }
        var select_dfd;
        function load_file() {
            var files, file, reader;
            files = $file_input[0].files;
            if (files.length != 1) {
                select_dfd.reject();
                return;
            }
            file = files[0];
            if (!file.type.match("image.*")) {
                select_dfd.reject();
                return;
            }
            reader = new window.FileReader();
            reader.onerror = function() {
                select_dfd.reject();
            };
            reader.onload = function() {
                load_data(reader.result).done(function() {
                    select_dfd.resolve();
                }).fail(function() {
                    select_dfd.reject();
                });
            };
            reader.readAsDataURL(file);
        }
        function select_file() {
            select_dfd = $.Deferred();
            if (window.File && window.FileReader) $file_input.trigger("click"); else select_dfd.reject();
            return select_dfd.promise();
        }
        setup();
        return self;
    };
})(jQuery, shell);

var shell = shell || {};

(function($, cockpit, shell) {
    shell.esc = function esc(str) {
        if (str === null || str === undefined) return "";
        var pre = document.createElement("pre");
        var text = document.createTextNode(str);
        pre.appendChild(text);
        return pre.innerHTML.replace(/"/g, "&quot;").replace(/'/g, "&#39;");
    };
    shell.format_delay = function format_delay(d) {
        var seconds = Math.round(d / 1e3);
        var minutes = Math.floor(seconds / 60);
        var hours = Math.floor(minutes / 60);
        seconds = seconds - minutes * 60;
        minutes = minutes - hours * 60;
        var s = seconds + " seconds";
        if (minutes > 0) s = minutes + " minutes, " + s;
        if (hours > 0) s = hours + " hours, " + s;
        return s;
    };
    shell.find_in_array = function find_in_array(array, elt) {
        for (var i = 0; i < array.length; i++) {
            if (array[i] == elt) return true;
        }
        return false;
    };
    shell.action_btn = function action_btn(func, spec, btn_classes) {
        var direct_btn, indirect_btns, btn;
        var direct_action, disabled;
        direct_btn = $("<button>", {
            "data-container": "body",
            "class": "btn btn-default"
        }).text("").addClass(btn_classes);
        indirect_btns = [];
        disabled = [];
        spec.forEach(function(s, i) {
            indirect_btns[i] = $("<li>", {
                "class": "presentation"
            }).append($("<a>", {
                role: "menuitem",
                on: {
                    click: function(e) {
                        if (!disabled[i] && !$(e.currentTarget).hasClass("disabled")) func(s.action);
                    }
                }
            }).addClass(btn_classes).addClass(s.btn_classes).append($("<span>", {
                "class": s.danger ? "text-danger" : ""
            }).text(s.title)));
            disabled[i] = false;
        });
        btn = $("<div>", {
            "class": "btn-group"
        }).append(direct_btn, $("<button>", {
            "class": "btn btn-default dropdown-toggle",
            "data-toggle": "dropdown"
        }).append($("<span>", {
            "class": "caret"
        })), $("<ul>", {
            "class": "dropdown-menu",
            style: "right:0px;left:auto;min-width:0;text-align:left",
            role: "menu"
        }).append(indirect_btns));
        function select(a) {
            spec.forEach(function(s, i) {
                if (s.action == a || a == "default" && s.is_default) {
                    direct_action = s.action;
                    direct_btn.text(s.title);
                    direct_btn.toggleClass("btn-danger", s.danger);
                    direct_btn.toggleClass("btn-default", !s.danger);
                    direct_btn.off("click");
                    direct_btn.on("click", function() {
                        func(s.action);
                    });
                    direct_btn.prop("disabled", disabled[i]);
                }
            });
        }
        function enable(a, val) {
            if (direct_action == a) direct_btn.prop("disabled", !val);
            spec.forEach(function(s, i) {
                if (s.action == a) {
                    disabled[i] = !val;
                    indirect_btns[i].toggleClass("disabled", !val);
                }
            });
        }
        select("default");
        $.data(btn[0], "cockpit-action-btn-funcs", {
            select: select,
            enable: enable
        });
        return btn;
    };
    shell.action_btn_select = function action_btn_select(btn, action) {
        $.data(btn[0], "cockpit-action-btn-funcs").select(action);
    };
    shell.action_btn_enable = function action_btn_enable(btn, action, val) {
        $.data(btn[0], "cockpit-action-btn-funcs").enable(action, val);
    };
    shell.select_btn = function select_btn(func, spec) {
        var div, btn;
        function option_mapper(opt) {
            if (opt.group && Array.isArray(opt.group)) {
                var group = $("<optgroup>").attr("label", opt.title ? opt.title : "").append(opt.group.map(option_mapper));
                return group;
            } else {
                return $("<option>", {
                    value: opt.choice
                }).text(opt.title);
            }
        }
        btn = $('<select class="form-control">').append(spec.map(option_mapper));
        btn.on("change", function() {
            func(btn.val());
        });
        function select(a) {
            btn.val(a);
            btn.selectpicker("render");
        }
        function selected() {
            return btn.val();
        }
        div = $("<div>").append(btn);
        btn.selectpicker();
        $.data(div[0], "cockpit-select-btn-funcs", {
            select: select,
            selected: selected
        });
        return div;
    };
    shell.select_btn_select = function select_btn_select(btn, choice) {
        $.data(btn[0], "cockpit-select-btn-funcs").select(choice);
    };
    shell.select_btn_selected = function select_btn_selected(btn) {
        return $.data(btn[0], "cockpit-select-btn-funcs").selected();
    };
    shell.util = shell.util || {};
    function cache_debug() {
        if (window.debugging == "all" || window.debugging == "dbus") console.debug.apply(console, arguments);
    }
    shell.util.make_resource_cache = make_resource_cache;
    function make_resource_cache() {
        var resources = {};
        function get(key, create) {
            var handle;
            handle = resources[key];
            if (!handle) {
                cache_debug("Creating %s", key);
                handle = {
                    refcount: 1,
                    resource: create()
                };
                resources[key] = handle;
                handle.resource.release = function() {
                    cache_debug("Releasing %s", key);
                    window.setTimeout(function() {
                        if (!handle.refcount) {
                            console.warn("Releasing unreffed resource");
                        } else {
                            handle.refcount -= 1;
                            if (handle.refcount === 0) {
                                delete resources[key];
                                cache_debug("Closing %s", key);
                                handle.resource.close("unused");
                            }
                        }
                    }, 1e4);
                };
            } else {
                cache_debug("Getting %s", key);
                handle.refcount += 1;
            }
            return handle.resource;
        }
        return {
            get: get
        };
    }
    var machine_info_promises = {};
    shell.util.machine_info = machine_info;
    function machine_info(address) {
        var pr = machine_info_promises[address];
        var dfd;
        if (!pr) {
            dfd = $.Deferred();
            machine_info_promises[address] = pr = dfd.promise();
            cockpit.spawn([ "cat", "/proc/meminfo", "/proc/cpuinfo" ], {
                host: address
            }).done(function(text) {
                var info = {};
                var match = text.match(/MemTotal:[^0-9]*([0-9]+) [kK]B/);
                var total_kb = match && parseInt(match[1], 10);
                if (total_kb) info.memory = total_kb * 1024;
                info.cpus = 0;
                var re = new RegExp("^processor", "gm");
                while (re.test(text)) info.cpus += 1;
                dfd.resolve(info);
            }).fail(function() {
                dfd.reject();
            });
        }
        return pr;
    }
    shell.util.hostname_for_display = hostname_for_display;
    function hostname_for_display(iface) {
        if (iface.PrettyHostname) return iface.PrettyHostname; else if (iface.StaticHostname && iface.StaticHostname != "localhost" && iface.StaticHostname != "localhost.localdomain") return iface.StaticHostname; else if (iface._client && iface._client.target != "localhost") return iface._client.target; else if (iface.client && iface.client.options.host && iface.client.options.host != "localhost") return iface.client.options.host; else return window.location.hostname;
    }
})(jQuery, cockpit, shell);

var shell = shell || {};

var modules = modules || {};

(function($, cockpit, shell, modules) {
    shell.pages = [];
    shell.dialogs = [];
    var visited_dialogs = {};
    shell.dbus = dbus;
    var loaded = false;
    modules.loaded = false;
    function maybe_init() {
        if (loaded && modules.loaded) init();
    }
    require([ "system/server", "shell/machines", "manifests" ], function(server, machines, manifests) {
        modules.server = server;
        modules.machines = machines;
        if (manifests["docker"]) {
            require([ "docker/docker" ], function(docker) {
                modules.docker = docker;
                modules.loaded = true;
                maybe_init();
            });
        } else {
            modules.loaded = true;
            maybe_init();
        }
    });
    function init() {
        $(".dropdown-toggle").dropdown();
        content_init();
        content_show();
    }
    require([ "translated!base1/po" ], function(po) {
        cockpit.locale(po);
    });
    var dbus_clients = shell.util.make_resource_cache();
    function make_dict_key(dict) {
        function stringify_elt(k) {
            return JSON.stringify(k) + ":" + JSON.stringify(dict[k]);
        }
        return Object.keys(dict).sort().map(stringify_elt).join(";");
    }
    function dbus(address, options) {
        return dbus_clients.get(make_dict_key($.extend({
            host: address
        }, options)), function() {
            return shell.dbus_client(address, options);
        });
    }
    var current_params;
    var current_page_element;
    var current_legacy_page;
    function content_init() {
        var current_visible_dialog = null;
        var pages = $("#content > div");
        pages.each(function(i, p) {
            $(p).hide();
        });
        current_params = null;
        $('div[role="dialog"]').on("show.bs.modal", function() {
            current_visible_dialog = $(this).attr("id");
            dialog_enter($(this).attr("id"));
        });
        $('div[role="dialog"]').on("shown.bs.modal", function() {
            dialog_show($(this).attr("id"));
        });
        $('div[role="dialog"]').on("hidden.bs.modal", function() {
            current_visible_dialog = null;
            dialog_leave($(this).attr("id"));
        });
        $("#content-navbar a[data-page-id]").click(function() {
            cockpit.location.go([ cockpit.location.path[0], $(this).attr("data-page-id") ]);
        });
        $("#content-navbar a[data-task-id]").click(function() {
            cockpit.location.go([ cockpit.location.path[0], $(this).attr("data-task-id") ]);
        });
        $(cockpit).on("locationchanged", function() {
            if (current_visible_dialog) $("#" + current_visible_dialog).modal("hide");
            display_location();
        });
        $(window).on("resize", function() {
            recalculate_layout();
        });
        shell.content_refresh();
        $(".selectpicker").selectpicker();
    }
    function content_show() {
        display_location();
        phantom_checkpoint();
    }
    shell.content_refresh = function content_refresh() {
        if (current_params) display_params(current_params);
    };
    function recalculate_layout() {
        var $extra = $("#shell-header-extra");
        var $body = $("body");
        $body.css("padding-top", $extra.height());
        $(cockpit).trigger("resize");
    }
    shell.default_permission = cockpit.permission({
        group: "wheel"
    });
    function update_global_nav() {
        var page_title = null;
        if (current_legacy_page) {
            var section_id = current_legacy_page.section_id || current_legacy_page.id;
            page_title = current_legacy_page.getTitle();
        } else if (current_page_element) {
            if (current_page_element[0].contentDocument) page_title = current_page_element[0].contentDocument.title;
        }
        var doc_title;
        if (page_title) doc_title = page_title; else doc_title = _("Cockpit");
        document.title = doc_title;
        recalculate_layout();
    }
    var components = {};
    var page_iframes = {};
    var visited_legacy_pages = {};
    function register_component(prefix, package_name, entry) {
        var key = JSON.stringify(prefix);
        components[key] = {
            pkg: package_name,
            entry: entry
        };
    }
    function register_tool(ident, label) {
        var link = $("<a>").attr("data-task-id", ident).text(label);
        $("<li>").append(link).appendTo("#tools-menu");
    }
    function get_location_hash(location) {
        return location.href.split("#")[1] || "";
    }
    function legacy_page_from_id(id) {
        var n;
        for (n = 0; n < shell.pages.length; n++) {
            if (shell.pages[n].id == id) return shell.pages[n];
        }
        return null;
    }
    function display_params(params) {
        var element = null;
        var legacy_page = null;
        var id = params.path[0];
        legacy_page = visited_legacy_pages[id];
        if (!legacy_page) {
            legacy_page = legacy_page_from_id(id);
            if (legacy_page) {
                if (legacy_page.setup) legacy_page.setup();
                visited_legacy_pages[id] = legacy_page;
            }
        }
        if (legacy_page) element = $("#" + id);
        if (!element) {
            cockpit.location.go([ "server" ]);
            return;
        }
        var old_element = current_page_element;
        var old_legacy_page = current_legacy_page;
        if (old_legacy_page) old_legacy_page.leave();
        $("#shell-header-extra").empty();
        current_params = params;
        current_page_element = element;
        current_legacy_page = legacy_page;
        if (legacy_page) legacy_page.enter();
        if (old_element) old_element.hide();
        update_global_nav();
        element.show();
        if (legacy_page) legacy_page.show();
    }
    function display_location() {
        var host = null;
        var path = cockpit.location.path.slice(0);
        var options = cockpit.location.options;
        if (path.length < 1) path = [ "dashboard" ];
        display_params({
            host: host,
            path: path,
            options: options
        });
    }
    function dialog_from_id(id) {
        var n;
        for (n = 0; n < shell.dialogs.length; n++) {
            if (shell.dialogs[n].id == id) return shell.dialogs[n];
        }
        return null;
    }
    function dialog_enter(id) {
        var dialog = dialog_from_id(id);
        var first_visit = true;
        if (visited_dialogs[id]) first_visit = false;
        if (dialog) {
            if (first_visit && dialog.setup) dialog.setup();
            if (!dialog.entered) dialog.enter();
            dialog.entered = true;
        }
        visited_dialogs[id] = true;
        phantom_checkpoint();
    }
    function dialog_leave(id) {
        var dialog = dialog_from_id(id);
        if (dialog) {
            if (dialog.entered) dialog.leave();
            dialog.entered = false;
        }
        phantom_checkpoint();
    }
    function dialog_show(id) {
        var dialog = dialog_from_id(id);
        if (dialog) {
            dialog.show();
        }
        phantom_checkpoint();
    }
    shell.get_page_param = function get_page_param(key) {
        return current_params.options[key];
    };
    shell.set_page_param = function set_page_param(key, val) {
        if (val) {
            if (val == current_params.options[key]) return;
            current_params.options[key] = val;
        } else {
            if (current_params.options[key] === undefined) return;
            delete current_params.options[key];
        }
        cockpit.location.replace(cockpit.location.path, current_params.options);
        current_params.options = cockpit.location.options;
    };
    shell.show_error_dialog = function show_error_dialog(title, message) {
        if (message) {
            $("#error-popup-title").text(title);
            $("#error-popup-message").text(message);
        } else {
            $("#error-popup-title").text(_("Error"));
            $("#error-popup-message").text(title);
        }
        $('.modal[role="dialog"]').modal("hide");
        $("#error-popup").modal("show");
    };
    shell.show_unexpected_error = function show_unexpected_error(error) {
        shell.show_error_dialog(_("Unexpected error"), error.message || error);
    };
    shell.confirm = function confirm(title, body, action_text) {
        var deferred = $.Deferred();
        $("#confirmation-dialog-title").text(title);
        if (typeof body == "string") $("#confirmation-dialog-body").text(body); else $("#confirmation-dialog-body").html(body);
        $("#confirmation-dialog-confirm").text(action_text);
        function close() {
            $("#confirmation-dialog button").off("click");
            $("#confirmation-dialog").modal("hide");
        }
        $("#confirmation-dialog-confirm").click(function() {
            close();
            deferred.resolve();
        });
        $("#confirmation-dialog-cancel").click(function() {
            close();
            deferred.reject();
        });
        $("#confirmation-dialog").modal("show");
        return deferred.promise();
    };
    $(function() {
        loaded = true;
        maybe_init();
    });
})(jQuery, cockpit, shell, modules);

function N_(str) {
    return str;
}

var _ = cockpit.gettext;

var C_ = cockpit.gettext;

var shell = shell || {};

(function($, cockpit, shell) {
    function setup_plot(graph_id, resmon, data, user_options, store_samples) {
        var options = {
            colors: [ "#0099d3" ],
            legend: {
                show: false
            },
            series: {
                shadowSize: 0,
                lines: {
                    lineWidth: 0,
                    fill: 1
                }
            },
            xaxis: {
                tickFormatter: function() {
                    return "";
                }
            },
            yaxis: {
                tickFormatter: function() {
                    return "";
                }
            },
            points: {
                radius: 0
            },
            grid: {
                borderWidth: 1,
                aboveData: true,
                color: "black",
                borderColor: $.color.parse("black").scale("a", .22).toString(),
                labelMargin: 0
            }
        };
        var num_series = data.length;
        var num_points;
        var got_historical_data = false;
        var plot;
        var running = false;
        var ready = false;
        var self;
        $.extend(true, options, user_options);
        var outer_div = $(graph_id);
        var inner_div = $("<div/>");
        outer_div.empty();
        outer_div.append(inner_div);
        function init() {
            if (!ready && resmon.NumSamples !== undefined) {
                num_points = resmon.NumSamples;
                for (var n = 0; n < num_series; n++) {
                    var series = [];
                    for (var m = 0; m < num_points; m++) {
                        series[m] = [ m, 0 ];
                    }
                    data[n].data = series;
                }
                $(resmon).on("NewSample", new_sample_handler);
                resmon.call("GetSamples", {}, function(error, result) {
                    if (!error) {
                        got_historical_samples(result);
                        got_historical_data = true;
                        refresh();
                    }
                });
                $(cockpit).on("resize", resize);
                ready = true;
                maybe_start();
            }
        }
        function sync_divs() {
            inner_div.width(outer_div.width());
            inner_div.height(outer_div.height());
        }
        function set_yaxis_max(max) {
            if (plot) {
                plot.getAxes().yaxis.options.max = max;
                refresh();
            } else {
                options.yaxis.max = max;
            }
        }
        function start() {
            running = true;
            maybe_start();
        }
        function maybe_start() {
            if (ready && running && outer_div.width() !== 0 && outer_div.height() !== 0) {
                if (!plot) {
                    sync_divs();
                    plot = $.plot(inner_div, data, options);
                } else resize();
            }
        }
        function stop() {
            running = false;
        }
        function refresh() {
            if (plot && running) {
                plot.setData(data);
                if (user_options.setup_hook) user_options.setup_hook(plot);
                plot.setupGrid();
                plot.draw();
            }
        }
        function resize() {
            if (plot && running) {
                sync_divs();
                plot.resize();
                refresh();
            }
        }
        function new_samples(samples) {
            var series;
            var i, n, m, floor;
            for (n = 0; n < data.length; n++) {
                series = data[n].data;
                for (m = 0; m < series.length - 1; m++) {
                    series[m][1] = series[m + 1][1];
                    series[m][2] = series[m + 1][2];
                }
            }
            store_samples(samples, data[0].data.length - 1);
        }
        function got_historical_samples(history) {
            var n, offset;
            offset = data[0].data.length - history.length;
            for (n = 0; n < history.length; n++) {
                store_samples(history[n][1], n + offset);
            }
        }
        function new_sample_handler(event, timestampUsec, samples) {
            if (got_historical_data) {
                new_samples(samples);
                refresh();
            }
        }
        function destroy() {
            $(self).trigger("destroyed");
            $(resmon).off("notify:NumSamples", init);
            $(resmon).off("NewSample", new_sample_handler);
            $(window).off("resize", resize);
            $(outer_div).empty();
            plot = null;
        }
        if (resmon.NumSamples !== undefined) init(); else $(resmon).on("notify:NumSamples", init);
        self = {
            start: start,
            stop: stop,
            resize: resize,
            element: inner_div[0],
            set_yaxis_max: set_yaxis_max,
            destroy: destroy
        };
        return self;
    }
    shell.setup_complicated_plot = function setup_complicated_plot(graph_id, resmon, data, options) {
        var i;
        var plot = null;
        var my_options = $.extend({
            legend: {
                show: true
            }
        }, options);
        function store_samples(samples, index) {
            var value, series;
            var n, i, floor;
            for (n = 0; n < data.length; n++) {
                series = data[n].data;
                value = samples[n];
                if (options.x_rh_stack_graphs) {
                    floor = 0;
                    for (i = n + 1; i < samples.length; i++) {
                        floor += samples[i];
                    }
                    series[index][1] = value + floor;
                    series[index][2] = floor;
                } else {
                    series[index][1] = value;
                }
            }
        }
        function setup_legends() {
            for (i = 0; i < data.length; i++) data[i].label = resmon.Legends[i];
            if (plot) plot.resize();
        }
        if (resmon.Legends) setup_legends(); else $(resmon).on("notify:Legends", setup_legends);
        plot = setup_plot(graph_id, resmon, data, my_options, store_samples);
        return plot;
    };
    shell.setup_simple_plot = function setup_simple_plot(plot_id, text_id, resmon, options, series_combine_func, series_text_func) {
        var data = [ {} ];
        function store_samples(samples, index) {
            var series = data[0].data;
            series[index][1] = series_combine_func(samples);
            if (index == series.length - 1) $(text_id).html(series_text_func(samples));
        }
        return setup_plot(plot_id, resmon, data, options, store_samples);
    };
    shell.setup_multi_plot = function setup_multi_plot(element, monitor, sample_index, colors, is_interesting, setup_hook) {
        var self = this;
        var max_consumers = colors.length - 1;
        var data = new Array(max_consumers + 1);
        var consumers = new Array(max_consumers);
        var plot;
        var i;
        for (i = 0; i < data.length; i++) data[i] = {};
        function update_consumers() {
            var mcons = monitor.Consumers || [];
            consumers.forEach(function(c, i) {
                if (c && mcons.indexOf(c) < 0 || !is_interesting(c)) {
                    consumers[i] = null;
                }
            });
            mcons.forEach(function(mc) {
                if (!is_interesting(mc)) return;
                if (consumers.indexOf(mc) < 0) {
                    for (i = 0; i < max_consumers; i++) {
                        if (!consumers[i]) {
                            consumers[i] = mc;
                            return;
                        }
                    }
                    console.warn("Too many consumers");
                }
            });
        }
        function store_samples(samples, index) {
            var total = 0;
            for (var c in samples) {
                if (is_interesting(c)) total += samples[c][sample_index];
            }
            function store(i, value) {
                var series = data[i].data;
                var floor = i > 0 ? data[i - 1].data[index][2] : 0;
                series[index][1] = floor;
                series[index][2] = floor + value;
            }
            consumers.forEach(function(c, i) {
                store(i, c && samples[c] ? samples[c][sample_index] : 0);
            });
            store(max_consumers, total);
            if (index == monitor.NumSamples - 1) $(plot).trigger("update-total", [ total ]);
        }
        plot = setup_plot(element, monitor, data, {
            colors: colors,
            grid: {
                hoverable: true,
                autoHighlight: false
            },
            setup_hook: setup_hook
        }, store_samples);
        $(monitor).on("notify:Consumers", update_consumers);
        var cur_highlight = null;
        function highlight(consumer) {
            if (consumer != cur_highlight) {
                cur_highlight = consumer;
                $(plot).trigger("highlight", [ consumer ]);
            }
        }
        function highlight_on(event, pos, item) {
            var i, index;
            index = Math.round(pos.x);
            if (index < 0) index = 0;
            if (index > monitor.NumSamples - 1) index = monitor.NumSamples - 1;
            for (i = 0; i < max_consumers; i++) {
                if (i < max_consumers && data[i].data[index][1] <= pos.y && pos.y <= data[i].data[index][2]) break;
            }
            if (i < max_consumers) highlight(consumers[i]); else highlight(null);
        }
        function highlight_off(event) {
            highlight(null);
        }
        $(plot.element).on("plothover", highlight_on);
        $(plot.element).on("mouseleave", highlight_off);
        $(plot).on("destroyed", function() {
            $(monitor).off("notify:Consumers", update_consumers);
            $(plot.element).off("plothover", highlight_on);
            $(plot.element).off("mouseleave", highlight_off);
        });
        update_consumers();
        return plot;
    };
})(jQuery, cockpit, shell);

var shell = shell || {};

var modules = modules || {};

(function($, cockpit, shell, modules) {
    var month_names = [ "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" ];
    function format_date_tick(val, axis) {
        function pad(n) {
            var str = n.toFixed();
            if (str.length == 1) str = "0" + str;
            return str;
        }
        var d = new Date(val);
        var n = new Date();
        var time = pad(d.getHours()) + ":" + pad(d.getMinutes());
        if (d.getFullYear() == n.getFullYear() && d.getMonth() == n.getMonth() && d.getDate() == n.getDate()) {
            return time;
        } else {
            var day = C_("month-name", month_names[d.getMonth()]) + " " + d.getDate().toFixed();
            return day + ", " + time;
        }
    }
    var common_plot_options = {
        legend: {
            show: false
        },
        series: {
            shadowSize: 0
        },
        xaxis: {
            tickColor: "#d1d1d1",
            mode: "time",
            tickFormatter: format_date_tick,
            minTickSize: [ 1, "minute" ]
        },
        points: {
            radius: 0
        },
        grid: {
            borderWidth: 1,
            borderColor: "#e1e6ed",
            hoverable: true,
            autoHighlight: false
        }
    };
    function memory_ticks(opts) {
        var size = Math.pow(2, Math.ceil(Math.log(opts.max / 5) / Math.LN2));
        var ticks = [];
        for (var t = 0; t < opts.max; t += size) ticks.push(t);
        return ticks;
    }
    var resource_monitors = [ {
        selector: "#dashboard-plot-0",
        plot: {
            direct: [ "kernel.all.cpu.nice", "kernel.all.cpu.user", "kernel.all.cpu.sys" ],
            internal: [ "cpu.basic.nice", "cpu.basic.user", "cpu.basic.system" ],
            units: "millisec",
            derive: "rate",
            factor: .1
        },
        options: {
            yaxis: {
                tickColor: "#e1e6ed",
                tickFormatter: function(v) {
                    return v + "%";
                }
            }
        },
        ymax_unit: 100
    }, {
        selector: "#dashboard-plot-1",
        plot: {
            direct: [ "mem.util.used" ],
            internal: [ "memory.used" ],
            units: "bytes"
        },
        options: {
            yaxis: {
                ticks: memory_ticks,
                tickColor: "#e1e6ed",
                tickFormatter: function(v) {
                    return cockpit.format_bytes(v);
                }
            }
        },
        ymax_unit: 1e8
    }, {
        selector: "#dashboard-plot-2",
        plot: {
            direct: [ "network.interface.total.bytes" ],
            internal: [ "network.all.rx" ],
            units: "bytes",
            "omit-instances": [ "lo" ],
            derive: "rate"
        },
        options: {
            yaxis: {
                tickColor: "#e1e6ed",
                tickFormatter: function(v) {
                    return cockpit.format_bits_per_sec(v * 8);
                }
            }
        },
        ymax_min: 1e5
    }, {
        selector: "#dashboard-plot-3",
        plot: {
            direct: [ "disk.dev.total_bytes" ],
            internal: [ "block.device.read", "block.device.written" ],
            units: "bytes",
            derive: "rate"
        },
        options: {
            yaxis: {
                tickColor: "#e1e6ed",
                tickFormatter: function(v) {
                    return cockpit.format_bytes_per_sec(v);
                }
            }
        },
        ymax_min: 1e5
    } ];
    var avatar_editor;
    $(function() {
        avatar_editor = shell.image_editor($("#host-edit-avatar"), 256, 256);
        $("#host-edit-color").parent().on("show.bs.dropdown", function() {
            var $div = $("#host-edit-color");
            var $pop = $("#host-edit-color-popover");
            var div_pos = $div.position();
            var div_width = $div.width();
            var div_height = $div.height();
            var pop_width = $pop.width();
            var pop_height = $pop.height();
            $pop.css("left", div_pos.left + (div_width - pop_width) / 2);
            $pop.css("top", div_pos.top - pop_height + 10);
            $pop.show();
        }).on("hide.bs.dropdown", function() {
            $("#host-edit-color-popover").hide();
        });
    });
    function show_problem_dialog(machine) {
        $("#reconnect-dialog-summary").text(cockpit.format(_("Couldn't establish connection to $0."), machine.display_name));
        $("#reconnect-dialog-problem").text(cockpit.message(machine.problem));
        $("#reconnect-dialog-reconnect").off("click");
        $("#reconnect-dialog-reconnect").on("click", function() {
            $("#reconnect-dialog").modal("hide");
            machine.connect();
        });
        $("#reconnect-dialog").modal("show");
    }
    function host_edit_dialog(machine) {
        if (!machine) return;
        $("#host-edit-fail").text("").hide();
        $("#host-edit-name").val(machine.label);
        $("#host-edit-name").prop("disabled", machine.state == "failed");
        $("#host-edit-color").css("background-color", machine.color);
        $("#host-edit-apply").off("click");
        $("#host-edit-apply").on("click", function() {
            $("#host-edit-dialog").modal("hide");
            var values = {
                avatar: avatar_editor.changed ? avatar_editor.get_data(128, 128, "image/png") : null,
                color: $("#host-edit-color").css("background-color"),
                label: $("#host-edit-name").val()
            };
            machine.change(values).fail(shell.show_unexpected_error);
        });
        $("#host-edit-avatar").off("click");
        $("#host-edit-avatar").on("click", function() {
            $("#host-edit-fail").text("").hide();
            avatar_editor.select_file().done(function() {
                $("#host-edit-avatar").off("click");
                avatar_editor.changed = true;
                avatar_editor.start_cropping();
            });
        });
        $("#host-edit-dialog").modal("show");
        avatar_editor.stop_cropping();
        avatar_editor.load_data(machine.avatar || "images/server-large.png").fail(function() {
            $("#host-edit-fail").text("Can't load image").show();
        });
    }
    function update_servers_privileged() {
        shell.update_privileged_ui(shell.default_permission, ".servers-privileged", cockpit.format(_("The user <b>$0</b> is not permitted to manage servers"), cockpit.user.name));
    }
    $(shell.default_permission).on("changed", update_servers_privileged);
    PageDashboard.prototype = {
        _init: function() {
            this.id = "dashboard";
            this.edit_enabled = false;
        },
        getTitle: function() {
            return null;
        },
        toggle_edit: function(val) {
            var self = this;
            self.edit_enabled = val;
            $("#dashboard-enable-edit").toggleClass("active", self.edit_enabled);
            $("#dashboard-hosts").toggleClass("editable", self.edit_enabled);
        },
        setup: function() {
            var self = this;
            this.machines = modules.machines.instance();
            function make_color_div(c) {
                return $('<div class="color-cell">').css("background-color", c);
            }
            var rows = [];
            for (var i = 0; i < modules.machines.colors.length; i += 6) {
                var part = modules.machines.colors.slice(i, i + 6);
                rows.push($("<div>").append(part.map(make_color_div)));
            }
            $("#host-edit-color-popover .popover-content").append(rows);
            $("#host-edit-color-popover .popover-content .color-cell").click(function() {
                $("#host-edit-color").css("background-color", $(this).css("background-color"));
            });
            var current_monitor = 0;
            var plot_x_range = 5 * 60;
            var plot_x_stop;
            $("#dashboard-add").click(function() {
                shell.host_setup(self.machines);
            });
            $("#dashboard-enable-edit").click(function() {
                self.toggle_edit(!self.edit_enabled);
            });
            var renderer = host_renderer($("#dashboard-hosts .list-group"));
            $(self.machines).on("added.dashboard", function(ev, machine) {
                machine.connect();
                renderer();
            });
            $(self.machines).on("removed.dashboard", renderer);
            $(self.machines).on("changed.dashboard", renderer);
            $("#dashboard .nav-tabs li").click(function() {
                set_monitor(parseInt($(this).data("monitor-id"), 10));
            });
            function set_monitor(id) {
                $("#dashboard .nav-tabs li").removeClass("active");
                $("#dashboard .nav-tabs li[data-monitor-id=" + id + "]").addClass("active");
                current_monitor = id;
                $(".dashboard-plot").hide();
                $(resource_monitors[id].selector).show();
                plot_refresh();
            }
            $("#dashboard-range-buttons button").click(function() {
                set_plot_x_range(parseInt($(this).data("seconds"), 10));
            });
            $("#dashboard-scroll-left").click(function() {
                scroll_plot_left();
            });
            $("#dashboard-scroll-right").click(function() {
                scroll_plot_right();
            });
            function update_scroll_buttons() {
                $("#dashboard-scroll-right").attr("disabled", plot_x_stop === undefined);
            }
            function set_plot_x_range(val) {
                $("#dashboard-range-buttons button").removeClass("active");
                $("#dashboard-range-buttons button[data-seconds=" + val + "]").addClass("active");
                plot_x_range = val;
                plot_x_stop = undefined;
                plot_reset_soft();
                update_scroll_buttons();
            }
            function scroll_plot_left() {
                var step = plot_x_range / 10;
                if (plot_x_stop === undefined) plot_x_stop = new Date().getTime() / 1e3;
                plot_x_stop -= step;
                plot_reset_soft();
                update_scroll_buttons();
            }
            function scroll_plot_right() {
                var step = plot_x_range / 10;
                if (plot_x_stop !== undefined) {
                    plot_x_stop += step;
                    if (plot_x_stop >= new Date().getTime() / 1e3 - 10) plot_x_stop = undefined;
                    plot_reset_soft();
                }
                update_scroll_buttons();
            }
            plot_init();
            set_monitor(current_monitor);
            set_plot_x_range(plot_x_range);
            $("#dashboard-hosts").on("click", "a.list-group-item", function() {
                if (self.edit_enabled) return false;
                var addr = $(this).attr("data-address");
                var machine = self.machines.lookup(addr);
                if (machine.state == "failed") {
                    show_problem_dialog(machine);
                    return false;
                }
            }).on("click", "button.pficon-delete", function() {
                var item = $(this).parent(".list-group-item");
                self.toggle_edit(false);
                var machine = self.machines.lookup(item.attr("data-address"));
                if (machine) {
                    machine.change({
                        visible: false
                    });
                    machine.close();
                }
                return false;
            }).on("click", "button.pficon-edit", function() {
                var item = $(this).parent(".list-group-item");
                self.toggle_edit(false);
                host_edit_dialog(self.machines.lookup(item.attr("data-address")));
                return false;
            }).on("mouseenter", "a.list-group-item", function() {
                highlight($(this), true);
            }).on("mouseleave", "a.list-group-item", function() {
                highlight($(this), false);
            });
            var series = {};
            function update_series() {
                var refresh = false;
                var seen = {};
                $.each(series, function(addr) {
                    seen[addr] = true;
                });
                $("#dashboard-hosts .list-group-item").each(function() {
                    var item = $(this);
                    var addr = item.attr("data-address");
                    var machine = self.machines.lookup(addr);
                    if (!machine || machine.state == "failed") return;
                    delete seen[addr];
                    if (!series[addr]) {
                        series[addr] = plot_add(addr);
                    }
                    series[addr].forEach(function(s) {
                        $(s).off("hover").on("hover", function(event, val) {
                            highlight(item, val);
                        });
                        var color = shell.esc(machine.color);
                        if (s.options.color != color) {
                            refresh = true;
                            s.options.color = color;
                        }
                    });
                });
                $.each(seen, function(addr) {
                    series[addr].forEach(function(s) {
                        s.remove();
                    });
                    delete series[addr];
                });
                if (refresh) plot_refresh();
            }
            function highlight(item, val) {
                item.toggleClass("highlight", val);
                var ser = series[item.attr("data-address")];
                if (ser) {
                    ser.forEach(function(s) {
                        s.options.lines.lineWidth = val ? 3 : 2;
                        if (val) s.move_to_front();
                    });
                    plot_refresh();
                }
            }
            function host_renderer(target) {
                var template = $("#dashboard-hosts-tmpl").html();
                Mustache.parse(template);
                function render_avatar() {
                    if (this.state == "failed") return "images/server-error.png"; else if (this.avatar) return this.avatar; else return "images/server-small.png";
                }
                function render() {
                    var text = Mustache.render(template, {
                        machines: self.machines.list,
                        render_avatar: render_avatar
                    });
                    target.html(text);
                    update_servers_privileged();
                    update_series();
                }
                return render;
            }
            function plot_refresh() {
                self.plots.forEach(function(p) {
                    p.refresh();
                });
            }
            function plot_add(addr) {
                var machine = self.machines.lookup(addr);
                if (!machine || machine.state == "failed") return null;
                var series = [];
                var i = 0;
                resource_monitors.forEach(function(rm) {
                    if (self.plots[i]) {
                        series.push(self.plots[i].add_metrics_sum_series($.extend({
                            host: addr
                        }, rm.plot), {
                            color: machine.color,
                            lines: {
                                lineWidth: 2
                            }
                        }));
                    }
                    i += 1;
                });
                return series;
            }
            function plot_init() {
                self.plots = [];
                $("#dashboard-toolbar").hide();
                resource_monitors.forEach(function(rm) {
                    function setup_hook(flot) {
                        var axes = flot.getAxes();
                        var config = rm;
                        if (rm.ymax_unit) {
                            if (axes.yaxis.datamax) axes.yaxis.options.max = Math.ceil(axes.yaxis.datamax / config.ymax_unit) * rm.ymax_unit; else axes.yaxis.options.max = rm.ymax_unit;
                        }
                        if (rm.ymax_min) {
                            if (axes.yaxis.datamax < rm.ymax_min) axes.yaxis.options.max = rm.ymax_min; else axes.yaxis.options.max = null;
                        }
                        axes.yaxis.options.min = 0;
                    }
                    if (!rm.selector) return;
                    var options = $.extend({
                        setup_hook: setup_hook
                    }, common_plot_options, rm.options);
                    var plot = shell.plot($(rm.selector), plot_x_range, plot_x_stop);
                    plot.set_options(options);
                    self.plots.push(plot);
                    $(plot).on("changed", function() {
                        if (plot.archives) $("#dashboard-toolbar").show();
                    });
                });
                series = {};
                update_series();
            }
            function plot_reset_soft() {
                self.plots.forEach(function(p) {
                    p.stop_walking();
                    p.reset(plot_x_range, plot_x_stop);
                    p.refresh();
                    if (plot_x_stop === undefined) p.start_walking();
                });
            }
            $(cockpit).on("resize.dashboard", function() {
                self.plots.forEach(function(p) {
                    p.resize();
                });
            });
            renderer();
        },
        show: function() {
            update_servers_privileged();
            this.plots[0].resize();
            this.toggle_edit(false);
        },
        enter: function() {
            this.old_sidebar_state = $("#cockpit-sidebar").is(":visible");
            $("#content-navbar").hide();
            $("#cockpit-sidebar").hide();
        },
        leave: function() {
            $("#content-navbar").show();
            $("#cockpit-sidebar").toggle(this.old_sidebar_state);
        }
    };
    function PageDashboard() {
        this._init();
    }
    shell.pages.push(new PageDashboard());
})(jQuery, cockpit, shell, modules);

var shell = shell || {};

(function($, cockpit, shell) {
    PageSetupServer.prototype = {
        _init: function() {
            this.id = "dashboard_setup_server_dialog";
        },
        show: function() {
            $("#dashboard_setup_address").focus();
        },
        leave: function() {
            $(this.local_client).off(".setup");
            this.local_client.release();
            this.local_client = null;
        },
        setup: function() {
            $("#dashboard_setup_cancel").on("click", $.proxy(this, "cancel"));
            $("#dashboard_setup_prev").on("click", $.proxy(this, "prev"));
            $("#dashboard_setup_next").on("click", $.proxy(this, "next"));
        },
        highlight_error: function(container) {
            $(container).addClass("has-error");
        },
        hide_error: function(container) {
            $(container).removeClass("has-error");
        },
        highlight_error_message: function(id, message) {
            $(id).text(message);
            $(id).css("visibility", "visible");
        },
        hide_error_message: function(id) {
            $(id).css("visibility", "hidden");
        },
        check_empty_address: function() {
            var addr = $("#dashboard_setup_address").val();
            if (addr === "") {
                $("#dashboard_setup_next").prop("disabled", true);
                this.hide_error("#dashboard_setup_address_tab");
                this.hide_error_message("#dashboard_setup_address_error");
            } else if (addr.search(/\s+/) === -1) {
                $("#dashboard_setup_next").prop("disabled", false);
                this.hide_error("#dashboard_setup_address_tab");
                this.hide_error_message("#dashboard_setup_address_error");
            } else {
                $("#dashboard_setup_next").prop("disabled", true);
                this.highlight_error("#dashboard_setup_address_tab");
                this.highlight_error_message("#dashboard_setup_address_error", _("IP address or host name cannot contain whitespace."));
            }
            $("#dashboard_setup_next").text(_("Next"));
        },
        check_empty_name: function() {
            var name = $("#dashboard_setup_login_user").val();
            if (name === "") {
                this.name_is_done = false;
                $("#dashboard_setup_next").prop("disabled", true);
                this.hide_error("#login_user_cell");
                this.hide_error_message("#dashboard_setup_login_error");
            } else if (name.search(/\s+/) === -1) {
                this.name_is_done = true;
                $("#dashboard_setup_next").prop("disabled", false);
                this.hide_error("#login_user_cell");
                this.hide_error_message("#dashboard_setup_login_error");
            } else {
                this.name_is_done = false;
                $("#dashboard_setup_next").prop("disabled", true);
                this.highlight_error("#login_user_cell");
                this.highlight_error_message("#dashboard_setup_login_error", _("User name cannot contain whitespace."));
            }
            $("#dashboard_setup_next").text(_("Next"));
        },
        enter: function() {
            var self = this;
            self.machines = PageSetupServer.machines;
            self.client = null;
            self.address = null;
            self.options = {
                "host-key": ""
            };
            self.name_is_done = false;
            $("#dashboard_setup_address")[0].placeholder = _("Enter IP address or host name");
            $("#dashboard_setup_address").on("keyup change", $.proxy(this, "update_discovered"));
            $("#dashboard_setup_address").on("input change focus", $.proxy(this, "check_empty_address"));
            $("#dashboard_setup_login_user").on("input change focus", $.proxy(this, "check_empty_name"));
            $("#dashboard_setup_login_password").on("input focus", function() {
                if (self.name_is_done) self.hide_error_message("#dashboard_setup_login_error");
            });
            $("#dashboard_setup_address").on("keyup", function(event) {
                if (event.which === 13) {
                    var disable = $("#dashboard_setup_next").prop("disabled");
                    if (!disable) self.next();
                }
            });
            $("#dashboard_setup_login_user").on("keyup", function(event) {
                if (event.which === 13) $("#dashboard_setup_login_password").focus();
            });
            $("#dashboard_setup_login_password").on("keyup", function(event) {
                if (event.which === 13) {
                    var disable = $("#dashboard_setup_next").prop("disabled");
                    if (!disable) self.next();
                }
            });
            self.local_client = shell.dbus("localhost");
            $(self.local_client).on("objectAdded.setup objectRemoved.setup", function(event, object) {
                if (object.lookup("com.redhat.Cockpit.Machine")) self.update_discovered();
            });
            $(self.local_client).on("propertiesChanged.setup", function(event, object, iface) {
                if (iface._iface_name == "com.redhat.Cockpit.Machine") self.update_discovered();
            });
            $("#dashboard_setup_address").val("");
            $("#dashboard_setup_login_user").val("");
            $("#dashboard_setup_login_password").val("");
            $("#dashboard_setup_address_reuse_creds").prop("checked", true);
            self.show_tab("address");
            self.update_discovered();
            $("#dashboard_setup_next").prop("disabled", true);
        },
        update_discovered: function() {
            var self = this;
            var filter = $("#dashboard_setup_address").val();
            var discovered = $("#dashboard_setup_address_discovered");
            function render_address(address) {
                if (!address.trim()) return null;
                if (!filter) return $("<span/>").text(address);
                var index = address.indexOf(filter);
                if (index == -1) return null;
                return $("<span/>").append($("<span/>").text(address.substring(0, index)), $("<b/>").text(address.substring(index, index + filter.length)), $("<span/>").text(address.substring(index + filter.length)));
            }
            discovered.empty();
            var rendered_address, item;
            var address, machine, addresses = self.machines.addresses;
            for (var i = 0; i < addresses.length; i++) {
                address = addresses[i];
                machine = self.machines.lookup(address);
                if (!machine.visible) {
                    rendered_address = render_address(address);
                    if (rendered_address) {
                        item = $("<li>", {
                            "class": "list-group-item",
                            on: {
                                click: $.proxy(this, "discovered_clicked", address)
                            }
                        }).html(rendered_address);
                        discovered.append(item);
                    }
                }
            }
        },
        discovered_clicked: function(address) {
            $("#dashboard_setup_address").val(address);
            this.update_discovered();
            $("#dashboard_setup_address").focus();
        },
        show_tab: function(tab) {
            $(".cockpit-setup-tab").hide();
            $("#dashboard_setup_next").text(_("Next"));
            if (tab == "address") {
                $("#dashboard_setup_address_tab").show();
                $("#dashboard_setup_address").focus();
                this.hide_error_message("#dashboard_setup_address_error");
                this.next_action = this.next_select;
                this.prev_tab = null;
            } else if (tab == "login") {
                $("#dashboard_setup_login_tab").show();
                $("#dashboard_setup_login_user").focus();
                this.hide_error_message("#dashboard_setup_login_error");
                this.next_action = this.next_login;
                this.prev_tab = "address";
            } else if (tab == "action") {
                $("#dashboard_setup_action_tab").show();
                $("#dashboard_setup_next").text(_("Add host"));
                this.next_action = this.next_setup;
                var reuse = $("#dashboard_setup_address_reuse_creds").prop("checked");
                if (reuse) this.prev_tab = "address"; else this.prev_tab = "login";
            } else if (tab == "close") {
                $("#dashboard_setup_action_tab").show();
                $("#dashboard_setup_next").text(_("Close"));
                this.next_action = this.next_close;
                this.prev_tab = null;
            }
            if (this.next_action === this.next_login) this.check_empty_name(); else $("#dashboard_setup_next").prop("disabled", false);
            $("#dashboard_setup_prev").prop("disabled", !this.prev_tab);
        },
        close: function() {
            if (this.client) this.client.close("cancelled");
            $("#dashboard_setup_server_dialog").modal("hide");
        },
        cancel: function() {
            this.close();
        },
        prev: function() {
            if (this.prev_tab) this.show_tab(this.prev_tab);
        },
        next: function() {
            $("#dashboard_setup_next").html('<div class="waiting"/>');
            $("#dashboard_setup_next").prop("disabled", true);
            this.next_action();
        },
        connect_server: function() {
            var self = this;
            var client = shell.dbus(self.address, self.options);
            $(client).on("state-change", function() {
                if (client.state == "closed") {
                    if (!self.options.host_key && client.error == "unknown-hostkey") {
                        self.options["host-key"] = client.error_details["host-key"];
                        $("#dashboard_setup_action_fingerprint").text(client.error_details["host-fingerprint"]);
                        self.connect_server();
                        return;
                    } else if (client.error == "authentication-failed") {
                        self.show_tab("login");
                        self.highlight_error_message("#dashboard_setup_login_error", cockpit.message(client.error));
                        return;
                    }
                    var problem = client.error || "disconnected";
                    self.highlight_error_message("#dashboard_setup_address_error", cockpit.message(problem));
                    self.highlight_error_message("#dashboard_setup_login_error", cockpit.message(problem));
                    $("#dashboard_setup_next").prop("disabled", false);
                    $("#dashboard_setup_next").text(_("Next"));
                    return;
                } else if (client.state == "ready") {
                    self.client = client;
                    self.prepare_setup();
                }
            });
        },
        next_select: function() {
            var me = this;
            var reuse_creds;
            me.hide_error_message("#dashboard_setup_address_error");
            me.address = $("#dashboard_setup_address").val();
            if (me.address.trim() !== "") {
                $("#dashboard_setup_login_address").text(me.address);
                reuse_creds = $("#dashboard_setup_address_reuse_creds").prop("checked");
                if (!reuse_creds) me.show_tab("login"); else {
                    me.options.user = null;
                    me.options.password = null;
                    me.options.host_key = null;
                    me.connect_server();
                }
            } else {
                $("#dashboard_setup_next").text(_("Next"));
                me.highlight_error_message("#dashboard_setup_address_error", _("IP address or host name cannot be empty."));
            }
        },
        next_login: function() {
            var me = this;
            var user = $("#dashboard_setup_login_user").val();
            var pass = $("#dashboard_setup_login_password").val();
            me.hide_error_message("#dashboard_setup_login_error");
            me.options.user = user;
            me.options.password = pass;
            if (user.trim() !== "") {
                me.connect_server();
            } else {
                $("#dashboard_setup_next").text(_("Next"));
                me.highlight_error_message("#dashboard_setup_login_error", _("User name cannot be empty."));
            }
        },
        reset_tasks: function() {
            var $tasks = $("#dashboard_setup_action_tasks");
            this.tasks = [];
            $tasks.empty();
        },
        add_task: function(desc, func) {
            var $tasks = $("#dashboard_setup_action_tasks");
            var $entry = $("<li/>", {
                "class": "list-group-item"
            }).append($("<table/>", {
                "class": "cockpit-setup-task-table",
                style: "width:100%"
            }).append($("<tr/>").append($("<td/>").text(desc), $('<td style="width:16px"/>').append($("<div>", {
                "class": "cockpit-setup-task-spinner waiting",
                style: "display:none"
            }), $("<img/>", {
                "class": "cockpit-setup-task-error",
                src: "images/dialog-error.png",
                style: "display:none"
            }), $("<img/>", {
                "class": "cockpit-setup-task-done",
                src: "images/face-smile.png",
                style: "display:none"
            })))));
            var task = {
                entry: $entry,
                func: func,
                error: function(msg) {
                    this.had_error = true;
                    this.entry.find(".cockpit-setup-task-table").append($("<tr/>").append($("<td/>", {
                        style: "color:red"
                    }).text(msg)));
                }
            };
            this.tasks.push(task);
            $tasks.append($entry);
        },
        run_tasks: function(done) {
            var me = this;
            function run(i) {
                var t;
                if (i < me.tasks.length) {
                    t = me.tasks[i];
                    t.entry.find(".cockpit-setup-task-spinner").show();
                    t.func(t, function() {
                        t.entry.find(".cockpit-setup-task-spinner").hide();
                        if (t.had_error) t.entry.find(".cockpit-setup-task-error").show(); else t.entry.find(".cockpit-setup-task-done").show();
                        run(i + 1);
                    });
                } else done();
            }
            run(0);
        },
        prepare_setup: function() {
            var me = this;
            function get_role_accounts(client, roles) {
                var i;
                var accounts = client.getInterfacesFrom("/com/redhat/Cockpit/Accounts/", "com.redhat.Cockpit.Account");
                var map = {};
                function groups_contain_roles(groups) {
                    for (var i = 0; i < roles.length; i++) {
                        if (shell.find_in_array(groups, roles[i][0])) return true;
                    }
                    return false;
                }
                for (i = 0; i < accounts.length; i++) {
                    if (roles === null || groups_contain_roles(accounts[i].Groups)) map[accounts[i].UserName] = accounts[i];
                }
                return map;
            }
            var manager = this.local_client.lookup("/com/redhat/Cockpit/Accounts", "com.redhat.Cockpit.Accounts");
            var local = get_role_accounts(this.local_client, manager.Roles);
            var remote = get_role_accounts(this.client, null);
            function needs_update(l) {
                return true;
            }
            function update_account(templ, task, done) {
                var acc;
                function std_cont(func) {
                    return function(error, result) {
                        if (error) {
                            task.error(error.message);
                            done();
                        } else {
                            func(result);
                        }
                    };
                }
                function create() {
                    var accounts = me.client.getInterfacesFrom("/com/redhat/Cockpit/Accounts", "com.redhat.Cockpit.Account");
                    for (var i = 0; i < accounts.length; i++) {
                        if (accounts[i].UserName == templ.UserName) {
                            acc = accounts[i];
                            set_groups();
                            return;
                        }
                    }
                    var manager = me.client.lookup("/com/redhat/Cockpit/Accounts", "com.redhat.Cockpit.Accounts");
                    manager.call("CreateAccount", templ.UserName, templ.RealName, "", false, std_cont(set_groups_path));
                }
                function set_groups_path(path) {
                    acc = me.client.lookup(path, "com.redhat.Cockpit.Account");
                    if (!acc) {
                        task.error("Account object not found");
                        done();
                    } else set_groups();
                }
                function set_groups() {
                    acc.call("ChangeGroups", templ.Groups, [], std_cont(get_avatar));
                }
                function get_avatar() {
                    templ.call("GetIconDataURL", std_cont(set_avatar));
                }
                function set_avatar(dataurl) {
                    if (dataurl) {
                        acc.call("SetIconDataURL", dataurl, std_cont(get_password_hash));
                    } else get_password_hash();
                }
                function get_password_hash() {
                    templ.call("GetPasswordHash", std_cont(set_password_hash));
                }
                function set_password_hash(hash) {
                    acc.call("SetPasswordHash", hash, std_cont(done));
                }
                create();
            }
            this.reset_tasks();
            function create_update_task(templ) {
                return function(task, done) {
                    update_account(templ, task, done);
                };
            }
            for (var l in local) {
                if (needs_update(l)) this.add_task("Synchronize administrator " + local[l].UserName, create_update_task(local[l]));
            }
            $("#dashboard_setup_action_address").text(this.address);
            this.show_tab("action");
        },
        next_setup: function() {
            var self = this;
            self.machines.add(self.address, self.options["host-key"]).fail(function(ex) {
                self.highlight_error_message("#dashboard_setup_address_error", ex.toString());
                self.show_tab("address");
            }).done(function(machine) {
                self.run_tasks(function() {
                    machine.change({
                        visible: true
                    }).fail(function(ex) {
                        shell.show_unexpected_error(ex);
                    });
                    self.show_tab("close");
                });
            });
        },
        next_close: function() {
            this.close();
        }
    };
    function PageSetupServer() {
        this._init();
    }
    shell.dialogs.push(new PageSetupServer());
    shell.host_setup = function host_setup(machines) {
        PageSetupServer.machines = machines;
        $("#dashboard_setup_server_dialog").modal("show");
    };
})(jQuery, cockpit, shell);

var shell = shell || {};

var modules = modules || {};

(function($, cockpit, shell, modules) {
    function update_hostname_privileged() {
        shell.update_privileged_ui(shell.default_permission, ".hostname-privileged", cockpit.format(_("The user <b>$0</b> is not permitted to modify hostnames"), cockpit.user.name));
    }
    function update_realm_privileged() {
        shell.update_privileged_ui(shell.default_permission, ".realm-privileged", cockpit.format(_("The user <b>$0</b> is not permitted to modify realms"), cockpit.user.name));
    }
    $(shell.default_permission).on("changed", update_realm_privileged);
    $(shell.default_permission).on("changed", update_hostname_privileged);
    function ServerTime() {
        var self = this;
        var client = cockpit.dbus("org.freedesktop.timedate1");
        var timedate = client.proxy();
        var time_offset = null;
        var remote_offset = null;
        self.timedate = timedate;
        Object.defineProperty(self, "now", {
            enumerable: true,
            get: function get() {
                var offset = time_offset + remote_offset;
                return new Date(offset + new Date().valueOf());
            }
        });
        self.format = function format(and_time) {
            var string = self.now.toISOString();
            if (!and_time) return string.split("T")[0];
            var pos = string.lastIndexOf(":");
            if (pos !== -1) string = string.substring(0, pos);
            return string.replace("T", " ");
        };
        var interval = window.setInterval(function() {
            $(self).triggerHandler("changed");
        }, 3e4);
        function offsets(timems, offsetms) {
            var now = new Date();
            time_offset = timems - now.valueOf();
            remote_offset = offsetms;
            $(self).triggerHandler("changed");
        }
        self.update = function update() {
            cockpit.spawn([ "/usr/bin/date", "+%s:%:z" ]).done(function(data) {
                var parts = data.trim().split(":").map(function(x) {
                    return parseInt(x, 10);
                });
                if (parts[1] < 0) parts[2] = -parts[2];
                offsets(parts[0] * 1e3, parts[1] * 36e5 + parts[2] * 6e4);
            });
        };
        self.change_time = function change_time(datestr, hourstr, minstr) {
            var dfd = $.Deferred();
            cockpit.spawn([ "/usr/bin/date", "--date=" + datestr + " " + hourstr + ":" + minstr, "+%s" ]).fail(function(ex) {
                dfd.reject(ex);
            }).done(function(data) {
                var seconds = parseInt(data.trim(), 10);
                timedate.SetTime(seconds * 1e3 * 1e3, false, true).fail(function(ex) {
                    dfd.reject(ex);
                }).done(function() {
                    self.update();
                    dfd.resolve();
                });
            });
            return dfd;
        };
        self.close = function close() {
            client.close();
        };
        self.update();
    }
    PageServer.prototype = {
        _init: function() {
            this.id = "server";
            this.server_time = null;
        },
        getTitle: function() {
            return null;
        },
        setup: function() {
            var self = this;
            update_realm_privileged();
            update_hostname_privileged();
            self.timedate = cockpit.dbus("org.freedesktop.timedate1").proxy();
            $("#shutdown-group").append(shell.action_btn(function(op) {
                self.shutdown(op);
            }, [ {
                title: _("Restart"),
                action: "default"
            }, {
                title: _("Shutdown"),
                action: "shutdown"
            } ]));
            $("#server-avatar").on("click", $.proxy(this, "trigger_change_avatar"));
            $("#server-avatar-uploader").on("change", $.proxy(this, "change_avatar"));
            $("#system_information_hostname_button").on("click", function() {
                PageSystemInformationChangeHostname.client = self.client;
                $("#system_information_change_hostname").modal("show");
            });
            $("#system_information_systime_button").on("click", function() {
                PageSystemInformationChangeSystime.server_time = self.server_time;
                $("#system_information_change_systime").modal("show");
            });
            $("#system_information_realms_button").on("click", function() {
                if (self.realms.Joined && self.realms.Joined.length > 0) {
                    var name = self.realms.Joined[0][0];
                    var details = self.realms.Joined[0][1];
                    shell.realms_op_set_parameters(self.realms, "leave", name, details);
                    $("#realms-op").modal("show");
                } else {
                    shell.realms_op_set_parameters(self.realms, "join", "", {});
                    $("#realms-op").modal("show");
                }
            });
            self.server_time = new ServerTime();
            $(self.server_time).on("changed", function() {
                $("#system_information_systime_button").text(self.server_time.format(true));
            });
        },
        enter: function() {
            var self = this;
            self.client = shell.dbus(null);
            self.manager = self.client.get("/com/redhat/Cockpit/Manager", "com.redhat.Cockpit.Manager");
            $(self.manager).on("AvatarChanged.server", $.proxy(this, "update_avatar"));
            $("#server-avatar").attr("src", "images/server-large.png");
            var monitor;
            var series;
            var cpu_data = {
                direct: [ "kernel.all.cpu.nice", "kernel.all.cpu.user", "kernel.all.cpu.sys" ],
                internal: [ "cpu.basic.nice", "cpu.basic.user", "cpu.basic.system" ],
                units: "millisec",
                derive: "rate",
                factor: .1
            };
            var cpu_options = shell.plot_simple_template();
            $.extend(cpu_options.yaxis, {
                max: 100
            });
            self.cpu_plot = shell.plot($("#server_cpu_graph"), 300);
            self.cpu_plot.set_options(cpu_options);
            series = self.cpu_plot.add_metrics_sum_series(cpu_data, {});
            $(series).on("value", function(ev, value) {
                $("#server_cpu_text").text(value.toFixed(1) + "%");
            });
            var memory_data = {
                direct: [ "mem.util.used" ],
                internal: [ "memory.used" ],
                units: "bytes"
            };
            var memory_options = shell.plot_simple_template();
            self.memory_plot = shell.plot($("#server_memory_graph"), 300);
            self.memory_plot.set_options(memory_options);
            series = self.memory_plot.add_metrics_sum_series(memory_data, {});
            $(series).on("value", function(ev, value) {
                $("#server_memory_text").text(cockpit.format_bytes(value));
            });
            var network_data = {
                direct: [ "network.interface.total.bytes" ],
                internal: [ "network.all.tx", "network.all.rx" ],
                units: "bytes",
                derive: "rate"
            };
            var network_options = shell.plot_simple_template();
            network_options.setup_hook = function network_setup_hook(plot) {
                var axes = plot.getAxes();
                if (axes.yaxis.datamax < 1e5) axes.yaxis.options.max = 1e5; else axes.yaxis.options.max = null;
                axes.yaxis.options.min = 0;
            };
            self.network_plot = shell.plot($("#server_network_traffic_graph"), 300);
            self.network_plot.set_options(network_options);
            series = self.network_plot.add_metrics_sum_series(network_data, {});
            $(series).on("value", function(ev, value) {
                $("#server_network_traffic_text").text(cockpit.format_bits_per_sec(value * 8));
            });
            var disk_data = {
                direct: [ "disk.dev.total_bytes" ],
                internal: [ "block.device.read", "block.device.written" ],
                units: "bytes",
                derive: "rate"
            };
            self.disk_plot = shell.plot($("#server_disk_io_graph"), 300);
            self.disk_plot.set_options(shell.plot_simple_template());
            series = self.disk_plot.add_metrics_sum_series(disk_data, {});
            $(series).on("value", function(ev, value) {
                $("#server_disk_io_text").text(cockpit.format_bytes_per_sec(value));
            });
            shell.util.machine_info(null).done(function(info) {
                cpu_options.yaxis.max = info.cpus * 100;
                self.cpu_plot.set_options(cpu_options);
                memory_options.yaxis.max = info.memory;
                self.memory_plot.set_options(memory_options);
            });
            self.update_avatar();
            function bindf(sel, object, prop, func) {
                function update() {
                    $(sel).text(func(object[prop]));
                }
                $(object).on("notify:" + prop + ".server", update);
                update();
            }
            function bind(sel, object, prop) {
                bindf(sel, object, prop, function(s) {
                    return s;
                });
            }
            function parse_lines(output) {
                var ret = {};
                $.each(output.split("\n"), function(i, line) {
                    var pos = line.indexOf(":");
                    if (pos !== -1) ret[line.substring(0, pos)] = line.substring(pos + 1);
                });
                return ret;
            }
            cockpit.spawn([ "grep", "\\w", "bios_vendor", "bios_version", "bios_date", "sys_vendor", "product_name" ], {
                directory: "/sys/devices/virtual/dmi/id",
                err: "ignore"
            }).done(function(output) {
                var fields = parse_lines(output);
                $("#system_information_bios_text").text(fields.bios_vendor + " " + fields.bios_version + " (" + fields.bios_date + ")");
                $("#system_information_hardware_text").text(fields.sys_vendor + " " + fields.product_name);
            }).fail(function(ex) {
                console.warn("couldn't read dmi info: " + ex);
            });
            cockpit.spawn([ "grep", "\\w", "product_serial", "chassis_serial" ], {
                directory: "/sys/devices/virtual/dmi/id",
                superuser: true,
                err: "ignore"
            }).done(function(output) {
                var fields = parse_lines(output);
                $("#system_information_asset_tag_text").text(fields.product_serial || fields.chassis_serial);
            }).fail(function(ex) {
                if (ex.problem != "access-denied") console.warn("couldn't read serial dmi info: " + ex);
            });
            bind("#system_information_os_text", self.manager, "OperatingSystem");
            function hostname_text() {
                var pretty_hostname = self.manager.PrettyHostname;
                var static_hostname = self.manager.StaticHostname;
                var str;
                if (!pretty_hostname || pretty_hostname == static_hostname) str = static_hostname; else str = pretty_hostname + " (" + static_hostname + ")";
                if (str === "") str = _("Set Host name");
                return str;
            }
            bindf("#system_information_hostname_button", self.manager, "StaticHostname", hostname_text);
            bindf("#system_information_hostname_button", self.manager, "PrettyHostname", hostname_text);
            self.realms = self.client.get("/com/redhat/Cockpit/Realms", "com.redhat.Cockpit.Realms");
            $(self.realms).on("notify:Joined.server", $.proxy(self, "update_realms"));
            self.update_realms();
        },
        show: function() {
            this.cpu_plot.start_walking();
            this.memory_plot.start_walking();
            this.disk_plot.start_walking();
            this.network_plot.start_walking();
        },
        leave: function() {
            var self = this;
            self.cpu_plot.destroy();
            self.memory_plot.destroy();
            self.disk_plot.destroy();
            self.network_plot.destroy();
            $(self.manager).off(".server");
            self.manager = null;
            $(self.realms).off(".server");
            self.realms = null;
            $(self.client).off(".server");
            self.client.release();
            self.client = null;
        },
        shutdown: function(action_type) {
            PageShutdownDialog.type = action_type;
            $("#shutdown-dialog").modal("show");
        },
        update_avatar: function() {
            this.manager.call("GetAvatarDataURL", function(error, result) {
                if (result) $("#server-avatar").attr("src", result);
            });
        },
        trigger_change_avatar: function() {
            if (window.File && window.FileReader) $("#server-avatar-uploader").trigger("click");
        },
        change_avatar: function() {
            var me = this;
            shell.show_change_avatar_dialog("#server-avatar-uploader", function(data) {
                me.manager.call("SetAvatarDataURL", data, function(error) {
                    if (error) shell.show_unexpected_error(error);
                });
            });
        },
        update_realms: function() {
            var self = this;
            var joined = self.realms.Joined;
            function realms_text(val) {
                if (!val || val.length === 0) return _("Join Domain");
                var res = [];
                for (var i = 0; i < val.length; i++) res.push(val[i][0]);
                return res.join(", ");
            }
            $("#system_information_realms_button").text(realms_text(joined));
        }
    };
    function PageServer() {
        this._init();
    }
    shell.pages.push(new PageServer());
    PageSystemInformationChangeHostname.prototype = {
        _init: function() {
            this.id = "system_information_change_hostname";
        },
        setup: function() {
            $("#sich-pretty-hostname").on("input change", $.proxy(this._on_full_name_changed, this));
            $("#sich-hostname").on("input change", $.proxy(this._on_name_changed, this));
            $("#sich-apply-button").on("click", $.proxy(this._on_apply_button, this));
        },
        enter: function() {
            var self = this;
            self.manager = PageSystemInformationChangeHostname.client.get("/com/redhat/Cockpit/Manager", "com.redhat.Cockpit.Manager");
            self._initial_hostname = self.manager.StaticHostname || "";
            self._initial_pretty_hostname = self.manager.PrettyHostname || "";
            $("#sich-pretty-hostname").val(self._initial_pretty_hostname);
            $("#sich-hostname").val(self._initial_hostname);
            this._always_update_from_pretty = false;
            this._update();
        },
        show: function() {
            $("#sich-pretty-hostname").focus();
        },
        leave: function() {},
        _on_apply_button: function(event) {
            var self = this;
            var new_full_name = $("#sich-pretty-hostname").val();
            var new_name = $("#sich-hostname").val();
            self.manager.call("SetHostname", new_full_name, new_name, {}, function(error, reply) {
                $("#system_information_change_hostname").modal("hide");
                if (error) {
                    shell.show_unexpected_error(error);
                }
            });
        },
        _on_full_name_changed: function(event) {
            var pretty_hostname = $("#sich-pretty-hostname").val();
            if (this._always_update_from_pretty || this._initial_pretty_hostname != pretty_hostname) {
                var old_hostname = $("#sich-hostname").val();
                var first_dot = old_hostname.indexOf(".");
                var new_hostname = pretty_hostname.toLowerCase().replace(/['".]+/g, "").replace(/[^a-zA-Z0-9]+/g, "-");
                new_hostname = new_hostname.substr(0, 64);
                if (first_dot >= 0) new_hostname = new_hostname + old_hostname.substr(first_dot);
                $("#sich-hostname").val(new_hostname);
                this._always_update_from_pretty = true;
            }
            this._update();
        },
        _on_name_changed: function(event) {
            this._update();
        },
        _update: function() {
            var apply_button = $("#sich-apply-button");
            var note1 = $("#sich-note-1");
            var note2 = $("#sich-note-2");
            var changed = false;
            var valid = false;
            var can_apply = false;
            var charError = "Real host name can only contain lower-case characters, digits, dashes, and periods (with populated subdomains)";
            var lengthError = "Real host name must be 64 characters or less";
            var validLength = $("#sich-hostname").val().length <= 64;
            var hostname = $("#sich-hostname").val();
            var pretty_hostname = $("#sich-pretty-hostname").val();
            var validSubdomains = true;
            var periodCount = 0;
            for (var i = 0; i < $("#sich-hostname").val().length; i++) {
                if ($("#sich-hostname").val()[i] == ".") periodCount++; else periodCount = 0;
                if (periodCount > 1) {
                    validSubdomains = false;
                    break;
                }
            }
            var validName = hostname.match(/[.a-z0-9-]*/) == hostname && validSubdomains;
            if ((hostname != this._initial_hostname || pretty_hostname != this._initial_pretty_hostname) && (hostname !== "" || pretty_hostname !== "")) changed = true;
            if (validLength && validName) valid = true;
            if (changed && valid) can_apply = true;
            if (valid) {
                $(note1).css("visibility", "hidden");
                $(note2).css("visibility", "hidden");
                $("#sich-hostname-error").removeClass("has-error");
            } else if (!validLength && validName) {
                $("#sich-hostname-error").addClass("has-error");
                $(note1).text(lengthError);
                $(note1).css("visibility", "visible");
                $(note2).css("visibility", "hidden");
            } else if (validLength && !validName) {
                $("#sich-hostname-error").addClass("has-error");
                $(note1).text(charError);
                $(note1).css("visibility", "visible");
                $(note2).css("visibility", "hidden");
            } else {
                $("#sich-hostname-error").addClass("has-error");
                if ($(note1).text() === lengthError) $(note2).text(charError); else if ($(note1).text() === charError) $(note2).text(lengthError); else {
                    $(note1).text(lengthError);
                    $(note2).text(charError);
                }
                $(note1).css("visibility", "visible");
                $(note2).css("visibility", "visible");
            }
            apply_button.prop("disabled", !can_apply);
        }
    };
    function PageSystemInformationChangeHostname() {
        this._init();
    }
    shell.dialogs.push(new PageSystemInformationChangeHostname());
    PageSystemInformationChangeSystime.prototype = {
        _init: function() {
            this.id = "system_information_change_systime";
        },
        setup: function() {
            $("#systime-apply-button").on("click", $.proxy(this._on_apply_button, this));
            $("#change_systime").on("change", $.proxy(this, "update"));
            $("#systime-time-minutes").on("focusout", $.proxy(this, "update_minutes"));
            $("#systime-time-minutes").on("change", $.proxy(this, "check_input"));
            $("#systime-time-hours").on("change", $.proxy(this, "check_input"));
            $("#systime-date-input").on("change", $.proxy(this, "check_input"));
            $("#systime-date-input").datepicker({
                autoclose: true,
                todayHighlight: true,
                format: "yyyy-mm-dd"
            });
        },
        enter: function() {
            var server_time = PageSystemInformationChangeSystime.server_time;
            $("#systime-date-input").val(server_time.format());
            $("#systime-time-minutes").val(server_time.now.getUTCMinutes());
            $("#systime-time-hours").val(server_time.now.getUTCHours());
            $("#change_systime").val(server_time.timedate.NTP ? "ntp_time" : "manual_time");
            $("#change_systime").selectpicker("refresh");
            $("#systime-parse-error").css("visibility", "hidden");
            $("#systime-apply-button").prop("disabled", false);
            this.update();
            this.update_minutes();
        },
        show: function() {},
        leave: function() {},
        _on_apply_button: function(event) {
            var server_time = PageSystemInformationChangeSystime.server_time;
            var manual_time = $("#change_systime").val() == "manual_time";
            if (manual_time && !this.check_input()) return;
            server_time.timedate.SetNTP($("#change_systime").val() == "ntp_time", true).fail(function(err) {
                shell.show_unexpected_error(err);
                $("#system_information_change_systime").modal("hide");
            }).done(function() {
                if (!manual_time) {
                    $("#system_information_change_systime").modal("hide");
                    return;
                }
                server_time.change_time($("#systime-date-input").val(), $("#systime-time-hours").val(), $("#systime-time-minutes").val()).fail(function(err) {
                    shell.show_unexpected_error(err);
                }).always(function() {
                    $("#system_information_change_systime").modal("hide");
                });
            });
        },
        check_input: function() {
            var time_error = false;
            var date_error = false;
            var new_date;
            var hours = parseInt($("#systime-time-hours").val(), 10);
            var minutes = parseInt($("#systime-time-minutes").val(), 10);
            if (isNaN(hours) || hours < 0 || hours > 23 || isNaN(minutes) || minutes < 0 || minutes > 59) {
                time_error = true;
            }
            new_date = new Date($("#systime-date-input").val());
            if (isNaN(new_date.getTime()) || new_date.getTime() < 0) date_error = true;
            if (time_error && date_error) $("#systime-parse-error").text(_("Invalid date format and invalid time format")); else if (time_error) $("#systime-parse-error").text(_("Invalid time format")); else if (date_error) $("#systime-parse-error").text(_("Invalid date format")); else $("#systime-parse-error").css("visibility", "hidden");
            if (time_error || date_error) {
                $("#systime-parse-error").css("visibility", "visible");
                $("#systime-apply-button").prop("disabled", true);
                return false;
            } else {
                $("#systime-parse-error").css("visibility", "hidden");
                $("#systime-apply-button").prop("disabled", false);
                return true;
            }
        },
        update: function() {
            var ntp_time = $("#change_systime").val() === "ntp_time";
            $("#systime-date-input").prop("disabled", ntp_time);
            $("#systime-time-hours").prop("disabled", ntp_time);
            $("#systime-time-minutes").prop("disabled", ntp_time);
        },
        update_minutes: function() {
            var val = parseInt($("#systime-time-minutes").val(), 10);
            if (val < 10) $("#systime-time-minutes").val("0" + val);
        }
    };
    function PageSystemInformationChangeSystime() {
        this._init();
    }
    shell.dialogs.push(new PageSystemInformationChangeSystime());
    PageShutdownDialog.prototype = {
        _init: function() {
            this.id = "shutdown-dialog";
        },
        setup: function() {
            $("#shutdown-delay").html(this.delay_btn = shell.select_btn($.proxy(this, "update"), [ {
                choice: "1",
                title: _("1 Minute")
            }, {
                choice: "5",
                title: _("5 Minutes")
            }, {
                choice: "20",
                title: _("20 Minutes")
            }, {
                choice: "40",
                title: _("40 Minutes")
            }, {
                choice: "60",
                title: _("60 Minutes")
            }, {
                group: [ {
                    choice: "0",
                    title: _("No Delay")
                }, {
                    choice: "x",
                    title: _("Specific Time")
                } ]
            } ]).css("display", "inline"));
            $("#shutdown-time input").change($.proxy(this, "update"));
        },
        enter: function(event) {
            $("#shutdown-message").val("").attr("placeholder", _("Message to logged in users")).attr("rows", 5);
            shell.select_btn_select(this.delay_btn, "1");
            if (PageShutdownDialog.type == "shutdown") {
                $("#shutdown-dialog .modal-title").text(_("Shutdown"));
                $("#shutdown-action").click($.proxy(this, "shutdown"));
                $("#shutdown-action").text(_("Shutdown"));
            } else {
                $("#shutdown-dialog .modal-title").text(_("Restart"));
                $("#shutdown-action").click($.proxy(this, "restart"));
                $("#shutdown-action").text(_("Restart"));
            }
            this.update();
        },
        show: function(e) {},
        leave: function() {},
        update: function() {
            var disabled = false;
            var delay = shell.select_btn_selected(this.delay_btn);
            $("#shutdown-time").toggle(delay == "x");
            if (delay == "x") {
                var h = parseInt($("#shutdown-time input:nth-child(1)").val(), 10);
                var m = parseInt($("#shutdown-time input:nth-child(3)").val(), 10);
                var valid = h >= 0 && h < 24 && (m >= 0 && m < 60);
                $("#shutdown-time").toggleClass("has-error", !valid);
                if (!valid) disabled = true;
            }
            $("#shutdown-action").prop("disabled", disabled);
        },
        do_action: function(op) {
            var delay = shell.select_btn_selected(this.delay_btn);
            var message = $("#shutdown-message").val();
            var when;
            if (delay == "x") when = $("#shutdown-time input:nth-child(1)").val() + ":" + $("#shutdown-time input:nth-child(3)").val(); else when = "+" + delay;
            var arg = op == "shutdown" ? "--poweroff" : "--reboot";
            cockpit.spawn([ "shutdown", arg, when, message ], {
                superuser: true
            }).fail(function(ex) {
                $("#shutdown-dialog").modal("hide");
                shell.show_unexpected_error(ex);
            }).done(function(ex) {
                $("#shutdown-dialog").modal("hide");
            });
        },
        restart: function() {
            this.do_action("restart");
        },
        shutdown: function() {
            this.do_action("shutdown");
        }
    };
    function PageShutdownDialog() {
        this._init();
    }
    shell.dialogs.push(new PageShutdownDialog());
})(jQuery, cockpit, shell, modules);

var shell = shell || {};

(function($, cockpit, shell) {
    PageCpuStatus.prototype = {
        _init: function() {
            this.id = "cpu_status";
        },
        getTitle: function() {
            return C_("page-title", "CPU Status");
        },
        enter: function() {
            var self = this;
            this.client = shell.dbus(null);
            var resmon = this.client.get("/com/redhat/Cockpit/CpuMonitor", "com.redhat.Cockpit.ResourceMonitor");
            var options = {
                series: {
                    shadowSize: 0,
                    lines: {
                        lineWidth: 0,
                        fill: true
                    }
                },
                yaxis: {
                    min: 0,
                    max: 100,
                    show: true,
                    ticks: 5,
                    tickFormatter: function(v) {
                        return v + "%";
                    }
                },
                xaxis: {
                    show: true,
                    ticks: [ [ 0 * 60, "5 min" ], [ 1 * 60, "4 min" ], [ 2 * 60, "3 min" ], [ 3 * 60, "2 min" ], [ 4 * 60, "1 min" ] ]
                },
                x_rh_stack_graphs: true
            };
            this.plot = shell.setup_complicated_plot("#cpu_status_graph", resmon, [ {
                color: "rgb(200,200,200)"
            }, {
                color: "rgb(150,150,150)"
            }, {
                color: "rgb(100,100,100)"
            }, {
                color: "rgb( 50, 50, 50)"
            } ], options);
            shell.util.machine_info().done(function(info) {
                self.plot.set_yaxis_max(info.cpus * 100);
            });
        },
        show: function() {
            this.plot.start();
        },
        leave: function() {
            this.plot.destroy();
            this.client.release();
            this.client = null;
        }
    };
    function PageCpuStatus() {
        this._init();
    }
    shell.pages.push(new PageCpuStatus());
})(jQuery, cockpit, shell);

var shell = shell || {};

(function($, cockpit, shell) {
    PageMemoryStatus.prototype = {
        _init: function() {
            this.id = "memory_status";
        },
        getTitle: function() {
            return C_("page-title", "Memory");
        },
        enter: function() {
            this.client = shell.dbus(null);
            var resmon = this.client.get("/com/redhat/Cockpit/MemoryMonitor", "com.redhat.Cockpit.ResourceMonitor");
            var options = {
                series: {
                    shadowSize: 0,
                    lines: {
                        lineWidth: 0,
                        fill: true
                    }
                },
                yaxis: {
                    min: 0,
                    ticks: 5,
                    tickFormatter: function(v) {
                        return cockpit.format_bytes(v);
                    }
                },
                xaxis: {
                    show: true,
                    ticks: [ [ 0 * 60, "5 min" ], [ 1 * 60, "4 min" ], [ 2 * 60, "3 min" ], [ 3 * 60, "2 min" ], [ 4 * 60, "1 min" ] ]
                },
                x_rh_stack_graphs: true
            };
            this.plot = shell.setup_complicated_plot("#memory_status_graph", resmon, [ {
                color: "rgb(200,200,200)"
            }, {
                color: "rgb(150,150,150)"
            }, {
                color: "rgb(100,100,100)"
            }, {
                color: "rgb( 50, 50, 50)"
            } ], options);
        },
        show: function() {
            this.plot.start();
        },
        leave: function() {
            this.plot.destroy();
            this.client.release();
            this.client = null;
        }
    };
    function PageMemoryStatus() {
        this._init();
    }
    shell.pages.push(new PageMemoryStatus());
})(jQuery, cockpit, shell);

var shell = shell || {};

var modules = modules || {};

(function($, cockpit, shell, modules) {
    function nm_debug() {
        if (window.debugging == "all" || window.debugging == "nm") console.debug.apply(console, arguments);
    }
    function generate_uuid() {
        return "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g, function(c) {
            var r = Math.random() * 16 | 0, v = c === "x" ? r : r & 3 | 8;
            return v.toString(16);
        });
    }
    function NetworkManagerModel() {
        var self = this;
        var byteorder = null;
        var client = cockpit.dbus("org.freedesktop.NetworkManager");
        self.client = client;
        var objects = {};
        function complain() {
            console.warn.apply(console, arguments);
        }
        function conv_Object(type) {
            return function(path) {
                return get_object(path, type);
            };
        }
        function conv_Array(conv) {
            return function(elts) {
                return elts.map(conv);
            };
        }
        function priv(obj) {
            return obj[" priv"];
        }
        var outstanding_refreshes = 0;
        function push_refresh() {
            outstanding_refreshes += 1;
        }
        function pop_refresh() {
            outstanding_refreshes -= 1;
            if (outstanding_refreshes === 0) export_model();
        }
        function get_object(path, type) {
            if (path == "/") return null;
            function constructor() {
                this[" priv"] = {};
                priv(this).type = type;
                priv(this).path = path;
                for (var p in type.props) this[p] = type.props[p].def;
            }
            if (!objects[path]) {
                constructor.prototype = type.prototype;
                objects[path] = new constructor();
                if (type.refresh) type.refresh(objects[path]);
                if (type.exporters && type.exporters[0]) type.exporters[0](objects[path]);
            }
            return objects[path];
        }
        function peek_object(path) {
            return objects[path] || null;
        }
        function drop_object(path) {
            var obj = objects[path];
            if (obj) {
                if (priv(obj).type.drop) priv(obj).type.drop(obj);
                delete objects[path];
                export_model();
            }
        }
        function set_object_properties(obj, props) {
            var p, decl, val;
            decl = priv(obj).type.props;
            for (p in decl) {
                val = props[decl[p].prop || p];
                if (val !== undefined) {
                    if (decl[p].conv) val = decl[p].conv(val);
                    if (val !== obj[p]) {
                        obj[p] = val;
                        if (decl[p].trigger) decl[p].trigger(obj);
                    }
                }
            }
        }
        function remove_signatures(props_with_sigs) {
            var props = {};
            for (var p in props_with_sigs) {
                if (props_with_sigs.hasOwnProperty(p)) {
                    props[p] = props_with_sigs[p].v;
                }
            }
            return props;
        }
        function objpath(obj) {
            if (obj && priv(obj).path) return priv(obj).path; else return "/";
        }
        function call_object_method(obj, iface, method) {
            var dfd = new $.Deferred();
            client.call(objpath(obj), iface, method, Array.prototype.slice.call(arguments, 3)).fail(function(ex) {
                dfd.reject(ex);
            }).done(function(reply) {
                dfd.resolve.apply(dfd, reply);
            });
            return dfd.promise();
        }
        var interface_types = {};
        function set_object_types(all_types) {
            all_types.forEach(function(type) {
                if (type.exporters && type.exporters.length > max_export_phases) max_export_phases = type.exporters.length;
                type.interfaces.forEach(function(iface) {
                    interface_types[iface] = type;
                });
            });
        }
        function signal_emitted(path, iface, signal, args) {
            var obj = peek_object(path);
            if (obj) {
                var type = priv(obj).type;
                if (signal == "PropertiesChanged") {
                    push_refresh();
                    set_object_properties(obj, remove_signatures(args[0]));
                    pop_refresh();
                } else if (type.signals && type.signals[signal]) type.signals[signal](obj, args);
            }
        }
        function interface_properties(path, iface, props) {
            var type = interface_types[iface];
            if (type) set_object_properties(get_object(path, type), props);
        }
        function interface_removed(path, iface) {
            drop_object(path);
        }
        var max_export_phases = 0;
        var export_pending;
        function export_model() {
            function doit() {
                var phase, path, obj, exp;
                for (phase = 0; phase < max_export_phases; phase++) {
                    for (path in objects) {
                        obj = objects[path];
                        exp = priv(obj).type.exporters;
                        if (exp && exp[phase]) exp[phase](obj);
                    }
                }
                $(self).trigger("changed");
            }
            if (!export_pending) {
                export_pending = true;
                window.setTimeout(function() {
                    export_pending = false;
                    doit();
                }, 300);
            }
        }
        client.call("/org/freedesktop/NetworkManager", "org.freedesktop.DBus.Properties", "Get", [ "org.freedesktop.NetworkManager", "State" ], {
            flags: ""
        }).fail(complain).done(function(reply, options) {
            if (options.flags) {
                if (options.flags.indexOf(">") !== -1) byteorder = "be"; else if (options.flags.indexOf("<") !== -1) byteorder = "le";
            }
        });
        var subscription = client.subscribe({}, signal_emitted);
        var watch = client.watch({});
        $(client).on("notify", function(event, data) {
            $.each(data, function(path, ifaces) {
                $.each(ifaces, function(iface, props) {
                    if (props) interface_properties(path, iface, props); else interface_removed(path, iface);
                });
            });
        });
        self.close = function close() {
            subscription.remove();
            watch.remove();
            $(client).off("notify");
            client.close("unused");
        };
        function toDec(n) {
            return n.toString(10);
        }
        function bytes_from_nm32(num) {
            var bytes = [], i;
            if (byteorder == "be") {
                for (i = 3; i >= 0; i--) {
                    bytes[i] = num & 255;
                    num = num >>> 8;
                }
            } else {
                for (i = 0; i < 4; i++) {
                    bytes[i] = num & 255;
                    num = num >>> 8;
                }
            }
            return bytes;
        }
        function bytes_to_nm32(bytes) {
            var num = 0, i;
            if (byteorder == "be") {
                for (i = 0; i < 4; i++) {
                    num = 256 * num + bytes[i];
                }
            } else {
                for (i = 3; i >= 0; i--) {
                    num = 256 * num + bytes[i];
                }
            }
            return num;
        }
        function ip4_to_text(num) {
            return bytes_from_nm32(num).map(toDec).join(".");
        }
        function ip4_from_text(text) {
            var parts = text.split(".");
            if (parts.length == 4) return bytes_to_nm32(parts.map(function(s) {
                return parseInt(s, 10);
            })); else return 0;
        }
        var text_to_prefix_bits = {
            "255": 8,
            "254": 7,
            "252": 6,
            "248": 5,
            "240": 4,
            "224": 3,
            "192": 2,
            "128": 1,
            "0": 0
        };
        function ip4_prefix_from_text(text) {
            if (/^[0-9]+$/.test(text.trim())) return parseInt(text, 10);
            var parts = text.split(".");
            if (parts.length != 4) return -1;
            var prefix = 0;
            var i;
            for (i = 0; i < 4; i++) {
                var p = text_to_prefix_bits[parts[i].trim()];
                if (p !== undefined) {
                    prefix += p;
                    if (p < 8) break;
                } else return -1;
            }
            for (i += 1; i < 4; i++) {
                if (/^0+$/.test(parts[i].trim()) === false) return -1;
            }
            return prefix;
        }
        function ip4_address_from_nm(addr) {
            return [ ip4_to_text(addr[0]), addr[1].toString(), ip4_to_text(addr[2]) ];
        }
        function ip4_address_to_nm(addr) {
            return [ ip4_from_text(addr[0]), ip4_prefix_from_text(addr[1]), ip4_from_text(addr[2]) ];
        }
        function ip4_route_from_nm(addr) {
            return [ ip4_to_text(addr[0]), addr[1].toString(), ip4_to_text(addr[2]), addr[3].toString() ];
        }
        function ip4_route_to_nm(addr) {
            return [ ip4_from_text(addr[0]), ip4_prefix_from_text(addr[1]), ip4_from_text(addr[2]), parseInt(addr[3], 10) || 0 ];
        }
        function ip6_from_text(text) {
            var parts = text.split(":");
            var bytes = [];
            for (var i = 0; i < 8; i++) {
                var num = parseInt(parts[i], 16) || 0;
                bytes[2 * i] = num >> 8;
                bytes[2 * i + 1] = num & 255;
            }
            return cockpit.base64_encode(bytes);
        }
        function ip6_to_text(data) {
            var parts = [];
            var bytes = cockpit.base64_decode(data);
            for (var i = 0; i < 8; i++) parts[i] = ((bytes[2 * i] << 8) + bytes[2 * i + 1]).toString(16);
            return parts.join(":");
        }
        function ip6_address_from_nm(addr) {
            return [ ip6_to_text(addr[0]), addr[1].toString(), ip6_to_text(addr[2]) ];
        }
        function ip6_address_to_nm(addr) {
            return [ ip6_from_text(addr[0]), parseInt(addr[1], 10) || 64, ip6_from_text(addr[2]) ];
        }
        function ip6_route_from_nm(addr) {
            return [ ip6_to_text(addr[0]), addr[1].toString(), ip6_to_text(addr[2]), addr[3].toString() ];
        }
        function ip6_route_to_nm(addr) {
            return [ ip6_from_text(addr[0]), parseInt(addr[1], 10) || 64, ip6_from_text(addr[2]), parseInt(addr[3], 10) || 0 ];
        }
        function settings_from_nm(settings) {
            function get(first, second, def) {
                if (settings[first] && settings[first][second]) return settings[first][second].v; else return def;
            }
            function get_ip(first, addr_from_nm, route_from_nm, ip_to_text) {
                return {
                    method: get(first, "method", "auto"),
                    ignore_auto_dns: get(first, "ignore-auto-dns", false),
                    ignore_auto_routes: get(first, "ignore-auto-routes", false),
                    addresses: get(first, "addresses", []).map(addr_from_nm),
                    dns: get(first, "dns", []).map(ip_to_text),
                    dns_search: get(first, "dns-search", []),
                    routes: get(first, "routes", []).map(route_from_nm)
                };
            }
            var result = {
                connection: {
                    type: get("connection", "type"),
                    uuid: get("connection", "uuid"),
                    interface_name: get("connection", "interface-name"),
                    timestamp: get("connection", "timestamp"),
                    id: get("connection", "id", _("Unknown")),
                    autoconnect: get("connection", "autoconnect", true),
                    slave_type: get("connection", "slave-type"),
                    master: get("connection", "master")
                }
            };
            if (!settings.connection.master) {
                result.ipv4 = get_ip("ipv4", ip4_address_from_nm, ip4_route_from_nm, ip4_to_text);
                result.ipv6 = get_ip("ipv6", ip6_address_from_nm, ip6_route_from_nm, ip6_to_text);
            }
            if (settings.bond) {
                result.bond = {
                    options: $.extend({}, get("bond", "options", {})),
                    interface_name: get("bond", "interface-name")
                };
            }
            if (settings.bridge) {
                result.bridge = {
                    interface_name: get("bridge", "interface-name"),
                    stp: get("bridge", "stp", true),
                    priority: get("bridge", "priority", 32768),
                    forward_delay: get("bridge", "forward-delay", 15),
                    hello_time: get("bridge", "hello-time", 2),
                    max_age: get("bridge", "max-age", 20),
                    ageing_time: get("bridge", "ageing-time", 300)
                };
            }
            if (settings["bridge-port"] || result.connection.slave_type == "bridge") {
                result.bridge_port = {
                    priority: get("bridge-port", "priority", 32),
                    path_cost: get("bridge-port", "path-cost", 100),
                    hairpin_mode: get("bridge-port", "hairpin-mode", false)
                };
            }
            if (settings.vlan) {
                result.vlan = {
                    parent: get("vlan", "parent"),
                    id: get("vlan", "id"),
                    interface_name: get("vlan", "interface-name")
                };
            }
            return result;
        }
        function settings_to_nm(settings, orig) {
            var result = $.extend(true, {}, orig);
            function set(first, second, sig, val, def) {
                if (val === undefined) val = def;
                if (!result[first]) result[first] = {};
                if (val !== undefined) result[first][second] = cockpit.variant(sig, val); else delete result[first][second];
            }
            function set_ip(first, addrs_sig, addr_to_nm, routes_sig, route_to_nm, ips_sig, ip_from_text) {
                set(first, "method", "s", settings[first].method);
                set(first, "ignore-auto-dns", "b", settings[first].ignore_auto_dns);
                set(first, "ignore-auto-routes", "b", settings[first].ignore_auto_routes);
                set(first, "addresses", addrs_sig, settings[first].addresses.map(addr_to_nm));
                set(first, "dns", ips_sig, settings[first].dns.map(ip_from_text));
                set(first, "dns-search", "as", settings[first].dns_search);
                set(first, "routes", routes_sig, settings[first].routes.map(route_to_nm));
                delete result[first]["address-labels"];
            }
            set("connection", "id", "s", settings.connection.id);
            set("connection", "autoconnect", "b", settings.connection.autoconnect);
            set("connection", "uuid", "s", settings.connection.uuid);
            set("connection", "interface-name", "s", settings.connection.interface_name);
            set("connection", "type", "s", settings.connection.type);
            set("connection", "slave-type", "s", settings.connection.slave_type);
            set("connection", "master", "s", settings.connection.master);
            delete result.ipv4;
            if (settings.ipv4) set_ip("ipv4", "aau", ip4_address_to_nm, "aau", ip4_route_to_nm, "au", ip4_from_text);
            delete result.ipv6;
            if (settings.ipv6) set_ip("ipv6", "a(ayuay)", ip6_address_to_nm, "a(ayuayu)", ip6_route_to_nm, "aay", ip6_from_text);
            delete result.bond;
            if (settings.bond) {
                set("bond", "options", "a{ss}", settings.bond.options);
                set("bond", "interface-name", "s", settings.bond.interface_name);
            }
            delete result.bridge;
            if (settings.bridge) {
                set("bridge", "interface-name", "s", settings.bridge.interface_name);
                set("bridge", "stp", "b", settings.bridge.stp);
                set("bridge", "priority", "u", settings.bridge.priority);
                set("bridge", "forward-delay", "u", settings.bridge.forward_delay);
                set("bridge", "hello-time", "u", settings.bridge.hello_time);
                set("bridge", "max-age", "u", settings.bridge.max_age);
                set("bridge", "ageing-time", "u", settings.bridge.ageing_time);
            }
            delete result["bridge-port"];
            if (settings.bridge_port) {
                set("bridge-port", "priority", "u", settings.bridge_port.priority);
                set("bridge-port", "path-cost", "u", settings.bridge_port.path_cost);
                set("bridge-port", "hairpin-mode", "b", settings.bridge_port.hairpin_mode);
            }
            delete result.vlan;
            if (settings.vlan) {
                set("vlan", "parent", "s", settings.vlan.parent);
                set("vlan", "id", "u", settings.vlan.id);
                set("vlan", "interface-name", "s", settings.vlan.interface_name);
            }
            if (settings["802-3-ethernet"]) {
                if (!result["802-3-ethernet"]) result["802-3-ethernet"] = {};
            }
            return result;
        }
        function device_type_to_symbol(type) {
            switch (type) {
              case 0:
                return "unknown";

              case 1:
                return "ethernet";

              case 2:
                return "wifi";

              case 3:
                return "unused1";

              case 4:
                return "unused2";

              case 5:
                return "bt";

              case 6:
                return "olpc_mesh";

              case 7:
                return "wimax";

              case 8:
                return "modem";

              case 9:
                return "infiniband";

              case 10:
                return "bond";

              case 11:
                return "vlan";

              case 12:
                return "adsl";

              case 13:
                return "bridge";

              default:
                return "";
            }
        }
        function device_state_to_text(state) {
            switch (state) {
              case 0:
                return "?";

              case 10:
                return "";

              case 20:
                return _("Not available");

              case 30:
                return _("Inactive");

              case 40:
                return _("Preparing");

              case 50:
                return _("Configuring");

              case 60:
                return _("Authenticating");

              case 70:
                return _("Configuring IP");

              case 80:
                return _("Checking IP");

              case 90:
                return _("Waiting");

              case 100:
                return _("Active");

              case 110:
                return _("Deactivating");

              case 120:
                return _("Failed");

              default:
                return "";
            }
        }
        var connections_by_uuid = {};
        function set_settings(obj, settings) {
            if (obj.Settings && obj.Settings.connection && obj.Settings.connection.uuid) delete connections_by_uuid[obj.Settings.connection.uuid];
            obj.Settings = settings;
            if (obj.Settings && obj.Settings.connection && obj.Settings.connection.uuid) connections_by_uuid[obj.Settings.connection.uuid] = obj;
        }
        function refresh_settings(obj) {
            push_refresh();
            client.call(objpath(obj), "org.freedesktop.NetworkManager.Settings.Connection", "GetSettings").always(pop_refresh).fail(complain).done(function(reply) {
                var result = reply[0];
                if (result) {
                    priv(obj).orig = result;
                    if (!priv(obj).frozen) {
                        set_settings(obj, settings_from_nm(result));
                    }
                }
            });
        }
        function refresh_udev(obj) {
            if (!obj.Udi.startsWith("/sys/")) return;
            push_refresh();
            cockpit.spawn([ "udevadm", "info", obj.Udi ]).done(function(res) {
                var props = {};
                function snarf_prop(line, env, prop) {
                    var prefix = "E: " + env + "=";
                    if (line.startsWith(prefix)) {
                        props[prop] = line.substr(prefix.length);
                    }
                }
                res.split("\n").forEach(function(line) {
                    snarf_prop(line, "ID_MODEL_FROM_DATABASE", "IdModel");
                    snarf_prop(line, "ID_VENDOR_FROM_DATABASE", "IdVendor");
                });
                set_object_properties(obj, props);
            }).fail(function(ex) {
                console.warn(ex);
            }).always(pop_refresh);
        }
        function handle_updated(obj) {
            refresh_settings(obj);
        }
        var type_Ipv4Config = {
            interfaces: [ "org.freedesktop.NetworkManager.IP4Config" ],
            props: {
                Addresses: {
                    conv: conv_Array(ip4_address_from_nm),
                    def: []
                }
            }
        };
        var type_Ipv6Config = {
            interfaces: [ "org.freedesktop.NetworkManager.IP6Config" ],
            props: {
                Addresses: {
                    conv: conv_Array(ip6_address_from_nm),
                    def: []
                }
            }
        };
        var type_Connection = {
            interfaces: [ "org.freedesktop.NetworkManager.Settings.Connection" ],
            props: {
                Unsaved: {}
            },
            signals: {
                Updated: handle_updated
            },
            refresh: refresh_settings,
            drop: function(obj) {
                set_settings(obj, null);
            },
            prototype: {
                freeze: function() {
                    priv(this).frozen = true;
                },
                apply: function() {
                    var self = this;
                    return call_object_method(this, "org.freedesktop.NetworkManager.Settings.Connection", "Update", settings_to_nm(this.Settings, priv(this).orig)).done(function() {
                        priv(self).frozen = false;
                    });
                },
                reset: function() {
                    set_settings(this, settings_from_nm(priv(this).orig));
                    priv(this).frozen = false;
                    export_model();
                },
                activate: function(dev, specific_object) {
                    return call_object_method(get_object("/org/freedesktop/NetworkManager", type_Manager), "org.freedesktop.NetworkManager", "ActivateConnection", objpath(this), objpath(dev), objpath(specific_object));
                },
                delete_: function() {
                    return call_object_method(this, "org.freedesktop.NetworkManager.Settings.Connection", "Delete");
                }
            },
            exporters: [ function(obj) {
                obj.Slaves = [];
                obj.Interfaces = [];
            }, null, function(obj) {
                var master, iface;
                function check_con(con) {
                    if (con.Settings.connection.type == obj.Settings.connection.slave_type) {
                        obj.Masters.push(con);
                        con.Slaves.push(obj);
                    }
                }
                obj.Masters = [];
                if (obj.Settings && obj.Settings.connection && obj.Settings.connection.slave_type) {
                    master = connections_by_uuid[obj.Settings.connection.master];
                    if (master) {
                        obj.Masters.push(master);
                        master.Slaves.push(obj);
                    } else {
                        iface = peek_interface(obj.Settings.connection.master);
                        if (iface) {
                            if (iface.Device) iface.Device.AvailableConnections.forEach(check_con); else iface.Connections.forEach(check_con);
                        }
                    }
                }
            } ]
        };
        var type_ActiveConnection = {
            interfaces: [ "org.freedesktop.NetworkManager.Connection.Active" ],
            props: {
                Connection: {
                    conv: conv_Object(type_Connection)
                },
                Ip4Config: {
                    conv: conv_Object(type_Ipv4Config)
                },
                Ip6Config: {
                    conv: conv_Object(type_Ipv6Config)
                }
            },
            prototype: {
                deactivate: function() {
                    return call_object_method(get_object("/org/freedesktop/NetworkManager", type_Manager), "org.freedesktop.NetworkManager", "DeactivateConnection", objpath(this));
                }
            }
        };
        var type_Device = {
            interfaces: [ "org.freedesktop.NetworkManager.Device", "org.freedesktop.NetworkManager.Device.Wired", "org.freedesktop.NetworkManager.Device.Bond", "org.freedesktop.NetworkManager.Device.Bridge", "org.freedesktop.NetworkManager.Device.Vlan" ],
            props: {
                DeviceType: {
                    conv: device_type_to_symbol
                },
                Interface: {},
                StateText: {
                    prop: "State",
                    conv: device_state_to_text,
                    def: _("Unknown")
                },
                State: {},
                HwAddress: {},
                AvailableConnections: {
                    conv: conv_Array(conv_Object(type_Connection)),
                    def: []
                },
                ActiveConnection: {
                    conv: conv_Object(type_ActiveConnection)
                },
                Udi: {
                    trigger: refresh_udev
                },
                IdVendor: {
                    def: ""
                },
                IdModel: {
                    def: ""
                },
                Driver: {
                    def: ""
                },
                Carrier: {
                    def: true
                },
                Speed: {}
            },
            prototype: {
                activate: function(connection, specific_object) {
                    return call_object_method(get_object("/org/freedesktop/NetworkManager", type_Manager), "org.freedesktop.NetworkManager", "ActivateConnection", objpath(connection), objpath(this), objpath(specific_object));
                },
                disconnect: function() {
                    return call_object_method(this, "org.freedesktop.NetworkManager.Device", "Disconnect");
                }
            }
        };
        var type_Interface = {
            interfaces: [],
            exporters: [ function(obj) {
                obj.Device = null;
                obj.Connections = [];
            }, null, function(obj) {
                if (!obj.Device && obj.Connections.length === 0) {
                    drop_object(priv(obj).path);
                    return;
                }
                if (obj.Device) {
                    obj.Device.AvailableConnections.forEach(function(con) {
                        con.Interfaces.push(obj);
                    });
                } else {
                    obj.Connections.forEach(function(con) {
                        con.Interfaces.push(obj);
                    });
                }
            } ]
        };
        function get_interface(iface) {
            var obj = get_object(":interface:" + iface, type_Interface);
            obj.Name = iface;
            return obj;
        }
        function peek_interface(iface) {
            return peek_object(":interface:" + iface);
        }
        var type_Settings = {
            interfaces: [ "org.freedesktop.NetworkManager.Settings" ],
            props: {
                Connections: {
                    conv: conv_Array(conv_Object(type_Connection)),
                    def: []
                }
            },
            prototype: {
                add_connection: function(conf) {
                    var dfd = $.Deferred();
                    call_object_method(this, "org.freedesktop.NetworkManager.Settings", "AddConnection", settings_to_nm(conf, {})).done(function(path) {
                        dfd.resolve(get_object(path, type_Connection));
                    }).fail(function(error) {
                        dfd.reject(error);
                    });
                    return dfd.promise();
                }
            },
            exporters: [ null, function(obj) {
                if (obj.Connections) {
                    obj.Connections.forEach(function(con) {
                        function add_to_interface(name) {
                            get_interface(name).Connections.push(con);
                        }
                        if (con.Settings) {
                            if (con.Settings.bond) add_to_interface(con.Settings.bond.interface_name);
                            if (con.Settings.bridge) add_to_interface(con.Settings.bridge.interface_name);
                            if (con.Settings.vlan) add_to_interface(con.Settings.vlan.interface_name);
                        }
                    });
                }
            } ]
        };
        var type_Manager = {
            interfaces: [ "org.freedesktop.NetworkManager" ],
            props: {
                Devices: {
                    conv: conv_Array(conv_Object(type_Device)),
                    def: []
                },
                ActiveConnections: {
                    conv: conv_Array(conv_Object(type_ActiveConnection)),
                    def: []
                }
            },
            exporters: [ null, function(obj) {
                obj.Devices.forEach(function(dev) {
                    if (dev.Interface) {
                        var iface = get_interface(dev.Interface);
                        iface.Device = dev;
                    }
                });
            } ]
        };
        type_ActiveConnection.props.Master = {
            conv: conv_Object(type_Device)
        };
        type_Device.props.Slaves = {
            conv: conv_Array(conv_Object(type_Device)),
            def: []
        };
        self.list_interfaces = function list_interfaces() {
            var path, obj;
            var result = [];
            for (path in objects) {
                obj = objects[path];
                if (priv(obj).type === type_Interface) result.push(obj);
            }
            return result.sort(function(a, b) {
                return a.Name.localeCompare(b.Name);
            });
        };
        self.find_interface = peek_interface;
        self.get_manager = function() {
            return get_object("/org/freedesktop/NetworkManager", type_Manager);
        };
        self.get_settings = function() {
            return get_object("/org/freedesktop/NetworkManager/Settings", type_Settings);
        };
        set_object_types([ type_Manager, type_Device, type_Ipv4Config, type_Ipv6Config, type_Connection, type_ActiveConnection ]);
        get_object("/org/freedesktop/NetworkManager", type_Manager);
        get_object("/org/freedesktop/NetworkManager/Settings", type_Settings);
        return self;
    }
    var nm_models = shell.util.make_resource_cache();
    function get_nm_model() {
        return nm_models.get("default", function() {
            return new NetworkManagerModel();
        });
    }
    function render_interface_link(iface) {
        return $("<a>").text(iface).click(function() {
            cockpit.location.go("network-interface", {
                dev: iface
            });
        });
    }
    function device_state_text(dev) {
        if (!dev) return _("Inactive");
        if (dev.State == 100 && dev.Carrier === false) return _("No carrier");
        return dev.StateText;
    }
    function render_connection_link(con) {
        var res = $("<span>").append(array_join(con.Interfaces.map(function(iface) {
            return $("<a>").text(iface.Name).click(function() {
                cockpit.location.go("network-interface", {
                    dev: iface.Name
                });
            });
        }), ", "));
        return res;
    }
    function array_join(elts, sep) {
        var result = [];
        for (var i = 0; i < elts.length; i++) {
            result.push(elts[i]);
            if (i < elts.length - 1) result.push(sep);
        }
        return result;
    }
    function render_active_connection(dev, with_link, hide_link_local) {
        var parts = [];
        var con;
        if (!dev) return "";
        con = dev.ActiveConnection;
        if (con && con.Master) {
            return $("<span>").append($("<span>").text(_("Part of ")), with_link ? render_interface_link(con.Master.Interface) : con.Master.Interface);
        }
        if (con && con.Ip4Config) {
            con.Ip4Config.Addresses.forEach(function(a) {
                parts.push(a[0] + "/" + a[1]);
            });
        }
        function is_ipv6_link_local(addr) {
            return addr.startsWith("fe8") || addr.startsWith("fe9") || addr.startsWith("fea") || addr.startsWith("feb");
        }
        if (con && con.Ip6Config) {
            con.Ip6Config.Addresses.forEach(function(a) {
                if (!(hide_link_local && is_ipv6_link_local(a[0]))) parts.push(a[0] + "/" + a[1]);
            });
        }
        return $("<span>").text(parts.join(", "));
    }
    function network_plot_setup_hook(plot) {
        var axes = plot.getAxes();
        if (axes.yaxis.datamax < 1e5) axes.yaxis.options.max = 1e5; else axes.yaxis.options.max = null;
        axes.yaxis.options.min = 0;
    }
    function update_network_privileged() {
        shell.update_privileged_ui(shell.default_permission, ".network-privileged", cockpit.format(_("The user <b>$0</b> is not permitted to modify network settings"), cockpit.user.name));
    }
    $(shell.default_permission).on("changed", update_network_privileged);
    PageNetworking.prototype = {
        _init: function() {
            this.id = "networking";
        },
        getTitle: function() {
            return C_("page-title", "Networking");
        },
        setup: function() {
            update_network_privileged();
            $("#networking-add-bond").click($.proxy(this, "add_bond"));
            $("#networking-add-bridge").click($.proxy(this, "add_bridge"));
            $("#networking-add-vlan").click($.proxy(this, "add_vlan"));
        },
        enter: function() {
            var self = this;
            this.model = get_nm_model();
            this.ifaces = {};
            var blues = [ "#006bb4", "#008ff0", "#2daaff", "#69c2ff", "#a5daff", "#e1f3ff", "#00243c", "#004778" ];
            function is_interesting_netdev(netdev) {
                return netdev && self.ifaces[netdev];
            }
            function highlight_netdev_row(event, id) {
                $("#networking-interfaces tr").removeClass("highlight");
                if (id) {
                    $('#networking-interfaces tr[data-interface="' + shell.esc(id) + '"]').addClass("highlight");
                }
            }
            function render_samples(event, timestamp, samples) {
                for (var id in samples) {
                    var row = $('#networking-interfaces tr[data-sample-id="' + shell.esc(id) + '"]');
                    if (row.length > 0) {
                        row.find("td:nth-child(3)").text(cockpit.format_bits_per_sec(samples[id][1] * 8));
                        row.find("td:nth-child(4)").text(cockpit.format_bits_per_sec(samples[id][0] * 8));
                    }
                }
            }
            this.cockpitd = shell.dbus(null);
            this.monitor = this.cockpitd.get("/com/redhat/Cockpit/NetdevMonitor", "com.redhat.Cockpit.MultiResourceMonitor");
            $(this.monitor).on("NewSample.networking", render_samples);
            this.rx_plot = shell.setup_multi_plot("#networking-rx-graph", this.monitor, 0, blues.concat(blues), is_interesting_netdev, network_plot_setup_hook);
            $(this.rx_plot).on("update-total", function(event, total) {
                $("#networking-rx-text").text(cockpit.format_bits_per_sec(total * 8));
            });
            $(this.rx_plot).on("highlight", highlight_netdev_row);
            this.tx_plot = shell.setup_multi_plot("#networking-tx-graph", this.monitor, 1, blues.concat(blues), is_interesting_netdev, network_plot_setup_hook);
            $(this.tx_plot).on("update-total", function(event, total) {
                $("#networking-tx-text").text(cockpit.format_bits_per_sec(total * 8));
            });
            $(this.tx_plot).on("highlight", highlight_netdev_row);
            this.log_box = modules.server.logbox([ "_SYSTEMD_UNIT=NetworkManager.service", "_SYSTEMD_UNIT=firewalld.service" ], 10);
            $("#networking-log").empty().append(this.log_box);
            $(this.model).on("changed.networking", $.proxy(this, "update_devices"));
            this.update_devices();
        },
        show: function() {
            this.rx_plot.start();
            this.tx_plot.start();
        },
        leave: function() {
            this.rx_plot.destroy();
            this.tx_plot.destroy();
            if (this.log_box) this.log_box.stop();
            $(this.model).off(".networking");
            this.model.release();
            this.model = null;
            $(this.monitor).off(".networking");
            this.cockpitd.release();
            this.cockpitd = null;
        },
        update_devices: function() {
            var self = this;
            var tbody;
            var new_ifaces = {};
            var ifaces_changed = false;
            tbody = $("#networking-interfaces tbody");
            tbody.empty();
            self.model.list_interfaces().forEach(function(iface) {
                function has_master(iface) {
                    var connections = iface.Device ? iface.Device.AvailableConnections : iface.Connections;
                    return connections.some(function(c) {
                        return c.Masters.length > 0;
                    });
                }
                if (has_master(iface)) return;
                if (iface.Device && iface.Device.DeviceType != "ethernet" && iface.Device.DeviceType != "bond" && iface.Device.DeviceType != "vlan" && iface.Device.DeviceType != "bridge") return;
                var dev = iface.Device;
                var is_active = dev && dev.State == 100 && dev.Carrier === true;
                new_ifaces[iface.Name] = true;
                if (!self.ifaces[iface.Name]) ifaces_changed = true;
                tbody.append($("<tr>", {
                    "data-interface": iface.Name,
                    "data-sample-id": is_active ? iface.Name : null
                }).append($("<td>").text(iface.Name), $("<td>").html(render_active_connection(dev, false, true)), is_active ? [ $("<td>").text(""), $("<td>").text("") ] : $('<td colspan="2">').text(device_state_text(dev))).click(function() {
                    cockpit.location.go("network-interface", {
                        dev: iface.Name
                    });
                }));
            });
            if (!ifaces_changed) {
                for (var name in self.ifaces) {
                    if (!new_ifaces[name]) ifaces_changed = true;
                }
            }
            if (ifaces_changed) {
                self.ifaces = new_ifaces;
                $(self.monitor).trigger("notify:Consumers");
            }
        },
        add_bond: function() {
            var iface, i, uuid;
            uuid = generate_uuid();
            for (i = 0; i < 100; i++) {
                iface = "bond" + i;
                if (!this.model.find_interface(iface)) break;
            }
            PageNetworkBondSettings.model = this.model;
            PageNetworkBondSettings.done = null;
            PageNetworkBondSettings.connection = null;
            PageNetworkBondSettings.settings = {
                connection: {
                    id: uuid,
                    autoconnect: false,
                    type: "bond",
                    uuid: uuid,
                    interface_name: iface
                },
                bond: {
                    options: {
                        mode: "balance-rr"
                    },
                    interface_name: iface
                }
            };
            $("#network-bond-settings-dialog").modal("show");
        },
        add_bridge: function() {
            var iface, i, uuid;
            uuid = generate_uuid();
            for (i = 0; i < 100; i++) {
                iface = "bridge" + i;
                if (!this.model.find_interface(iface)) break;
            }
            PageNetworkBridgeSettings.model = this.model;
            PageNetworkBridgeSettings.done = null;
            PageNetworkBridgeSettings.connection = null;
            PageNetworkBridgeSettings.settings = {
                connection: {
                    id: uuid,
                    autoconnect: false,
                    type: "bridge",
                    uuid: uuid,
                    interface_name: iface
                },
                bridge: {
                    interface_name: iface,
                    stp: false,
                    priority: 32768,
                    forward_delay: 15,
                    hello_time: 2,
                    max_age: 20,
                    ageing_time: 300
                }
            };
            $("#network-bridge-settings-dialog").modal("show");
        },
        add_vlan: function() {
            var iface, i, uuid;
            uuid = generate_uuid();
            PageNetworkVlanSettings.model = this.model;
            PageNetworkVlanSettings.done = null;
            PageNetworkVlanSettings.connection = null;
            PageNetworkVlanSettings.settings = {
                connection: {
                    id: uuid,
                    autoconnect: false,
                    type: "vlan",
                    uuid: uuid,
                    interface_name: ""
                },
                vlan: {
                    interface_name: "",
                    parent: ""
                }
            };
            $("#network-vlan-settings-dialog").modal("show");
        }
    };
    function PageNetworking() {
        this._init();
    }
    shell.pages.push(new PageNetworking());
    var ipv4_method_choices = [ {
        choice: "auto",
        title: _("Automatic (DHCP)")
    }, {
        choice: "link-local",
        title: _("Link local")
    }, {
        choice: "manual",
        title: _("Manual")
    }, {
        choice: "shared",
        title: _("Shared")
    }, {
        choice: "disabled",
        title: _("Disabled")
    } ];
    var ipv6_method_choices = [ {
        choice: "auto",
        title: _("Automatic")
    }, {
        choice: "dhcp",
        title: _("Automatic (DHCP only)")
    }, {
        choice: "link-local",
        title: _("Link local")
    }, {
        choice: "manual",
        title: _("Manual")
    }, {
        choice: "ignore",
        title: _("Ignore")
    } ];
    var bond_mode_choices = [ {
        choice: "balance-rr",
        title: _("Round Robin")
    }, {
        choice: "active-backup",
        title: _("Active Backup")
    }, {
        choice: "balance-xor",
        title: _("XOR")
    }, {
        choice: "broadcast",
        title: _("Broadcast")
    }, {
        choice: "802.3ad",
        title: _("802.3ad")
    }, {
        choice: "balance-tlb",
        title: _("Adaptive transmit load balancing")
    }, {
        choice: "balance-alb",
        title: _("Adaptive load balancing")
    } ];
    var bond_monitoring_choices = [ {
        choice: "mii",
        title: _("MII (Recommended)")
    }, {
        choice: "arp",
        title: _("ARP")
    } ];
    function choice_title(choices, choice, def) {
        for (var i = 0; i < choices.length; i++) {
            if (choices[i].choice == choice) return choices[i].title;
        }
        return def;
    }
    PageNetworkInterface.prototype = {
        _init: function() {
            this.id = "network-interface";
            this.section_id = "networking";
        },
        getTitle: function() {
            return C_("page-title", "Networking");
        },
        setup: function() {
            var self = this;
            $("#network-interface-delete").click($.proxy(this, "delete_connections"));
            $("#network-interface-delete").parent().append(this.device_onoff = shell.OnOff(false, $.proxy(this, "connect"), $.proxy(this, "disconnect"), null, "network-privileged"));
        },
        enter: function() {
            var self = this;
            self.model = get_nm_model();
            $(self.model).on("changed.network-interface", $.proxy(self, "update"));
            self.dev_name = shell.get_page_param("dev");
            $("#network-interface .breadcrumb .active").text(self.dev_name);
            var blues = [ "#006bb4", "#008ff0", "#2daaff", "#69c2ff", "#a5daff", "#e1f3ff", "#00243c", "#004778" ];
            self.graph_ifaces = {};
            function is_interesting_netdev(netdev) {
                return netdev && self.graph_ifaces[netdev];
            }
            function highlight_netdev_row(event, id) {
                $("#network-interface-slaves tr").removeClass("highlight");
                if (id) {
                    $('#network-interface-slaves tr[data-interface="' + shell.esc(id) + '"]').addClass("highlight");
                }
            }
            function render_samples(event, timestamp, samples) {
                for (var id in samples) {
                    var row = $('#network-interface-slaves tr[data-sample-id="' + shell.esc(id) + '"]');
                    if (row.length > 0) {
                        row.find("td:nth-child(2)").text(cockpit.format_bits_per_sec(samples[id][1] * 8));
                        row.find("td:nth-child(3)").text(cockpit.format_bits_per_sec(samples[id][0] * 8));
                    }
                }
            }
            this.cockpitd = shell.dbus(null);
            this.monitor = this.cockpitd.get("/com/redhat/Cockpit/NetdevMonitor", "com.redhat.Cockpit.MultiResourceMonitor");
            $(this.monitor).on("NewSample.networking", render_samples);
            this.rx_plot = shell.setup_multi_plot("#network-interface-rx-graph", this.monitor, 0, blues.concat(blues), is_interesting_netdev, network_plot_setup_hook);
            $(this.rx_plot).on("update-total", function(event, total) {
                $("#network-interface-rx-text").text(cockpit.format_bits_per_sec(total * 8));
            });
            $(this.rx_plot).on("highlight", highlight_netdev_row);
            this.tx_plot = shell.setup_multi_plot("#network-interface-tx-graph", this.monitor, 1, blues.concat(blues), is_interesting_netdev, network_plot_setup_hook);
            $(this.tx_plot).on("update-total", function(event, total) {
                $("#network-interface-tx-text").text(cockpit.format_bits_per_sec(total * 8));
            });
            $(this.tx_plot).on("highlight", highlight_netdev_row);
            $("#network-interface-delete").hide();
            self.dev = null;
            self.update();
        },
        show: function() {
            this.rx_plot.start();
            this.tx_plot.start();
        },
        leave: function() {
            this.rx_plot.destroy();
            this.tx_plot.destroy();
            $(this.model).off(".network-interface");
            this.model.release();
            this.model = null;
            this.dev = null;
            this.cockpitd.release();
            this.cockpitd = null;
        },
        delete_connections: function() {
            var self = this;
            function delete_connection_and_slaves(con) {
                return $.when(con.delete_(), $.when.apply($, con.Slaves.map(function(s) {
                    return free_slave_connection(s);
                })));
            }
            function delete_connections(cons) {
                return $.when.apply($, cons.map(delete_connection_and_slaves));
            }
            function delete_iface_connections(iface) {
                if (iface.Device) return delete_connections(iface.Device.AvailableConnections); else return delete_connections(iface.Connections);
            }
            if (this.iface) {
                var location = cockpit.location;
                delete_iface_connections(this.iface).done(function() {
                    location.go("networking");
                }).fail(shell.show_unexpected_error);
            }
        },
        connect: function() {
            var self = this;
            var settings_manager = self.model.get_settings();
            function fail(error) {
                shell.show_unexpected_error(error);
                self.update();
            }
            function activate(con) {
                con.activate(self.dev, null).fail(fail);
            }
            if (self.ghost_settings) {
                settings_manager.add_connection(self.ghost_settings).done(activate).fail(fail);
            } else if (self.main_connection) {
                activate(self.main_connection);
            } else self.update();
        },
        disconnect: function() {
            var self = this;
            if (!self.dev) {
                self.update();
                return;
            }
            self.dev.disconnect().fail(function(error) {
                shell.show_unexpected_error(error);
                self.update();
            });
        },
        update: function() {
            var self = this;
            var iface = self.model.find_interface(self.dev_name);
            var dev = iface && iface.Device;
            self.iface = iface;
            self.dev = dev;
            var desc;
            if (dev) {
                if (dev.DeviceType == "ethernet") {
                    desc = cockpit.format("$IdVendor $IdModel $Driver)", dev);
                } else if (dev.DeviceType == "bond") {
                    desc = _("Bond");
                } else if (dev.DeviceType == "vlan") {
                    desc = _("VLAN");
                } else if (dev.DeviceType == "bridge") {
                    desc = _("Bridge");
                }
            } else if (iface) {
                if (iface.Connections[0] && iface.Connections[0].Settings.connection.type == "bond") desc = _("Bond"); else if (iface.Connections[0] && iface.Connections[0].Settings.connection.type == "vlan") desc = _("VLAN"); else if (iface.Connections[0] && iface.Connections[0].Settings.connection.type == "bridge") desc = _("Bridge"); else desc = _("Unknown");
            } else desc = _("Unknown");
            $("#network-interface-name").text(self.dev_name);
            $("#network-interface-hw").text(desc);
            $("#network-interface-mac").text(dev ? dev.HwAddress : "");
            this.device_onoff.prop("disabled", !dev);
            this.device_onoff.set(!!(dev && dev.ActiveConnection));
            $("#network-interface-disconnect").prop("disabled", !dev || !dev.ActiveConnection);
            var is_deletable = iface && !dev || dev && (dev.DeviceType == "bond" || dev.DeviceType == "vlan" || dev.DeviceType == "bridge");
            $("#network-interface-delete").toggle(!!is_deletable);
            function render_carrier_status_row() {
                if (dev && dev.Carrier !== undefined) {
                    return $("<tr>").append($("<td>").text(_("Carrier")), $("<td>").append(dev.Carrier ? dev.Speed ? cockpit.format_bits_per_sec(dev.Speed * 1e6) : _("Yes") : _("No")));
                } else return null;
            }
            function render_active_status_row() {
                var state;
                if (self.main_connection && self.main_connection.Masters.length > 0) return null;
                if (!dev) state = _("Inactive"); else if (dev.State != 100) state = dev.StateText; else state = null;
                return $("<tr>").append($("<td>").text(_("Status")), $("<td>").append(render_active_connection(dev, true, false), " ", state ? $("<span>").text(state) : null));
            }
            function render_connection_settings_rows(con, settings) {
                if (!settings) return [];
                function apply() {
                    if (con) con.apply().fail(shell.show_unexpected_error); else {
                        var settings_manager = self.model.get_settings();
                        settings_manager.add_connection(settings).fail(shell.show_unexpected_error);
                    }
                }
                function reactivate_connection() {
                    if (con && dev && dev.ActiveConnection && dev.ActiveConnection.Connection === con) {
                        con.activate(dev, null).fail(shell.show_unexpected_error);
                    }
                }
                function render_ip_settings(topic) {
                    var params = settings[topic];
                    var parts = [];
                    if (params.method != "manual") parts.push(choice_title(topic == "ipv4" ? ipv4_method_choices : ipv6_method_choices, params.method, _("Unknown configuration")));
                    var addr_is_extra = params.method != "manual";
                    var addrs = [];
                    params.addresses.forEach(function(a) {
                        var addr = a[0] + "/" + a[1];
                        if (a[2] && a[2] != "0.0.0.0" && a[2] != "0:0:0:0:0:0:0:0") addr += " via " + a[2];
                        addrs.push(addr);
                    });
                    if (addrs.length > 0) parts.push(cockpit.format(addr_is_extra ? "Additional address $val" : "Address $val", {
                        val: addrs.join(", ")
                    }));
                    var dns_is_extra = !params["ignore-auto-dns"] && params.method != "manual";
                    if (params.dns.length > 0) parts.push(cockpit.format(dns_is_extra ? "Additional DNS $val" : "DNS $val", {
                        val: params.dns.join(", ")
                    }));
                    if (params.dns_search.length > 0) parts.push(cockpit.format(dns_is_extra ? "Additional DNS Search Domains $val" : "DNS Search Domains $val", {
                        val: params.dns_search.join(", ")
                    }));
                    return parts.map(function(p) {
                        return $("<div>").text(p);
                    });
                }
                function configure_ip_settings(topic) {
                    PageNetworkIpSettings.model = self.model;
                    PageNetworkIpSettings.connection = con;
                    PageNetworkIpSettings.settings = $.extend({}, settings);
                    PageNetworkIpSettings.topic = topic;
                    PageNetworkIpSettings.done = reactivate_connection;
                    $("#network-ip-settings-dialog").modal("show");
                }
                function configure_bond_settings() {
                    PageNetworkBondSettings.model = self.model;
                    PageNetworkBondSettings.connection = con;
                    PageNetworkBondSettings.settings = settings;
                    PageNetworkBondSettings.done = reactivate_connection;
                    $("#network-bond-settings-dialog").modal("show");
                }
                function configure_bridge_settings() {
                    PageNetworkBridgeSettings.model = self.model;
                    PageNetworkBridgeSettings.connection = con;
                    PageNetworkBridgeSettings.settings = con.Settings;
                    PageNetworkBridgeSettings.done = reactivate_connection;
                    $("#network-bridge-settings-dialog").modal("show");
                }
                function configure_bridge_port_settings() {
                    PageNetworkBridgePortSettings.model = self.model;
                    PageNetworkBridgePortSettings.connection = con;
                    PageNetworkBridgePortSettings.settings = con.Settings;
                    PageNetworkBridgePortSettings.done = reactivate_connection;
                    $("#network-bridgeport-settings-dialog").modal("show");
                }
                function configure_vlan_settings() {
                    PageNetworkVlanSettings.model = self.model;
                    PageNetworkVlanSettings.connection = con;
                    PageNetworkVlanSettings.settings = con.Settings;
                    PageNetworkVlanSettings.done = reactivate_connection;
                    $("#network-vlan-settings-dialog").modal("show");
                }
                function render_settings_row(title, rows, configure) {
                    return $("<tr>").append($("<td>").text(title).css("vertical-align", rows.length > 1 ? "top" : "center"), $("<td>").append(rows), $('<td style="text-align:right;vertical-align:top">').append($('<button class="btn btn-default network-privileged">').text(_("Configure")).click(function() {
                        configure();
                    })));
                }
                function render_ip_settings_row(topic, title) {
                    if (!settings[topic]) return null;
                    return render_settings_row(title, render_ip_settings(topic), function() {
                        configure_ip_settings(topic);
                    });
                }
                function render_master() {
                    if (con && con.Masters.length > 0) {
                        return $("<tr>").append($("<td>").text(_("Master")), $("<td>").append(array_join(con.Masters.map(render_connection_link), ", ")));
                    } else return null;
                }
                function render_bond_settings_row() {
                    var parts = [];
                    var rows = [];
                    var options;
                    if (!settings.bond) return null;
                    options = settings.bond.options;
                    parts.push(choice_title(bond_mode_choices, options.mode, options.mode));
                    if (options.arp_interval) parts.push(_("ARP Monitoring"));
                    if (parts.length > 0) rows.push($("<div>").text(parts.join(", ")));
                    return render_settings_row(_("Bond"), rows, configure_bond_settings);
                }
                function render_bridge_settings_row() {
                    var rows = [];
                    var options = settings.bridge;
                    if (!options) return null;
                    function add_row(fmt, args) {
                        rows.push($("<div>").text(cockpit.format(fmt, args)));
                    }
                    if (options.stp) {
                        add_row(_("Spanning Tree Protocol"));
                        if (options.priority != 32768) add_row(_("Priority $priority"), options);
                        if (options.forward_delay != 15) add_row(_("Forward delay $forward_delay"), options);
                        if (options.hello_time != 2) add_row(_("Hello time $hello_time"), options);
                        if (options.max_age != 20) add_row(_("Maximum message age $max_age"), options);
                    }
                    return render_settings_row(_("Bridge"), rows, configure_bridge_settings);
                }
                function render_bridge_port_settings_row() {
                    var rows = [];
                    var options = settings.bridge_port;
                    if (!options) return null;
                    function add_row(fmt, args) {
                        rows.push($("<div>").text(cockpit.format(fmt, args)));
                    }
                    if (options.priority != 32) add_row(_("Priority $priority"), options);
                    if (options.path_cost != 100) add_row(_("Path cost $path_cost"), options);
                    if (options.hairpin_mode) add_row(_("Hairpin mode"));
                    return render_settings_row(_("Bridge port"), rows, configure_bridge_port_settings);
                }
                function render_vlan_settings_row() {
                    var rows = [];
                    var options = settings.vlan;
                    if (!options) return null;
                    function add_row(fmt, args) {
                        rows.push($("<div>").text(cockpit.format(fmt, args)));
                    }
                    add_row(_("Parent $parent"), options);
                    add_row(_("Id $id"), options);
                    return render_settings_row(_("VLAN"), rows, configure_vlan_settings);
                }
                return [ render_master(), $("<tr>").append($("<td>").text(_("General")), $("<td>").append($('<label style="font-weight:inherit">').append($('<input type="checkbox" style="margin-left:0px">').prop("checked", settings.connection.autoconnect).change(function() {
                    settings.connection.autoconnect = $(this).prop("checked");
                    apply();
                }), _("Connect automatically")))), render_ip_settings_row("ipv4", _("IPv4")), render_ip_settings_row("ipv6", _("IPv6")), render_vlan_settings_row(), render_bridge_settings_row(), render_bridge_port_settings_row(), render_bond_settings_row() ];
            }
            function create_ghost_connection_settings() {
                var uuid = generate_uuid();
                return {
                    connection: {
                        id: uuid,
                        uuid: uuid,
                        autoconnect: false,
                        type: "802-3-ethernet",
                        interface_name: iface.Name
                    },
                    "802-3-ethernet": {},
                    ipv4: {
                        method: "auto",
                        addresses: [],
                        dns: [],
                        dns_search: [],
                        routes: []
                    },
                    ipv6: {
                        method: "auto",
                        addresses: [],
                        dns: [],
                        dns_search: [],
                        routes: []
                    }
                };
            }
            self.ghost_settings = null;
            self.main_connection = null;
            self.connection_settings = null;
            function find_main_connection(cons) {
                cons.forEach(function(c) {
                    if (!self.main_connection || self.main_connection.Settings.connection.timestamp < c.Settings.connection.timestamp) self.main_connection = c;
                });
                if (self.main_connection) {
                    self.connection_settings = self.main_connection.Settings;
                } else {
                    self.ghost_settings = create_ghost_connection_settings();
                    self.connection_settings = self.ghost_settings;
                }
            }
            if (iface) {
                if (iface.Device) find_main_connection(iface.Device.AvailableConnections); else find_main_connection(iface.Connections);
            }
            $("#network-interface-settings").empty().append(render_active_status_row()).append(render_carrier_status_row()).append(render_connection_settings_rows(self.main_connection, self.connection_settings));
            update_network_privileged();
            function update_connection_slaves(con) {
                var tbody = $("#network-interface-slaves tbody");
                var rows = {};
                self.graph_ifaces = {};
                tbody.empty();
                if (!con || con.Settings.connection.type != "bond" && con.Settings.connection.type != "bridge") {
                    self.graph_ifaces[self.dev_name] = true;
                    $(self.monitor).trigger("notify:Consumers");
                    $("#network-interface-slaves").hide();
                    return;
                }
                $("#network-interface-slaves thead th:first-child").text(con.Settings.connection.type == "bond" ? _("Members") : _("Ports"));
                con.Slaves.forEach(function(slave_con) {
                    slave_con.Interfaces.forEach(function(iface) {
                        var dev = iface.Device;
                        var is_active = dev && dev.State == 100 && dev.Carrier === true;
                        self.graph_ifaces[iface.Name] = true;
                        rows[iface.Name] = $("<tr>", {
                            "data-interface": iface.Name,
                            "data-sample-id": is_active ? iface.Name : null
                        }).append($("<td>").text(iface.Name), is_active ? [ $("<td>").text(""), $("<td>").text("") ] : $('<td colspan="2">').text(device_state_text(dev)), $('<td style="text-align:right">').append(shell.OnOff(is_active, function() {
                            slave_con.activate(iface.Device).fail(shell.show_unexpected_error);
                        }, function() {
                            if (dev) {
                                dev.disconnect().fail(shell.show_unexpected_error);
                            }
                        }, null, "network-privileged")), $('<td width="28px">').append($('<button class="btn btn-default btn-control network-privileged">').text("-").click(function() {
                            slave_con.delete_().fail(shell.show_unexpected_error);
                            return false;
                        }))).click(function() {
                            cockpit.location.go("network-interface", {
                                dev: iface.Name
                            });
                        });
                    });
                });
                Object.keys(rows).sort().forEach(function(name) {
                    tbody.append(rows[name]);
                });
                var add_btn = $("<div>", {
                    "class": "dropdown"
                }).append($("<button>", {
                    "class": "network-privileged btn btn-default btn-control dropdown-toggle",
                    "data-toggle": "dropdown"
                }).text("+"), $("<ul>", {
                    "class": "dropdown-menu",
                    style: "right:0px;left:auto;min-width:0;text-align:left",
                    role: "menu"
                }).append(self.model.list_interfaces().map(function(iface) {
                    if (is_interesting_interface(iface) && !self.graph_ifaces[iface.Name] && iface != self.iface) {
                        return $('<li role="presentation">').append($('<a role="menuitem" class="network-privileged">').text(iface.Name).click(function() {
                            set_slave(self.model, con, con.Settings, con.Settings.connection.type, iface.Name, true).fail(shell.show_unexpected_error);
                        }));
                    }
                    return null;
                })));
                $("#network-interface-slaves thead th:nth-child(5)").html(add_btn);
                $(self.monitor).trigger("notify:Consumers");
                $("#network-interface-slaves").show();
                update_network_privileged();
            }
            update_connection_slaves(self.main_connection);
        }
    };
    function PageNetworkInterface() {
        this._init();
    }
    shell.pages.push(new PageNetworkInterface());
    PageNetworkIpSettings.prototype = {
        _init: function() {
            this.id = "network-ip-settings-dialog";
        },
        setup: function() {
            $("#network-ip-settings-cancel").click($.proxy(this, "cancel"));
            $("#network-ip-settings-apply").click($.proxy(this, "apply"));
        },
        enter: function() {
            $("#network-ip-settings-error").text("");
            if (PageNetworkIpSettings.connection) PageNetworkIpSettings.connection.freeze();
            this.update();
        },
        show: function() {},
        leave: function() {},
        update: function() {
            var self = this;
            var con = PageNetworkIpSettings.connection;
            var settings = PageNetworkIpSettings.settings;
            var topic = PageNetworkIpSettings.topic;
            var params = settings[topic];
            var method_btn, addresses_table;
            var auto_dns_btn, dns_table;
            var auto_dns_search_btn, dns_search_table;
            var auto_routes_btn, routes_table;
            function choicebox(p, choices) {
                var btn = shell.select_btn(function(choice) {
                    params[p] = choice;
                    self.update();
                }, choices);
                shell.select_btn_select(btn, params[p]);
                return btn;
            }
            function inverted_switchbox(title, p) {
                var onoff;
                var btn = $("<span>").append($('<span style="margin-right:10px">').text(title), onoff = shell.OnOff(!params[p], function(val) {
                    params[p] = !val;
                    self.update();
                }));
                btn.enable = function enable(val) {
                    onoff.enable(val);
                };
                return btn;
            }
            function tablebox(title, p, columns, def, header_buttons) {
                var direct = false;
                var add_btn;
                if (typeof columns == "string") {
                    direct = true;
                    columns = [ columns ];
                }
                function get(i, j) {
                    if (direct) return params[p][i]; else return params[p][i][j];
                }
                function set(i, j, val) {
                    if (direct) params[p][i] = val; else params[p][i][j] = val;
                }
                function add() {
                    return function() {
                        params[p].push(def);
                        self.update();
                    };
                }
                function remove(index) {
                    return function() {
                        params[p].splice(index, 1);
                        self.update();
                    };
                }
                var panel = $('<div class="network-ip-settings-row">').append($("<div>").append($('<span style="font-weight:bold">').text(title), $('<div style="float:right">').append(header_buttons, add_btn = $('<button class="btn btn-default" style="width:2em">').text("+").css("margin-left", "10px").click(add()))), $('<table width="100%">').append(params[p].map(function(a, i) {
                    return $("<tr>").append(columns.map(function(c, j) {
                        return $("<td>").append($('<input class="form-control">').val(get(i, j)).attr("placeholder", c).change(function(event) {
                            set(i, j, $(event.target).val());
                        }));
                    }), $('<td style="text-align:right; padding-right: 0;">').append($('<button class="btn btn-default" style="width:2em">').text(_("-")).click(remove(i))));
                })));
                panel.enable_add = function enable_add(val) {
                    add_btn.prop("disabled", !val);
                };
                return panel;
            }
            function render_ip_settings() {
                var prefix_text = topic == "ipv4" ? _("Prefix length or Netmask") : _("Prefix length");
                var body = $("<div>").append(addresses_table = tablebox(_("Addresses"), "addresses", [ "Address", prefix_text, "Gateway" ], [ "", "", "" ], method_btn = choicebox("method", topic == "ipv4" ? ipv4_method_choices : ipv6_method_choices).css("display", "inline-block")), $("<br>"), dns_table = tablebox(_("DNS"), "dns", "Server", "", auto_dns_btn = inverted_switchbox(_("Automatic"), "ignore_auto_dns")), $("<br>"), dns_search_table = tablebox(_("DNS Search Domains"), "dns_search", "Search Domain", "", auto_dns_search_btn = inverted_switchbox(_("Automatic"), "ignore_auto_dns")), $("<br>"), routes_table = tablebox(_("Routes"), "routes", [ "Address", prefix_text, "Gateway", "Metric" ], [ "", "", "", "" ], auto_routes_btn = inverted_switchbox(_("Automatic"), "ignore_auto_routes")));
                return body;
            }
            if (params.method == "manual" && params.addresses.length === 0) params.addresses = [ [ "", "", "" ] ];
            var is_off = params.method == "disabled" || params.method == "ignore";
            var can_have_extra = !(params.method == "link-local" || params.method == "shared" || is_off);
            if (!can_have_extra) {
                params.addresses = [];
                params.dns = [];
                params.dns_search = [];
            }
            if (is_off) {
                params.routes = [];
            }
            $("#network-ip-settings-dialog .modal-title").text(topic == "ipv4" ? _("IPv4 Settings") : _("IPv6 Settings"));
            $("#network-ip-settings-body").html(render_ip_settings());
            var can_auto = params.method == "auto" || params.method == "dhcp";
            auto_dns_btn.enable(can_auto);
            auto_dns_search_btn.enable(can_auto);
            auto_routes_btn.enable(can_auto);
            addresses_table.enable_add(can_have_extra);
            dns_table.enable_add(can_have_extra);
            dns_search_table.enable_add(can_have_extra);
            routes_table.enable_add(!is_off);
        },
        cancel: function() {
            if (PageNetworkIpSettings.connection) PageNetworkIpSettings.connection.reset();
            $("#network-ip-settings-dialog").modal("hide");
        },
        apply: function() {
            function apply_or_create() {
                if (PageNetworkIpSettings.connection) return PageNetworkIpSettings.connection.apply(); else {
                    var settings_manager = PageNetworkIpSettings.model.get_settings();
                    return settings_manager.add_connection(PageNetworkIpSettings.settings);
                }
            }
            apply_or_create().done(function() {
                $("#network-ip-settings-dialog").modal("hide");
                if (PageNetworkIpSettings.done) PageNetworkIpSettings.done();
            }).fail(function(error) {
                $("#network-ip-settings-error").text(error.message || error.toString());
            });
        }
    };
    function PageNetworkIpSettings() {
        this._init();
    }
    shell.dialogs.push(new PageNetworkIpSettings());
    function is_interface_connection(iface, connection) {
        return connection && connection.Interfaces.indexOf(iface) != -1;
    }
    function is_interesting_interface(iface) {
        return !iface.Device || iface.Device.DeviceType == "ethernet" || iface.Device.DeviceType == "bond" || iface.Device.DeviceType == "vlan" || iface.Device.DeviceType == "bridge";
    }
    function slave_connection_for_interface(master, iface) {
        return master && master.Slaves.find(function(s) {
            return is_interface_connection(iface, s);
        });
    }
    function slave_interface_choices(model, master) {
        return model.list_interfaces().filter(function(iface) {
            return !is_interface_connection(iface, master) && is_interesting_interface(iface);
        });
    }
    function render_slave_interface_choices(model, master) {
        return $('<ul class="list-group available-interfaces-group">').append(slave_interface_choices(model, master).map(function(iface) {
            return $('<li class="list-group-item">').append($('<div class="checkbox">').css("margin", "0px").append($("<label>").append($("<input>", {
                type: "checkbox",
                "data-iface": iface.Name
            }).prop("checked", !!slave_connection_for_interface(master, iface)), $("<span>").text(iface.Name))));
        }));
    }
    function slave_chooser_btn(change, slave_choices) {
        var choices = [ {
            title: "-",
            choice: "",
            is_default: true
        } ];
        slave_choices.find("input[data-iface]").each(function(i, elt) {
            var name = $(elt).attr("data-iface");
            if ($(elt).prop("checked")) choices.push({
                title: name,
                choice: name
            });
        });
        return shell.select_btn(change, choices);
    }
    function free_slave_connection(con) {
        if (con.Settings.connection.slave_type) {
            delete con.Settings.connection.slave_type;
            delete con.Settings.connection.master;
            return con.apply();
        }
    }
    function set_slave(model, master_connection, master_settings, slave_type, iface_name, val) {
        var iface, uuid;
        var main_connection;
        iface = model.find_interface(iface_name);
        if (!iface) return false;
        function find_main_connection(cons) {
            cons.forEach(function(c) {
                if (!main_connection || main_connection.Settings.connection.timestamp < c.Settings.connection.timestamp) main_connection = c;
            });
        }
        if (iface.Device) find_main_connection(iface.Device.AvailableConnections); else find_main_connection(iface.Connections);
        if (val) {
            if (!main_connection) {
                uuid = generate_uuid();
                return model.get_settings().add_connection({
                    connection: {
                        id: uuid,
                        uuid: uuid,
                        autoconnect: true,
                        type: "802-3-ethernet",
                        interface_name: iface.Name,
                        slave_type: slave_type,
                        master: master_settings.connection.uuid
                    },
                    "802-3-ethernet": {}
                });
            } else if (main_connection.Settings.connection.master != master_settings.connection.uuid) {
                main_connection.Settings.connection.slave_type = slave_type;
                main_connection.Settings.connection.master = master_settings.connection.uuid;
                delete main_connection.Settings.ipv4;
                delete main_connection.Settings.ipv6;
                return main_connection.apply().then(function() {
                    var dev = iface.Device;
                    if (dev && dev.ActiveConnection && dev.ActiveConnection.Connection === main_connection) return main_connection.activate(dev, null);
                });
            }
        } else {
            if (main_connection && main_connection.Settings.connection.master == master_settings.connection.uuid) {
                free_slave_connection(main_connection);
            }
        }
        return true;
    }
    function apply_master_slave(choices, model, master_connection, master_settings, slave_type) {
        var settings_manager = model.get_settings();
        function set_all_slaves() {
            var deferreds = choices.find("input[data-iface]").map(function(i, elt) {
                return set_slave(model, master_connection, master_settings, slave_type, $(elt).attr("data-iface"), $(elt).prop("checked"));
            });
            return $.when.apply($, deferreds.get());
        }
        function update_master() {
            if (master_connection) return master_connection.apply(); else return settings_manager.add_connection(master_settings);
        }
        return update_master().then(set_all_slaves);
    }
    PageNetworkBondSettings.prototype = {
        _init: function() {
            this.id = "network-bond-settings-dialog";
        },
        setup: function() {
            $("#network-bond-settings-cancel").click($.proxy(this, "cancel"));
            $("#network-bond-settings-apply").click($.proxy(this, "apply"));
        },
        enter: function() {
            $("#network-bond-settings-error").text("");
            if (PageNetworkBondSettings.connection) PageNetworkBondSettings.connection.freeze();
            this.update();
        },
        show: function() {},
        leave: function() {},
        find_slave_con: function(iface) {
            if (!PageNetworkBondSettings.connection) return null;
            return PageNetworkBondSettings.connection.Slaves.find(function(s) {
                return s.Interfaces.indexOf(iface) >= 0;
            }) || null;
        },
        update: function() {
            var self = this;
            var model = PageNetworkBondSettings.model;
            var master = PageNetworkBondSettings.connection;
            var settings = PageNetworkBondSettings.settings;
            var options = settings.bond.options;
            var slaves_element;
            var mode_btn, primary_btn;
            var monitoring_btn, interval_input, targets_input, updelay_input, downdelay_input;
            function change_slaves() {
                var btn = slave_chooser_btn(change_mode, slaves_element);
                primary_btn.replaceWith(btn);
                primary_btn = btn;
                shell.select_btn_select(primary_btn, options.primary);
                change_mode();
            }
            function change_mode() {
                options.mode = shell.select_btn_selected(mode_btn);
                primary_btn.parents("tr").toggle(options.mode == "active-backup");
                if (options.mode == "active-backup") options.primary = shell.select_btn_selected(primary_btn); else delete options.primary;
            }
            function change_monitoring() {
                var use_mii = shell.select_btn_selected(monitoring_btn) == "mii";
                targets_input.parents("tr").toggle(!use_mii);
                updelay_input.parents("tr").toggle(use_mii);
                downdelay_input.parents("tr").toggle(use_mii);
                if (use_mii) {
                    options.miimon = interval_input.val();
                    options.updelay = updelay_input.val();
                    options.downdelay = downdelay_input.val();
                    delete options.arp_interval;
                    delete options.arp_ip_target;
                } else {
                    delete options.miimon;
                    delete options.updelay;
                    delete options.downdelay;
                    options.arp_interval = interval_input.val();
                    options.arp_ip_target = targets_input.val();
                }
            }
            var body = $('<table class="cockpit-form-table">').append($("<tr>").append($("<td>").text(_("Name")), $("<td>").append($('<input class="form-control">').val(settings.bond.interface_name).change(function(event) {
                var val = $(event.target).val();
                settings.bond.interface_name = val;
                settings.connection.interface_name = val;
            }))), $("<tr>").append($('<td class="top">').text(_("Members")), $("<td>").append(slaves_element = render_slave_interface_choices(model, master).change(change_slaves))).toggle(!master), $("<tr>").append($("<td>").text(_("Mode")), $("<td>").append(mode_btn = shell.select_btn(change_mode, bond_mode_choices))), $("<tr>").append($("<td>").text(_("Primary")), $("<td>").append(primary_btn = slave_chooser_btn(change_mode, slaves_element))), $("<tr>").append($("<td>").text(_("Link Monitoring")), $("<td>").append(monitoring_btn = shell.select_btn(change_monitoring, bond_monitoring_choices))), $("<tr>").append($("<td>").text(_("Monitoring Interval")), $("<td>").append(interval_input = $('<input class="form-control network-number-field" type="text" maxlength="4">').val(options.miimon || options.arp_interval || "100").change(change_monitoring))), $("<tr>").append($("<td>").text(_("Monitoring Targets")), $("<td>").append(targets_input = $('<input class="form-control">').val(options.arp_ip_targets).change(change_monitoring))), $("<tr>").append($("<td>").text(_("Link up delay")), $("<td>").append(updelay_input = $('<input class="form-control network-number-field" type="text" maxlength="4">').val(options.updelay || "0").change(change_monitoring))), $("<tr>").append($("<td>").text(_("Link down delay")), $("<td>").append(downdelay_input = $('<input class="form-control network-number-field" type="text" maxlength="4">').val(options.downdelay || "0").change(change_monitoring))));
            shell.select_btn_select(mode_btn, options.mode);
            shell.select_btn_select(monitoring_btn, options.miimon !== 0 ? "mii" : "arp");
            change_slaves();
            change_mode();
            change_monitoring();
            $("#network-bond-settings-body").html(body);
        },
        cancel: function() {
            if (PageNetworkBondSettings.connection) PageNetworkBondSettings.connection.reset();
            $("#network-bond-settings-dialog").modal("hide");
        },
        apply: function() {
            apply_master_slave($("#network-bond-settings-body"), PageNetworkBondSettings.model, PageNetworkBondSettings.connection, PageNetworkBondSettings.settings, "bond").done(function() {
                $("#network-bond-settings-dialog").modal("hide");
                if (PageNetworkBondSettings.done) PageNetworkBondSettings.done();
            }).fail(function(error) {
                $("#network-bond-settings-error").text(error.message || error.toString());
            });
        }
    };
    function PageNetworkBondSettings() {
        this._init();
    }
    shell.dialogs.push(new PageNetworkBondSettings());
    PageNetworkBridgeSettings.prototype = {
        _init: function() {
            this.id = "network-bridge-settings-dialog";
        },
        setup: function() {
            $("#network-bridge-settings-cancel").click($.proxy(this, "cancel"));
            $("#network-bridge-settings-apply").click($.proxy(this, "apply"));
        },
        enter: function() {
            $("#network-bridge-settings-error").text("");
            if (PageNetworkBridgeSettings.connection) PageNetworkBridgeSettings.connection.freeze();
            this.update();
        },
        show: function() {},
        leave: function() {},
        find_slave_con: function(iface) {
            if (!PageNetworkBridgeSettings.connection) return null;
            return PageNetworkBridgeSettings.connection.Slaves.find(function(s) {
                return s.Interfaces.indexOf(iface) >= 0;
            }) || null;
        },
        update: function() {
            var self = this;
            var model = PageNetworkBridgeSettings.model;
            var settings = PageNetworkBridgeSettings.settings;
            var options = settings.bridge;
            var stp_input, priority_input, forward_delay_input, hello_time_input, max_age_input;
            function change_stp() {
                options.stp = stp_input.prop("checked");
                options.priority = parseInt(priority_input.val(), 10);
                options.forward_delay = parseInt(forward_delay_input.val(), 10);
                options.hello_time = parseInt(hello_time_input.val(), 10);
                options.max_age = parseInt(max_age_input.val(), 10);
                priority_input.parents("tr").toggle(options.stp);
                forward_delay_input.parents("tr").toggle(options.stp);
                hello_time_input.parents("tr").toggle(options.stp);
                max_age_input.parents("tr").toggle(options.stp);
            }
            var body = $('<table class="cockpit-form-table">').append($("<tr>").append($("<td>").text(_("Name")), $("<td>").append($('<input class="form-control">').val(options.interface_name).change(function(event) {
                var val = $(event.target).val();
                options.interface_name = val;
                settings.connection.interface_name = val;
            }))), $("<tr>").append($('<td class="top">').text(_("Ports")), $("<td>").append(render_slave_interface_choices(model, PageNetworkBridgeSettings.connection))).toggle(!PageNetworkBridgeSettings.connection), $("<tr>").append($("<td>").text(_("Spanning Tree Protocol (STP)")), $("<td>").append(stp_input = $('<input type="checkbox">').prop("checked", options.stp).change(change_stp))), $("<tr>").append($("<td>").text(_("STP Priority")), $("<td>").append(priority_input = $('<input class="form-control" type="text">').val(options.priority).change(change_stp))), $("<tr>").append($("<td>").text(_("STP Forward delay")), $("<td>").append(forward_delay_input = $('<input class="form-control" type="text">').val(options.forward_delay).change(change_stp))), $("<tr>").append($("<td>").text(_("STP Hello time")), $("<td>").append(hello_time_input = $('<input class="form-control" type="text">').val(options.hello_time).change(change_stp))), $("<tr>").append($("<td>").text(_("STP Maximum message age")), $("<td>").append(max_age_input = $('<input class="form-control" type="text">').val(options.max_age).change(change_stp))));
            change_stp();
            $("#network-bridge-settings-body").html(body);
        },
        cancel: function() {
            if (PageNetworkBridgeSettings.connection) PageNetworkBridgeSettings.connection.reset();
            $("#network-bridge-settings-dialog").modal("hide");
        },
        apply: function() {
            apply_master_slave($("#network-bridge-settings-body"), PageNetworkBridgeSettings.model, PageNetworkBridgeSettings.connection, PageNetworkBridgeSettings.settings, "bridge").done(function() {
                $("#network-bridge-settings-dialog").modal("hide");
                if (PageNetworkBridgeSettings.done) PageNetworkBridgeSettings.done();
            }).fail(function(error) {
                $("#network-bridge-settings-error").text(error.message || error.toString());
            });
        }
    };
    function PageNetworkBridgeSettings() {
        this._init();
    }
    shell.dialogs.push(new PageNetworkBridgeSettings());
    PageNetworkBridgePortSettings.prototype = {
        _init: function() {
            this.id = "network-bridgeport-settings-dialog";
        },
        setup: function() {
            $("#network-bridgeport-settings-cancel").click($.proxy(this, "cancel"));
            $("#network-bridgeport-settings-apply").click($.proxy(this, "apply"));
        },
        enter: function() {
            $("#network-bridgeport-settings-error").text("");
            if (PageNetworkBridgePortSettings.connection) PageNetworkBridgePortSettings.connection.freeze();
            this.update();
        },
        show: function() {},
        leave: function() {},
        update: function() {
            var self = this;
            var model = PageNetworkBridgePortSettings.model;
            var settings = PageNetworkBridgePortSettings.settings;
            var options = settings.bridge_port;
            var priority_input, path_cost_input, hairpin_mode_input;
            function change() {
                options.priority = parseInt(priority_input.val(), 10);
                options.path_cost = parseInt(path_cost_input.val(), 10);
                options.hairpin_mode = hairpin_mode_input.prop("checked");
            }
            var body = $('<table class="cockpit-form-table">').append($("<tr>").append($("<td>").text(_("Priority")), $("<td>").append(priority_input = $('<input class="form-control network-number-field" type="text">').val(options.priority).change(change))), $("<tr>").append($("<td>").text(_("Path cost")), $("<td>").append(path_cost_input = $('<input class="form-control network-number-field" type="text">').val(options.path_cost).change(change))), $("<tr>").append($("<td>").text(_("Hair Pin mode")), $("<td>").append(hairpin_mode_input = $('<input type="checkbox">').prop("checked", options.hairpin_mode).change(change))));
            $("#network-bridgeport-settings-body").html(body);
        },
        cancel: function() {
            if (PageNetworkBridgePortSettings.connection) PageNetworkBridgePortSettings.connection.reset();
            $("#network-bridgeport-settings-dialog").modal("hide");
        },
        apply: function() {
            var self = this;
            var model = PageNetworkBridgePortSettings.model;
            var master_settings = PageNetworkBridgePortSettings.settings;
            var settings_manager = model.get_settings();
            function show_error(error) {
                $("#network-bridgeport-settings-error").text(error.message || error.toString());
            }
            function update_master() {
                if (PageNetworkBridgePortSettings.connection) return PageNetworkBridgePortSettings.connection.apply(); else return settings_manager.add_connection(master_settings);
            }
            update_master().done(function() {
                $("#network-bridgeport-settings-dialog").modal("hide");
                if (PageNetworkBridgePortSettings.done) PageNetworkBridgePortSettings.done();
            }).fail(show_error);
        }
    };
    function PageNetworkBridgePortSettings() {
        this._init();
    }
    shell.dialogs.push(new PageNetworkBridgePortSettings());
    PageNetworkVlanSettings.prototype = {
        _init: function() {
            this.id = "network-vlan-settings-dialog";
        },
        setup: function() {
            $("#network-vlan-settings-cancel").click($.proxy(this, "cancel"));
            $("#network-vlan-settings-apply").click($.proxy(this, "apply"));
        },
        enter: function() {
            $("#network-vlan-settings-error").text("");
            if (PageNetworkVlanSettings.connection) PageNetworkVlanSettings.connection.freeze();
            this.update();
        },
        show: function() {},
        leave: function() {},
        update: function() {
            var self = this;
            var model = PageNetworkVlanSettings.model;
            var settings = PageNetworkVlanSettings.settings;
            var options = settings.vlan;
            var auto_update_name = true;
            var parent_btn, id_input, name_input;
            function change() {
                options.parent = shell.select_btn_selected(parent_btn);
                $("#network-vlan-settings-apply").toggleClass("disabled", !options.parent);
                options.id = parseInt(id_input.val(), 10);
                if (auto_update_name && options.parent && options.id) name_input.val(options.parent + "." + options.id);
                options.interface_name = name_input.val();
                settings.connection.interface_name = options.interface_name;
            }
            function change_name() {
                auto_update_name = false;
                change();
            }
            var parent_choices = [];
            model.list_interfaces().forEach(function(i) {
                if (!is_interface_connection(i, PageNetworkVlanSettings.connection) && is_interesting_interface(i)) parent_choices.push({
                    title: i.Name,
                    choice: i.Name
                });
            });
            var body = $('<table class="cockpit-form-table">').append($("<tr>").append($("<td>").text(_("Parent")), $("<td>").append(parent_btn = shell.select_btn(change, parent_choices))), $("<tr>").append($("<td>").text(_("VLAN Id")), $("<td>").append(id_input = $('<input class="form-control" type="text">').val(options.id || "1").change(change).on("input", change))), $("<tr>").append($("<td>").text(_("Name")), $("<td>").append(name_input = $('<input class="form-control" type="text">').val(options.interface_name).change(change_name).on("input", change_name))));
            shell.select_btn_select(parent_btn, options.parent || (parent_choices[0] ? parent_choices[0].choice : ""));
            change();
            $("#network-vlan-settings-body").html(body);
        },
        cancel: function() {
            if (PageNetworkVlanSettings.connection) PageNetworkVlanSettings.connection.reset();
            $("#network-vlan-settings-dialog").modal("hide");
        },
        apply: function() {
            var self = this;
            var model = PageNetworkVlanSettings.model;
            var master_settings = PageNetworkVlanSettings.settings;
            var settings_manager = model.get_settings();
            function show_error(error) {
                $("#network-vlan-settings-error").text(error.message || error.toString());
            }
            function update_master() {
                if (PageNetworkVlanSettings.connection) return PageNetworkVlanSettings.connection.apply(); else return settings_manager.add_connection(master_settings);
            }
            update_master().done(function() {
                $("#network-vlan-settings-dialog").modal("hide");
                if (PageNetworkVlanSettings.done) PageNetworkVlanSettings.done();
            }).fail(show_error);
        }
    };
    function PageNetworkVlanSettings() {
        this._init();
    }
    shell.dialogs.push(new PageNetworkVlanSettings());
})(jQuery, cockpit, shell, modules);

var shell = shell || {};

var modules = modules || {};

(function($, cockpit, shell, modules) {
    function fmt_size(bytes) {
        return cockpit.format_bytes(bytes, 1024);
    }
    function fmt_size_long(bytes) {
        var with_unit = cockpit.format_bytes(bytes, 1024);
        return with_unit + " (" + bytes + " " + C_("format-bytes", "bytes") + ")";
    }
    function format_temperature(kelvin) {
        var celcius = kelvin - 273.15;
        var fahrenheit = 9 * celcius / 5 + 32;
        return celcius.toFixed(1) + "° C / " + fahrenheit.toFixed(1) + "° F";
    }
    function esc_id_attr(str) {
        return shell.esc(str).replace(/ /g, "&#20;").replace(/\x09/g, "&#09;").replace(/\x0a/g, "&#0a;").replace(/\x0c/g, "&#0c;").replace(/\x0d/g, "&#0d;");
    }
    var active_targets = [];
    function job_target_class(target) {
        return "spinner-" + target.replace(/[^a-zA-Z0-9_-]/g, "-");
    }
    function prepare_as_target(elt) {
        $(elt).hide();
    }
    function mark_as_target(elt, target) {
        var i;
        var cl = job_target_class(target);
        elt = $(elt);
        elt.addClass(cl);
        for (i = 0; i < active_targets.length; i++) {
            if (active_targets[i] == cl) elt.show();
        }
    }
    function watch_jobs(client) {
        function update() {
            var objs = client.getObjectsFrom("/com/redhat/Cockpit/Jobs/");
            var job;
            var i, j;
            for (i = 0; i < active_targets.length; i++) {
                $("." + active_targets[i]).hide();
            }
            active_targets = [];
            for (i = 0; i < objs.length; i++) {
                job = objs[i].lookup("com.redhat.Cockpit.Job");
                if (job) {
                    for (j = 0; j < job.Targets.length; j++) {
                        var t = job_target_class(job.Targets[j]);
                        active_targets.push(t);
                        $("." + t).show();
                    }
                }
            }
        }
        if (!client._job_watchers) {
            client._job_watchers = 1;
            $(client).on("objectAdded.watch-jobs", update);
            $(client).on("objectRemoved.watch-jobs", update);
            update();
        }
    }
    function unwatch_jobs(client) {
        client._job_watchers = client._job_watchers - 1;
        if (!client._job_watchers) {
            $(client).off(".watch-jobs");
        }
    }
    function job_box(client, tbody, domain, descriptions, target_describer) {
        function update() {
            var objs = client.getObjectsFrom("/com/redhat/Cockpit/Jobs/");
            var i, j, t, tdesc;
            var target_desc, desc, progress, remaining, cancel;
            var some_added = false;
            function make_click_handler() {
                return function(event) {
                    j.call("Cancel", function(error) {
                        if (error) shell.show_unexpected_error(error);
                    });
                };
            }
            tbody.empty();
            for (i = 0; i < objs.length; i++) {
                j = objs[i].lookup("com.redhat.Cockpit.Job");
                if (j && j.Domain == domain) {
                    target_desc = "";
                    for (t = 0; t < j.Targets.length; t++) {
                        tdesc = target_describer(j.Targets[t]);
                        if (tdesc) {
                            if (target_desc) target_desc += ", ";
                            target_desc += tdesc;
                        }
                    }
                    desc = cockpit.format(descriptions[j.Operation] || _("Unknown operation on $0"), target_desc);
                    if (j.ProgressValid) progress = (j.Progress * 100).toFixed() + "%"; else progress = "";
                    if (j.RemainingUSecs) remaining = shell.format_delay(j.RemainingUSecs / 1e3); else remaining = "";
                    if (j.Cancellable) {
                        cancel = $('<button class="btn btn-default">').text(_("Cancel"));
                        cancel.on("click", make_click_handler());
                    } else cancel = "";
                    tbody.append($("<tr>").append($('<td style="width:50%"/>').text(desc), $('<td style="width:15%text-align:right"/>').text(progress), $('<td style="width:15%text-align:right"/>').text(remaining), $('<td style="text-align:right"/>').append(cancel)));
                    some_added = true;
                }
            }
            tbody.parents(".panel").toggle(some_added);
        }
        function update_props(event, obj, iface) {
            if (iface._iface_name == "com.redhat.Cockpit.Job") update();
        }
        function stop() {
            $(client).off("objectAdded", update);
            $(client).off("objectRemoved", update);
            $(client).off("propertiesChanged", update_props);
        }
        function start() {
            $(client).on("objectAdded", update);
            $(client).on("objectRemoved", update);
            $(client).on("propertiesChanged", update_props);
            update();
        }
        start();
        return {
            stop: stop
        };
    }
    function get_block_devices_for_drive(drive) {
        var drive_obj = drive.getObject();
        var ret = [];
        var objs = drive._client.getObjectsFrom("/com/redhat/Cockpit/Storage/block_devices/");
        for (var n = 0; n < objs.length; n++) {
            var obj = objs[n];
            var block = obj.lookup("com.redhat.Cockpit.Storage.Block");
            if (block.PartitionNumber === 0 && block.Drive == drive_obj.objectPath) {
                ret.push(block);
            }
        }
        return ret;
    }
    function get_block_devices_for_mdraid(mdraid) {
        var mdraid_obj = mdraid.getObject();
        var ret = [];
        var objs = mdraid._client.getObjectsFrom("/com/redhat/Cockpit/Storage/block_devices/");
        for (var n = 0; n < objs.length; n++) {
            var obj = objs[n];
            var block = obj.lookup("com.redhat.Cockpit.Storage.Block");
            if (block.PartitionNumber === 0 && block.MDRaid == mdraid_obj.objectPath) {
                ret.push(block);
            }
        }
        return ret;
    }
    function find_block_device_for_drive(drive) {
        var blocks = get_block_devices_for_drive(drive);
        return blocks.length > 0 ? blocks[0] : undefined;
    }
    function find_block_device_for_mdraid(mdraid) {
        var blocks = get_block_devices_for_mdraid(mdraid);
        return blocks.length > 0 ? blocks[0] : undefined;
    }
    function find_logical_volume_block(lv) {
        var lv_obj = lv.getObject();
        var objs = lv._client.getObjectsFrom("/com/redhat/Cockpit/Storage/block_devices/");
        for (var n = 0; n < objs.length; n++) {
            var obj = objs[n];
            var block = obj.lookup("com.redhat.Cockpit.Storage.Block");
            if (block.LogicalVolume == lv_obj.objectPath) return block;
        }
        return null;
    }
    function mark_as_block_target(elt, block) {
        mark_as_target(elt, block.getObject().objectPath);
        for (var i = 0; i < block.Partitions.length; i++) {
            var b = block._client.lookup(block.Partitions[i][0], "com.redhat.Cockpit.Storage.Block");
            if (b) mark_as_block_target(elt, b);
        }
        if (block.IdUsage == "crypto") {
            var cleartext_device = find_cleartext_device(block);
            if (cleartext_device) mark_as_block_target(elt, cleartext_device);
        }
    }
    function storage_job_box(client, elt) {
        return job_box(client, elt, "storage", {
            "format-mkfs": _("Creating filesystem on $target"),
            "format-erase": _("Erasing $target"),
            "lvm-vg-empty-device": _("Emptying $target")
        }, function(target) {
            var block = client.lookup(target, "com.redhat.Cockpit.Storage.Block");
            return block ? block.Device : null;
        });
    }
    function storage_log_box(elt) {
        var logbox = modules.server.logbox([ "_SYSTEMD_UNIT=udisks2.service", "+", "_SYSTEMD_UNIT=dm-event.service", "+", "_SYSTEMD_UNIT=smartd.service", "+", "COCKPIT_DOMAIN=storage" ], 10);
        elt.empty().append(logbox);
        return logbox;
    }
    function highlight_error(container) {
        $(container).addClass("has-error");
    }
    function hide_error(container) {
        $(container).removeClass("has-error");
    }
    function highlight_error_message(id, message) {
        $(id).text(message);
        $(id).css("visibility", "visible");
    }
    function hide_error_message(id) {
        $(id).css("visibility", "hidden");
    }
    function format_fsys_usage(used, total) {
        var text = "";
        var units = 1024;
        var parts = cockpit.format_bytes(total, units, true);
        text = " / " + parts.join(" ");
        units = parts[1];
        parts = cockpit.format_bytes(used, units, true);
        return parts[0] + text;
    }
    function update_storage_privileged() {
        shell.update_privileged_ui(shell.default_permission, ".storage-privileged", cockpit.format(_("The user <b>$0</b> is not permitted to manage storage"), cockpit.user.name));
    }
    $(shell.default_permission).on("changed", update_storage_privileged);
    PageStorage.prototype = {
        _init: function() {
            this.id = "storage";
        },
        getTitle: function() {
            return C_("page-title", "Storage");
        },
        setup: function() {
            var self = this;
            $("#storage_create_raid").on("click", function() {
                PageCreateRaid.client = self.client;
                $("#create-raid-dialog").modal("show");
            });
            $("#storage_create_volume_group").on("click", function() {
                PageCreateVolumeGroup.client = self.client;
                $("#create-volume-group-dialog").modal("show");
            });
        },
        enter: function() {
            var self = this;
            this.client = shell.dbus(null);
            watch_jobs(this.client);
            this._drives = $("#storage_drives");
            this._raids = $("#storage_raids");
            this._vgs = $("#storage_vgs");
            this._other_devices = $("#storage_other_devices");
            this._mounts = $("#storage_mounts");
            this._coldplug();
            $(this.client).on("objectAdded.storage", $.proxy(this._onObjectAdded, this));
            $(this.client).on("objectRemoved.storage", $.proxy(this._onObjectRemoved, this));
            $(this.client).on("propertiesChanged.storage", $.proxy(this._onPropertiesChanged, this));
            this.job_box = storage_job_box(this.client, $("#storage-jobs"));
            this.log_box = storage_log_box($("#storage-log"));
            var blues = [ "#006bb4", "#008ff0", "#2daaff", "#69c2ff", "#a5daff", "#e1f3ff", "#00243c", "#004778" ];
            function is_interesting_blockdev(dev) {
                return dev && self.blockdevs[dev];
            }
            function highlight_blockdev_row(event, id) {
                $("#storage tr").removeClass("highlight");
                if (id) {
                    $('#storage tr[data-blockdev~="' + shell.esc(id) + '"]').addClass("highlight");
                }
            }
            function render_samples(event, timestamp, samples) {
                for (var id in samples) {
                    var row = $('#storage tr[data-blockdev="' + shell.esc(id) + '"]');
                    if (row.length > 0) {
                        row.find("span.reading").text("R: " + cockpit.format_bytes_per_sec(samples[id][0]));
                        row.find("span.writing").text("W: " + cockpit.format_bytes_per_sec(samples[id][1]));
                    }
                }
            }
            this.monitor = this.client.get("/com/redhat/Cockpit/BlockdevMonitor", "com.redhat.Cockpit.MultiResourceMonitor");
            $(this.monitor).on("NewSample.storage", render_samples);
            this.rx_plot = shell.setup_multi_plot("#storage-reading-graph", this.monitor, 0, blues.concat(blues), is_interesting_blockdev);
            $(this.rx_plot).on("update-total", function(event, total) {
                $("#storage-reading-text").text(cockpit.format_bytes_per_sec(total));
            });
            $(this.rx_plot).on("highlight", highlight_blockdev_row);
            this.tx_plot = shell.setup_multi_plot("#storage-writing-graph", this.monitor, 1, blues.concat(blues), is_interesting_blockdev);
            $(this.tx_plot).on("update-total", function(event, total) {
                $("#storage-writing-text").text(cockpit.format_bytes_per_sec(total));
            });
            $(this.tx_plot).on("highlight", highlight_blockdev_row);
            function render_mount_samples(event, timestamp, samples) {
                for (var id in samples) {
                    var used = samples[id][0];
                    var total = samples[id][1];
                    var row = self.mount_bar_rows[id];
                    if (row) {
                        row.attr("value", used + "/" + total);
                        row.toggleClass("bar-row-danger", used > .95 * total);
                    }
                    var text = self.mount_texts[id];
                    if (text) text.text(format_fsys_usage(samples[id][0], samples[id][1]));
                }
            }
            this.mount_monitor = this.client.get("/com/redhat/Cockpit/MountMonitor", "com.redhat.Cockpit.MultiResourceMonitor");
            $(this.mount_monitor).on("NewSample.storage", render_mount_samples);
        },
        show: function() {
            this.rx_plot.start();
            this.tx_plot.start();
        },
        leave: function() {
            this.rx_plot.destroy();
            this.tx_plot.destroy();
            $(this.client).off(".storage");
            $(this.monitor).off(".storage");
            $(this.mount_monitor).off(".storage");
            this.job_box.stop();
            if (this.log_box) this.log_box.stop();
            this.log_box = null;
            unwatch_jobs(this.client);
            this.client.release();
            this.client = null;
        },
        _onObjectAdded: function(event, obj) {
            if (obj.objectPath.indexOf("/com/redhat/Cockpit/Storage/") !== 0) return;
            this._delayed_coldplug();
        },
        _onObjectRemoved: function(event, obj) {
            if (obj.objectPath.indexOf("/com/redhat/Cockpit/Storage/") !== 0) return;
            this._delayed_coldplug();
        },
        _onPropertiesChanged: function(event, obj, iface) {
            if (obj.objectPath.indexOf("/com/redhat/Cockpit/Storage/") !== 0) return;
            this._delayed_coldplug();
        },
        _coldplug: function() {
            this._drives.empty();
            this._drives.closest(".panel").hide();
            this._raids.empty();
            this._raids.closest(".panel").hide();
            this._vgs.empty();
            this._vgs.closest(".panel").hide();
            this._other_devices.empty();
            this._other_devices.closest(".panel").hide();
            this._mounts.empty();
            this.blockdevs = {};
            this.mount_bar_rows = {};
            this.mount_texts = {};
            var objs = this.client.getObjectsFrom("/com/redhat/Cockpit/Storage/");
            for (var n = 0; n < objs.length; n++) {
                this._add(objs[n]);
            }
            $(this.monitor).trigger("notify:Consumers");
        },
        _delayed_coldplug: function() {
            var self = this;
            if (!self._coldplug_pending) {
                self._coldplug_pending = true;
                window.setTimeout(function() {
                    self._coldplug_pending = false;
                    if (self.client) self._coldplug();
                }, 0);
            }
        },
        _monitor_block: function(block) {
            if (!block) return "";
            var blockdev = block.Device;
            if (blockdev.startsWith("/dev/")) blockdev = blockdev.substr(5);
            this.blockdevs[blockdev] = true;
            return blockdev;
        },
        _add: function(obj) {
            if (obj.lookup("com.redhat.Cockpit.Storage.Drive")) this._addDrive(obj); else if (obj.lookup("com.redhat.Cockpit.Storage.MDRaid")) this._addRaid(obj); else if (obj.lookup("com.redhat.Cockpit.Storage.VolumeGroup")) this._addVG(obj); else if (obj.lookup("com.redhat.Cockpit.Storage.Block")) {
                this._addOtherDevice(obj);
                this._addMount(obj);
            }
        },
        _addDrive: function(obj) {
            var drive = obj.lookup("com.redhat.Cockpit.Storage.Drive");
            var id = esc_id_attr(obj.objectPath.substr(obj.objectPath.lastIndexOf("/") + 1));
            var sort_key = drive.SortKey;
            var n;
            var blockdev = this._monitor_block(find_block_device_for_drive(drive));
            var size_str = fmt_size(drive.Size);
            var desc;
            if (drive.Classification == "hdd") {
                desc = size_str + " " + C_("storage", "Hard Disk");
            } else if (drive.Classification == "ssd") {
                desc = size_str + " " + C_("storage", "Solid-State Disk");
            } else if (drive.Classification == "removable") {
                if (drive.Size === 0) desc = C_("storage", "Removable Drive"); else desc = size_str + " " + C_("storage", "Removable Drive");
            } else if (drive.Classification == "optical") {
                desc = C_("storage", "Optical Drive");
            } else {
                if (drive.Size === 0) desc = C_("storage", "Drive"); else desc = size_str + " " + C_("storage", "Drive");
            }
            var tr = $("<tr>", {
                id: "storage-drive-" + id,
                Sort: sort_key,
                "data-blockdev": blockdev
            }).click(function() {
                cockpit.location.go("storage-detail", {
                    type: "drive",
                    id: id
                });
            }).append($('<td style="width: 48px">').append($("<img>", {
                src: "images/storage-disk.png"
            })), $('<td class="row">').append($('<span class="col-md-12">').text(drive.Name), $("<br>"), $('<span class="col-md-12 col-lg-5 storage-disk-size">').text(desc), $('<span class="col-md-12 col-lg-7">').append($('<span class="reading">').text(""), $('<span style="display:inline-block;width:1em">'), $('<span class="writing">').text(""))), $('<td style="width:28px">').append($("<div>", {
                id: "storage-spinner-" + id,
                "class": "waiting"
            })));
            var children = this._drives[0].childNodes;
            var insert_before = null;
            var child, child_sort_key;
            for (n = 0; n < children.length; n++) {
                child = children[n];
                child_sort_key = child.getAttribute("sort");
                if (child_sort_key > sort_key) {
                    insert_before = child;
                    break;
                }
            }
            this._drives[0].insertBefore(tr[0], insert_before);
            this._drives.closest(".panel").show();
            prepare_as_target($("#storage-spinner-" + id));
            var blocks = get_block_devices_for_drive(drive);
            for (n = 0; n < blocks.length; n++) mark_as_block_target($("#storage-spinner-" + id), blocks[n]);
        },
        _addRaid: function(obj) {
            var raid = obj.lookup("com.redhat.Cockpit.Storage.MDRaid");
            var id = esc_id_attr(obj.objectPath.substr(obj.objectPath.lastIndexOf("/") + 1));
            var desc = raid_get_desc(raid);
            var sort_key = desc;
            var n;
            var tr = $("<tr>", {
                id: "storage-raid-" + id,
                Sort: sort_key
            }).click(function() {
                cockpit.location.go("storage-detail", {
                    type: "mdraid",
                    id: id
                });
            }).append($('<td class="row">').append($('<span class="col-xs-3 col-xs-push-9 raid-size">').text(fmt_size(raid.Size)), $('<span class="col-xs-9 col-xs-pull-3">').text(raid_get_desc(raid))), $('<td style="width:28px">').append($("<div>", {
                id: "storage-spinner-" + id,
                "class": "waiting"
            })));
            var children = this._raids[0].childNodes;
            var insert_before = null;
            var child, child_sort_key;
            for (n = 0; n < children.length; n++) {
                child = children[n];
                child_sort_key = child.getAttribute("sort");
                if (child_sort_key.localeCompare(sort_key) > 0) {
                    insert_before = child;
                    break;
                }
            }
            this._raids[0].insertBefore(tr[0], insert_before);
            this._raids.closest(".panel").show();
            prepare_as_target($("#storage-spinner-" + id));
            var blocks = get_block_devices_for_mdraid(raid);
            for (n = 0; n < blocks.length; n++) mark_as_block_target($("#storage-spinner-" + id), blocks[n]);
        },
        _addVG: function(obj) {
            var vg = obj.lookup("com.redhat.Cockpit.Storage.VolumeGroup");
            var id = esc_id_attr(obj.objectPath.substr(obj.objectPath.lastIndexOf("/") + 1));
            var sort_key = vg.Name;
            var n;
            var tr = $("<tr>", {
                id: "storage-vg-" + id,
                Sort: sort_key
            }).click(function() {
                cockpit.location.go("storage-detail", {
                    type: "vg",
                    id: id
                });
            }).append($('<td class="row">').append($('<span class="col-xs-3 col-xs-push-9 vg-size">').text(fmt_size(vg.Size)), $('<span class="col-xs-9 col-xs-pull-3">').text(vg.Name)), $('<td style="width:28px">').append($("<div>", {
                id: "storage-spinner-" + id,
                "class": "waiting"
            })));
            var children = this._vgs[0].childNodes;
            var insert_before = null;
            var child, child_sort_key;
            for (n = 0; n < children.length; n++) {
                child = children[n];
                child_sort_key = child.getAttribute("sort");
                if (child_sort_key.localeCompare(sort_key) > 0) {
                    insert_before = child;
                    break;
                }
            }
            this._vgs[0].insertBefore(tr[0], insert_before);
            this._vgs.closest(".panel").show();
            prepare_as_target($("#storage-spinner-" + id));
        },
        _addOtherDevice: function(obj) {
            var block = obj.lookup("com.redhat.Cockpit.Storage.Block");
            if (block.PartitionNumber !== 0 || block.Drive != "/" || block.CryptoBackingDevice != "/" || block.MDRaid != "/" || block.LogicalVolume != "/" || block.HintIgnore) return;
            var blockdev = this._monitor_block(block);
            var id = esc_id_attr(obj.objectPath.substr(obj.objectPath.lastIndexOf("/") + 1));
            var sort_key = block.DeviceNumber;
            var tr = $("<tr>", {
                id: "storage-block-" + id,
                Sort: sort_key,
                "data-blockdev": blockdev
            }).click(function() {
                cockpit.location.go("storage-detail", {
                    type: "block",
                    id: id
                });
            }).append($("<td>").append($("<span>").text(block.Device), $("<br>"), $("<span>").text(fmt_size(block.Size) + " " + C_("storage", "Block Device")), $("<br>"), $('<span class="reading">').text(""), $('<span style="display:inline-block;width:2em">'), $('<span class="writing">').text("")), $('<td style="width:28px">').append($("<div>", {
                id: "storage-spinner-" + id,
                "class": "waiting"
            })));
            var children = this._other_devices[0].childNodes;
            var insert_before = null;
            var child, child_sort_key;
            for (var n = 0; n < children.length; n++) {
                child = children[n];
                child_sort_key = child.getAttribute("sort");
                if (parseInt(child_sort_key, 10) > sort_key) {
                    insert_before = child;
                    break;
                }
            }
            this._other_devices[0].insertBefore(tr[0], insert_before);
            this._other_devices.closest(".panel").show();
            prepare_as_target($("#storage-spinner-" + id));
            mark_as_block_target($("#storage-spinner-" + id), block);
        },
        _addMount: function(obj) {
            var block = obj.lookup("com.redhat.Cockpit.Storage.Block");
            if (block.IdUsage != "filesystem" || block.HintIgnore) return;
            var id = esc_id_attr(obj.objectPath.substr(obj.objectPath.lastIndexOf("/") + 1));
            var sort_key = id;
            var bar_row = null;
            var text = $('<td style="text-align:right">');
            if (block.MountedAt && block.MountedAt.length > 0) {
                bar_row = shell.BarRow();
                for (var i = 0; i < block.MountedAt.length; i++) {
                    this.mount_bar_rows[block.MountedAt[i]] = bar_row;
                    this.mount_texts[block.MountedAt[i]] = text;
                }
            } else text.text(cockpit.format_bytes(block.Size, 1024));
            var tr = $("<tr>", {
                Sort: sort_key
            }).click(function() {
                block_go(block);
            }).append($("<td>").text(block.IdLabel || block.Device), $("<td>").text(block.MountedAt || "-"), $("<td>").append(bar_row), text);
            var children = this._mounts[0].childNodes;
            var insert_before = null;
            var child, child_sort_key;
            for (var n = 0; n < children.length; n++) {
                child = children[n];
                child_sort_key = child.getAttribute("sort");
                if (child_sort_key > sort_key) {
                    insert_before = child;
                    break;
                }
            }
            this._mounts[0].insertBefore(tr[0], insert_before);
        }
    };
    function PageStorage() {
        this._init();
    }
    shell.pages.push(new PageStorage());
    function lvol_get_desc(lv) {
        var type;
        if (lv.Type == "pool") type = _("Pool for Thin Logical Volumes"); else if (lv.ThinPool != "/") type = _("Thin Logical Volume"); else if (lv.Origin != "/") type = _("Logical Volume (Snapshot)"); else type = _("Logical Volume");
        return cockpit.format('$type "$name"', {
            type: type,
            name: shell.esc(lv.Name)
        });
    }
    function block_get_desc(block, partition_label, cleartext_device) {
        var ret, lv;
        if (block.IdUsage == "filesystem") {
            ret = $("<span>").text(cockpit.format(C_("storage-id-desc", "$0 File System"), block.IdType));
        } else if (block.IdUsage == "raid") {
            if (block.IdType == "linux_raid_member") {
                ret = $("<span>").text(C_("storage-id-desc", "Linux MD-RAID Component"));
            } else if (block.IdType == "LVM2_member") {
                ret = $("<span>").text(C_("storage-id-desc", "LVM2 Physical Volume"));
            } else {
                ret = $("<span>").text(C_("storage-id-desc", "RAID Member"));
            }
            if (block.PvGroup != "/") {
                var vg = block._client.get(block.PvGroup, "com.redhat.Cockpit.Storage.VolumeGroup");
                ret.append(" of ", $("<a>").text(vg.Name).click(function() {
                    cockpit.location.go("storage-detail", {
                        type: "vg",
                        id: vg.Name
                    });
                }));
            } else if (block.MDRaidMember != "/") {
                var id = block.MDRaidMember.substr(block.MDRaidMember.lastIndexOf("/") + 1);
                var raid = block._client.get(block.MDRaidMember, "com.redhat.Cockpit.Storage.MDRaid");
                ret.append(" of ", $("<a>").text(raid_get_desc(raid)).click(function() {
                    cockpit.location.go("storage-detail", {
                        type: "mdraid",
                        id: id
                    });
                }));
            }
        } else if (block.IdUsage == "crypto") {
            if (block.IdType == "crypto_LUKS") {
                ret = $("<span>").text(C_("storage-id-desc", "LUKS Encrypted"));
            } else {
                ret = $("<span>").text(C_("storage-id-desc", "Encrypted"));
            }
        } else if (block.IdUsage == "other") {
            if (block.IdType == "swap") {
                ret = $("<span>").text(C_("storage-id-desc", "Swap Space"));
            } else {
                ret = $("<span>").text(C_("storage-id-desc", "Other Data"));
            }
        } else {
            ret = $("<span>").text(C_("storage-id-desc", "Unrecognized Data"));
        }
        if (block.PartitionNumber > 0) {
            ret = $("<span>").append(cockpit.format(_("$size $partition"), {
                size: fmt_size(block.Size),
                partition: partition_label
            }), " (", ret, ")");
        }
        if (block.LogicalVolume != "/") {
            lv = block._client.lookup(block.LogicalVolume, "com.redhat.Cockpit.Storage.LogicalVolume");
            ret = $("<span>").append(cockpit.format(_("$size $partition"), {
                size: fmt_size(block.Size),
                partition: lvol_get_desc(lv)
            }), " (", ret, ")");
        }
        ret.append($("<br/>"), shell.esc(block.Device));
        if (block.IdUsage == "filesystem") {
            ret.append(", ");
            if (block.MountedAt.length > 0) ret.append(cockpit.format(_("mounted on $0"), shell.esc(block.MountedAt[0]))); else ret.append(_("not mounted"));
        } else if (block.IdUsage == "crypto") {
            ret.append(", ");
            if (cleartext_device) ret.append(_("unlocked")); else ret.append(_("locked"));
        }
        return ret;
    }
    function block_get_short_desc(block) {
        if (block.PartitionNumber > 0) return "Partition"; else if (block.LogicalVolume != "/") {
            var lv = block._client.lookup(block.LogicalVolume, "com.redhat.Cockpit.Storage.LogicalVolume");
            return lvol_get_desc(lv);
        } else if (block.Drive != "/") {
            var drive = block._client.lookup(block.Drive, "com.redhat.Cockpit.Storage.Drive");
            return drive ? shell.esc(drive.Name) : block.Device;
        } else return "Block Device";
    }
    function block_go(block) {
        var id, lv, vg, path;
        while (true) {
            if (block.PartitionTable && block.PartitionTable != "/") block = block._client.get(block.PartitionTable, "com.redhat.Cockpit.Storage.Block"); else if (block.CryptoBackingDevice && block.CryptoBackingDevice != "/") block = block._client.get(block.CryptoBackingDevice, "com.redhat.Cockpit.Storage.Block"); else break;
        }
        if (block.Drive != "/") {
            id = block.Drive.substr(block.Drive.lastIndexOf("/") + 1);
            cockpit.location.go("storage-detail", {
                type: "drive",
                id: id
            });
        } else if (block.MDRaid != "/") {
            id = block.MDRaid.substr(block.MDRaid.lastIndexOf("/") + 1);
            cockpit.location.go("storage-detail", {
                type: "mdraid",
                id: id
            });
        } else if (block.LogicalVolume != "/") {
            lv = block._client.get(block.LogicalVolume, "com.redhat.Cockpit.Storage.LogicalVolume");
            if (lv.VolumeGroup != "/") {
                vg = lv._client.get(lv.VolumeGroup, "com.redhat.Cockpit.Storage.VolumeGroup");
                cockpit.location.go("storage-detail", {
                    type: "vg",
                    id: vg.Name
                });
            }
        } else {
            path = block.getObject().objectPath;
            id = path.substr(path.lastIndexOf("/") + 1);
            cockpit.location.go("storage-detail", {
                type: "block",
                id: id
            });
        }
    }
    function block_get_link_desc(block) {
        var is_part = false;
        var is_crypt = false;
        var link = null;
        while (true) {
            if (block.PartitionTable && block.PartitionTable != "/") {
                block = block._client.get(block.PartitionTable, "com.redhat.Cockpit.Storage.Block");
                is_part = true;
            } else if (block.CryptoBackingDevice && block.CryptoBackingDevice != "/") {
                block = block._client.get(block.CryptoBackingDevice, "com.redhat.Cockpit.Storage.Block");
                is_crypt = true;
            } else break;
        }
        if (block.Drive != "/") {
            var drive = block._client.get(block.Drive, "com.redhat.Cockpit.Storage.Drive");
            link = $("<a>").text(drive.Name || block.Device).click(function() {
                var id = block.Drive.substr(block.Drive.lastIndexOf("/") + 1);
                cockpit.location.go("storage-detail", {
                    type: "drive",
                    id: id
                });
            });
        } else if (block.MDRaid != "/") {
            var raid = block._client.get(block.MDRaid, "com.redhat.Cockpit.Storage.MDRaid");
            link = $("<span>").append(_("RAID Device"), " ", $("<a>").text(raid_get_desc(raid)).click(function() {
                var id = block.MDRaid.substr(block.MDRaid.lastIndexOf("/") + 1);
                cockpit.location.go("storage-detail", {
                    type: "mdraid",
                    id: id
                });
            }));
        } else if (block.LogicalVolume != "/") {
            var lv = block._client.get(block.LogicalVolume, "com.redhat.Cockpit.Storage.LogicalVolume");
            if (lv.VolumeGroup != "/") {
                var vg = lv._client.get(lv.VolumeGroup, "com.redhat.Cockpit.Storage.VolumeGroup");
                link = $("<span>").append(lvol_get_desc(lv), " of ", $("<a>").text(vg.Name).click(function() {
                    cockpit.location.go("storage-detail", {
                        type: "vg",
                        id: vg.Name
                    });
                }));
            } else {
                link = $("<span>").text(lvol_get_desc(lv));
            }
        } else {
            link = $("<a>").text(block.Device).click(function() {
                var path = block.getObject().objectPath;
                var id = path.substr(path.lastIndexOf("/") + 1);
                cockpit.location.go("storage-detail", {
                    type: "block",
                    id: id
                });
            });
        }
        var res = link;
        if (is_part) res = $("<span>").append(_("Partition of "), res);
        if (is_crypt) res = $("<span>").append(res, _(", encrypted"));
        return res;
    }
    function find_cleartext_device(block) {
        var objpath = block.getObject().objectPath;
        var objs = block._client.getObjectsFrom("/com/redhat/Cockpit/Storage/block_devices/");
        for (var n = 0; n < objs.length; n++) {
            var o = objs[n];
            var b = o.lookup("com.redhat.Cockpit.Storage.Block");
            if (b && b.CryptoBackingDevice == objpath) return b;
        }
        return null;
    }
    function raid_get_desc(raid) {
        if (!raid.Name) return "";
        var parts = raid.Name.split(":");
        if (parts.length != 2) return raid.Name;
        var manager = raid._client.lookup("/com/redhat/Cockpit/Manager", "com.redhat.Cockpit.Manager");
        if (manager && parts[0] == manager.StaticHostname) return parts[1]; else return cockpit.format(_("$name (on $host)"), {
            name: parts[1],
            host: parts[0]
        });
    }
    function get_free_block_devices(client, filter) {
        function is_extended_partition(b) {
            var b_objpath = b.getObject().objectPath;
            var part_block, table;
            var i;
            if (b.PartitionTable) part_block = client.lookup(b.PartitionTable, "com.redhat.Cockpit.Storage.Block");
            if (part_block) table = part_block.Partitions;
            if (table) {
                for (i = 0; i < table.length; i++) {
                    if (table[i][0] == b_objpath) return table[i][3] == "x";
                }
            }
            return false;
        }
        function has_fs_label(b) {
            return b.IdUsage !== "" && (b.IdType != "LVM2_member" || b.PvGroup != "/");
        }
        var result = [];
        var objs = client.getObjectsFrom("/com/redhat/Cockpit/Storage/block_devices/");
        for (var n = 0; n < objs.length; n++) {
            var o = objs[n];
            var b = o.lookup("com.redhat.Cockpit.Storage.Block");
            if (b && !b.HintIgnore && b.Size > 0 && !has_fs_label(b) && !b.PartitionTableType && !is_extended_partition(b) && !(filter && filter(b))) result.push(b);
        }
        return result;
    }
    PageStorageDetail.prototype = {
        _init: function() {
            this.id = "storage-detail";
            this.section_id = "storage";
            this.watched_objects = [];
        },
        getTitle: function() {
            return C_("page-title", "Storage");
        },
        get_page_title: function() {
            var ret;
            if (this._drive) {
                if (this._drive.Vendor && this._drive.Vendor.length > 0) ret = this._drive.Vendor + " " + this._drive.Model; else ret = this._drive.Model;
            } else if (this._mdraid) {
                ret = raid_get_desc(this._mdraid);
            } else if (this._vg) {
                ret = this._vg.Name;
            } else ret = this._block.Device;
            return ret || "?";
        },
        show: function() {},
        watch_object: function(obj) {
            if (obj) {
                this.watched_objects.push(obj);
                $(obj).on("notify.storage-details", $.proxy(this, "_update"));
            }
        },
        unwatch_all_objects: function() {
            for (var i = 0; i < this.watched_objects.length; i++) $(this.watched_objects[i]).off(".storage-details");
            this.watched_objects = [];
        },
        leave: function() {
            this.unwatch_all_objects();
            this.job_box.stop();
            if (this.log_box) this.log_box.stop();
            this.log_box = null;
            this.stop_vg_polling();
            unwatch_jobs(this.client);
            $(this.client).off(".storage-details");
            this.client.release();
            this.client = null;
        },
        setup: function() {
            var self = this;
            self.raid_action_btn = shell.action_btn(function(op) {
                self.action(op);
            }, [ {
                title: _("Start"),
                action: "start"
            }, {
                title: _("Stop"),
                action: "stop"
            }, {
                title: _("Format"),
                action: "format"
            }, {
                title: _("Start Scrubbing"),
                action: "start-scrub"
            }, {
                title: _("Stop Scrubbing"),
                action: "stop-scrub"
            }, {
                title: _("Delete"),
                action: "delete"
            } ]);
            $("#raid_action_btn").html(self.raid_action_btn, "storage-privileged");
            $("#raid-disks-add").on("click", $.proxy(this, "raid_disk_add"));
            function change_bitmap(val) {
                self._mdraid.call("SetBitmapLocation", val ? "internal" : "none", function(error, result) {
                    if (error) {
                        self._update();
                        shell.show_unexpected_error(error);
                    }
                });
            }
            this.bitmap_onoff = shell.OnOff(false, change_bitmap, undefined, null, "storage-privileged");
            $("#raid_detail_bitmap").append(this.bitmap_onoff);
            $("#drive_format").on("click", function() {
                self.action("format");
            });
            var btn = shell.action_btn(function(op) {
                self.volume_group_action(op);
            }, [ {
                title: _("Rename"),
                action: "rename",
                is_default: true
            }, {
                title: _("Delete"),
                action: "delete"
            } ], "storage-privileged");
            $("#vg_action_btn").html(btn);
            $("#vg-pv-add").on("click", $.proxy(this, "add_physical_volume"));
            $("#block_format").on("click", function() {
                self.action("format");
            });
        },
        enter: function() {
            var me = this;
            var type = shell.get_page_param("type");
            var id = shell.get_page_param("id");
            this.client = shell.dbus(null);
            watch_jobs(this.client);
            this._drive = null;
            this._mdraid = null;
            this._vg = null;
            this._block = null;
            $("#disk_detail_list").hide();
            $("#raid_detail_list").hide();
            $("#vg_detail_list").hide();
            $("#block_detail_list").hide();
            if (type == "drive") {
                this._drive = this.client.get("/com/redhat/Cockpit/Storage/drives/" + id, "com.redhat.Cockpit.Storage.Drive");
                $("#disk_detail_list").show();
            } else if (type == "mdraid") {
                this._mdraid = this.client.get("/com/redhat/Cockpit/Storage/raids/" + id, "com.redhat.Cockpit.Storage.MDRaid");
                $("#raid_detail_list").show();
            } else if (type == "vg") {
                this._vg = this.client.get("/com/redhat/Cockpit/Storage/lvm/" + id, "com.redhat.Cockpit.Storage.VolumeGroup");
                $("#vg_detail_list").show();
            } else {
                this._block = this.client.get("/com/redhat/Cockpit/Storage/block_devices/" + id, "com.redhat.Cockpit.Storage.Block");
                $("#block_detail_list").show();
            }
            this.job_box = storage_job_box(this.client, $("#storage-detail-jobs"));
            this.log_box = storage_log_box($("#storage-detail-log"));
            this._update();
            $("#storage-detail-title").text(this.get_page_title());
            $(this.client).on("objectAdded.storage-details", $.proxy(this._update, this));
            $(this.client).on("objectRemoved.storage-details", $.proxy(this._update, this));
            $(this.client).on("propertiesChanged.storage-details", $.proxy(this._onPropertiesChanged, this));
            update_storage_privileged();
        },
        _onPropertiesChanged: function(event, obj, iface) {
            if (obj.objectPath.indexOf("/com/redhat/Cockpit/Storage/") !== 0) return;
            this._update();
        },
        _update: function() {
            this.unwatch_all_objects();
            if (this._drive) this._updateDrive(); else if (this._mdraid) this._updateMDRaid(); else if (this._vg) this._updateVG(); else if (this._block) this._updateBlock();
            $("#storage-detail .breadcrumb .active").text(this.get_page_title());
        },
        _updateBlock: function() {
            var val;
            var block = this._block;
            this.watch_object(block);
            this._updateContent(block);
            val = shell.esc(block.Device);
            $("#block_detail_device").html(val);
            val = block.Size > 0 ? fmt_size_long(block.Size) : C_("storage", "No Media Inserted");
            $("#block_detail_capacity").html(val);
            update_storage_privileged();
        },
        _updateContent: function(block_or_vg) {
            var me = this;
            var id = 0;
            function block_action_func(target) {
                return function(op) {
                    me.block_action(target, op);
                };
            }
            function create_block_action_btn(target, is_crypto_locked, is_partition) {
                var btn;
                var filesystem_action_spec = [ {
                    title: _("Mount"),
                    action: "mount"
                }, {
                    title: _("Unmount"),
                    action: "unmount"
                }, {
                    title: _("Filesystem Options"),
                    action: "fsys-options"
                } ];
                var crypto_action_spec = [ {
                    title: _("Lock"),
                    action: "lock"
                }, {
                    title: _("Unlock"),
                    action: "unlock"
                }, {
                    title: _("Encryption Options"),
                    action: "crypto-options"
                } ];
                var lvol_action_spec = [ {
                    title: _("Resize"),
                    action: "resize"
                }, {
                    title: _("Rename"),
                    action: "rename"
                } ];
                var lvol_block_action_spec = [ {
                    title: _("Create Snapshot"),
                    action: "create-snapshot"
                }, {
                    title: _("Activate"),
                    action: "activate"
                }, {
                    title: _("Deactivate"),
                    action: "deactivate"
                } ];
                var lvol_pool_action_spec = [ {
                    title: _("Create Thin Volume"),
                    action: "create-thin-volume"
                } ];
                var format_action_spec = [ {
                    title: _("Format"),
                    action: "format"
                } ];
                var delete_action_spec = [ {
                    title: _("Delete"),
                    action: "delete"
                } ];
                var is_filesystem = target.IdUsage == "filesystem";
                var is_filesystem_mounted = target.MountedAt && target.MountedAt.length !== 0;
                var is_crypto = target.IdUsage == "crypto";
                var is_lvol = target._iface_name == "com.redhat.Cockpit.Storage.LogicalVolume" || target.LogicalVolume != "/";
                var is_lvol_pool = target.Type == "pool";
                var is_lvol_active = target._iface_name == "com.redhat.Cockpit.Storage.Block" || target.Active;
                var is_formattable = target._iface_name == "com.redhat.Cockpit.Storage.Block" && !target.ReadOnly;
                var default_op = null;
                var action_spec = [];
                if (is_filesystem) {
                    action_spec = action_spec.concat(filesystem_action_spec);
                    default_op = is_filesystem_mounted ? "unmount" : "mount";
                } else if (is_crypto) {
                    action_spec = action_spec.concat(crypto_action_spec);
                    default_op = is_crypto_locked ? "unlock" : "lock";
                }
                if (is_formattable) {
                    action_spec = action_spec.concat(format_action_spec);
                    if (!default_op) default_op = "format";
                }
                if (is_lvol) {
                    action_spec = action_spec.concat(lvol_action_spec);
                    if (is_lvol_pool) {
                        action_spec = action_spec.concat(lvol_pool_action_spec);
                        default_op = "create-thin-volume";
                    } else {
                        action_spec = action_spec.concat(lvol_block_action_spec);
                        if (!default_op) default_op = is_lvol_active ? "deactivate" : "activate";
                    }
                }
                if (is_partition || is_lvol) {
                    action_spec = action_spec.concat(delete_action_spec);
                    if (!default_op) default_op = "delete";
                }
                if (action_spec.length > 0) {
                    btn = shell.action_btn(function(op) {
                        me.block_action(target, op);
                    }, action_spec, "storage-privileged");
                    shell.action_btn_select(btn, default_op);
                    shell.action_btn_enable(btn, "mount", !is_filesystem_mounted);
                    shell.action_btn_enable(btn, "unmount", is_filesystem_mounted);
                    shell.action_btn_enable(btn, "lock", !is_crypto_locked);
                    shell.action_btn_enable(btn, "unlock", is_crypto_locked);
                    shell.action_btn_enable(btn, "activate", !is_lvol_active);
                    shell.action_btn_enable(btn, "deactivate", is_lvol_active);
                }
                return btn;
            }
            function create_simple_btn(title, func) {
                return $("<button>", {
                    "class": "btn btn-default storage-privileged",
                    on: {
                        click: func
                    }
                }).text(title);
            }
            function append_entry(level, name, desc, button) {
                id += 1;
                if (button === true) button = null;
                var tr = $("<tr>");
                if (level > 0) tr.append($("<td>", {
                    Width: 30 * level
                }));
                tr.append($("<td>", {
                    style: "width:50%"
                }).html(desc));
                if (name) tr.append($("<td>", {
                    style: "text-align:left"
                }).html(name));
                if (button) tr.append($("<td>", {
                    style: "float:right"
                }).append(button));
                tr.append($("<td>", {
                    style: "float:right"
                }).append($("<div>", {
                    id: "entry-spinner-" + id,
                    "class": "waiting"
                })));
                list.append($("<li>", {
                    "class": "list-group-item"
                }).append($("<table>", {
                    style: "width:100%"
                }).append(tr)));
                prepare_as_target("#entry-spinner-" + id);
                return id;
            }
            function append_non_partitioned_block(level, block, part_desc) {
                var id, name, desc, btn;
                var cleartext_device;
                if (block.IdUsage == "crypto") cleartext_device = find_cleartext_device(block);
                btn = create_block_action_btn(block, !cleartext_device, !!part_desc);
                if (block.IdLabel.length > 0) name = shell.esc(block.IdLabel); else if (!btn) name = null; else name = "—";
                desc = block_get_desc(block, part_desc, cleartext_device);
                id = append_entry(level, name, desc, btn);
                mark_as_target("#entry-spinner-" + id, block.getObject().objectPath);
                me.watch_object(block);
                if (cleartext_device) append_device(level + 1, cleartext_device);
            }
            function append_partitions(level, block) {
                var device_level = level;
                var is_dos_partitioned = block.PartitionTableType == "dos";
                var wanted = block.Partitions.sort(function(a, b) {
                    return a[1] - b[1];
                });
                function append_free_space(level, start, size) {
                    var desc;
                    var real_start = (Math.floor(start / (1024 * 1024)) + 1) * 1024 * 1024;
                    var enable_dos_extended = false;
                    if (start + size - real_start >= 1024 * 1024) {
                        if (is_dos_partitioned) {
                            if (level > device_level) {
                                desc = cockpit.format(_("$0 Free Space for Logical Partitions"), fmt_size(size));
                            } else {
                                desc = cockpit.format(_("$0 Free Space for Primary Partitions"), fmt_size(size));
                                enable_dos_extended = true;
                            }
                        } else {
                            desc = cockpit.format(_("$0 Free Space"), fmt_size(size));
                        }
                        append_entry(level, null, desc, create_simple_btn(_("Create Partition"), $.proxy(me, "create_partition", block, start, size, enable_dos_extended)));
                    }
                }
                function append_extended_partition(level, block, start, size) {
                    var desc = cockpit.format(_("$0 Extended Partition"), fmt_size(size));
                    var btn = create_block_action_btn(block, false, true);
                    append_entry(level, null, desc, btn);
                    me.watch_object(block);
                    process_level(level + 1, start, size);
                }
                function process_level(level, container_start, container_size) {
                    var n;
                    var last_end = container_start;
                    var total_end = container_start + container_size;
                    var block, start, size, type, part_desc;
                    for (n = 0; n < wanted.length; n++) {
                        block = me.client.lookup(wanted[n][0], "com.redhat.Cockpit.Storage.Block");
                        start = wanted[n][1];
                        size = wanted[n][2];
                        type = wanted[n][3];
                        if (block === null) continue;
                        if (level === device_level && type == "l") continue;
                        if (level == device_level + 1 && type != "l") continue;
                        if (start < container_start || start + size > container_start + container_size) continue;
                        append_free_space(level, last_end, start - last_end);
                        if (type == "x") append_extended_partition(level, block, start, size); else {
                            if (is_dos_partitioned) {
                                if (level > device_level) part_desc = _("Logical Partition"); else part_desc = _("Primary Partition");
                            } else part_desc = _("Partition");
                            append_non_partitioned_block(level, block, part_desc);
                        }
                        last_end = start + size;
                    }
                    append_free_space(level, last_end, total_end - last_end);
                }
                process_level(device_level, 0, block.Size);
            }
            function append_device(level, block, desc) {
                var is_partitioned = !!block.PartitionTableType;
                if (block.PartitionTableType) append_partitions(level, block); else append_non_partitioned_block(level, block, null);
            }
            function append_volume_group(level, vg) {
                var i, id, lvs, objs, block, lv, vg_obj, desc;
                function append_logical_volume_block(level, block, lv) {
                    var btn, id, desc;
                    if (block.PartitionTableType) {
                        desc = cockpit.format(_("$size $desc"), {
                            desc: lvol_get_desc(lv),
                            size: fmt_size(block.Size)
                        });
                        desc += "<br/>" + shell.esc(block.Device);
                        btn = create_block_action_btn(block, false, false);
                        id = append_entry(level, null, desc, btn);
                        append_partitions(level + 1, block);
                    } else append_non_partitioned_block(level, block, null);
                }
                function append_logical_volume(level, lv) {
                    var block, desc, btn, id, ratio;
                    var lv_obj, objs, i, llv;
                    if (lv.Type == "pool") {
                        ratio = Math.max(lv.DataAllocatedRatio, lv.MetadataAllocatedRatio);
                        desc = cockpit.format(_("$size $desc<br/>${percent}% full"), {
                            size: fmt_size(lv.Size),
                            desc: lvol_get_desc(lv),
                            percent: (ratio * 100).toFixed(0)
                        });
                        btn = create_block_action_btn(lv, false, false);
                        id = append_entry(level, null, desc, btn);
                        lv_obj = lv.getObject();
                        objs = me.client.getObjectsFrom(lv.VolumeGroup);
                        objs.sort(function(a, b) {
                            return a.objectPath.localeCompare(b.objectPath);
                        });
                        for (i = 0; i < objs.length; i++) {
                            llv = objs[i].lookup("com.redhat.Cockpit.Storage.LogicalVolume");
                            if (llv && llv.ThinPool == lv_obj.objectPath) {
                                append_logical_volume(level + 1, llv);
                            }
                        }
                    } else {
                        block = find_logical_volume_block(lv);
                        if (block) append_logical_volume_block(level, block, lv); else {
                            desc = cockpit.format(_("$size $desc<br/>($state)"), {
                                size: fmt_size(lv.Size),
                                desc: lvol_get_desc(lv),
                                state: lv.Active ? _("active, but unsupported") : _("inactive")
                            });
                            btn = create_block_action_btn(lv, false, false);
                            id = append_entry(level, null, desc, btn);
                        }
                    }
                }
                lvs = [];
                vg_obj = vg.getObject();
                objs = me.client.getObjectsFrom(vg_obj.objectPath);
                objs.sort(function(a, b) {
                    return a.objectPath.localeCompare(b.objectPath);
                });
                for (i = 0; i < objs.length; i++) {
                    lv = objs[i].lookup("com.redhat.Cockpit.Storage.LogicalVolume");
                    if (lv && lv.VolumeGroup == vg_obj.objectPath && lv.ThinPool == "/") {
                        append_logical_volume(level, lv);
                    }
                }
                if (vg.FreeSize > 0) {
                    desc = cockpit.format(_("$0 Free Space for Logical Volumes"), fmt_size(vg.FreeSize));
                    var btn = shell.action_btn(function(op) {
                        me.volume_group_action(op);
                    }, [ {
                        title: _("Create Plain Logical Volume"),
                        action: "create-plain",
                        is_default: true
                    }, {
                        title: _("Create RAID Logical Volume"),
                        action: "create-raid"
                    }, {
                        title: _("Create Pool for Thin Logical Volumes"),
                        action: "create-thin-pool"
                    } ], "storage-privileged");
                    id = append_entry(level, null, desc, btn);
                }
            }
            if (!block_or_vg) {
                $("#storage_detail_content").hide();
                return;
            }
            $("#storage_detail_content").show();
            var list = $("#storage_detail_partition_list");
            list.empty();
            if (block_or_vg._iface_name == "com.redhat.Cockpit.Storage.Block") {
                $("#storage_detail_content_title").text(_("Content"));
                append_device(0, block_or_vg);
            } else if (block_or_vg._iface_name == "com.redhat.Cockpit.Storage.VolumeGroup") {
                $("#storage_detail_content_title").text(_("Logical Volumes"));
                append_volume_group(0, block_or_vg);
            }
        },
        _updateDrive: function() {
            var val;
            var drive = this._drive;
            var blocks = get_block_devices_for_drive(drive);
            var block = blocks.length > 0 ? blocks[0] : undefined;
            this.watch_object(drive);
            this.watch_object(block);
            this._updateContent(block);
            if (drive.Vendor && drive.Vendor.length > 0) val = shell.esc(drive.Vendor) + " " + shell.esc(drive.Model); else val = shell.esc(drive.Model);
            $("#disk_detail_model").html(val);
            val = drive.Revision ? shell.esc(drive.Revision) : "—";
            $("#disk_detail_firmware_version").html(val);
            val = drive.Serial ? shell.esc(drive.Serial) : "—";
            $("#disk_detail_serial_number").html(val);
            val = drive.WWN ? shell.esc(drive.WWN) : "—";
            $("#disk_detail_world_wide_name").html(val);
            val = drive.Size > 0 ? fmt_size_long(drive.Size) : C_("disk-drive", "No Media Inserted");
            $("#disk_detail_capacity").html(val);
            if (drive.FailingValid) {
                if (drive.Failing) {
                    val = '<div class="cockpit-disk-failing">' + C_("disk-drive", "DISK IS FAILING") + "</div>";
                } else {
                    val = C_("disk-drive", "Disk is OK");
                }
            } else {
                val = "—";
            }
            if (drive.Temperature > 0) {
                val += " (" + format_temperature(drive.Temperature) + ")";
            }
            $("#disk_detail_assessment").html(val);
            val = "";
            for (var n = 0; n < blocks.length; n++) {
                var b = blocks[n];
                if (n > 0) {
                    val += " ";
                }
                val += shell.esc(b.Device);
            }
            $("#disk_detail_device_file").html(val);
            if (drive.Classification === "optical") $("#drive_format").hide(); else $("#drive_format").show();
            update_storage_privileged();
        },
        _updateMDRaid: function() {
            function format_level(str) {
                return {
                    raid0: _("RAID 0"),
                    raid1: _("RAID 1"),
                    raid4: _("RAID 4"),
                    raid5: _("RAID 5"),
                    raid6: _("RAID 6"),
                    raid10: _("RAID 10")
                }[str] || cockpit.format(_("RAID ($level)"), str);
            }
            var raid = this._mdraid;
            var block = find_block_device_for_mdraid(raid);
            this.watch_object(raid);
            this.watch_object(block);
            this._updateContent(block);
            if (block) $("#raid_detail_device").html(shell.esc(block.Device)); else $("#raid_detail_device").html("--");
            var val = raid.Size > 0 ? fmt_size_long(raid.Size) : "--";
            $("#raid_detail_capacity").html(val);
            $("#raid_detail_name").text(raid_get_desc(raid));
            $("#raid_detail_uuid").html(shell.esc(raid.UUID));
            var level = format_level(raid.Level);
            if (raid.NumDevices > 0) level += ", " + cockpit.format(_("$0 Disks"), raid.NumDevices);
            if (raid.ChunkSize > 0) level += ", " + cockpit.format(_("%0 Chunk Size"), fmt_size(raid.ChunkSize));
            $("#raid_detail_level").html(shell.esc(level));
            var state, action_state = "", is_running;
            var action, percent, rate, remaining;
            var degraded = null;
            var loc = raid.BitmapLocation;
            if (loc) {
                this.bitmap_onoff.set(loc != "none");
                $("#raid_detail_bitmap_row").show();
            } else {
                $("#raid_detail_bitmap_row").hide();
            }
            is_running = !!block;
            if (raid.Degraded > 0) {
                degraded = '<span style="color:red">' + _("ARRAY IS DEGRADED") + "</span> -- " + cockpit.format(_("%0 disks are missing"), raid.Degraded);
            }
            if (!raid.SyncAction) {
                if (block) {
                    state = _("Running");
                } else {
                    state = _("Not running");
                }
            } else {
                if (degraded) state = degraded; else state = _("Running");
                action = {
                    idle: "",
                    check: _("Data Scrubbing"),
                    repair: _("Data Scrubbing and Repair"),
                    resync: _("Resyncing"),
                    recover: _("Recovering "),
                    frozen: _("Frozen")
                }[raid.SyncAction] || raid.SyncAction;
                if (action && action != "idle") {
                    percent = Math.round(raid.SyncCompleted * 100).toString();
                    if (raid.SyncRate > 0) {
                        action_state = cockpit.format(_("$action, ${percent}% complete at $rate"), {
                            action: action,
                            percent: percent,
                            rate: fmt_size(raid.SyncRate) + "/s"
                        });
                    } else {
                        action_state = cockpit.format(_("$action, ${percent}% complete"), {
                            action: action,
                            percent: percent
                        });
                    }
                    state = state + "<br/>" + action_state;
                    if (raid.SyncRemainingTime > 0) {
                        remaining = cockpit.format(_("$0 remaining"), shell.format_delay(raid.SyncRemainingTime / 1e3));
                        state = state + "<br/>" + remaining;
                    }
                }
            }
            $("#raid_detail_state").html(state);
            shell.action_btn_select(this.raid_action_btn, is_running ? "stop" : "start");
            shell.action_btn_enable(this.raid_action_btn, "stop", is_running);
            shell.action_btn_enable(this.raid_action_btn, "start", !is_running);
            $("#raid-disks").closest(".panel").toggle(is_running);
            function render_state(slot, state) {
                if (state == "faulty") return '<span style="color:red">' + _("FAILED") + "</span>"; else if (state == "in_sync") return _("In Sync"); else if (state == "spare") return slot < 0 ? _("Spare") : _("Recovering"); else if (state == "write_mostly") return _("Write-mostly"); else if (state == "blocked") return _("Blocked"); else return cockpit.format(_("Unknown ($0)"), state);
            }
            $("#raid-disks-add").css("visibility", raid.Level === "raid0" ? "hidden" : "visible");
            var disks = $("#raid-disks");
            var info = this._mdraid.ActiveDevices || [];
            var i, j, slot, drive, states, state_html, num_errors;
            disks.empty();
            for (i = 0; i < info.length; i++) {
                slot = info[i][1];
                block = this.client.lookup(info[i][0], "com.redhat.Cockpit.Storage.Block");
                states = info[i][2];
                num_errors = info[i][3];
                if (!block) continue;
                state_html = "";
                for (j = 0; j < states.length; j++) {
                    if (j > 0) state_html += "<br/>";
                    state_html += render_state(slot, states[j]);
                }
                if (num_errors > 0) {
                    if (states.length > 0) state_html += "<br/>";
                    state_html += '<span style="color:red">' + cockpit.format(_("$0 Read Errors"), num_errors) + "</span>";
                }
                disks.append($('<li class="list-group-item">').append($('<table style="width:100%">').append($("<tr>").append($('<td style="width:20px;text-align:center">').text(slot < 0 ? "--" : slot), $("<td>").append(block_get_link_desc(block)), $('<td style="width:100px;text-align:right">').html(state_html), $('<td style="text-align:right">').append($("<button>", {
                    "class": "btn btn-default storage-privileged",
                    on: {
                        click: $.proxy(this, "raid_disk_remove", block)
                    }
                }).text(_("Remove")).css("visibility", raid.Level === "raid0" ? "hidden" : "visible"))))));
            }
            update_storage_privileged();
        },
        _updateVG: function() {
            var me = this;
            var vg = this._vg;
            var vg_obj = vg.getObject();
            var i, val, block, drive, objs, pvs_list, pvs, lvs_list, lvs, desc;
            this.watch_object(vg);
            if (vg.NeedsPolling) this.start_vg_polling(); else this.stop_vg_polling();
            val = vg.Size > 0 ? fmt_size_long(vg.Size) : "--";
            $("#vg_detail_capacity").html(val);
            $("#vg_detail_name").text(vg.Name);
            $("#vg_detail_uuid").text(vg.UUID);
            pvs_list = $("#vg-physical-volumes");
            pvs = [];
            objs = this.client.getObjectsFrom("/com/redhat/Cockpit/Storage/block_devices/");
            for (i = 0; i < objs.length; i++) {
                block = objs[i].lookup("com.redhat.Cockpit.Storage.Block");
                if (block && block.PvGroup == vg_obj.objectPath) {
                    pvs.push(block);
                }
            }
            pvs.sort(function(a, b) {
                return a.DeviceNumber - b.DeviceNumber;
            });
            function physical_action_func(block) {
                return function(op) {
                    me.physical_volume_action(block, op);
                };
            }
            var physical_action_spec = [ {
                title: _("Remove"),
                action: "remove",
                is_default: true
            }, {
                title: _("Empty"),
                action: "empty"
            } ];
            pvs_list.empty();
            for (i = 0; i < pvs.length; i++) {
                block = pvs[i];
                drive = block && block.PartitionNumber === 0 && this.client.lookup(block.Drive, "com.redhat.Cockpit.Storage.Drive");
                pvs_list.append($('<li class="list-group-item">').append($('<table style="width:100%">').append($("<tr>").append($("<td>").append(block_get_link_desc(block), $("<br>"), cockpit.format(_("$size, $free free"), {
                    size: fmt_size(block.PvSize),
                    free: fmt_size(block.PvFreeSize)
                })), $('<td style="text-align:right">').html(shell.action_btn(physical_action_func(block), physical_action_spec, "storage-privileged"))))));
            }
            this._updateContent(vg);
            update_storage_privileged();
        },
        action: function(op) {
            if (op == "format") this.format_disk(); else if (op == "delete") this.delete_raid(); else if (op == "start") this.start(); else if (op == "stop") this.stop(); else if (op == "start-scrub") this.start_scrub(); else if (op == "stop-scrub") this.stop_scrub(); else console.log("Unknown op %s", op);
        },
        block_action: function(target, op) {
            function is_lv(block_or_lv) {
                return block_or_lv._iface_name == "com.redhat.Cockpit.Storage.LogicalVolume" || block_or_lv.LogicalVolume != "/";
            }
            function get_lv(block_or_lv) {
                if (block_or_lv._iface_name == "com.redhat.Cockpit.Storage.LogicalVolume") return block_or_lv; else return block_or_lv._client.lookup(block_or_lv.LogicalVolume, "com.redhat.Cockpit.Storage.LogicalVolume");
            }
            if (op == "format") this.format(target); else if (op == "delete") {
                if (is_lv(target)) {
                    this.delete_logical_volume(get_lv(target));
                } else this.delete_partition(target);
            } else if (op == "mount") this.mount(target); else if (op == "unmount") this.unmount(target); else if (op == "lock") this.lock(target); else if (op == "unlock") this.unlock(target); else if (op == "fsys-options") this.fsys_options(target); else if (op == "crypto-options") this.crypto_options(target); else if (op == "resize") this.resize_logical_volume(get_lv(target)); else if (op == "rename") this.rename_logical_volume(get_lv(target)); else if (op == "activate") this.activate_logical_volume(get_lv(target)); else if (op == "deactivate") this.deactivate_logical_volume(get_lv(target)); else if (op == "create-thin-volume") this.create_thin_volume(get_lv(target)); else if (op == "create-snapshot") this.create_snapshot(get_lv(target)); else console.log("Unknown block op %s", op);
        },
        raid_action: function() {
            this.action(this.raid_op);
        },
        start: function() {
            this._mdraid.call("Start", function(error, result) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        stop: function() {
            this._mdraid.call("Stop", function(error, result) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        delete_raid: function() {
            var self = this;
            var location = cockpit.location;
            shell.confirm(cockpit.format(_("Please confirm deletion of $0"), raid_get_desc(this._mdraid)), _("Deleting a RAID Device will erase all data on it."), _("Delete")).done(function() {
                self._mdraid.call("Delete", function(error, result) {
                    if (error) shell.show_unexpected_error(error); else location.go("storage");
                });
            });
        },
        start_scrub: function() {
            this._mdraid.call("RequestSyncAction", "repair", function(error, result) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        stop_scrub: function() {
            this._mdraid.call("RequestSyncAction", "idle", function(error, result) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        format_disk: function(block) {
            PageFormatDisk.block = null;
            if (this._drive) PageFormatDisk.block = find_block_device_for_drive(this._drive); else if (this._mdraid) PageFormatDisk.block = find_block_device_for_mdraid(this._mdraid); else if (this._block) PageFormatDisk.block = this._block;
            if (PageFormatDisk.block) $("#storage_format_disk_dialog").modal("show");
        },
        format: function(block) {
            PageFormat.mode = "format";
            PageFormat.block = block;
            $("#storage_format_dialog").modal("show");
        },
        delete_partition: function(block) {
            shell.confirm(cockpit.format(_("Please confirm deletion of $0"), block.Device), _("Deleting a partition will delete all data in it."), _("Delete")).done(function() {
                block.call("DeletePartition", function(error) {
                    if (error) shell.show_unexpected_error(error);
                });
            });
        },
        create_partition: function(block, start, size, enable_dos_extended) {
            PageFormat.block = block;
            PageFormat.mode = "create-partition";
            PageFormat.start = start;
            PageFormat.size = size;
            PageFormat.enable_dos_extended = enable_dos_extended;
            $("#storage_format_dialog").modal("show");
        },
        mount: function(block) {
            block.call("Mount", function(error) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        unmount: function(block) {
            block.call("Unmount", function(error) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        lock: function(block) {
            block.call("Lock", function(error) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        unlock: function(block) {
            PageUnlock.block = block;
            $("#storage_unlock_dialog").modal("show");
        },
        fsys_options: function(block) {
            PageFilesystemOptions.block = block;
            $("#filesystem_options_dialog").modal("show");
        },
        crypto_options: function(block) {
            PageCryptoOptions.block = block;
            $("#crypto_options_dialog").modal("show");
        },
        raid_disk_remove: function(block) {
            this._mdraid.call("RemoveDevices", [ block.getObject().objectPath ], function(error) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        raid_disk_add: function() {
            PageRaidDiskAdd.mdraid = this._mdraid;
            $("#raid_disk_add_dialog").modal("show");
        },
        start_vg_polling: function() {
            var me = this;
            function poll() {
                if (me._vg) me._vg.call("Poll", function(error) {
                    if (error) console.log(error.message);
                });
            }
            if (!this.vg_polling_id) {
                poll();
                this.vg_polling_id = window.setInterval(poll, 5e3);
            }
        },
        stop_vg_polling: function() {
            if (this.vg_polling_id) {
                window.clearInterval(this.vg_polling_id);
                this.vg_polling_id = null;
            }
        },
        volume_group_action: function(op) {
            if (op == "delete") this.delete_volume_group(); else if (op == "rename") this.rename_volume_group(); else if (op == "create-plain") this.create_plain_volume(this._vg); else if (op == "create-thin-pool") this.create_thin_pool(this._vg); else if (op == "create-raid") this.create_raid_volume(this._vg); else console.log("Unknown volume group op %s", op);
        },
        delete_volume_group: function() {
            var self = this;
            var location = cockpit.location;
            shell.confirm(cockpit.format(_("Please confirm deletion of $0"), self._vg.Name), _("Deleting a volume group will erase all data on it."), _("Delete")).done(function() {
                self._vg.call("Delete", function(error, result) {
                    if (error) shell.show_unexpected_error(error); else location.go("storage");
                });
            });
        },
        physical_volume_action: function(target, op) {
            if (op == "remove") this.remove_physical_volume(target); else if (op == "empty") this.empty_physical_volume(target); else console.log("Unknown physical volume action %s", op);
        },
        remove_physical_volume: function(block) {
            if (block.PvFreeSize != block.PvSize) {
                shell.show_error_dialog("Error", "Volume is in use.");
                return;
            }
            var n = 0;
            var objs = this.client.getObjectsFrom("/com/redhat/Cockpit/Storage/block_devices/");
            for (var i = 0; i < objs.length; i++) {
                var b = objs[i].lookup("com.redhat.Cockpit.Storage.Block");
                if (b && b.PvGroup == this._vg.getObject().objectPath) {
                    n += 1;
                }
            }
            if (n == 1) {
                shell.show_error_dialog("Error", "Can't remove the last physical volume.");
                return;
            }
            this._vg.call("RemoveDevice", block.getObject().objectPath, function(error) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        empty_physical_volume: function(block) {
            var used = block.PvSize - block.PvFreeSize;
            if (used === 0) {
                shell.show_error_dialog("Dude", "Volume is already empty.");
                return;
            }
            if (used > this._vg.FreeSize) {
                shell.show_error_dialog("Error", "Not enough free space.");
                return;
            }
            this._vg.call("EmptyDevice", block.getObject().objectPath, function(error) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        add_physical_volume: function() {
            PageVGDiskAdd.volume_group = this._vg;
            $("#vg_disk_add_dialog").modal("show");
        },
        create_plain_volume: function(volume_group) {
            PageCreatePlainVolume.volume_group = volume_group;
            $("#storage_create_plain_volume_dialog").modal("show");
        },
        create_thin_pool: function(volume_group) {
            PageCreateThinPool.volume_group = volume_group;
            $("#storage_create_thin_pool_dialog").modal("show");
        },
        create_thin_volume: function(pool) {
            PageCreateThinVolume.pool = pool;
            $("#storage_create_thin_volume_dialog").modal("show");
        },
        create_raid_volume: function(volume_group) {
            shell.show_error_dialog("Sorry", "Not yet.");
        },
        create_snapshot: function(origin) {
            if (origin.Origin != "/") {
                shell.show_error_dialog("Error", "Can't take a snapshot of a snapshot.");
                return;
            }
            PageCreateSnapshot.origin = origin;
            $("#storage_create_snapshot_dialog").modal("show");
        },
        delete_logical_volume: function(lv) {
            var self = this;
            shell.confirm(cockpit.format(_("Please confirm deletion of $0"), self._vg.Name + "/" + lv.Name), _("Deleting a logical volume will erase all data in it."), _("Delete")).done(function() {
                lv.call("Delete", function(error, result) {
                    if (error) shell.show_unexpected_error(error);
                });
            });
        },
        resize_logical_volume: function(lv) {
            PageResizeVolume.volume = lv;
            $("#storage_resize_volume_dialog").modal("show");
        },
        rename_volume_group: function() {
            PageRenameGroup.group = this._vg;
            $("#storage_rename_group_dialog").modal("show");
        },
        rename_logical_volume: function(lv) {
            PageRenameVolume.volume = lv;
            $("#storage_rename_volume_dialog").modal("show");
        },
        activate_logical_volume: function(lv) {
            lv.call("Activate", function(error, result) {
                if (error) shell.show_unexpected_error(error);
            });
        },
        deactivate_logical_volume: function(lv) {
            lv.call("Deactivate", function(error, result) {
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageStorageDetail() {
        this._init();
    }
    shell.pages.push(new PageStorageDetail());
    PageCreateRaid.prototype = {
        _init: function() {
            this.id = "create-raid-dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#create-raid-create").on("click", $.proxy(this, "create"));
            $("#create-raid-level").on("change", $.proxy(this, "update"));
        },
        enter: function() {
            this.client = PageCreateRaid.client;
            this.blocks = fill_free_devices_list(this.client, "create-raid-drives", null);
            $("#create-raid-drives input").on("change", $.proxy(this, "update"));
            $("#create-raid-name").val("");
            $("#create-raid-level").selectpicker("val", "raid5");
            $("#create-raid-chunk").selectpicker("val", "512");
            $("#accounts-create-locked").prop("checked", false);
            this.update();
        },
        update: function() {
            var me = this;
            var n_disks, disk_size, raid_size, level, n_disks_needed;
            var n, b, i;
            var blocks = get_selected_devices_objpath($("#create-raid-drives"), me.blocks);
            n_disks = blocks.length;
            disk_size = Infinity;
            for (i = 0; i < blocks.length; i++) {
                b = this.client.lookup(blocks[i], "com.redhat.Cockpit.Storage.Block");
                if (b.Size < disk_size) disk_size = b.Size;
            }
            $("#create-raid-chunk").parents("tr").toggle($("#create-raid-level").val() !== "raid1");
            switch ($("#create-raid-level").val()) {
              case "raid0":
                n_disks_needed = 2;
                raid_size = disk_size * n_disks;
                break;

              case "raid1":
                n_disks_needed = 2;
                raid_size = disk_size;
                break;

              case "raid4":
                n_disks_needed = 2;
                raid_size = disk_size * (n_disks - 1);
                break;

              case "raid5":
                n_disks_needed = 2;
                raid_size = disk_size * (n_disks - 1);
                break;

              case "raid6":
                n_disks_needed = 4;
                raid_size = disk_size * (n_disks - 2);
                break;

              case "raid10":
                n_disks_needed = 2;
                var num_far_copies = 1;
                var num_near_copies = 2;
                raid_size = disk_size / num_far_copies;
                raid_size *= n_disks;
                raid_size /= num_near_copies;
                break;

              default:
                console.log("Unexpected RAID level %s", $("#create-raid-level").val());
                n_disks_needed = 0;
                raid_size = 0;
                break;
            }
            if (n_disks >= n_disks_needed) {
                $("#create-raid-summary-drives").text(cockpit.format(_("$n disks of $size each"), {
                    n: n_disks,
                    size: fmt_size(disk_size)
                }));
                $("#create-raid-summary-size").text(fmt_size(raid_size));
                $("#create-raid-create").prop("disabled", false);
            } else {
                $("#create-raid-summary-drives").text(cockpit.format(_("$0 more disks needed"), n_disks_needed - n_disks));
                $("#create-raid-summary-size").text("--");
                $("#create-raid-create").prop("disabled", true);
            }
        },
        create: function() {
            var me = this;
            var level = $("#create-raid-level").val();
            var chunk = level === "raid1" ? 0 : $("#create-raid-chunk").val();
            var name = $("#create-raid-name").val();
            var blocks = get_selected_devices_objpath($("#create-raid-drives"), me.blocks);
            var manager = this.client.lookup("/com/redhat/Cockpit/Storage/Manager", "com.redhat.Cockpit.Storage.Manager");
            manager.call("MDRaidCreate", blocks, level, name, chunk * 1024, function(error) {
                $("#create-raid-dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageCreateRaid() {
        this._init();
    }
    shell.dialogs.push(new PageCreateRaid());
    function fill_free_devices_list(client, id, filter) {
        var blocks;
        var element = $("#" + id);
        blocks = get_free_block_devices(client, filter);
        blocks.sort(function(a, b) {
            return a.DeviceNumber - b.DeviceNumber;
        });
        var list = $("<ul/>", {
            "class": "list-group available-disks-group"
        });
        for (var n = 0; n < blocks.length; n++) {
            var block = blocks[n];
            var desc = cockpit.format("$size $desc $dev", {
                size: fmt_size(block.Size),
                desc: block_get_short_desc(block),
                dev: block.Device
            });
            var id_n = id + "-" + n;
            list.append($("<li>", {
                "class": "list-group-item"
            }).append($("<div>", {
                "class": "checkbox",
                style: "margin:0px"
            }).append($("<input/>", {
                type: "checkbox",
                name: id_n,
                id: id_n,
                "data-index": n
            }), $("<label/>", {
                "for": id_n
            }).text(desc))));
        }
        element.html(list);
        return blocks;
    }
    function get_selected_devices_objpath(element, blocks) {
        var selected = [];
        element.find("input").each(function(i, e) {
            if (e.checked) {
                var n = $(e).attr("data-index");
                selected.push(blocks[n].getObject().objectPath);
            }
        });
        return selected;
    }
    PageCreateVolumeGroup.prototype = {
        _init: function() {
            this.id = "create-volume-group-dialog";
        },
        show: function() {
            if (this.blocks.length > 0) {
                $("#create-vg-name").prop("disabled", false);
                $("#create-vg-name").focus();
            } else {
                $("#create-vg-name").prop("disabled", true);
            }
        },
        leave: function() {},
        setup: function() {
            $("#create-vg-create").on("click", $.proxy(this, "create"));
        },
        enter: function() {
            var disk_count = -1;
            this.client = PageCreateVolumeGroup.client;
            this.blocks = fill_free_devices_list(this.client, "create-vg-drives", null);
            $("#disks-not-found .close").on("click", function() {
                $("#disks-not-found").hide();
            });
            $("#create-vg-name").on("input", change_disk_count);
            $("#create-vg-name").on("input change focus", check_vg_condition);
            $('#create-vg-drives [type = "checkbox"]').on("change", change_checkbox_count);
            $('#create-vg-drives [type = "checkbox"]').on("click change", check_vg_condition);
            $("#create-vg-name").val("");
            $("#create-vg-create").prop("disabled", true);
            $("#create-vg-drives").prop("checked", false);
            control_warning(this.blocks);
            function check_vg_condition() {
                hide_error_message("#create-vg-error");
                if (check_vg_input()) {
                    if (check_checked_box()) {
                        $("#create-vg-create").prop("disabled", false);
                        hide_error_message("#create-vg-error");
                    } else {
                        if (disk_count === -1) return;
                        $("#create-vg-create").prop("disabled", true);
                        highlight_error_message("#create-vg-error", _("At least one disk needed."));
                    }
                } else {
                    $("#create-vg-create").prop("disabled", true);
                }
            }
            function change_disk_count() {
                if (disk_count === 0) disk_count = -1;
            }
            function check_vg_input() {
                function check_input(input, exclusive_pattern) {
                    var match_array = input.match(exclusive_pattern);
                    if (match_array) {
                        var last_subarray = match_array[match_array.length - 1];
                        return last_subarray[last_subarray.length - 1];
                    } else {
                        return null;
                    }
                }
                var addr = $("#create-vg-name").val();
                if (addr === "") {
                    hide_error("#creat-vg-name-cell");
                    hide_error_message("#create-vg-error");
                    return false;
                } else if (addr.length > 127) {
                    highlight_error("#creat-vg-name-cell");
                    highlight_error_message("#create-vg-error", _("Name length cannot exceed 127 characters."));
                    return false;
                } else {
                    var pattern = new RegExp("[^a-zA-Z0-9+._-]+", "g");
                    var match = check_input(addr, pattern);
                    if (!match) {
                        hide_error("#creat-vg-name-cell");
                        hide_error_message("#create-vg-error");
                        return true;
                    } else {
                        highlight_error("#creat-vg-name-cell");
                        if (match.search(/\s+/) === -1) highlight_error_message("#create-vg-error", _("Name cannot contain '" + match + "'.")); else highlight_error_message("#create-vg-error", _("Name cannot contain whitespace."));
                        return false;
                    }
                }
            }
            function count_checked_box(selector) {
                var group = $(selector);
                var count = 0;
                for (var i = 0; i < group.length; i++) {
                    if (group[i].checked) count++;
                }
                return count;
            }
            function change_checkbox_count() {
                disk_count = count_checked_box();
            }
            function check_checked_box() {
                var count = count_checked_box('#create-vg-drives [type = "checkbox"]');
                if (count > 0) return true; else return false;
            }
            function control_warning(blocks) {
                if (blocks.length > 0) {
                    $("#disks-not-found span.alert-message").text("");
                    $("#disks-not-found").hide();
                } else {
                    $("#disks-not-found span.alert-message").text(_("No available disks"));
                    $("#disks-not-found").show();
                }
            }
        },
        create: function() {
            var me = this;
            var name = $("#create-vg-name").val();
            if (name.trim() === "") {
                highlight_error("#creat-vg-name-cell");
                highlight_error_message("#create-vg-error", _("Name cannot contain whitespace."));
                return;
            }
            var blocks = get_selected_devices_objpath($("#create-vg-drives"), me.blocks);
            var manager = me.client.lookup("/com/redhat/Cockpit/Storage/Manager", "com.redhat.Cockpit.Storage.Manager");
            manager.call("VolumeGroupCreate", name, blocks, function(error) {
                $("#create-volume-group-dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageCreateVolumeGroup() {
        this._init();
    }
    shell.dialogs.push(new PageCreateVolumeGroup());
    PageFormatDisk.prototype = {
        _init: function() {
            this.id = "storage_format_disk_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#format-disk-format").on("click", $.proxy(this, "format"));
        },
        enter: function() {
            $("#format-disk-title").text(cockpit.format(_("Format Disk $0"), PageFormatDisk.block.Device));
            $("#format-disk-type").selectpicker("val", "gpt");
            $("#format-disk-erase").selectpicker("val", "no");
        },
        format: function() {
            PageFormatDisk.block.call("Format", $("#format-disk-type").val(), $("#format-disk-erase").val(), "", "", "", "", "", "", function(error) {
                $("#storage_format_disk_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageFormatDisk() {
        this._init();
    }
    shell.dialogs.push(new PageFormatDisk());
    PageFormat.prototype = {
        _init: function() {
            this.id = "storage_format_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#format-format").on("click", $.proxy(this, "format"));
            $("#format-type").on("change", $.proxy(this, "update"));
            $("#format-custom").on("keyup change", $.proxy(this, "update"));
            $("#format-passphrase").on("keyup change", $.proxy(this, "update"));
            $("#format-passphrase-2").on("keyup change", $.proxy(this, "update"));
            $("#format-mounting").on("change", $.proxy(this, "update"));
            $("#format-mount-point").on("keyup change", $.proxy(this, "update"));
        },
        enter: function() {
            $("#format-size-row").toggle(PageFormat.mode == "create-partition");
            function enable_dos_extended(flag) {
                $('#format-type option[value="dos-extended"]').toggle(flag);
                $("#format-type").selectpicker("refresh");
            }
            if (PageFormat.mode == "format") {
                $("#format-title").text(cockpit.format(_("Format $0"), PageFormat.block.Device));
                $("#format-warning").text(_("Formatting a storage device will erase all data on it."));
                $("#format-format").text(_("Format"));
                $("#format-format").addClass("btn-danger").removeClass("btn-primary");
                $("#format-mounting").selectpicker("val", PageFormat.block.MountPoint ? "custom" : "default");
                $("#format-mount-point").val(PageFormat.block.MountPoint);
                $("#format-mount-options").val(PageFormat.block.MountOptions);
                enable_dos_extended(false);
            } else {
                $("#format-title").text(cockpit.format(_("Create Partition on $0"), PageFormat.block.Device));
                $("#format-warning").text("");
                $("#format-format").text(_("Create partition"));
                $("#format-format").addClass("btn-primary").removeClass("btn-danger");
                $("#format-mounting").selectpicker("val", "default");
                $("#format-mount-point").val("");
                $("#format-mount-options").val("");
                enable_dos_extended(PageFormat.enable_dos_extended);
            }
            $("#format-size").val("");
            $("#format-erase").selectpicker("val", "no");
            $("#format-type").selectpicker("val", "xfs");
            $("#format-name").val("");
            $("#format-custom").val("");
            $("#format-crpyto-options").val("");
            $("#format-passphrase").val("");
            $("#format-passphrase-2").val("");
            $("#format-store-passphrase").prop("checked", false);
            this.update();
        },
        update: function() {
            var type = $("#format-type").val();
            var isFS = type != "empty" && type != "dos-extended";
            var isLuks = type == "luks+xfs" || type == "luks+ext4";
            var isDefaultMount = !isFS || $("#format-mounting").val() == "default";
            $("#format-custom-row").toggle(type == "custom");
            $("#format-name-row").toggle(isFS);
            $("#format-passphrase-row, #format-passphrase-row-2, #format-store-passphrase-row, #format-crypto-options-row").toggle(isLuks);
            $("#format-mounting-row").toggle(isFS);
            $("#format-mount-point-row, #format-mount-options-row").toggle(!isDefaultMount);
            if (type == "custom" && !$("#format-custom").val() || isLuks && (!$("#format-passphrase").val() || $("#format-passphrase").val() != $("#format-passphrase-2").val()) || !isDefaultMount && !$("#format-mount-point").val()) {
                $("#format-format").prop("disabled", true);
            } else {
                $("#format-format").prop("disabled", false);
            }
        },
        format: function() {
            var size = $("#format-size").val();
            if (!size) size = PageFormat.size; else size = size * 1024 * 1024;
            var type = $("#format-type").val();
            var isLuks = type == "luks+xfs" || type == "luks+ext4";
            if (type == "custom") type = $("#format-custom").val();
            var erase = $("#format-erase").val();
            var label = $("#format-name").val();
            var isFS = type != "empty" && type != "dos-extended";
            var isDefaultMount = !isFS || $("#format-mounting").val() == "default";
            var passphrase = "";
            var stored_passphrase = "";
            if (isLuks) {
                passphrase = $("#format-passphrase").val();
                if (type == "luks+ext4") type = "ext4"; else if (type == "luks+xfs") type = "xfs"; else throw new Error("Unhandled filesystem type " + type);
                if ($("#format-store-passphrase").prop("checked")) stored_passphrase = passphrase;
            }
            var mount_point = "";
            var mount_options = "";
            if (!isDefaultMount) {
                mount_point = $("#format-mount-point").val();
                mount_options = $("#format-mount-options").val();
            }
            var crypto_options = $("#format-crypto-options").val();
            if (PageFormat.mode == "create-partition") PageFormat.block.call("CreatePartition", PageFormat.start, size, type, erase, label, passphrase, mount_point, mount_options, stored_passphrase, crypto_options, function(error) {
                $("#storage_format_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            }); else PageFormat.block.call("Format", type, erase, label, passphrase, mount_point, mount_options, stored_passphrase, crypto_options, function(error) {
                $("#storage_format_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageFormat() {
        this._init();
    }
    shell.dialogs.push(new PageFormat());
    PageCreatePlainVolume.prototype = {
        _init: function() {
            this.id = "storage_create_plain_volume_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#create-pvol-create").on("click", $.proxy(this, "create"));
        },
        enter: function() {
            $("#create-pvol-name").val("");
            $("#create-pvol-size").val("");
        },
        create: function() {
            var size = $("#create-pvol-size").val();
            var name = $("#create-pvol-name").val();
            size = size * 1024 * 1024;
            PageCreatePlainVolume.volume_group.call("CreatePlainVolume", name, size, function(error) {
                $("#storage_create_plain_volume_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageCreatePlainVolume() {
        this._init();
    }
    shell.dialogs.push(new PageCreatePlainVolume());
    PageCreateThinPool.prototype = {
        _init: function() {
            this.id = "storage_create_thin_pool_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#create-tpool-create").on("click", $.proxy(this, "create"));
        },
        enter: function() {
            $("#create-tpool-name").val("");
            $("#create-tpool-size").val("");
        },
        create: function() {
            var size = $("#create-tpool-size").val();
            var name = $("#create-tpool-name").val();
            size = size * 1024 * 1024;
            PageCreateThinPool.volume_group.call("CreateThinPoolVolume", name, size, function(error) {
                $("#storage_create_thin_pool_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageCreateThinPool() {
        this._init();
    }
    shell.dialogs.push(new PageCreateThinPool());
    PageCreateThinVolume.prototype = {
        _init: function() {
            this.id = "storage_create_thin_volume_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#create-tvol-create").on("click", $.proxy(this, "create"));
        },
        enter: function() {
            $("#create-tvol-name").val("");
            $("#create-tvol-size").val("");
        },
        create: function() {
            var size = $("#create-tvol-size").val();
            var name = $("#create-tvol-name").val();
            size = size * 1024 * 1024;
            var vg = PageCreateThinVolume.pool._client.lookup(PageCreateThinVolume.pool.VolumeGroup, "com.redhat.Cockpit.Storage.VolumeGroup");
            vg.call("CreateThinVolume", name, size, PageCreateThinVolume.pool.getObject().objectPath, function(error) {
                $("#storage_create_thin_volume_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageCreateThinVolume() {
        this._init();
    }
    shell.dialogs.push(new PageCreateThinVolume());
    PageCreateSnapshot.prototype = {
        _init: function() {
            this.id = "storage_create_snapshot_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#create-svol-create").on("click", $.proxy(this, "create"));
        },
        enter: function() {
            $("#create-svol-name").val("");
            $("#create-svol-size").val("");
            $("#create-svol-size-row").toggle(PageCreateSnapshot.origin.ThinPool == "/");
        },
        create: function() {
            var name = $("#create-svol-name").val();
            var size = $("#create-svol-size").val();
            size = size * 1024 * 1024;
            if (PageCreateSnapshot.origin.ThinPool != "/") size = 0;
            PageCreateSnapshot.origin.call("CreateSnapshot", name, size, function(error) {
                $("#storage_create_snapshot_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageCreateSnapshot() {
        this._init();
    }
    shell.dialogs.push(new PageCreateSnapshot());
    PageResizeVolume.prototype = {
        _init: function() {
            this.id = "storage_resize_volume_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#resize-lvol-resize").on("click", $.proxy(this, "resize"));
        },
        enter: function() {
            var block = find_logical_volume_block(PageResizeVolume.volume);
            $("#resize-lvol-size").val((PageResizeVolume.volume.Size / (1024 * 1024)).toFixed(0));
            $("#resize-lvol-resize-fsys").prop("checked", block && block.IdUsage == "filesystem");
            $("#resize-lvol-resize-fsys").parents("tr").toggle(PageResizeVolume.volume.Type == "block");
        },
        resize: function() {
            var size = $("#resize-lvol-size").val();
            var resize_fsys = $("#resize-lvol-resize-fsys").prop("checked");
            size = size * 1024 * 1024;
            PageResizeVolume.volume.call("Resize", size, {
                resize_fsys: resize_fsys
            }, function(error) {
                $("#storage_resize_volume_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageResizeVolume() {
        this._init();
    }
    shell.dialogs.push(new PageResizeVolume());
    PageRenameVolume.prototype = {
        _init: function() {
            this.id = "storage_rename_volume_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#rename-lvol-rename").on("click", $.proxy(this, "rename"));
        },
        enter: function() {
            $("#rename-lvol-name").val(PageRenameVolume.volume.Name);
        },
        rename: function() {
            var name = $("#rename-lvol-name").val();
            PageRenameVolume.volume.call("Rename", name, function(error) {
                $("#storage_rename_volume_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageRenameVolume() {
        this._init();
    }
    shell.dialogs.push(new PageRenameVolume());
    PageRenameGroup.prototype = {
        _init: function() {
            this.id = "storage_rename_group_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#rename-vg-rename").on("click", $.proxy(this, "rename"));
        },
        enter: function() {
            $("#rename-vg-name").val(PageRenameGroup.group.Name);
        },
        rename: function() {
            var name = $("#rename-vg-name").val();
            PageRenameGroup.group.call("Rename", name, function(error) {
                $("#storage_rename_group_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error); else cockpit.location.go("storage-detail", {
                    type: "vg",
                    id: name
                });
            });
        }
    };
    function PageRenameGroup() {
        this._init();
    }
    shell.dialogs.push(new PageRenameGroup());
    PageFilesystemOptions.prototype = {
        _init: function() {
            this.id = "filesystem_options_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#fsysopts-apply").on("click", $.proxy(this, "apply"));
            $("#fsysopts-mounting").on("change", $.proxy(this, "update"));
            $("#fsysopts-mount-point").on("change keyup", $.proxy(this, "update"));
        },
        enter: function() {
            $("#fsysopts-name").val(PageFilesystemOptions.block.IdLabel);
            $("#fsysopts-mounting").selectpicker("val", PageFilesystemOptions.block.MountPoint ? "custom" : "default");
            $("#fsysopts-mount-point").val(PageFilesystemOptions.block.MountPoint);
            $("#fsysopts-mount-options").val(PageFilesystemOptions.block.MountOptions);
            this.update();
        },
        update: function() {
            var isDefaultMount = $("#fsysopts-mounting").val() == "default";
            $("#fsysopts-mount-point-row, #fsysopts-mount-options-row").toggle(!isDefaultMount);
            $("#fsysopts-apply").prop("disabled", !isDefaultMount && !$("#fsysopts-mount-point").val());
        },
        apply: function() {
            var name = $("#fsysopts-name").val();
            var isDefaultMount = $("#fsysopts-mounting").val() == "default";
            var mount_point = "";
            var mount_options = "";
            if (!isDefaultMount) {
                mount_point = $("#fsysopts-mount-point").val();
                mount_options = $("#fsysopts-mount-options").val();
            }
            PageFilesystemOptions.block.call("SetFilesystemOptions", name, mount_point, mount_options, function(error) {
                $("#filesystem_options_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageFilesystemOptions() {
        this._init();
    }
    shell.dialogs.push(new PageFilesystemOptions());
    PageCryptoOptions.prototype = {
        _init: function() {
            this.id = "crypto_options_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#crypto-options-apply").on("click", $.proxy(this, "apply"));
        },
        enter: function() {
            $("#crypto-options-passphrase").val("");
            $("#crypto-options-options").val(PageCryptoOptions.block.CryptoOptions);
            PageCryptoOptions.block.call("GetCryptoPassphrase", function(error, result) {
                if (result) $("#crypto-options-passphrase").val(result);
            });
        },
        apply: function() {
            var passphrase = $("#crypto-options-passphrase").val();
            var options = $("#crypto-options-options").val();
            PageCryptoOptions.block.call("SetCryptoOptions", passphrase, options, function(error) {
                $("#crypto_options_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageCryptoOptions() {
        this._init();
    }
    shell.dialogs.push(new PageCryptoOptions());
    PageUnlock.prototype = {
        _init: function() {
            this.id = "storage_unlock_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#unlock-unlock").on("click", $.proxy(this, "unlock"));
        },
        enter: function() {
            $("#unlock-passphrase").val("");
        },
        unlock: function() {
            var passphrase = $("#unlock-passphrase").val();
            PageUnlock.block.call("Unlock", passphrase, function(error) {
                $("#storage_unlock_dialog").modal("hide");
                if (error) shell.unexpected_error(error);
            });
        }
    };
    function PageUnlock() {
        this._init();
    }
    shell.dialogs.push(new PageUnlock());
    PageRaidDiskAdd.prototype = {
        _init: function() {
            this.id = "raid_disk_add_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#raid-disk-add-add").on("click", $.proxy(this, "add"));
        },
        enter: function() {
            function is_us(b) {
                var r = b._client.lookup(b.MDRaid, "com.redhat.Cockpit.Storage.MDRaid");
                return b.MDRaid == PageRaidDiskAdd.mdraid.getObject().objectPath;
            }
            this.blocks = fill_free_devices_list(PageRaidDiskAdd.mdraid._client, "raid-disk-add-drives", is_us);
            $("#raid-disk-add-drives input").on("change", $.proxy(this, "update"));
            this.update();
        },
        update: function() {
            var n_disks = get_selected_devices_objpath($("#raid-disk-add-drives"), this.blocks).length;
            $("#raid-disk-add-add").prop("disabled", n_disks === 0);
        },
        add: function() {
            var me = this;
            var blocks = get_selected_devices_objpath($("#raid-disk-add-drives"), this.blocks);
            PageRaidDiskAdd.mdraid.call("AddDevices", blocks, function(error) {
                $("#raid_disk_add_dialog").modal("hide");
                if (error) shell.show_unexpected_error(error);
            });
        }
    };
    function PageRaidDiskAdd() {
        this._init();
    }
    shell.dialogs.push(new PageRaidDiskAdd());
    PageVGDiskAdd.prototype = {
        _init: function() {
            this.id = "vg_disk_add_dialog";
        },
        show: function() {},
        leave: function() {},
        setup: function() {
            $("#vg-disk-add-add").on("click", $.proxy(this, "add"));
            $("#vg-disks-not-found .close").on("click", function() {
                $("#vg-disks-not-found").toggleClass("hide", true);
            });
        },
        enter: function() {
            function is_ours(b) {
                var lv = b._client.lookup(b.LogicalVolume, "com.redhat.Cockpit.Storage.LogicalVolume");
                return lv && lv.VolumeGroup == PageVGDiskAdd.volume_group.getObject().objectPath;
            }
            this.blocks = fill_free_devices_list(PageVGDiskAdd.volume_group._client, "vg-disk-add-drives", is_ours);
            var show = this.blocks.length > 0;
            var msg = show ? "" : _("No available disks");
            $("#vg-disks-not-found span.alert-message").text(msg);
            $("#vg-disks-not-found").toggleClass("hide", show);
            $("#vg-disk-add-drives input").on("change", $.proxy(this, "update"));
            this.update();
        },
        update: function() {
            var n_disks = get_selected_devices_objpath($("#vg-disk-add-drives"), this.blocks).length;
            $("#vg-disk-add-add").prop("disabled", n_disks === 0);
        },
        add: function() {
            var me = this;
            var blocks = get_selected_devices_objpath($("#vg-disk-add-drives"), this.blocks);
            function add_them(i) {
                if (i < blocks.length) PageVGDiskAdd.volume_group.call("AddDevice", blocks[i], function(error) {
                    if (error) {
                        $("#vg_disk_add_dialog").modal("hide");
                        shell.show_unexpected_error(error);
                    } else {
                        add_them(i + 1);
                    }
                }); else $("#vg_disk_add_dialog").modal("hide");
            }
            add_them(0);
        }
    };
    function PageVGDiskAdd() {
        this._init();
    }
    shell.dialogs.push(new PageVGDiskAdd());
})(jQuery, cockpit, shell, modules);

var shell = shell || {};

(function($, cockpit, shell) {
    shell.realms_op_set_parameters = function realms_op_set_parameters(manager, op, realm, details) {
        PageRealmsOp.manager = manager;
        PageRealmsOp.op = op;
        PageRealmsOp.realm = realm;
        PageRealmsOp.details = details;
    };
    PageRealmsOp.prototype = {
        _init: function() {
            this.id = "realms-op";
        },
        show: function() {},
        setup: function() {
            var self = this;
            $("#realms-op-apply").on("click", function(e) {
                self.apply();
            });
            $("#realms-op-cancel").on("click", function(e) {
                self.cancel();
            });
            $("#realms-op-address").on("keyup", function(e) {
                self.maybe_check_realm();
            });
            $("#realms-op-address").on("change", function(e) {
                self.maybe_check_realm();
            });
            $(".realms-op-field").on("keydown", function(e) {
                if (e.which == 13) self.apply();
            });
            $("#realms-op-auth").on("change", function(e) {
                self.update_cred_fields();
            });
            $("#realms-op-software").on("change", function(e) {
                self.update_auth_methods();
            });
        },
        enter: function() {
            var me = this;
            $("#realms-op-diagnostics").hide();
            me.realm_manager = PageRealmsOp.manager;
            me.op = PageRealmsOp.op;
            me.realm = PageRealmsOp.realm;
            me.given_details = PageRealmsOp.details;
            $(me.realm_manager).on("notify:Busy.realms-op", function() {
                me.update_busy();
            });
            if (me.op == "join") {
                me.never_show_software_choice = 1;
                me.title = C_("page-title", "Join a Domain");
                $("#realms-op-apply").text(_("Join"));
                $(".realms-op-join-only-row").show();
            } else if (me.op == "leave") {
                me.never_show_software_choice = 1;
                me.title = C_("page-title", "Leave Domain");
                $("#realms-op-apply").text(_("Leave"));
                $(".realms-op-join-only-row").hide();
            } else {
                $("#realms-op").modal("hide");
                return;
            }
            $("#realms-op-title").empty().append(me.title);
            $("#realms-op-spinner").hide();
            $("#realms-op-wait-message").hide();
            $("#realms-op-address-spinner").hide();
            $("#realms-op-address-error").hide();
            $("#realms-op-error").empty();
            $("#realms-op-diagnostics").empty();
            $(".realms-op-field").val("");
            me.checking = 0;
            me.checked = "";
            me.discovered_details = [];
            if (me.op == "join") {
                me.check_default_realm();
                me.maybe_check_realm();
                me.update_discovered_details();
            } else {
                me.update_auth_methods();
            }
            me.update_busy();
        },
        leave: function() {},
        update_discovered_details: function() {
            var me = this;
            var sel = $("#realms-op-software");
            sel.empty();
            for (var i = 0; i < me.discovered_details.length; i++) {
                var d = me.discovered_details[i];
                var txt = d["client-software"] + " / " + d["server-software"];
                sel.append('<option value="' + i + '">' + shell.esc(txt) + "</option>");
            }
            me.update_auth_methods();
            if (me.never_show_software_choice || me.discovered_details.length < 2) {
                $("#realms-op-software-row").hide();
            } else {
                $("#realms-op-software-row").show();
            }
        },
        update_auth_methods: function() {
            var me = this;
            var m = {};
            function add_from_details(d) {
                if (d) {
                    var c = d["supported-join-credentials"];
                    if (c) {
                        for (var i = 0; i < c.length; i++) m[c[i]] = 1;
                    }
                }
            }
            if (me.op == "leave") {
                add_from_details(me.given_details);
            } else if (me.never_show_software_choice) {
                for (var i = 0; i < me.discovered_details.length; i++) add_from_details(me.discovered_details[i]);
            } else {
                var s = $("#realms-op-software").val();
                add_from_details(me.discovered_details[s]);
            }
            var have_one = 0;
            var sel = $("#realms-op-auth");
            function add_choice(tag, text) {
                if (tag in m) {
                    sel.append('<option value="' + tag + '">' + shell.esc(text) + "</option>");
                    have_one = 1;
                }
            }
            sel.empty();
            add_choice("admin", _("Administrator Password"));
            add_choice("user", _("User Password"));
            add_choice("otp", _("One Time Password"));
            add_choice("none", _("Automatic"));
            if (!have_one) sel.append('<option value="admin">' + _("Administrator Password") + "</option>");
            if ($('[data-id="realms-op-auth"]').length <= 1) $("#realms-authentification-row").hide(); else $("#realms-authentification-row").show();
            $("#realms-op-auth").selectpicker();
            me.update_cred_fields();
        },
        update_cred_fields: function() {
            var me = this;
            var a = $("#realms-op-auth").val();
            $("#realms-op-admin-row").hide();
            $("#realms-op-admin-password-row").hide();
            $("#realms-op-user-row").hide();
            $("#realms-op-user-password-row").hide();
            $("#realms-op-otp-row").hide();
            if (a == "admin") {
                $("#realms-op-admin-row").show();
                $("#realms-op-admin-password-row").show();
                var admin;
                if (me.op == "join") {
                    var s = $("#realms-op-software").val();
                    var d = s && me.discovered_details[s];
                    admin = d && d["suggested-administrator"];
                } else {
                    admin = me.given_details["suggested-administrator"];
                }
                if (admin && !$("#realms-op-admin").val()) $("#realms-op-admin")[0].placeholder = _('e.g. "' + admin + '"');
            } else if (a == "user") {
                $("#realms-op-user-row").show();
                $("#realms-op-user-password-row").show();
            } else if (a == "otp") {
                $("#realms-op-otp-row").show();
            }
        },
        update_busy: function() {
            var me = this;
            var busy = me.realm_manager.Busy;
            if (busy && busy[0]) {
                $("#realms-op-spinner").show();
                $("#realms-op-wait-message").show();
                $(".realms-op-field").prop("disabled", true);
                $("#realms-op-apply").prop("disabled", true);
                $("#realms-op-software").prop("disabled", true);
                $('[data-id="realms-op-auth"]').prop("disabled", true);
            } else {
                $("#realms-op-spinner").hide();
                $("#realms-op-wait-message").hide();
                $(".realms-op-field").prop("disabled", false);
                $('[data-id="realms-op-auth"]').prop("disabled", false);
                $("#realms-op-apply").prop("disabled", false);
                $("#realms-op-software").prop("disabled", false);
            }
        },
        check_default_realm: function() {
            var me = this;
            me.realm_manager.call("Discover", "", {}, function(error, result, details) {
                if (result) {
                    $("#realms-op-address")[0].placeholder = cockpit.format(_("e.g. $0"), result);
                }
            });
        },
        maybe_check_realm: function() {
            var me = this;
            if ($("#realms-op-address").val() != me.checked) {
                $("#realms-op-address-error").hide();
                if (me.timeout) window.clearTimeout(me.timeout);
                me.timeout = window.setTimeout(function() {
                    me.check_realm();
                }, 1e3);
            }
        },
        check_realm: function() {
            var me = this;
            var name = $("#realms-op-address").val();
            if (me.checking || !name || me.checked == name) {
                return;
            }
            $("#realms-op-address-spinner").show();
            me.checking = 1;
            me.checked = name;
            me.realm_manager.call("Discover", name, {}, function(error, result, details) {
                if ($("#realms-op-address").val() != me.checked) {
                    me.checking = 0;
                    me.check_realm();
                } else {
                    $("#realms-op-address-spinner").hide();
                    me.checking = 0;
                    me.discovered_details = [];
                    if (error) $("#realms-op-error").empty().append(error.message); else if (!result) {
                        $("#realms-op-address-error").show();
                        $("#realms-op-address-error").attr("title", cockpit.format(_("Domain $0 could not be contacted"), shell.esc(name)));
                    } else {
                        me.discovered_details = details;
                    }
                    me.update_discovered_details();
                }
            });
        },
        apply: function() {
            var me = this;
            function handle_op_result(error, result) {
                me.working = false;
                if (error && error.name != "com.redhat.Cockpit.Error.Cancelled") {
                    $("#realms-op-error").empty().append(error.message);
                    $("#realms-op-error").append(' <button id="realms-op-more-diagnostics" data-inline="true">' + _("More") + "</button>");
                    $("#realms-op-more-diagnostics").click(function(e) {
                        me.realm_manager.call("GetDiagnostics", function(error, result) {
                            $("#realms-op-more-diagnostics").hide();
                            $("#realms-op-diagnostics").show();
                            $("#realms-op-diagnostics").empty().append(result);
                        });
                    });
                } else {
                    $("#realms-op").modal("hide");
                }
            }
            $("#realms-op-error").empty();
            $("#realms-op-diagnostics").empty();
            var a = $("#realms-op-auth").val();
            var creds;
            if (a == "user") creds = [ "user", $("#realms-op-user").val(), $("#realms-op-user-password").val() ]; else if (a == "admin") creds = [ "admin", $("#realms-op-admin").val(), $("#realms-op-admin-password").val() ]; else if (a == "otp") creds = [ "otp", "", $("#realms-op-ot-password").val() ]; else creds = [ "none", "", "" ];
            var options;
            if (me.op == "join") {
                var details;
                if (me.never_show_software_choice) details = {}; else {
                    var s = $("#realms-join-software").val();
                    details = me.discovered_details[s];
                }
                options = {
                    "computer-ou": $("#realms-join-computer-ou").val()
                };
                if (details["client-software"]) options["client-software"] = details["client-software"];
                if (details["server-software"]) options["server-software"] = details["server-software"];
                me.working = true;
                me.realm_manager.call("Join", $("#realms-op-address").val(), creds, options, handle_op_result);
            } else if (me.op == "leave") {
                options = {
                    "server-software": me.given_details["server-software"],
                    "client-software": me.given_details["client-software"]
                };
                me.working = true;
                me.realm_manager.call("Leave", me.realm, creds, options, handle_op_result);
            }
        },
        cancel: function() {
            var me = this;
            if (me.working) {
                me.realm_manager.call("Cancel", function(error, result) {});
            } else {
                $("#realms-op").modal("hide");
            }
            $("#realms-op-spinner").hide();
            $("#realms-op-wait-message").hide();
        }
    };
    function PageRealmsOp() {
        this._init();
    }
    shell.dialogs.push(new PageRealmsOp());
})(jQuery, cockpit, shell);

var shell = shell || {};

var modules = modules || {};

(function($, cockpit, shell, modules) {
    function resource_debug() {
        if (window.debugging == "all" || window.debugging == "resource" || true) console.debug.apply(console, arguments);
    }
    function format_cpu_usage(usage) {
        if (usage === undefined || isNaN(usage)) return "";
        return Math.round(usage) + "%";
    }
    function systemd_unit_name_esc(str) {
        var validchars = /[0-9a-zA-Z:-_.\\]/;
        var res = "";
        var i;
        for (i = 0; i < str.length; i++) {
            var c = str[i];
            if (c == "/") res += "-"; else if (c == "-" || c == "\\" || !validchars.test(c)) {
                res += "\\x";
                var h = c.charCodeAt(0).toString(16);
                while (h.length < 2) h = "0" + h;
                res += h;
            } else res += c;
        }
        return res;
    }
    function systemd_kill_slashes(str) {
        str = str.replace(/\/+/g, "/");
        if (str.length > 1) str = str.replace(/\/$/, "").replace(/^\//, "");
        return str;
    }
    function systemd_path_esc(str) {
        str = systemd_kill_slashes(str);
        if (str == "/") return "-"; else return systemd_unit_name_esc(str);
    }
    function systemd_param_esc(str) {
        if (str.length > 0 && str[0] == "/") return systemd_path_esc(str); else return systemd_unit_name_esc(str);
    }
    N_("loaded");
    N_("error");
    N_("masked");
    N_("not-found");
    N_("active");
    N_("reloading");
    N_("inactive");
    N_("failed");
    N_("activating");
    N_("deactivating");
    N_("running");
    N_("dead");
    N_("listening");
    N_("exited");
    N_("waiting");
    N_("elapsed");
    N_("enabled");
    N_("enabled-runtime");
    N_("linked");
    N_("linked-runtime");
    N_("masked");
    N_("masked-runtime");
    N_("static");
    N_("disabled");
    N_("invalid");
    function render_service(name, desc, load_state, active_state, sub_state, file_state, manager) {
        var waiting, active, color_style;
        if (!load_state) load_state = "";
        if (!active_state) active_state = "";
        if (!sub_state) sub_state = "";
        if (active_state == "failed" || load_state == "error") color_style = ";color:red"; else color_style = "";
        if (load_state == "loaded") load_state = "";
        waiting = active_state == "activating" || active_state == "deactivating" || active_state == "reloading";
        active = active_state == "active" || active_state == "reloading";
        load_state = _(load_state);
        active_state = _(active_state);
        sub_state = _(sub_state);
        file_state = _(file_state);
        if (sub_state !== "" && sub_state != active_state) active_state = active_state + " (" + sub_state + ")";
        if (load_state !== "") active_state = load_state + " / " + active_state;
        var tr = $("<tr>", {
            "data-unit": name
        }).click(function() {
            cockpit.location.go("service", {
                s: name
            });
        }).append($('<td style="font-weight:bold">').text(desc), $("<td>").text(name), $("<td>", {
            style: "text-align:right;white-space:nowrap" + color_style
        }).text(active_state));
        if (manager) {
            var img_waiting = $('<div class="waiting">');
            var btn_play = $('<button class="btn btn-default btn-control services-privileged btn-play">').on("click", function() {
                manager.call("ServiceAction", name, "start", function(error) {
                    if (error) shell.show_unexpected_error(error);
                });
                return false;
            });
            var btn_stop = $('<button class="btn btn-default btn-control services-privileged btn-stop">').on("click", function() {
                manager.call("ServiceAction", name, "stop", function(error) {
                    if (error) shell.show_unexpected_error(error);
                });
                return false;
            });
            img_waiting.toggle(waiting);
            btn_play.toggle(!waiting && !active);
            btn_stop.toggle(!waiting && active);
            tr.append($('<td class="cell-buttons" style="padding-left:20px;padding-right:5px">').append(btn_play, btn_stop, img_waiting));
            var geard_match = name.match(/ctr-(.*)\.service/);
            if (geard_match) {
                var btn_eject = $('<button class="btn btn-default btn-control services-privileged btn-eject">').on("click", function() {
                    cockpit.spawn([ "gear", "delete", geard_match[1] ]).fail(shell.show_unexpected_error);
                    return false;
                });
                tr.append($('<td class="cell-buttons" style="padding-left:5px;padding-right:5px">').append(btn_eject));
            }
        }
        return tr;
    }
    function update_service_privileged() {
        shell.update_privileged_ui(shell.default_permission, ".services-privileged", cockpit.format(_("The user <b>$0</b> is not permitted to modify services"), cockpit.user.name));
    }
    $(shell.default_permission).on("changed", update_service_privileged);
    PageServices.prototype = {
        _init: function() {
            this.id = "services";
            this.geard_check_done = false;
            this.geard_present = false;
        },
        getTitle: function() {
            return C_("page-title", "Services");
        },
        setup: function() {
            update_service_privileged();
            var self = this;
            $("#services-add").click(function() {
                $("#service-add-dialog").modal("show");
            });
        },
        enter: function() {
            var me = this;
            if (!me.geard_check_done) {
                me.geard_check_done = true;
                cockpit.spawn([ "which", "gear" ], {
                    error: "output"
                }).done(function() {
                    me.geard_present = true;
                    shell.content_refresh();
                });
            }
            function tabbtn(title, id, val, active, attrs) {
                var btn = $('<label class="btn btn-default">').append(title, $("<input>", $.extend({
                    id: id,
                    type: "radio",
                    name: "services-filter",
                    value: val,
                    checked: active ? "checked" : undefined
                }, attrs)));
                if (active) btn.addClass("active");
                return btn;
            }
            var my_services_btn = null;
            if (me.geard_present) {
                my_services_btn = tabbtn(_("Services"), "services-filter-my-services", "^ctr-.*\\.service$", true, {
                    "data-show-graphs": true,
                    "data-include-buttons": true
                });
            }
            $("#shell-header-extra").append($('<div class="btn-group" data-toggle="buttons">').append(my_services_btn, tabbtn(_("Targets"), "services-filter-targets", "\\.target$", false), tabbtn(_("System Services"), "services-filter-services", "\\.service$", !my_services_btn), tabbtn(_("Sockets"), "services-filter-sockets", "\\.socket$", false), tabbtn(_("Timers"), "services-filter-timers", "\\.timer$", false), tabbtn(_("Paths"), "services-filter-paths", "\\.path$", false)));
            $("#shell-header-extra input").on("change", function(event) {
                me.update();
            });
            me.client = shell.dbus(null);
            me.manager = me.client.get("/com/redhat/Cockpit/Services", "com.redhat.Cockpit.Services");
            $(me.manager).on("ServiceUpdate.services", function(event, service) {
                me.update_service(service[0], service[1], service[2], service[3], service[4], service[5]);
            });
            $(me.manager).on("ServiceUpdateAll.services", function(event) {
                me.update();
            });
            var blues = [ "#006bb4", "#008ff0", "#2daaff", "#69c2ff", "#a5daff", "#e1f3ff", "#00243c", "#004778" ];
            me.monitor = me.client.get("/com/redhat/Cockpit/LxcMonitor", "com.redhat.Cockpit.MultiResourceMonitor");
            function is_interesting_cgroup(cgroup) {
                return cgroup && !!cgroup.match("ctr-.*\\.service$");
            }
            function highlight_service_row(event, id) {
                $("#services .list-group-item").removeClass("highlight");
                if (id) {
                    id = id.split("/").pop();
                    $('[data-unit="' + shell.esc(id) + '"]').addClass("highlight");
                }
            }
            this.cpu_plot = shell.setup_multi_plot("#services-cpu-graph", me.monitor, 4, blues.concat(blues), is_interesting_cgroup);
            $(this.cpu_plot).on("update-total", function(event, total) {
                $("#services-cpu-text").text(format_cpu_usage(total));
            });
            $(this.cpu_plot).on("highlight", highlight_service_row);
            this.mem_plot = shell.setup_multi_plot("#services-mem-graph", me.monitor, 0, blues.concat(blues), is_interesting_cgroup);
            $(this.mem_plot).on("update-total", function(event, total) {
                $("#services-mem-text").text(cockpit.format_bytes(total, 1024));
            });
            $(this.mem_plot).on("highlight", highlight_service_row);
            $("#services-list-enabled, #services-list-disabled, #services-list-static").parents(".panel").hide();
            $("#services-list-enabled, #services-list-disabled, #services-list-static").empty();
            me.items = {};
            me.update();
        },
        show: function() {
            if ($("#services-graphs").is(":visible")) {
                this.cpu_plot.start();
                this.mem_plot.start();
            }
        },
        leave: function() {
            var self = this;
            this.cpu_plot.destroy();
            this.mem_plot.destroy();
            $(self.manager).off(".services");
            self.client.release();
            self.client = null;
            self.manager = null;
        },
        update_service: function(name, desc, load_state, active_state, sub_state, file_state) {
            var pattern = $('input[name="services-filter"]:checked').val();
            var include_buttons = $('input[name="services-filter"]:checked').attr("data-include-buttons") !== undefined;
            if (pattern && name.match(pattern)) {
                if (!include_buttons || load_state != "not-found") {
                    var item = $(render_service(name, desc, load_state, active_state, sub_state, file_state, include_buttons ? this.manager : null));
                    if (this.items[name]) this.items[name].replaceWith(item); else {
                        if (file_state == "enabled") {
                            item.appendTo($("#services-list-enabled"));
                            $("#services-list-enabled").parents(".panel").show();
                        } else if (file_state == "disabled") {
                            item.appendTo($("#services-list-disabled"));
                            $("#services-list-disabled").parents(".panel").show();
                        } else {
                            item.appendTo($("#services-list-static"));
                            $("#services-list-static").parents(".panel").show();
                        }
                    }
                    this.items[name] = item;
                } else {
                    if (this.items[name]) this.items[name].remove();
                    delete this.items[name];
                }
            }
        },
        update: function() {
            var me = this;
            if ($('input[name="services-filter"]:checked').attr("data-show-graphs") !== undefined) {
                $("#services-graphs").show();
                $("#services-add").show();
                if ($("#services-graphs").is(":visible")) {
                    this.cpu_plot.start();
                    this.mem_plot.start();
                }
            } else {
                $("#services-graphs").hide();
                $("#services-add").hide();
                this.cpu_plot.stop();
                this.mem_plot.stop();
            }
            function compare_service(a, b) {
                return a[1].localeCompare(b[1]);
            }
            function hack_get_service_info(name) {
                me.manager.call("GetServiceInfo", name, function(error, result) {
                    if (result) {
                        me.update_service(result.Id, result.Description, result.LoadState, result.ActiveState, result.SubState, result.UnitFileState);
                    }
                });
            }
            me.manager.call("ListServices", function(error, services) {
                var pattern;
                var service;
                var include_buttons;
                if (error) {
                    console.log("error %s", error.message);
                } else {
                    var list_enabled = $("#services-list-enabled"), enabled_added;
                    var list_disabled = $("#services-list-disabled"), disabled_added;
                    var list_static = $("#services-list-static"), static_added;
                    var i;
                    list_enabled.empty();
                    enabled_added = false;
                    list_disabled.empty();
                    disabled_added = false;
                    list_static.empty();
                    static_added = false;
                    me.items = {};
                    services.sort(compare_service);
                    pattern = $('input[name="services-filter"]:checked').val();
                    include_buttons = $('input[name="services-filter"]:checked').attr("data-include-buttons") !== undefined;
                    for (i = 0; i < services.length; i++) {
                        service = services[i];
                        if (!pattern || service[0].match(pattern) && (service[2] != "not-found" || !include_buttons)) {
                            if (service[1] == "Unknown" && me.manager) hack_get_service_info(service[0]);
                            var item = $(render_service(service[0], service[1], service[2], service[3], service[4], service[5], include_buttons ? me.manager : null));
                            if (me.items[service[0]]) {
                                me.items[service[0]].replaceWith(item);
                            } else if (service[5] == "enabled") {
                                item.appendTo(list_enabled);
                                enabled_added = true;
                            } else if (service[5] == "disabled") {
                                item.appendTo(list_disabled);
                                disabled_added = true;
                            } else {
                                item.appendTo(list_static);
                                static_added = true;
                            }
                            me.items[service[0]] = item;
                        }
                    }
                    list_enabled.parents(".panel").toggle(enabled_added);
                    list_disabled.parents(".panel").toggle(disabled_added);
                    list_static.parents(".panel").toggle(static_added);
                }
            });
            update_service_privileged();
        }
    };
    function PageServices() {
        this._init();
    }
    shell.pages.push(new PageServices());
    PageServiceAdd.prototype = {
        _init: function() {
            this.id = "service-add-dialog";
        },
        setup: function() {
            $("#service-add-add").click($.proxy(this, "add"));
        },
        enter: function() {
            this.docker = shell.docker();
            $(this.docker).on("image.services", $.proxy(this, "update"));
            $("#service-add-image, #service-add-name").val("");
            this.update();
        },
        show: function() {},
        leave: function() {
            $(shell.docker).off(".services");
            this.docker.release();
            this.docker = null;
        },
        update: function() {
            var $images = $("#service-add-images");
            var images = [];
            for (var id in this.docker.images) {
                var image = this.docker.images[id];
                if (image && image.RepoTags && image.RepoTags[0] != "<none>:<none>") images.push(image);
            }
            images.sort(function(a, b) {
                var an = a.RepoTags[0];
                var bn = b.RepoTags[0];
                return an > bn ? 1 : an < bn ? -1 : 0;
            });
            $images.html(images.map(function(image) {
                return $('<a class="list-group-item">').text(image.RepoTags[0]).click(function() {
                    $("#service-add-image").val(image.RepoTags[0]);
                });
            }));
        },
        add: function() {
            $("#service-add-dialog").modal("hide");
            cockpit.spawn([ "gear", "install", "--has-foreground", $("#service-add-image").val(), $("#service-add-name").val() ]).fail(shell.show_unexpected_error);
        }
    };
    function PageServiceAdd() {
        this._init();
    }
    shell.dialogs.push(new PageServiceAdd());
    PageService.prototype = {
        _init: function() {
            this.id = "service";
            this.section_id = "services";
        },
        getTitle: function() {
            return C_("page-title", "Services");
        },
        setup: function() {
            var self = this;
            var unit_action_spec = [ {
                title: _("Start"),
                action: "start",
                is_default: true
            }, {
                title: _("Stop"),
                action: "stop"
            }, {
                title: _("Restart"),
                action: "restart"
            }, {
                title: _("Reload"),
                action: "reload"
            }, {
                title: _("Reload or Restart"),
                action: "reload-or-restart"
            }, {
                title: _("Try Restart"),
                action: "try-restart"
            }, {
                title: _("Reload or Try Restart"),
                action: "reload-or-try-restart"
            }, {
                title: _("Isolate"),
                action: "isolate"
            } ];
            self.unit_action_btn = shell.action_btn(function(op) {
                self.action(op);
            }, unit_action_spec, "services-privileged");
            $("#service-unit-action-btn").html(self.unit_action_btn);
            var file_action_spec = [ {
                title: _("Enable"),
                action: "enable",
                is_default: true
            }, {
                title: _("Enable Forcefully"),
                action: "force-enable"
            }, {
                title: _("Disable"),
                action: "disable"
            }, {
                title: _("Preset"),
                action: "preset"
            }, {
                title: _("Preset Forcefully"),
                action: "force-preset"
            }, {
                title: _("Mask"),
                action: "mask"
            }, {
                title: _("Mask Forcefully"),
                action: "force-mask"
            }, {
                title: _("Unmask"),
                action: "unmask"
            } ];
            self.file_action_btn = shell.action_btn(function(op) {
                self.action(op);
            }, file_action_spec, "services-privileged");
            $("#service-file-action-btn").html(self.file_action_btn);
            $("#service-refresh").on("click", function() {
                self.update();
            });
            $("#service-instantiate").on("click", function() {
                var tp = self.service.indexOf("@");
                var sp = self.service.lastIndexOf(".");
                if (tp != -1) {
                    var s = self.service.substring(0, tp + 1);
                    s = s + systemd_param_esc($("#service-parameter").val());
                    if (sp != -1) s = s + self.service.substring(sp);
                    cockpit.location.go("service", {
                        s: s
                    });
                }
            });
        },
        enter: function() {
            var me = this;
            me.client = shell.dbus(null);
            me.manager = me.client.get("/com/redhat/Cockpit/Services", "com.redhat.Cockpit.Services");
            $(me.manager).on("ServiceUpdate", function(event, info) {
                if (info[0] == me.service) me.update();
            });
            $(me.manager).on("ServiceUpdateAll", function() {
                me.update();
            });
            me.service = shell.get_page_param("s") || "";
            me.update();
            me.watch_journal();
            var blues = [ "#006bb4", "#008ff0", "#2daaff", "#69c2ff", "#a5daff", "#e1f3ff", "#00243c", "#004778" ];
            me.monitor = me.client.get("/com/redhat/Cockpit/LxcMonitor", "com.redhat.Cockpit.MultiResourceMonitor");
            function is_interesting_cgroup(cgroup) {
                return cgroup && cgroup.endsWith(me.service);
            }
            this.cpu_plot = shell.setup_multi_plot("#service-cpu-graph", me.monitor, 4, blues.concat(blues), is_interesting_cgroup);
            $(this.cpu_plot).on("update-total", function(event, total) {
                $("#service-cpu-text").text(format_cpu_usage(total));
            });
            this.mem_plot = shell.setup_multi_plot("#service-mem-graph", me.monitor, 0, blues.concat(blues), is_interesting_cgroup);
            $(this.mem_plot).on("update-total", function(event, total) {
                $("#service-mem-text").text(cockpit.format_bytes(total, 1024));
            });
        },
        show: function() {
            this.cpu_plot.start();
            this.mem_plot.start();
        },
        leave: function() {
            this.cpu_plot.destroy();
            this.mem_plot.destroy();
            if (this.journal_watcher) this.journal_watcher.stop();
            delete this.journal_watcher;
            this.client.release();
            this.client = null;
        },
        watch_journal: function() {
            this.journal_watcher = modules.server.logbox([ "_SYSTEMD_UNIT=" + this.service, "+", "COREDUMP_UNIT=" + this.service, "+", "UNIT=" + this.service ], 10);
            $("#service-log").empty().append(this.journal_watcher);
        },
        update: function() {
            var me = this;
            var tp = me.service.indexOf("@");
            var sp = me.service.lastIndexOf(".");
            me.template = undefined;
            if (tp != -1 && tp + 1 != sp && tp + 1 != me.service.length) {
                me.template = me.service.substring(0, tp + 1);
                if (sp != -1) me.template = me.template + me.service.substring(sp);
            }
            $("#service .breadcrumb .active").text(me.service);
            function add_proc_info(procs, info, level) {
                var i;
                if (level > 0) procs.append('<div class="list-group-item">' + shell.esc(info[0]) + "</div>");
                for (i = 1; i < info.length; i++) {
                    if (true) {
                        procs.append('<div class="list-group-item">' + shell.esc(info[i].Pid) + " " + shell.esc(info[i].CmdLine) + "</div>");
                    } else {
                        add_proc_info(info[i], level + 1);
                    }
                }
            }
            me.manager.call("GetServiceInfo", me.service, function(error, info) {
                if (error) {
                    $("#service-unknown").show();
                    $("#service-known").hide();
                    $("#service-name").text(me.service);
                    return;
                }
                $("#service-unknown").hide();
                $("#service-known").show();
                $("#service-name").text(info.Description || info.Id || "--");
                if (info.IsTemplate) {
                    $("#service-unit-row").hide();
                    $("#service-instantiate-row").show();
                } else {
                    $("#service-unit-row").show();
                    $("#service-instantiate-row").hide();
                    var active_state = info.ActiveState;
                    var sub_state = info.SubState;
                    $("#service-active-state").text(_(active_state));
                    $("#service-sub-state").text(_(sub_state));
                    var timestamp;
                    if (active_state == "active" || active_state == "reloading") timestamp = info.ActiveEnterTimestamp; else if (active_state == "inactive" || active_state == "failed") timestamp = info.InactiveEnterTimestamp; else if (active_state == "activating") timestamp = info.InactiveExitTimestamp; else timestamp = info.ActiveExitTimestamp;
                    if (active_state == "active" || active_state == "reloading" || active_state == "activating") me.set_unit_action("stop"); else me.set_unit_action("start");
                    $("#service-since").text(new Date(timestamp / 1e3).toLocaleString());
                }
                if (me.template) {
                    $("#service-template-row").show();
                    var html = cockpit.format(_("This service is an instance of the $template service template."), {
                        template: cockpit.format('<a class="cockpit-link" onclick="${cmd}">$title</a>', {
                            cmd: shell.esc("cockpit.location.go('service', " + JSON.stringify({
                                s: me.template
                            }) + ");"),
                            title: shell.esc(me.template)
                        })
                    });
                    $("#service-template-link").html(html);
                } else $("#service-template-row").hide();
                var load_state = info.LoadState;
                var file_state = info.UnitFileState;
                $("#service-load-state").text(_(load_state));
                $("#service-file-row").show();
                $("#service-load-error-box").hide();
                $("#service-template-box").hide();
                $("#service-file-box").hide();
                if (info.IsTemplate) {
                    $("#service-template-box").show();
                    $("#service-template-state").text(_(file_state));
                } else if (load_state == "error") {
                    $("#service-load-error-box").show();
                    $("#service-load-error").text(shell.esc(info.LoadError[1]));
                } else {
                    var path = info.SourcePath || info.FragmentPath;
                    if (path || file_state) {
                        $("#service-file-box").show();
                        $("#service-file").text(shell.esc(path));
                        $("#service-file-state").text(_(file_state));
                    } else $("#service-file-row").hide();
                }
                if (load_state == "masked") me.set_file_action("unmask"); else if (file_state == "static") me.set_file_action("mask"); else if (file_state == "enabled") me.set_file_action("disable"); else me.set_file_action("enable");
                var procs = $("#service-processes");
                if (info.Processes) {
                    procs.closest(".panel").show();
                    procs.empty();
                    procs.append('<div class="list-group-item"> ' + _("CGroup") + ": " + shell.esc(info.DefaultControlGroup) + "</div>");
                    add_proc_info(procs, info.Processes, 0);
                } else {
                    procs.closest(".panel").hide();
                }
            });
            update_service_privileged();
        },
        set_unit_action: function(op) {
            shell.action_btn_select(this.unit_action_btn, op);
        },
        set_file_action: function(op) {
            shell.action_btn_select(this.file_action_btn, op);
        },
        action: function(op) {
            this.manager.call("ServiceAction", this.service, op, function(error) {
                if (error) shell.show_error_dialog(_("Error"), error.message);
            });
        }
    };
    function PageService() {
        this._init();
    }
    shell.pages.push(new PageService());
})(jQuery, cockpit, shell, modules);

var shell = shell || {};

(function($, cockpit, shell) {
    function update_accounts_privileged() {
        shell.update_privileged_ui(shell.default_permission, ".accounts-privileged", cockpit.format(_("The user <b>$0</b> is not permitted to modify accounts"), cockpit.user.name));
        $(".accounts-privileged").children("input").attr("disabled", shell.default_permission.allowed === false);
    }
    $(shell.default_permission).on("changed", update_accounts_privileged);
    function on_account_changes(client, id, func) {
        function object_added(event, obj) {
            if (obj.objectPath.indexOf("/com/redhat/Cockpit/Accounts/") === 0) func();
        }
        function object_removed(event, obj) {
            if (obj.objectPath.indexOf("/com/redhat/Cockpit/Accounts/") === 0) func();
        }
        function properties_changed(event, obj) {
            if (obj.objectPath.indexOf("/com/redhat/Cockpit/Accounts/") === 0) func();
        }
        function signal_emitted(event, iface, signal, args) {
            if (iface.getObject().objectPath.indexOf("/com/redhat/Cockpit/Accounts/") === 0 && signal == "Changed") func();
        }
        $(client).on("objectAdded." + id, object_added);
        $(client).on("objectRemoved." + id, object_removed);
        $(client).on("propertiesChanged." + id, properties_changed);
        $(client).on("signalEmitted." + id, signal_emitted);
    }
    function off_account_changes(client, id) {
        $(client).off("." + id);
    }
    function parse_passwd_content(content) {
        var ret = [];
        var lines = content.split("\n");
        var column;
        for (var i = 0; i < lines.length; i++) {
            if (!lines[i]) continue;
            column = lines[i].split(":");
            ret[i] = [];
            ret[i]["name"] = column[0];
            ret[i]["password"] = column[1];
            ret[i]["uid"] = column[2];
            ret[i]["gid"] = column[3];
            ret[i]["gecos"] = column[4];
            ret[i]["home"] = column[5];
            ret[i]["shell"] = column[6];
        }
        return ret;
    }
    function parse_group_content(content) {
        var ret = [];
        var lines = content.split("\n");
        var column;
        for (var i = 0; i < lines.length; i++) {
            if (!lines[i]) continue;
            column = lines[i].split(":");
            ret[i] = [];
            ret[i]["name"] = column[0];
            ret[i]["password"] = column[1];
            ret[i]["gid"] = column[2];
            ret[i]["userlist"] = column[3].split(",");
        }
        return ret;
    }
    function change_password(user_name, old_pass, new_pass, success_call, failure_call) {
        var expect_in;
        var call;
        var percent = /%/g;
        if (old_pass) old_pass = encodeURIComponent(old_pass).replace(percent, "\\x");
        if (new_pass) new_pass = encodeURIComponent(new_pass).replace(percent, "\\x");
        if (cockpit.user["user"] == user_name) {
            expect_in = "spawn passwd;" + 'expect "Changing password for user ' + user_name + '";' + 'expect "Changing password for ' + user_name + '";' + 'expect "(current) UNIX password: ";' + 'send -- "' + old_pass + '\\r";' + 'expect "New password: ";' + 'send -- "' + new_pass + '\\r";' + 'expect "Retype new password: ";' + 'send -- "' + new_pass + '\\r";' + 'expect "passwd: all authentication tokens updated successfully.";';
            call = cockpit.spawn([ "/usr/bin/expect" ], {
                environ: [ "LC_ALL=C" ]
            });
        } else {
            expect_in = "spawn passwd " + user_name + ";" + 'expect "Changing password for user ' + user_name + '";' + 'expect "New password: ";' + 'send -- "' + new_pass + '\\r";' + 'expect "Retype new password: ";' + 'send -- "' + new_pass + '\\r";' + 'expect "passwd: all authentication tokens updated successfully.";';
            call = cockpit.spawn([ "/usr/bin/expect" ], {
                superuser: true,
                environ: [ "LC_ALL=C" ]
            });
        }
        call.input(expect_in).fail(failure_call).done(success_call);
    }
    function is_user_in_group(user, group) {
        for (var i = 0; group["userlist"] && i < group["userlist"].length; i++) {
            if (group["userlist"][i] === user) return true;
        }
        return false;
    }
    PageAccounts.prototype = {
        _init: function() {
            this.id = "accounts";
        },
        getTitle: function() {
            return C_("page-title", "Accounts");
        },
        show: function() {},
        setup: function() {
            $("#accounts-create").on("click", $.proxy(this, "create"));
        },
        enter: function() {
            var self = this;
            function parse_accounts(content) {
                self.accounts = parse_passwd_content(content);
                self.update();
            }
            this.handle_passwd = cockpit.file("/etc/passwd");
            this.handle_passwd.read().done(parse_accounts).fail(shell.show_unexpected_error);
            this.handle_passwd.watch(parse_accounts);
        },
        leave: function() {
            if (this.handle_passwd) {
                this.handle_passwd.close();
                this.handle_passwd = null;
            }
        },
        update: function() {
            var list = $("#accounts-list");
            this.accounts.sort(function(a, b) {
                if (!a["gecos"]) return -1; else if (!b["gecos"]) return 1; else return a["gecos"].localeCompare(b["gecos"]);
            });
            list.empty();
            for (var i = 0; i < this.accounts.length; i++) {
                if (this.accounts[i]["uid"] < 1e3 || this.accounts[i]["shell"] == "/sbin/nologin") continue;
                var img = $("<img/>", {
                    "class": "cockpit-account-pic",
                    width: "48",
                    height: "48",
                    src: "images/avatar-default-48.png"
                });
                var div = $("<div/>", {
                    "class": "cockpit-account"
                }).append(img, $("<div/>", {
                    "class": "cockpit-account-real-name"
                }).text(this.accounts[i]["gecos"]), $("<div/>", {
                    "class": "cockpit-account-user-name"
                }).text(this.accounts[i]["name"]));
                div.on("click", $.proxy(this, "go", this.accounts[i]["name"]));
                list.append(div);
            }
        },
        create: function() {
            $("#accounts-create-dialog").modal("show");
        },
        go: function(user) {
            cockpit.location.go("account", {
                id: user
            });
        }
    };
    function PageAccounts() {
        this._init();
    }
    shell.pages.push(new PageAccounts());
    PageAccountsCreate.prototype = {
        _init: function() {
            this.error_timeout = null;
            this.id = "accounts-create-dialog";
        },
        show: function() {},
        setup: function() {
            $("#accounts-create-cancel").on("click", $.proxy(this, "cancel"));
            $("#accounts-create-create").on("click", $.proxy(this, "create"));
            $("#accounts-create-dialog .check-passwords").on("keydown", $.proxy(this, "update", "keydown"));
            $("#accounts-create-dialog .check-passwords").on("input", $.proxy(this, "update", "input"));
            $("#accounts-create-dialog input").on("focusout change", $.proxy(this, "update", "changeFocus"));
        },
        enter: function() {
            $("#accounts-create-user-name").val("");
            $("#accounts-create-real-name").val("");
            $("#accounts-create-pw1").val("");
            $("#accounts-create-pw2").val("");
            $("#accounts-create-locked").prop("checked", false);
            $("#accounts-create-message-password-mismatch").css("visibility", "hidden");
            $("#account-set-password-dialog .check-passwords").removeClass("has-error");
            this.update();
        },
        leave: function() {},
        update: function(behavior) {
            function check_params() {
                return $("#accounts-create-user-name").val() !== "" && $("#accounts-create-real-name").val() !== "" && $("#accounts-create-pw1").val() !== "" && $("#accounts-create-pw2").val() == $("#accounts-create-pw1").val();
            }
            function highlight_error() {
                $("#accounts-create-dialog .check-passwords").addClass("has-error");
                $("#accounts-create-message-password-mismatch").css("visibility", "visible");
            }
            function hide_error() {
                $("#accounts-create-dialog .check-passwords").removeClass("has-error");
                $("#accounts-create-message-password-mismatch").css("visibility", "hidden");
            }
            function check_password_match() {
                if ($("#accounts-create-pw2").val() !== "" && $("#accounts-create-pw1").val() !== $("#accounts-create-pw2").val()) highlight_error(); else hide_error();
            }
            window.clearTimeout(this.error_timeout);
            this.error_timeout = null;
            if (behavior == "changeFocus") {
                if ($("#accounts-create-pw2").val() !== "" && $("#accounts-create-pw1").val() !== $("#accounts-create-pw2").val()) highlight_error(); else hide_error();
            } else if (behavior == "input") {
                if ($("#accounts-create-pw2").val() !== "" && !$("#accounts-create-pw1").val().startsWith($("#accounts-create-pw2").val())) highlight_error(); else hide_error();
                this.error_timeout = window.setTimeout(check_password_match, 2e3);
                this.setTimeout = null;
            }
            $("#accounts-create-create").prop("disabled", !check_params());
        },
        cancel: function() {
            $("#accounts-create-dialog").modal("hide");
        },
        create: function() {
            var prog = [ "/usr/sbin/useradd" ];
            function adjust_locked() {
                if ($("#accounts-create-locked").prop("checked")) {
                    cockpit.spawn([ "/usr/sbin/usermod", $("#accounts-create-user-name").val(), "--lock" ], {
                        superuser: true
                    }).done(function() {
                        $("#accounts-create-dialog").modal("hide");
                    }).fail(shell.show_unexpected_error);
                } else {
                    $("#accounts-create-dialog").modal("hide");
                }
            }
            if ($("#accounts-create-real-name").val()) {
                prog.push("-c");
                prog.push($("#accounts-create-real-name").val());
            }
            prog.push($("#accounts-create-user-name").val());
            cockpit.spawn(prog, {
                superuser: true
            }).done(function() {
                if ($("#accounts-create-pw1").val()) {
                    change_password($("#accounts-create-user-name").val(), "", $("#accounts-create-pw1").val(), adjust_locked, function() {
                        shell.show_unexpected_error(_("Failed to change password"));
                    });
                } else {
                    adjust_locked();
                }
            }).fail(shell.show_unexpected_error);
        }
    };
    function PageAccountsCreate() {
        this._init();
    }
    shell.dialogs.push(new PageAccountsCreate());
    PageAccount.prototype = {
        _init: function() {
            this.id = "account";
            this.section_id = "accounts";
            this.roles = [];
        },
        getTitle: function() {
            return C_("page-title", "Accounts");
        },
        show: function() {},
        setup: function() {
            $("#account-real-name").on("change", $.proxy(this, "change_real_name"));
            $("#account-real-name").on("keydown", $.proxy(this, "real_name_edited"));
            $("#account-set-password").on("click", $.proxy(this, "set_password"));
            $("#account-delete").on("click", $.proxy(this, "delete_account"));
            $("#account-logout").on("click", $.proxy(this, "logout_account"));
            $("#account-change-roles").on("click", $.proxy(this, "change_roles"));
            $("#account-locked").on("change", $.proxy(this, "change_locked"));
        },
        get_user: function() {
            var self = this;
            function parse_user(content) {
                var accounts = parse_passwd_content(content);
                for (var i = 0; i < accounts.length; i++) {
                    if (accounts[i]["name"] !== shell.get_page_param("id")) continue;
                    self.account = accounts[i];
                    self.update();
                }
            }
            this.handle_passwd = cockpit.file("/etc/passwd");
            this.handle_passwd.read().done(parse_user).fail(shell.show_unexpected_error);
            this.handle_passwd.watch(parse_user);
        },
        get_roles: function() {
            var self = this;
            function parse_groups(content) {
                var i, j;
                self.groups = parse_group_content(content);
                while (self.roles.length > 0) self.roles.pop();
                for (i = 0, j = 0; i < self.groups.length; i++) {
                    if (self.groups[i]["name"] == "wheel" || self.groups[i]["name"] == "docker") {
                        self.roles[j] = {};
                        self.roles[j]["name"] = self.groups[i]["name"];
                        self.roles[j]["desc"] = self.groups[i]["name"] == "wheel" ? _("Server Administrator") : _("Container Administrator");
                        self.roles[j]["id"] = self.groups[i]["gid"];
                        self.roles[j]["member"] = is_user_in_group(shell.get_page_param("id"), self.groups[i]);
                        j++;
                    }
                }
                $(self).triggerHandler("roles");
                self.update();
            }
            this.handle_groups = cockpit.file("/etc/group");
            this.handle_groups.read().done(parse_groups).fail(shell.show_unexpected_error);
            this.handle_groups.watch(parse_groups);
        },
        get_last_login: function() {
            var self = this;
            function parse_last_login(data) {
                data = data.split("\n")[1];
                if (data.length === 0) return null;
                data = data.split("   ");
                if (data[data.length - 1].indexOf("**Never logged in**") > -1) return null; else return new Date(data[data.length - 1]);
            }
            cockpit.spawn([ "/usr/bin/lastlog", "-u", shell.get_page_param("id") ], {
                environ: [ "LC_ALL=C" ]
            }).done(function(data) {
                self.lastLogin = parse_last_login(data);
                self.update();
            }).fail(shell.show_unexpected_error);
        },
        get_locked: function() {
            var self = this;
            function parse_locked(content) {
                self.locked = content.indexOf("Password locked.") > -1;
                self.update();
            }
            cockpit.spawn([ "/usr/bin/passwd", "-S", shell.get_page_param("id") ], {
                environ: [ "LC_ALL=C" ],
                superuser: true
            }).done(parse_locked);
        },
        get_logged: function() {
            var self = this;
            function parse_logged(content) {
                self.logged = content.length > 0;
                if (!self.logged) self.get_last_login(); else self.update();
            }
            cockpit.spawn([ "/usr/bin/w", "-sh", shell.get_page_param("id") ]).done(parse_logged).fail(shell.show_unexpected_error);
        },
        enter: function() {
            this.real_name_dirty = false;
            this.get_user();
            this.get_roles();
            this.get_locked();
            this.get_logged();
        },
        leave: function() {
            if (this.handle_passwd) {
                this.handle_passwd.close();
                this.handle_passwd = null;
            }
            if (this.handle_groups) {
                this.handle_groups.close();
                this.handle_groups = null;
            }
        },
        update: function() {
            if (this.account) {
                var can_change = this.check_role_for_self_mod();
                $("#account-real-name").attr("disabled", !can_change);
                $("#account-logout").attr("disabled", !this.logged);
                if (!this.real_name_dirty) $("#account-real-name").val(this.account["gecos"]);
                $("#account-user-name").text(this.account["name"]);
                if (this.logged) $("#account-last-login").text(_("Logged In")); else if (!this.lastLogin) $("#account-last-login").text(_("Never")); else $("#account-last-login").text(this.lastLogin.toLocaleString());
                if (typeof this.locked != "undefined") {
                    $("#account-locked").prop("checked", this.locked);
                    $("#account-locked").parents("tr").show();
                } else {
                    $("#account-locked").parents("tr").hide();
                }
                var roles = "";
                for (var i = 0; this.roles && i < this.roles.length; i++) {
                    if (!this.roles[i]["member"]) continue;
                    if (roles !== "") roles += "<br/>";
                    roles += shell.esc(this.roles[i]["desc"]);
                }
                $("#account-roles").html(roles);
                $("#account .breadcrumb .active").text(this.account["gecos"]);
            } else {
                $("#account-real-name").val("");
                $("#account-user-name").text("");
                $("#account-last-login").text("");
                $("#account-locked").prop("checked", false);
                $("#account-roles").text("");
                $("#account .breadcrumb .active").text("?");
            }
            update_accounts_privileged();
        },
        real_name_edited: function() {
            this.real_name_dirty = true;
        },
        check_role_for_self_mod: function() {
            return this.account["name"] == cockpit.user["user"] || shell.default_permission.allowed !== false;
        },
        change_real_name: function() {
            var me = this;
            this.real_name_dirty = false;
            if (!me.check_role_for_self_mod()) {
                me.update();
                return;
            }
            cockpit.spawn([ "/usr/sbin/usermod", me.account["name"], "--comment", $("#account-real-name").val() ], {
                superuser: true
            }).done(function(data) {
                me.account["gecos"] = $("#account-real-name").val();
                me.update();
            }).fail(shell.show_unexpected_error);
        },
        change_locked: function() {
            cockpit.spawn([ "/usr/sbin/usermod", this.account["name"], $("#account-locked").prop("checked") ? "--lock" : "--unlock" ], {
                superuser: true
            }).done($.proxy(this, "get_locked")).fail(shell.show_unexpected_error);
        },
        set_password: function() {
            if (!this.check_role_for_self_mod()) return;
            PageAccountSetPassword.user_name = this.account["name"];
            $("#account-set-password-dialog").modal("show");
        },
        delete_account: function() {
            PageAccountConfirmDelete.user_name = this.account["name"];
            $("#account-confirm-delete-dialog").modal("show");
        },
        logout_account: function() {
            cockpit.spawn([ "/usr/bin/loginctl", "kill-user", this.account["name"] ], {
                superuser: true
            }).done($.proxy(this, "get_logged")).fail(shell.show_unexpected_error);
        },
        change_roles: function() {
            PageAccountChangeRoles.page = this;
            $("#account-change-roles-dialog").modal("show");
        }
    };
    function PageAccount() {
        this._init();
    }
    shell.pages.push(new PageAccount());
    var crop_handle_width = 20;
    PageAccountChangeRoles.prototype = {
        _init: function() {
            this.id = "account-change-roles-dialog";
        },
        show: function() {},
        setup: function() {
            $("#account-change-roles-apply").on("click", $.proxy(this, "apply"));
        },
        enter: function() {
            function update() {
                var r, u;
                var roles = $("#account-change-roles-roles");
                roles.empty();
                u = PageAccountChangeRoles.page.account;
                for (var i = 0; i < PageAccountChangeRoles.page.roles.length; i++) {
                    r = PageAccountChangeRoles.page.roles[i];
                    roles.append($("<li>", {
                        "class": "list-group-item"
                    }).append($("<div>", {
                        "class": "checkbox",
                        style: "margin:0px"
                    }).append($("<input/>", {
                        type: "checkbox",
                        name: "account-role-checkbox-" + r["id"],
                        id: "account-role-checkbox-" + r["id"],
                        checked: r["member"]
                    }), $("<label/>", {
                        "for": "account-role-checkbox-" + r["id"]
                    }).text(r["desc"]))));
                }
            }
            $(PageAccountChangeRoles.page).on("roles", update);
            update();
        },
        leave: function() {
            $(PageAccountChangeRoles.page).off("roles");
        },
        apply: function() {
            var checked, r;
            var new_roles = [];
            var del_roles = [];
            for (var i = 0; i < PageAccountChangeRoles.page.roles.length; i++) {
                r = PageAccountChangeRoles.page.roles[i];
                if ($("#account-role-checkbox-" + r["id"]).prop("checked") && !r["member"]) new_roles.push(r["id"]);
                if (!$("#account-role-checkbox-" + r["id"]).prop("checked") && r["member"]) del_roles.push(r["name"]);
            }
            if (new_roles.length > 0) {
                cockpit.spawn([ "/usr/sbin/usermod", PageAccountChangeRoles.page.account["name"], "-G", new_roles.toString(), "-a" ], {
                    superuser: true
                }).fail(shell.show_unexpected_error);
            }
            for (var j = 0; j < del_roles.length; j++) {
                cockpit.spawn([ "/usr/bin/gpasswd", "-d", PageAccountChangeRoles.page.account["name"], del_roles[j] ], {
                    superuser: true
                }).fail(shell.show_unexpected_error);
            }
            $("#account-change-roles-dialog").modal("hide");
        }
    };
    function PageAccountChangeRoles() {
        this._init();
    }
    shell.dialogs.push(new PageAccountChangeRoles());
    PageAccountConfirmDelete.prototype = {
        _init: function() {
            this.id = "account-confirm-delete-dialog";
        },
        show: function() {},
        setup: function() {
            $("#account-confirm-delete-apply").on("click", $.proxy(this, "apply"));
        },
        enter: function() {
            $("#account-confirm-delete-files").prop("checked", false);
            $("#account-confirm-delete-title").text(cockpit.format(_("Delete $0"), PageAccountConfirmDelete.user_name));
        },
        leave: function() {},
        apply: function() {
            var prog = [ "/usr/sbin/userdel" ];
            if ($("#account-confirm-delete-files").prop("checked")) prog.push("-r");
            prog.push(PageAccountConfirmDelete.user_name);
            cockpit.spawn(prog, {
                superuser: true
            }).done(function() {
                $("#account-confirm-delete-dialog").modal("hide");
                cockpit.location = "accounts";
            }).fail(shell.show_unexpected_error);
        }
    };
    function PageAccountConfirmDelete() {
        this._init();
    }
    shell.dialogs.push(new PageAccountConfirmDelete());
    PageAccountSetPassword.prototype = {
        _init: function() {
            this.error_timeout = null;
            this.id = "account-set-password-dialog";
        },
        show: function() {
            if (cockpit.user["user"] !== PageAccountSetPassword.user_name) {
                $("#account-set-password-old").parents("tr").toggle(false);
                $("#account-set-password-pw1").focus();
            } else {
                $("#account-set-password-old").parents("tr").toggle(true);
                $("#account-set-password-old").focus();
            }
        },
        setup: function() {
            $("#account-set-password-apply").on("click", $.proxy(this, "apply"));
            $("#account-set-password-dialog .check-passwords").on("keydown", $.proxy(this, "update", "keydown"));
            $("#account-set-password-dialog .check-passwords").on("input", $.proxy(this, "update", "input"));
            $("#account-set-password-dialog input").on("focusout change", $.proxy(this, "update", "changeFocus"));
        },
        enter: function() {
            $("#account-set-password-old").val("");
            $("#account-set-password-pw1").val("");
            $("#account-set-password-pw2").val("");
            $("#account-set-password-message-password-mismatch").css("visibility", "hidden");
            $("#account-set-password-dialog .check-passwords").removeClass("has-error");
            this.update();
        },
        leave: function() {},
        update: function(behavior) {
            function check_params() {
                return $("#account-set-password-pw1").val() !== "" && $("#account-set-password-pw2").val() == $("#account-set-password-pw1").val();
            }
            function highlight_error() {
                $("#account-set-password-dialog .check-passwords").addClass("has-error");
                $("#account-set-password-message-password-mismatch").css("visibility", "visible");
            }
            function hide_error() {
                $("#account-set-password-dialog .check-passwords").removeClass("has-error");
                $("#account-set-password-message-password-mismatch").css("visibility", "hidden");
            }
            function check_password_match() {
                if ($("#account-set-password-pw2").val() !== "" && $("#account-set-password-pw1").val() !== $("#account-set-password-pw2").val()) highlight_error(); else hide_error();
            }
            window.clearTimeout(this.error_timeout);
            this.error_timeout = null;
            if (behavior == "changeFocus") {
                if ($("#account-set-password-pw2").val() !== "" && $("#account-set-password-pw1").val() !== $("#account-set-password-pw2").val()) highlight_error(); else hide_error();
            } else if (behavior == "input") {
                if ($("#account-set-password-pw2").val() !== "" && !$("#account-set-password-pw1").val().startsWith($("#account-set-password-pw2").val())) highlight_error(); else hide_error();
                this.error_timeout = window.setTimeout(check_password_match, 2e3);
                this.setTimeout = null;
            }
            $("#account-set-password-apply").prop("disabled", !check_params());
        },
        apply: function() {
            change_password(PageAccountSetPassword.user_name, $("#account-set-password-old").val(), $("#account-set-password-pw1").val(), function() {
                $("#account-set-password-dialog").modal("hide");
            }, function() {
                shell.show_unexpected_error(_("Failed to change password"));
            });
        }
    };
    function PageAccountSetPassword() {
        this._init();
    }
    shell.dialogs.push(new PageAccountSetPassword());
    shell.change_password = function change_password() {
        PageAccountSetPassword.user_name = cockpit.user["user"];
        $("#account-set-password-dialog").modal("show");
    };
})(jQuery, cockpit, shell);

var shell = shell || {};

var modules = modules || {};

(function($, cockpit, shell, modules) {
    function resource_debug() {
        if (window.debugging == "all" || window.debugging == "resource") console.debug.apply(console, arguments);
    }
    function docker_debug() {
        if (window.debugging == "all" || window.debugging == "docker") console.debug.apply(console, arguments);
    }
    var docker_clients = shell.util.make_resource_cache();
    shell.docker = docker;
    function docker() {
        return docker_clients.get("default", function() {
            return new DockerClient();
        });
    }
    function quote_cmdline(cmds) {
        return modules.docker.quote_cmdline(cmds || []);
    }
    function unquote_cmdline(string) {
        return modules.docker.unquote_cmdline(string);
    }
    function render_container_cmdline(container) {
        if (container.Config) return quote_cmdline((container.Config.Entrypoint || []).concat(container.Config.Cmd || [])); else return container.Command;
    }
    function render_container_name(name) {
        if (name.length > 0 && name[0] == "/") return name.slice(1); else return name;
    }
    function render_container_state(state) {
        if (state.Running) return cockpit.format(_("Up since $StartedAt"), state); else return cockpit.format(_("Exited $ExitCode"), state);
    }
    function multi_line(strings) {
        return strings.map(shell.esc).join("<br/>");
    }
    function format_cpu_shares(priority) {
        if (!priority) return _("default");
        return cockpit.format(_("$0 shares"), Math.round(priority));
    }
    function format_cpu_usage(usage) {
        if (usage === undefined || isNaN(usage)) return "";
        return Math.round(usage) + "%";
    }
    function update_memory_bar(bar, usage, limit) {
        var parts = [ usage ];
        if (limit) parts.push(limit);
        $(bar).attr("value", parts.join("/")).toggleClass("bar-row-danger", !!(limit && usage > .9 * limit));
    }
    function format_memory_and_limit(usage, limit) {
        var mtext = "";
        var units = 1024;
        var parts;
        if (limit) {
            parts = cockpit.format_bytes(limit, units, true);
            mtext = " / " + parts.join(" ");
            units = parts[1];
        }
        if (usage) {
            parts = cockpit.format_bytes(usage, units, true);
            if (mtext) return parts[0] + mtext; else return parts.join(" ");
        } else {
            return "?" + mtext;
        }
    }
    function insert_table_sorted(table, row) {
        insert_table_sorted_generic(table, row, function(row1, row2) {
            return row1.text().localeCompare(row2.text());
        });
    }
    function insert_table_sorted_generic(table, row, cmp) {
        var rows = $(table).find("tbody tr");
        for (var j = 0; j < rows.length; j++) {
            if (cmp($(rows[j]), row) > 0) {
                $(row).insertBefore(rows[j]);
                row = null;
                break;
            }
        }
        if (row !== null) $(table).find("tbody").append(row);
    }
    function MemorySlider(sel, min, max) {
        var self = this;
        var slider, desc;
        var limit;
        function update_limit() {
            if (slider.disabled) {
                limit = undefined;
                return _("unlimited");
            }
            limit = Math.round(slider.value * max);
            if (limit < min) limit = min;
            return cockpit.format_bytes(limit, 1024);
        }
        slider = sel.find("div.slider").on("change", function() {
            $(desc).text(update_limit());
        })[0];
        desc = sel.find("span")[0];
        var check = sel.find("input[type='checkbox']").on("change", function() {
            $(slider).attr("disabled", !this.checked);
            $(desc).toggleClass("disabled", !this.checked);
            $(desc).text(update_limit());
        })[0];
        Object.defineProperty(this, "value", {
            get: function() {
                return limit;
            },
            set: function(v) {
                if (v !== undefined) {
                    $(slider).prop("value", v / max).trigger("change");
                }
                $(check).prop("checked", v !== undefined).trigger("change");
            }
        });
        Object.defineProperty(this, "max", {
            get: function() {
                return max;
            },
            set: function(v) {
                var old_max = max;
                max = v;
                $(slider).prop("value", slider.value * old_max / max).trigger("change");
            }
        });
        return this;
    }
    function CpuSlider(sel, min, max) {
        var self = this;
        var slider, desc;
        var priority;
        var minv = Math.log(min);
        var maxv = Math.log(max);
        var scale = maxv - minv;
        function update_priority() {
            if (slider.disabled) priority = undefined; else priority = Math.round(Math.exp(minv + scale * slider.value));
            return format_cpu_shares(priority);
        }
        slider = sel.find("div.slider").on("change", function() {
            $(desc).text(update_priority());
        })[0];
        desc = sel.find("span")[0];
        var check = sel.find("input[type='checkbox']").on("change", function() {
            $(slider).attr("disabled", !this.checked);
            $(desc).toggleClass("disabled", !this.checked);
            $(desc).text(update_priority());
        });
        Object.defineProperty(this, "value", {
            get: function() {
                return priority;
            },
            set: function(v) {
                if (v !== undefined) {
                    $(slider).prop("value", (Math.log(v) - minv) / scale).trigger("change");
                }
                $(check).prop("checked", v !== undefined).trigger("change");
            }
        });
        return this;
    }
    function setup_for_failure(page, client) {
        var $failure = $("#containers-failure");
        var $page = $("#" + page.id);
        function show_failure(ex) {
            var msg;
            var show_start = false;
            if (typeof ex == "string") {
                msg = ex;
                console.warn(ex);
            } else if (ex.problem == "not-found") {
                msg = _("Docker is not installed or activated on the system");
                show_start = true;
            } else if (ex.problem == "access-denied") {
                msg = _("Not authorized to access Docker on this system");
            } else {
                msg = cockpit.format(_("Can't connect to Docker: $0"), ex.toString());
                console.warn(ex);
            }
            $("#containers-failure-waiting").hide();
            $("#containers-failure-message").text(msg);
            $("#containers-failure-start").toggle(show_start);
            $("#containers-failure-retry").toggle(!show_start);
            $page.children().hide();
            $failure.show();
        }
        function hide_failure() {
            $page.children().show();
            $failure.hide();
        }
        $(client).on("failure.failure", function(event, ex) {
            if (ex.problem != "disconnected") show_failure(ex, page);
        });
        $("#containers-failure-retry").on("click.failure", function() {
            client.close();
            client.connect().done(function() {
                hide_failure();
                page.show();
            });
        });
        $("#containers-failure-start").on("click.failure", function() {
            $("#containers-failure-start").hide();
            $("#containers-failure-message").text(_("Starting docker"));
            $("#containers-failure-waiting").show();
            cockpit.spawn([ "systemctl", "start", "docker" ], {
                superuser: true
            }).done(function() {
                client.close();
                client.connect().done(function() {
                    hide_failure();
                    page.show();
                });
            }).fail(function(error) {
                show_failure(cockpit.format(_("Failed to start Docker: $0"), error));
            });
        });
        $page.prepend($failure);
        hide_failure();
        client.maybe_reconnect();
    }
    function unsetup_for_failure(client) {
        $(client).off(".failure");
        $("#containers-failure-start").off(".failure");
    }
    function render_container(client, $panel, filter_button, prefix, id, container, danger_mode) {
        var tr = $("#" + prefix + id);
        if (!container) {
            tr.remove();
            if (!$panel.find("table > tbody > tr").length) {
                $panel.find("button.enable-danger").toggle(false);
            }
            return;
        }
        var cputext;
        var memuse, memlimit;
        var membar, memtext, memtextstyle;
        var barvalue;
        if (container.State && container.State.Running) {
            cputext = format_cpu_usage(container.CpuUsage);
            memuse = container.MemoryUsage || 0;
            memlimit = container.MemoryLimit || 0;
            memtext = format_memory_and_limit(memuse, memlimit);
            membar = true;
            memtextstyle = {
                color: "inherit"
            };
        } else {
            cputext = "";
            membar = false;
            memtext = _("Stopped");
            memtextstyle = {
                color: "grey",
                "text-align": "right"
            };
        }
        var added = false;
        if (!tr.length) {
            $panel.find("button.enable-danger").toggle(true);
            var img_waiting = $('<div class="waiting">');
            var btn_delete = $('<button class="btn btn-danger pficon pficon-delete btn-delete">').on("click", function() {
                var self = this;
                $(self).hide().siblings("div.waiting").show();
                client.rm(id, true).fail(function(ex) {
                    shell.show_unexpected_error(ex);
                    $(self).show().siblings("div.waiting").hide();
                });
                return false;
            });
            var btn_play = $('<button class="btn btn-default btn-control btn-play">').on("click", function() {
                $(this).hide().siblings("div.waiting").show();
                client.start(id).fail(function(ex) {
                    shell.show_unexpected_error(ex);
                });
                return false;
            });
            var btn_stop = $('<button class="btn btn-default btn-control btn-stop">').on("click", function() {
                $(this).hide().siblings("div.waiting").show();
                client.stop(id).fail(function(ex) {
                    shell.show_unexpected_error(ex);
                });
                return false;
            });
            tr = $('<tr id="' + prefix + id + '">').append($('<td class="container-col-name">'), $('<td class="container-col-image">'), $('<td class="container-col-command">'), $('<td class="container-col-cpu">'), $('<td class="container-col-memory-graph">').append(shell.BarRow("containers-containers")), $('<td class="container-col-memory-text">'), $('<td class="container-col-danger cell-buttons">').append(btn_delete, img_waiting), $('<td class="container-col-actions cell-buttons">').append(btn_play, btn_stop, img_waiting.clone()));
            tr.on("click", function(event) {
                cockpit.location.go("container-details", {
                    id: id
                });
            });
            added = true;
        }
        var row = tr.children("td");
        $(row[0]).text(render_container_name(container.Name));
        $(row[1]).text(container.Image);
        $(row[2]).text(render_container_cmdline(container));
        $(row[3]).text(cputext);
        update_memory_bar($(row[4]).children("div").toggle(membar), memuse, memlimit);
        $(row[5]).css(memtextstyle).text(memtext);
        var waiting = id in client.waiting;
        $(row[6]).children("div.waiting").toggle(waiting);
        $(row[6]).children("button.btn-delete").toggle(!waiting).toggleClass("disabled", container.State.Running);
        var title = waiting || container.State.Running ? "You can only delete<br/> stopped containers" : "Delete immediately";
        $(row[6]).children("button.btn-delete").tooltip("destroy").attr("title", title).tooltip({
            html: true
        });
        $(row[7]).children("div.waiting").toggle(waiting);
        $(row[7]).children("button.btn-play").toggle(!waiting && !container.State.Running);
        $(row[7]).children("button.btn-stop").toggle(!waiting && container.State.Running);
        $(row[6]).toggle(danger_mode);
        $(row[7]).toggle(!danger_mode);
        if (filter_button) {
            var filter = shell.select_btn_selected(filter_button);
            tr.toggleClass("unimportant", !container.State.Running);
        }
        if (added) insert_table_sorted($panel.find("table"), tr);
    }
    function setup_danger_button(id, parent_id, callback) {
        var danger_button = $('<button class="btn btn-default btn-control fa fa-check enable-danger">').toggle(false).on("click", callback);
        $(id + " th.container-col-actions").append(danger_button);
        $(parent_id)[0].addEventListener("click", function(ev) {
            if ($(ev.target).parents(id).length === 0 && $(id + " button.enable-danger").hasClass("active")) callback();
        }, true);
    }
    PageContainers.prototype = {
        _init: function() {
            this.id = "containers";
            this.danger_enabled = false;
        },
        getTitle: function() {
            return C_("page-title", "Containers");
        },
        toggle_danger: function(val) {
            var self = this;
            self.danger_enabled = val;
            $("#containers-containers button.enable-danger").toggleClass("active", self.danger_enabled);
            $("#containers-containers td.container-col-actions").toggle(!self.danger_enabled);
            $("#containers-containers td.container-col-danger").toggle(self.danger_enabled);
        },
        setup: function() {
            var self = this;
            setup_danger_button("#containers-containers", "#" + this.id, function() {
                self.toggle_danger(!self.danger_enabled);
            });
            this.container_filter_btn = shell.select_btn($.proxy(this, "filter"), [ {
                title: _("All"),
                choice: "all",
                is_default: true
            }, {
                title: _("Running"),
                choice: "running"
            } ]);
            $("#containers-containers .panel-heading span").append(this.container_filter_btn);
            $("#containers-images-search").on("click", function() {
                PageSearchImage.display();
                return false;
            });
        },
        enter: function() {
            var self = this;
            this.client = shell.docker();
            this.dbus_client = shell.dbus(null);
            var reds = [ "#250304", "#5c080c", "#970911", "#ce0e15", "#ef2930", "#f36166", "#f7999c", "#fbd1d2" ];
            var blues = [ "#006bb4", "#008ff0", "#2daaff", "#69c2ff", "#a5daff", "#e1f3ff", "#00243c", "#004778" ];
            function highlight_container_row(event, id) {
                id = self.client.container_from_cgroup(id) || id;
                $("#containers-containers tr").removeClass("highlight");
                $("#" + id).addClass("highlight");
            }
            this.cpu_plot = this.client.setup_cgroups_plot("#containers-cpu-graph", 4, blues.concat(blues));
            $(this.cpu_plot).on("update-total", function(event, total) {
                $("#containers-cpu-text").text(format_cpu_usage(total));
            });
            $(this.cpu_plot).on("highlight", highlight_container_row);
            this.mem_plot = this.client.setup_cgroups_plot("#containers-mem-graph", 0, blues.concat(blues));
            $(this.mem_plot).on("update-total", function(event, total) {
                $("#containers-mem-text").text(cockpit.format_bytes(total, 1024));
            });
            $(this.mem_plot).on("highlight", highlight_container_row);
            $("#containers-containers table tbody tr").remove();
            $("#containers-images table tbody tr").remove();
            $(this.client).on("container.containers", function(event, id, container) {
                self.render_container(id, container);
            });
            $(this.client).on("image.containers", function(event, id, image) {
                self.render_image(id, image);
            });
            var id;
            $("#containers-containers button.enable-danger").toggle(false);
            for (id in this.client.containers) {
                this.render_container(id, this.client.containers[id]);
            }
            for (id in this.client.images) {
                this.render_image(id, this.client.images[id]);
            }
            setup_for_failure(self, self.client, self.address);
            self.render_storage();
            $(this.client).on("event.containers", this.throttled_render_storage());
        },
        show: function() {
            this.cpu_plot.start();
            this.mem_plot.start();
        },
        leave: function() {
            unsetup_for_failure(this.client);
            this.dbus_client.release();
            this.dbus_client = null;
            this.cpu_plot.destroy();
            this.mem_plot.destroy();
            $(this.client).off(".containers");
            this.client.release();
            this.client = null;
        },
        throttled_render_storage: function() {
            var self = this;
            var timer = null;
            var missed = false;
            var throttle = function() {
                if (!timer) {
                    self.render_storage();
                    timer = window.setTimeout(function() {
                        var need_call = missed;
                        missed = false;
                        timer = null;
                        if (need_call && self.client) throttle();
                    }, 1e4);
                } else {
                    missed = true;
                }
            };
            return throttle;
        },
        render_storage: function() {
            this.client.info().done(function(data) {
                var resp = data && JSON.parse(data);
                if (resp["Driver"] !== "devicemapper") {
                    $("#containers-storage .bar").html();
                    $("#containers-storage .data").html("Unknown");
                }
                var used;
                var total;
                var avail;
                $.each(resp["DriverStatus"], function(index, value) {
                    if (value && value[0] == "Data Space Total") total = value[1]; else if (value && value[0] == "Data Space Used") used = value[1]; else if (value && value[0] == "Data Space Available") avail = value[1];
                });
                if (used && total) {
                    var b_used = modules.docker.bytes_from_format(used);
                    var b_total = modules.docker.bytes_from_format(total);
                    if (avail) {
                        $("#containers-storage").tooltip("destroy");
                        b_total = modules.docker.bytes_from_format(avail);
                        total = cockpit.format_bytes(b_used + b_total);
                    } else {
                        var warning = _("WARNING: Docker may be reporting the size it has allocated to it's storage pool using sparse files, not the actual space available to the underlying storage device.");
                        $("#containers-storage").tooltip({
                            title: warning
                        });
                    }
                    var formated = used + " / " + total;
                    var bar_row = shell.BarRow();
                    bar_row.attr("value", b_used + "/" + b_total);
                    bar_row.toggleClass("bar-row-danger", used > .95 * total);
                    $("#containers-storage .bar").html(bar_row);
                    $("#containers-storage .data").html(formated);
                } else {
                    $("#containers-storage .bar").html();
                    $("#containers-storage .data").html("Unknown");
                }
            });
        },
        render_container: function(id, container) {
            render_container(this.client, $("#containers-containers"), this.container_filter_btn, "", id, container, this.danger_enabled);
        },
        render_image: function(id, image) {
            var self = this;
            var tr = $("#" + id);
            if (!image || !image.RepoTags || image.RepoTags[0] == "<none>:<none>") {
                tr.remove();
                return;
            }
            var added = false;
            if (!tr.length) {
                var button = $('<button class="btn btn-default btn-control btn-play">').on("click", function() {
                    PageRunImage.display(self.client, id);
                    return false;
                });
                tr = $('<tr id="' + id + '">').append($('<td class="image-col-tags">'), $('<td class="image-col-created">'), $('<td class="image-col-size-graph">'), $('<td class="image-col-size-text">'), $('<td class="cell-buttons">').append(button));
                tr.on("click", function(event) {
                    cockpit.location.go("image-details", {
                        id: id
                    });
                });
                added = true;
            }
            var row = tr.children("td");
            $(row[0]).html(multi_line(image.RepoTags));
            var threshold_date = new Date(image.Created * 1e3);
            threshold_date.setDate(threshold_date.getDate() + 2);
            if (threshold_date > new Date()) {
                $(row[1]).text(new Date(image.Created * 1e3).toLocaleString());
            } else {
                var creation_date = new Date(image.Created * 1e3);
                $(row[1]).text(creation_date.toLocaleDateString()).attr("title", creation_date.toLocaleString());
            }
            $(row[2]).children("div").attr("value", image.VirtualSize);
            $(row[3]).text(cockpit.format_bytes(image.VirtualSize, 1024));
            if (added) {
                insert_table_sorted($("#containers-images table"), tr);
            }
        },
        filter: function() {
            var filter = shell.select_btn_selected(this.container_filter_btn);
            $("#containers-containers table").toggleClass("filter-unimportant", filter === "running");
        }
    };
    function PageContainers() {
        this._init();
    }
    shell.pages.push(new PageContainers());
    PageRunImage.prototype = {
        _init: function() {
            this.error_timeout = null;
            this.id = "containers_run_image_dialog";
        },
        show: function() {},
        leave: function() {
            this.containers = null;
        },
        setup: function() {
            $("#containers-run-image-run").on("click", $.proxy(this, "run"));
            $("#containers-run-image-command").on("keydown", $.proxy(this, "update", "keydown", "command"));
            $("#containers-run-image-command").on("input", $.proxy(this, "update", "input", "command"));
            $("#containers-run-image-command").on("focusout change", $.proxy(this, "update", "changeFocus", "command"));
            this.memory_slider = new MemorySlider($("#containers-run-image-memory"), 10 * 1024 * 1024, 2 * 1024 * 1024 * 1024);
            this.cpu_slider = new CpuSlider($("#containers-run-image-cpu"), 2, 1e6);
            var table = $("#containers_run_image_dialog .modal-body table");
            var port_renderer = this.port_renderer();
            var self = this;
            $("#expose-ports").on("change", function() {
                var items = $("#select-exposed-ports");
                if ($(this).prop("checked")) {
                    if (items.children().length === 0) {
                        port_renderer();
                    }
                    items.show();
                } else {
                    items.hide();
                }
                self.update("changeFocus", "ports");
            });
            var renderer = this.link_renderer();
            $("#link-containers").change(function() {
                var items = $("#select-linked-containers");
                if ($(this).prop("checked")) {
                    if (items.children().length === 0) {
                        renderer();
                    }
                    items.show();
                } else {
                    items.hide();
                }
                self.update("changeFocus", "links");
            });
            this.validator = this.configuration_validator();
        },
        update: function(behavior, section) {
            if (this.perform_checks !== true && behavior !== "clear") return;
            this.validator(behavior, section);
        },
        enter: function() {
            var page = this;
            var info = PageRunImage.image_info;
            docker_debug("run-image", info);
            var checked;
            var value;
            PageRunImage.client.machine_info().done(function(info) {
                page.memory_slider.max = info.memory;
            });
            page.containers = [];
            var id;
            for (id in PageRunImage.client.containers) {
                page.containers.push(render_container_name(PageRunImage.client.containers[id].Name));
            }
            this.perform_checks = false;
            this.update("clear");
            $("#select-linked-containers").empty();
            $("#link-containers").prop("checked", false);
            if (info.container_config.Memory) {
                this.memory_slider.value = info.config_container.Memory;
            } else {
                this.memory_slider.value = 512 * 1024 * 1024;
                this.memory_slider.value = undefined;
            }
            if (info.container_config.CpuShares) {
                this.cpu_slider.value = info.container_config.CpuShares;
            } else {
                this.cpu_slider.value = 1024;
                this.cpu_slider.value = undefined;
            }
            var left = [ "happy", "jolly", "dreamy", "sad", "angry", "pensive", "focused", "sleepy", "grave", "distracted", "determined", "stoic", "stupefied", "sharp", "agitated", "cocky", "tender", "goofy", "furious", "desperate", "hopeful", "compassionate", "silly", "lonely", "condescending", "naughty", "kickass", "drunk", "boring", "nostalgic", "ecstatic", "insane", "cranky", "mad", "jovial", "sick", "hungry", "thirsty", "elegant", "backstabbing", "clever", "trusting", "loving", "suspicious", "berserk", "high", "romantic", "prickly", "evil" ];
            var right = [ "lovelace", "franklin", "tesla", "einstein", "bohr", "davinci", "pasteur", "nobel", "curie", "darwin", "turing", "ritchie", "torvalds", "pike", "thompson", "wozniak", "galileo", "euclid", "newton", "fermat", "archimedes", "poincare", "heisenberg", "feynman", "hawking", "fermi", "pare", "mccarthy", "engelbart", "babbage", "albattani", "ptolemy", "bell", "wright", "lumiere", "morse", "mclean", "brown", "bardeen", "brattain", "shockley" ];
            function make_name() {
                function ranchoice(array) {
                    return array[Math.round(Math.random() * (array.length - 1))];
                }
                return ranchoice(left) + "_" + ranchoice(right);
            }
            $("#containers-run-image").text(PageRunImage.image_info.RepoTags[0]);
            $("#containers-run-image-name").val(make_name());
            var command_input = $("#containers-run-image-command");
            command_input.val(quote_cmdline(PageRunImage.image_info.config.Cmd));
            var portmapping = $("#select-exposed-ports");
            portmapping.empty();
            var port_renderer = this.port_renderer();
            for (var p in PageRunImage.image_info.config.ExposedPorts) port_renderer(parseInt(p), p.slice(-3), false);
            if (portmapping.children().length > 0) {
                $("#expose-ports").prop("checked", true);
                portmapping.show();
            } else {
                $("#expose-ports").prop("checked", false);
            }
        },
        configuration_validator: function() {
            var self = this;
            function check_entries_valid() {
                $("#containers-run-image-run").prop("disabled", $("#containers_run_image_dialog").find(".has-error").length > 0);
            }
            function help_item_for_control(control, help_index) {
                if (help_index === undefined) return $(control.closest(".form-inline").find(".help-block")); else return $(control.closest(".form-inline").find(".help-block")[help_index]);
            }
            function message_prefix(help_index) {
                if (help_index === undefined) return ""; else if (help_index === 0) return _("Container") + ": "; else return _("Host") + ": ";
            }
            function show_port_message(port, message_type, message, help_index) {
                port.parent().addClass(message_type);
                var help_item = help_item_for_control(port, help_index);
                help_item.text(message_prefix(help_index) + message);
                var err_item = help_item.parent();
                err_item.addClass(message_type);
                err_item.show();
            }
            function clear_control_error(control, help_index) {
                control.parent().removeClass("has-error");
                var err_item = help_item_for_control(control, help_index).parent();
                err_item.removeClass("has-error");
                err_item.hide();
            }
            function check_port(ports, protocols, port_index, help_index) {
                var exposed_port = ports[port_index];
                var port_value = exposed_port.val();
                clear_control_error(exposed_port, help_index);
                if (port_value === "") return;
                if (/\D/.test(port_value) || port_value < 0 || port_value > 65535) {
                    show_port_message(exposed_port, "has-error", _("Invalid port"), help_index);
                    return;
                }
                for (var i = 0; i < ports.length; ++i) {
                    if (i === port_index) continue;
                    var second_port = ports[i];
                    if (port_value === second_port.val() && protocols[port_index] === protocols[i]) {
                        show_port_message(exposed_port, "has-error", _("Duplicate port"), help_index);
                        return;
                    }
                }
            }
            function clear_port_errors() {
                $("#select-exposed-ports").children("form").each(function() {
                    var element = $(this);
                    var input_ports = element.find("input");
                    input_ports = [ $(input_ports[0]), $(input_ports[1]) ];
                    clear_control_error(input_ports[0], 0);
                    clear_control_error(input_ports[1], 1);
                });
            }
            function check_ports() {
                if (!$("#expose-ports").prop("checked")) {
                    clear_port_errors();
                    check_entries_valid();
                    return;
                }
                var exposed_ports = {
                    container: [],
                    host: [],
                    protocol: []
                };
                $("#select-exposed-ports").children("form").each(function() {
                    var element = $(this);
                    var input_ports = element.find("input");
                    input_ports = [ $(input_ports[0]), $(input_ports[1]) ];
                    if (input_ports[0].val() !== "" || input_ports[1].val() !== "") {
                        exposed_ports.container.push(input_ports[0]);
                        exposed_ports.host.push(input_ports[1]);
                        exposed_ports.protocol.push(element.find("select").val().toLowerCase());
                    } else {
                        clear_control_error(input_ports[0], 0);
                        clear_control_error(input_ports[1], 1);
                    }
                });
                for (var port_index = 0; port_index < exposed_ports.container.length; ++port_index) {
                    check_port(exposed_ports.container, exposed_ports.protocol, port_index, 0);
                    check_port(exposed_ports.host, exposed_ports.protocol, port_index, 1);
                }
                check_entries_valid();
            }
            function clear_command_error() {
                $("#containers-run-image-command-note").hide();
                $("#containers-run-image-command").parent().removeClass("has-error");
            }
            function check_command() {
                if ($("#containers-run-image-command").val() === "") {
                    $("#containers-run-image-command-note").show();
                    $("#containers-run-image-command").parent().addClass("has-error");
                } else {
                    clear_command_error();
                }
                check_entries_valid();
            }
            function show_link_message(control, message_type, message, help_index) {
                control.parent().addClass(message_type);
                var help_item = help_item_for_control(control, help_index);
                help_item.text(message);
                var err_item = help_item.parent();
                err_item.addClass(message_type);
                err_item.show();
            }
            function check_alias(aliases, alias_index) {
                var alias = aliases[alias_index];
                var alias_value = alias.val();
                clear_control_error(alias, 1);
                if (alias_value === "") {
                    show_link_message(alias, "has-error", _("No alias specified"), 1);
                    return;
                }
                for (var i = 0; i < aliases.length; ++i) {
                    if (i === alias_index) continue;
                    var second_alias = aliases[i];
                    if (alias_value === second_alias.val()) {
                        show_link_message(alias, "has-error", _("Duplicate alias"), 1);
                        return;
                    }
                }
            }
            function clear_link_errors() {
                $("#select-linked-containers").children("form").each(function() {
                    var element = $(this);
                    var container = element.find("select");
                    var alias = element.find('input[name="alias"]');
                    clear_control_error(container, 0);
                    clear_control_error(alias, 1);
                });
            }
            function check_links() {
                if (!$("#link-containers").prop("checked")) {
                    clear_link_errors();
                    check_entries_valid();
                    return;
                }
                var aliases = [];
                var containers = [];
                $("#select-linked-containers").children("form").each(function() {
                    var element = $(this);
                    var container = element.find("select");
                    var alias = element.find('input[name="alias"]');
                    if (alias.val() !== "" || container.val() !== "") {
                        if (container.val() === "") show_link_message(container, "has-error", _("No container specified"), 0); else clear_control_error(container, 0);
                        aliases.push(alias);
                    } else {
                        clear_control_error(container, 0);
                        clear_control_error(alias, 1);
                    }
                });
                for (var alias_index = 0; alias_index < aliases.length; ++alias_index) check_alias(aliases, alias_index);
                check_entries_valid();
            }
            function update(behavior, section) {
                window.clearTimeout(self.error_timeout);
                self.error_timeout = null;
                if (behavior === "clear") {
                    clear_command_error();
                    clear_port_errors();
                    clear_link_errors();
                    check_entries_valid();
                } else if (behavior === "all") {
                    check_command();
                    check_ports();
                    check_links();
                } else if (behavior === "changeFocus" || behavior === "changeOption") {
                    if (section === "command") check_command(); else if (section === "ports") check_ports(); else if (section === "links") check_links();
                } else if (behavior === "input" || behavior === "keydown") {
                    if (section === "command") self.error_timeout = window.setTimeout(check_command, 2e3); else if (section === "ports") self.error_timeout = window.setTimeout(check_ports, 2e3); else if (section === "links") self.error_timeout = window.setTimeout(check_links, 2e3);
                    self.setTimeout = null;
                }
            }
            return update;
        },
        port_renderer: function() {
            var self = this;
            var template = $("#port-expose-tmpl").html();
            Mustache.parse(template);
            function add_row() {
                render();
            }
            function remove_row(e) {
                var parent = $(e.target).closest("form");
                parent.remove();
                if ($("#select-exposed-ports").children().length === 0) {
                    $("#expose-ports").attr("checked", false);
                }
                self.validator("changeFocus", "ports");
            }
            function render(port_internal, port_protocol, port_internal_editable) {
                if (port_internal === undefined) port_internal = "";
                if (port_protocol === undefined) port_protocol = "TCP";
                if (port_internal_editable === undefined) port_internal_editable = true;
                var row = $(Mustache.render(template, {
                    host_port_label: _("to host port"),
                    placeholder: _("none")
                }));
                row.children("button.fa-plus").on("click", add_row);
                if (port_internal_editable) {
                    row.children("button.pficon-close").on("click", remove_row);
                } else {
                    row.children("button.pficon-close").attr("disabled", true);
                }
                var row_container_input = row.find('input[name="container"]');
                row_container_input.val(port_internal);
                if (port_internal_editable) {
                    row_container_input.on("keydown", $.proxy(self, "update", "keydown", "ports"));
                    row_container_input.on("input", $.proxy(self, "update", "input", "ports"));
                    row_container_input.on("focusout change", $.proxy(self, "update", "changeFocus", "ports"));
                } else {
                    row_container_input.attr("disabled", true);
                }
                var row_host_input = row.find('input[name="host"]');
                row_host_input.on("keydown", $.proxy(self, "update", "keydown", "ports"));
                row_host_input.on("input", $.proxy(self, "update", "input", "ports"));
                row_host_input.on("focusout change", $.proxy(self, "update", "changeFocus", "ports"));
                var protocol_select = row.find("div select.selectpicker");
                if (port_internal_editable) {
                    protocol_select.on("change", $.proxy(self, "update", "changeOption", "ports"));
                } else {
                    protocol_select.attr("disabled", true);
                }
                protocol_select.selectpicker("refresh");
                if (port_protocol.toUpperCase() === _("UDP")) protocol_select.selectpicker("val", _("UDP")); else protocol_select.selectpicker("val", _("TCP"));
                $("#select-exposed-ports").append(row);
            }
            return render;
        },
        link_renderer: function() {
            var self = this;
            var template = $("#container-link-tmpl").html();
            Mustache.parse(template);
            function add_row() {
                render();
            }
            function remove_row(e) {
                var parent = $(e.target).closest("form");
                parent.remove();
                if ($("#select-linked-containers").children().length === 0) {
                    $("#link-containers").attr("checked", false);
                }
                self.update("changeFocus", "links");
            }
            function render() {
                var row = $(Mustache.render(template, {
                    containers: self.containers,
                    alias_label: _("alias"),
                    placeholder: _("none")
                }));
                row.children("button.fa-plus").on("click", add_row);
                row.children("button.pficon-close").on("click", remove_row);
                var row_input = row.find("input");
                row_input.on("keydown", $.proxy(self, "update", "keydown", "links"));
                row_input.on("input", $.proxy(self, "update", "input", "links"));
                row_input.on("focusout change", $.proxy(self, "update", "changeFocus", "links"));
                var container_select = row.find("div select.selectpicker");
                container_select.on("change", $.proxy(self, "update", "changeOption", "links"));
                container_select.selectpicker("refresh");
                $("#select-linked-containers").append(row);
            }
            return render;
        },
        run: function() {
            this.perform_checks = true;
            this.update("all");
            if ($("#containers-run-image-run").prop("disabled")) return;
            var name = $("#containers-run-image-name").val();
            var cmd = $("#containers-run-image-command").val();
            var port_bindings = {};
            var p, mapping;
            var map_from, map_to, map_protocol;
            var links = [];
            var exposed_ports = {};
            if ($("#expose-ports").prop("checked")) {
                $("#select-exposed-ports").children("form").each(function() {
                    var input_ports = $(this).find("input").map(function(idx, elem) {
                        return $(elem).val();
                    }).get();
                    map_from = input_ports[0];
                    map_to = input_ports[1];
                    map_protocol = $(this).find("select").val().toLowerCase();
                    if (map_from === "" || map_to === "") return;
                    port_bindings[map_from + "/" + map_protocol] = [ {
                        HostPort: map_to
                    } ];
                    exposed_ports[map_from + "/" + map_protocol] = {};
                });
            }
            if ($("#link-containers").prop("checked")) {
                $("#select-linked-containers form").each(function() {
                    var element = $(this);
                    var container = element.find('select[name="container"]').val();
                    var alias = element.find('input[name="alias"]').val();
                    if (!container || !alias) {
                        return;
                    }
                    links.push(container + ":" + alias);
                });
            }
            $("#containers_run_image_dialog").modal("hide");
            var tty = $("#containers-run-image-with-terminal").prop("checked");
            var options = {
                Cmd: unquote_cmdline(cmd),
                Image: PageRunImage.image_info.id,
                Memory: this.memory_slider.value || 0,
                MemorySwap: this.memory_slider.value * 2 || 0,
                CpuShares: this.cpu_slider.value || 0,
                Tty: tty,
                ExposedPorts: exposed_ports,
                HostConfig: {
                    Links: links
                }
            };
            if (tty) {
                $.extend(options, {
                    AttachStderr: true,
                    AttachStdin: true,
                    AttachStdout: true,
                    OpenStdin: true,
                    StdinOnce: true
                });
            }
            PageRunImage.client.create(name, options).fail(function(ex) {
                shell.show_unexpected_error(ex);
            }).done(function(result) {
                PageRunImage.client.start(result.Id, {
                    PortBindings: port_bindings
                }).fail(function(ex) {
                    shell.show_unexpected_error(ex);
                });
            });
        }
    };
    PageRunImage.display = function(client, id) {
        PageRunImage.image_info = client.images[id];
        PageRunImage.client = client;
        $("#containers_run_image_dialog").modal("show");
    };
    function PageRunImage() {
        this._init();
    }
    shell.dialogs.push(new PageRunImage());
    PageSearchImage.prototype = {
        _init: function() {
            this.id = "containers-search-image-dialog";
        },
        show: function() {
            $("#containers-search-image-search").focus();
        },
        leave: function() {
            this.cancel_search();
            $(this.client).off(".containers-search-image-dialog");
            this.client.release();
            this.client = null;
        },
        setup: function() {
            $("#containers-search-image-search").on("keypress", $.proxy(this, "input"));
            $("#containers-search-image-search").attr("placeholder", "search by name, namespace or description");
            $("#containers-search-download").on("click", $.proxy(this, "start_download"));
            $("#containers-search-tag").prop("disabled", true);
            $("#containers-search-download").prop("disabled", true);
            this.search_timeout = null;
            this.search_request = null;
        },
        enter: function() {
            this.client = shell.docker();
            $("#containers-search-image-results tbody tr").remove();
            $("#containers-search-image-results").hide();
            $("#containers-search-image-no-results").hide();
            $("#containers-search-image-search")[0].value = "";
        },
        input: function(event) {
            this.cancel_search();
            if (event.target.value.length < 3 && event.which != 13) return;
            var self = this;
            this.search_timeout = window.setTimeout(function() {
                self.perform_search(self.client);
            }, event.which == 13 ? 0 : 2e3);
        },
        start_download: function(event) {
            var repo = $("#containers-search-download").data("repo");
            var tag = $("#containers-search-tag").val();
            $("#containers-search-tag").prop("disabled", true);
            $("#containers-search-download").data("repo", "");
            $("#containers-search-download").prop("disabled", true);
            var tr = $('<tr id="imagedl_' + repo.replace("/", "_") + '">').append($('<td class="container-col-tags">').text(repo + ":" + tag), $('<td class="container-col-created">').text("Downloading"), $('<td class="image-col-size-graph">').append($('<div class="progress progress-striped active">').append($('<div class="progress-bar" role="progressbar" aria-valuenow="1" aria-valuemin="0" aria-valuemax="1" style="width: 100%">'))), $('<td class="image-col-size-text">'), $('<td class="cell-buttons">'));
            insert_table_sorted($("#containers-images table"), tr);
            var created = tr.children("td").eq(1);
            var size = tr.children("td").eq(2);
            var failed = false;
            var layers = {};
            var buffer = "";
            this.client.pull(repo, tag).stream(function(data) {
                buffer += data;
                var next = modules.docker.json_skip(buffer, 0);
                if (next === 0) return;
                var progress = JSON.parse(buffer.substring(0, next));
                buffer = buffer.substring(next);
                if ("error" in progress) {
                    failed = true;
                    created.text = "Error downloading";
                    size.text("Error downloading: " + progress["errorDetail"]["message"]);
                    tr.on("click", function() {
                        tr.remove();
                    });
                } else if ("status" in progress) {
                    if ("id" in progress) {
                        var new_string = progress["status"];
                        if (progress["status"] == "Downloading") {
                            new_string += ": " + progress["progressDetail"]["current"] + "/" + progress["progressDetail"]["total"];
                        }
                        layers[progress["id"]] = new_string;
                        if (progress["status"] == "Download complete") {
                            delete layers[progress["id"]];
                        }
                    }
                    var full_status = "";
                    for (var layer in layers) {
                        full_status += layer + ": " + layers[layer] + "&nbsp;&nbsp;&nbsp;&nbsp;";
                    }
                    size.html(full_status);
                }
            }).done(function() {
                if (!failed) {
                    tr.remove();
                }
            });
            $("#containers-search-image-dialog").modal("hide");
        },
        perform_search: function(client) {
            var term = $("#containers-search-image-search")[0].value;
            $("#containers-search-image-waiting").addClass("waiting");
            $("#containers-search-image-no-results").hide();
            $("#containers-search-image-results").hide();
            $("#containers-search-image-results tbody tr").remove();
            this.search_request = client.search(term).done(function(data) {
                var resp = data && JSON.parse(data);
                $("#containers-search-image-waiting").removeClass("waiting");
                if (resp && resp.length > 0) {
                    $("#containers-search-image-results").show();
                    resp.forEach(function(entry) {
                        var row = $("<tr>").append($("<td>").text(entry.name), $("<td>").text(entry.description));
                        row.on("click", function(event) {
                            $("#containers-search-image-results tr").each(function() {
                                $(this).removeClass("active");
                            });
                            row.addClass("active");
                            $("#containers-search-tag").val("latest");
                            $("#containers-search-tag").prop("disabled", false);
                            $("#containers-search-download").data("repo", entry.name);
                            $("#containers-search-download").prop("disabled", false);
                        });
                        row.data("entry", entry);
                        insert_table_sorted_generic($("#containers-search-image-results"), row, function(row1, row2) {
                            if (row1.data("entry").is_official && !row2.data("entry").is_official) return -1;
                            if (!row1.data("entry").is_official && row2.data("entry").is_official) return 1;
                            if (row1.data("entry").is_trusted && !row2.data("entry").is_trusted) return -1;
                            if (!row1.data("entry").is_trusted && row2.data("entry").is_trusted) return 1;
                            if (row1.data("entry").star_count != row2.data("entry").star_count) return row2.data("entry").star_count - row1.data("entry").star_count;
                            return row1.data("entry").name.localeCompare(row2.data("entry").name);
                        });
                    });
                } else {
                    $("#containers-search-image-no-results").html("No results for " + term + "<br />Please try another term");
                    $("#containers-search-image-no-results").show();
                }
            });
        },
        cancel_search: function() {
            window.clearTimeout(this.search_timeout);
            $("#containers-search-image-no-results").hide();
            $("#containers-search-image-results").hide();
            $("#containers-search-image-results tbody tr").remove();
            if (this.search_request !== null) {
                this.search_request.close();
                this.search_request = null;
            }
            $("#containers-search-image-waiting").removeClass("waiting");
            $("#containers-search-tag").prop("disabled", true);
            $("#containers-search-download").prop("disabled", true);
        }
    };
    PageSearchImage.display = function(client) {
        $("#containers-search-image-dialog").modal("show");
    };
    function PageSearchImage() {
        this._init();
    }
    shell.dialogs.push(new PageSearchImage());
    PageContainerDetails.prototype = {
        _init: function() {
            this.id = "container-details";
            this.section_id = "containers";
            this.terminal = null;
        },
        getTitle: function() {
            return C_("page-title", "Containers");
        },
        show: function() {},
        leave: function() {
            unsetup_for_failure(this.client);
            this.dbus_client.release();
            this.dbus_client = null;
            $(this.client).off(".container-details");
            this.client.release();
            this.client = null;
            if (this.terminal) {
                this.terminal.close();
                this.terminal = null;
            }
            $("#container-terminal").hide();
        },
        setup: function() {
            var self = this;
            $("#container-details-start").on("click", $.proxy(this, "start_container"));
            $("#container-details-stop").on("click", $.proxy(this, "stop_container"));
            $("#container-details-restart").on("click", $.proxy(this, "restart_container"));
            $("#container-details-delete").on("click", $.proxy(this, "delete_container"));
            self.memory_limit = new MemorySlider($("#container-resources-dialog .memory-slider"), 10 * 1024 * 1024, 2 * 1024 * 1024 * 1024);
            self.cpu_priority = new CpuSlider($("#container-resources-dialog .cpu-slider"), 2, 1e6);
            self.memory_usage = $("#container-details-memory .bar-row");
            $("#container-resources-dialog").on("show.bs.modal", function() {
                var info = self.client.containers[self.container_id];
                $(this).find(".container-name").text(self.name);
                self.memory_limit.value = info.MemoryLimit || undefined;
                self.cpu_priority.value = info.CpuPriority || undefined;
            }).find(".btn-primary").on("click", function() {
                self.client.change_memory_limit(self.container_id, self.memory_limit.value);
                var swap = self.memory_limit.value;
                if (!isNaN(swap)) swap *= 2;
                self.client.change_swap_limit(self.container_id, swap);
                self.client.change_cpu_priority(self.container_id, self.cpu_priority.value);
            });
        },
        enter: function() {
            var self = this;
            var commit = $("#container-commit-dialog")[0];
            $(commit).on("show.bs.modal", function() {
                var info = self.client.containers[self.container_id];
                $(commit).find(".container-name").text(self.name);
                var image = self.client.images[info.Config.Image];
                var repo = "";
                if (image && image.RepoTags) repo = image.RepoTags[0].split(":", 1)[0];
                $(commit).find(".container-repository").attr("value", repo);
                $(commit).find(".container-tag").attr("value", "");
                var author = cockpit.user["name"] || cockpit.user["user"];
                $(commit).find(".container-author").attr("value", author);
                var command = "";
                if (info.Config) command = quote_cmdline(info.Config.Cmd);
                if (!command) command = info.Command;
                $(commit).find(".container-command").attr("value", command);
            }).find(".btn-primary").on("click", function() {
                var location = cockpit.location;
                var run = {
                    Cmd: unquote_cmdline($(commit).find(".container-command").val())
                };
                var options = {
                    author: $(commit).find(".container-author").val()
                };
                var tag = $(commit).find(".container-tag").val();
                if (tag) options["tag"] = tag;
                var repository = $(commit).find(".container-repository").val();
                self.client.commit(self.container_id, repository, options, run).fail(function(ex) {
                    shell.show_unexpected_error(ex);
                }).done(function() {
                    location.go("containers");
                });
            });
            this.client = shell.docker();
            this.container_id = shell.get_page_param("id");
            this.name = this.container_id.slice(0, 12);
            this.client.machine_info().done(function(info) {
                self.memory_limit.max = info.memory;
            });
            this.dbus_client = shell.dbus(null);
            $(this.client).on("container.container-details", function(event, id, container) {
                if (id == self.container_id) self.update();
            });
            setup_for_failure(this, this.client, null);
            this.update();
        },
        maybe_show_terminal: function(info) {
            if (!this.terminal) {
                this.terminal = modules.docker.console(this.container_id, info.Config.Tty);
                $("#container-terminal").empty().append(this.terminal);
            }
            if (this.terminal.connected) this.terminal.typeable(info.State.Running);
            $("#container-terminal").show();
        },
        maybe_reconnect_terminal: function() {
            if (this.terminal && !this.terminal.connected) {
                this.terminal.connect();
                this.terminal.typeable(true);
            }
        },
        update: function() {
            $("#container-details-names").text("");
            $("#container-details-id").text("");
            $("#container-details-created").text("");
            $("#container-details-image").text("");
            $("#container-details-command").text("");
            $("#container-details-state").text("");
            $("#container-details-ports-row").hide();
            $("#container-details-links-row").hide();
            $("#container-details-resource-row").hide();
            var info = this.client.containers[this.container_id];
            docker_debug("container-details", this.container_id, info);
            if (!info) {
                $("#container-details-names").text(_("Not found"));
                return;
            }
            var waiting = !!this.client.waiting[this.container_id];
            $("#container-details div.waiting").toggle(waiting);
            $("#container-details button").toggle(!waiting);
            $("#container-details-start").prop("disabled", info.State.Running);
            $("#container-details-stop").prop("disabled", !info.State.Running);
            $("#container-details-restart").prop("disabled", !info.State.Running);
            $("#container-details-delete").prop("disabled", info.State.Running);
            $("#container-details-commit").prop("disabled", !!info.State.Running);
            $("#container-details-memory-row").toggle(!!info.State.Running);
            $("#container-details-cpu-row").toggle(!!info.State.Running);
            $("#container-details-resource-row").toggle(!!info.State.Running);
            this.name = render_container_name(info.Name);
            $("#container-details .breadcrumb .active").text(this.name);
            var port_bindings = [];
            if (info.NetworkSettings) {
                for (var p in info.NetworkSettings.Ports) {
                    var h = info.NetworkSettings.Ports[p];
                    if (!h) continue;
                    for (var i = 0; i < h.length; i++) {
                        port_bindings.push(cockpit.format(_("${hip}:${hport} -> $cport"), {
                            hip: h[i].HostIp,
                            hport: h[i].HostPort,
                            cport: p
                        }));
                    }
                }
            }
            $("#container-details-id").text(info.ID);
            $("#container-details-names").text(render_container_name(info.Name));
            $("#container-details-created").text(info.Created);
            $("#container-details-image").text(info.Image);
            $("#container-details-command").text(render_container_cmdline(info));
            $("#container-details-state").text(render_container_state(info.State));
            $("#container-details-ports-row").toggle(port_bindings.length > 0);
            $("#container-details-ports").html(port_bindings.map(shell.esc).join("<br/>"));
            this.update_links(info);
            update_memory_bar(this.memory_usage, info.MemoryUsage, info.MemoryLimit);
            $("#container-details-memory-text").text(format_memory_and_limit(info.MemoryUsage, info.MemoryLimit));
            $("#container-details .cpu-usage").text(format_cpu_usage(info.CpuUsage));
            $("#container-details .cpu-shares").text(format_cpu_shares(info.CpuPriority));
            this.maybe_show_terminal(info);
        },
        update_links: function(info) {
            $("#container-details-links").empty();
            var links = info.HostConfig.Links;
            if (links) {
                $("#container-details-links-row").toggle(true);
                $("#container-details-links").html(links.join("<br/>"));
            }
        },
        start_container: function() {
            var self = this;
            this.client.start(this.container_id).fail(function(ex) {
                shell.show_unexpected_error(ex);
            }).done(function() {
                self.maybe_reconnect_terminal();
            });
        },
        stop_container: function() {
            this.client.stop(this.container_id).fail(function(ex) {
                shell.show_unexpected_error(ex);
            });
        },
        restart_container: function() {
            var self = this;
            this.client.restart(this.container_id).fail(function(ex) {
                shell.show_unexpected_error(ex);
            }).done(function() {
                self.maybe_reconnect_terminal();
            });
        },
        delete_container: function() {
            var self = this;
            var location = cockpit.location;
            shell.confirm(cockpit.format(_("Please confirm deletion of $0"), self.name), _("Deleting a container will erase all data in it."), _("Delete")).done(function() {
                self.client.rm(self.container_id).fail(function(ex) {
                    shell.show_unexpected_error(ex);
                }).done(function() {
                    location.go("containers");
                });
            });
        }
    };
    function PageContainerDetails() {
        this._init();
    }
    shell.pages.push(new PageContainerDetails());
    PageImageDetails.prototype = {
        _init: function() {
            this.id = "image-details";
            this.section_id = "containers";
            this.danger_enabled = false;
        },
        getTitle: function() {
            return C_("page-title", "Containers");
        },
        show: function() {},
        leave: function() {
            unsetup_for_failure(this.client);
            this.dbus_client.release();
            this.dbus_client = null;
            $(this.client).off(".image-details");
            this.client.release();
            this.client = null;
        },
        toggle_danger: function(val) {
            var self = this;
            self.danger_enabled = val;
            $("#image-details-containers button.enable-danger").toggleClass("active", self.danger_enabled);
            $("#image-details-containers td.container-col-actions").toggle(!self.danger_enabled);
            $("#image-details-containers td.container-col-danger").toggle(self.danger_enabled);
        },
        setup: function() {
            var self = this;
            setup_danger_button("#image-details-containers", "#" + this.id, function() {
                self.toggle_danger(!self.danger_enabled);
            });
            $("#image-details-run").on("click", $.proxy(this, "run_image"));
            $("#image-details-delete").on("click", $.proxy(this, "delete_image"));
        },
        enter: function() {
            var self = this;
            this.client = shell.docker();
            this.image_id = shell.get_page_param("id");
            this.name = cockpit.format(_("Image $0"), this.image_id.slice(0, 12));
            this.dbus_client = shell.dbus(null);
            $("#image-details-containers table tbody tr").remove();
            $("#image-details-containers button.enable-danger").toggle(false);
            $(this.client).on("image.image-details", function(event, id, image) {
                if (id == self.image_id) self.update();
            });
            $(this.client).on("container.image-details", function(event, id, container) {
                if (!container || container.Config && container.Config.Image == self.image_id) self.render_container(id, container);
            });
            for (var cid in this.client.containers) {
                var c = this.client.containers[cid];
                if (c.Config && c.Config.Image == self.image_id) self.render_container(c.Id, c);
            }
            setup_for_failure(this, this.client);
            this.update();
        },
        update: function() {
            $("#image-details-id").text("");
            $("#image-details-entrypoint").text("");
            $("#image-details-command").text("");
            $("#image-details-created").text("");
            $("#image-details-author").text("");
            $("#image-details-ports").text("");
            var info = this.client.images[this.image_id];
            docker_debug("image-details", this.image_id, info);
            if (!info) {
                $("#image-details-id").text(_("Not found"));
                return;
            }
            var waiting = !!this.client.waiting[this.image_id];
            $("#image-details-buttons div.waiting").toggle(waiting);
            $("#image-details-buttons button").toggle(!waiting);
            if (info.RepoTags && info.RepoTags.length > 0) this.name = info.RepoTags[0];
            $("#image-details .breadcrumb .active").text(this.name);
            $("#image-details-id").text(info.id);
            $("#image-details-tags").html(multi_line(info.RepoTags));
            $("#image-details-created").text(info.created);
            $("#image-details-author").text(info.author);
            var config = info.config;
            if (config) {
                var ports = [];
                for (var p in config.ExposedPorts) {
                    ports.push(p);
                }
                $("#image-details-entrypoint").text(quote_cmdline(config.Entrypoint));
                $("#image-details-command").text(quote_cmdline(config.Cmd));
                $("#image-details-ports").text(ports.join(", "));
            }
        },
        render_container: function(id, container) {
            render_container(this.client, $("#image-details-containers"), null, "I", id, container, this.danger_enabled);
        },
        run_image: function() {
            PageRunImage.display(this.client, this.image_id);
        },
        delete_image: function() {
            var self = this;
            var location = cockpit.location;
            shell.confirm(cockpit.format(_("Please confirm deletion of $0"), self.name), _("Deleting an image will delete it, but you can probably download it again if you need it later.  Unless this image has never been pushed to a repository, that is, in which case you probably can't download it again."), _("Delete")).done(function() {
                self.client.rmi(self.image_id).fail(function(ex) {
                    shell.show_unexpected_error(ex);
                }).done(function() {
                    location.go("containers");
                });
            });
        }
    };
    function PageImageDetails() {
        this._init();
    }
    shell.pages.push(new PageImageDetails());
    function DockerClient() {
        var me = this;
        var events;
        var watch;
        var http;
        var connected;
        var got_failure;
        var alive = true;
        var later;
        function trigger_event() {
            if (!later) {
                later = window.setTimeout(function() {
                    later = null;
                    $(me).trigger("event");
                }, 300);
            }
        }
        function connect_events() {
            events = http.get("/v1.10/events");
            events.stream(function(resp) {
                docker_debug("event:", resp);
                if (connected.state() == "pending") connected.resolve();
                trigger_event();
            }).always(function() {
                window.setTimeout(function() {
                    if (alive && events) {
                        connect_events();
                        alive = false;
                    }
                }, 1e3);
            });
        }
        this.containers = {};
        this.images = {};
        this.waiting = {};
        var containers_meta = {};
        var containers_by_name = {};
        var images_meta = {};
        var dbus_client;
        var monitor;
        function container_to_name(container) {
            if (!container.Name) return null;
            var name = container.Name;
            if (name[0] === "/") name = name.substring(1);
            return name;
        }
        function populate_container(id, container) {
            if (container.State === undefined) container.State = {};
            if (container.Config === undefined) container.Config = {};
            $.extend(container, containers_meta[id]);
            var name = container_to_name(container);
            if (name) containers_by_name[name] = id;
        }
        function remove_container(id) {
            var container = me.containers[id];
            if (container) {
                var name = container_to_name(container);
                if (name && containers_by_name[name] == id) delete containers_by_name[name];
                delete me.containers[id];
                $(me).trigger("container", [ id, undefined ]);
            }
        }
        function fetch_containers() {
            http.get("/v1.10/containers/json", {
                all: 1
            }).done(function(data) {
                var containers = JSON.parse(data);
                if (connected.state() == "pending") connected.resolve();
                alive = true;
                var seen = {};
                $(containers).each(function(i, item) {
                    var id = item.Id;
                    if (!id) return;
                    seen[id] = id;
                    containers_meta[id] = item;
                    http.get("/v1.10/containers/" + encodeURIComponent(id) + "/json").done(function(data) {
                        var container = JSON.parse(data);
                        populate_container(id, container);
                        me.containers[id] = container;
                        $(me).trigger("container", [ id, container ]);
                    });
                });
                var removed = [];
                $.each(me.containers, function(id) {
                    if (!seen[id]) removed.push(id);
                });
                $.each(removed, function(i, id) {
                    remove_container(id);
                });
            }).fail(function(ex) {
                if (connected.state() == "pending") connected.reject(ex);
                got_failure = true;
                $(me).trigger("failure", [ ex ]);
            });
        }
        function populate_image(id, image) {
            if (image.config === undefined) {
                if (image.container_config) image.config = image.container_config; else image.config = {};
            }
            $.extend(image, images_meta[id]);
            if (image.RepoTags) image.RepoTags.sort();
        }
        function remove_image(id) {
            if (me.images[id]) {
                delete me.images[id];
                $(me).trigger("image", [ id, undefined ]);
            }
        }
        function fetch_images() {
            http.get("/v1.10/images/json").done(function(data) {
                var images = JSON.parse(data);
                if (connected.state() == "pending") connected.resolve();
                alive = true;
                var seen = {};
                $.each(images, function(i, item) {
                    var id = item.Id;
                    if (!id) return;
                    seen[id] = id;
                    images_meta[id] = item;
                    http.get("/v1.10/images/" + encodeURIComponent(id) + "/json").done(function(data) {
                        var image = JSON.parse(data);
                        populate_image(id, image);
                        me.images[id] = image;
                        $(me).trigger("image", [ id, image ]);
                    });
                });
                var removed = [];
                $.each(me.images, function(id) {
                    if (!seen[id]) removed.push(id);
                });
                $.each(removed, function(i, id) {
                    remove_image(id);
                });
            }).fail(function(ex) {
                if (connected.state() == "pending") connected.reject(ex);
                got_failure = true;
                $(me).trigger("failure", [ ex ]);
            });
        }
        $(me).on("event", function() {
            fetch_containers();
            fetch_images();
        });
        function perform_connect() {
            got_failure = false;
            connected = $.Deferred();
            http = cockpit.http("/var/run/docker.sock", {
                superuser: true
            });
            connect_events();
            if (watch && watch.valid) watch.close();
            http.get("/v1.10/info").done(function(data) {
                var info = data && JSON.parse(data);
                watch = cockpit.channel({
                    payload: "fslist1",
                    path: info["DockerRootDir"]
                });
                $(watch).on("message", function(event, data) {
                    trigger_event();
                });
                $(watch).on("close", function(event, options) {
                    if (options.problem && options.problem != "not-found") console.warn("monitor for docker directory failed: " + options.problem);
                });
                $(me).triggerHandler("event");
            }).fail(function(err) {
                if (err != "not-found") console.warn("monitor for docker directory failed: " + err);
                $(me).triggerHandler("event");
            });
            dbus_client = shell.dbus(null);
            monitor = dbus_client.get("/com/redhat/Cockpit/LxcMonitor", "com.redhat.Cockpit.MultiResourceMonitor");
            $(monitor).on("NewSample", handle_new_samples);
        }
        var regex_docker_cgroup = /docker-([A-Fa-f0-9]{64})\.scope/;
        var regex_geard_cgroup = /.*\/ctr-(.+).service/;
        this.container_from_cgroup = container_from_cgroup;
        function container_from_cgroup(cgroup) {
            var match = regex_docker_cgroup.exec(cgroup);
            if (match) return match[1];
            match = regex_geard_cgroup.exec(cgroup);
            if (match) return containers_by_name[match[1]];
            return null;
        }
        this.pull = function pull(repo, tag) {
            docker_debug("pulling: " + repo + ", tag: " + tag);
            var params = {
                fromImage: repo
            };
            if (tag) params["tag"] = tag;
            return http.request({
                method: "POST",
                path: "/v1.10/images/create",
                params: params,
                body: ""
            });
        };
        function handle_new_samples(event, timestampUsec, samples) {
            resource_debug("samples", timestampUsec, samples);
            for (var cgroup in samples) {
                var id = container_from_cgroup(cgroup);
                if (!id) continue;
                var container = me.containers[id];
                if (!container) continue;
                var sample = samples[cgroup];
                container.CGroup = cgroup;
                var mem = sample[0];
                var limit = sample[1];
                var cpu = sample[4];
                var priority = sample[5];
                if (mem != container.MemoryUsage || limit != container.MemoryLimit || cpu != container.CpuUsage || priority != container.CpuPriority) {
                    container.MemoryUsage = mem;
                    container.MemoryLimit = limit;
                    container.CpuUsage = cpu;
                    container.CpuPriority = priority;
                    $(me).trigger("container", [ id, container ]);
                }
            }
        }
        function trigger_id(id) {
            if (id in me.containers) $(me).trigger("container", [ id, me.containers[id] ]); else if (id in me.images) $(me).trigger("image", [ id, me.images[id] ]);
        }
        function waiting(id, yes) {
            if (id in me.waiting) {
                me.waiting[id]++;
            } else {
                me.waiting[id] = 1;
                trigger_id(id);
            }
        }
        function not_waiting(id) {
            me.waiting[id]--;
            if (me.waiting[id] === 0) {
                delete me.waiting[id];
                trigger_id(id);
            }
        }
        perform_connect();
        this.start = function start(id, options) {
            waiting(id);
            docker_debug("starting:", id);
            return http.request({
                method: "POST",
                path: "/v1.10/containers/" + encodeURIComponent(id) + "/start",
                headers: {
                    "Content-Type": "application/json"
                },
                body: JSON.stringify(options || {})
            }).fail(function(ex) {
                docker_debug("start failed:", id, ex);
            }).done(function(resp) {
                docker_debug("started:", id, resp);
            }).always(function() {
                not_waiting(id);
            });
        };
        this.stop = function stop(id, timeout) {
            waiting(id);
            if (timeout === undefined) timeout = 10;
            docker_debug("stopping:", id, timeout);
            return http.request({
                method: "POST",
                path: "/v1.10/containers/" + encodeURIComponent(id) + "/stop",
                params: {
                    t: timeout
                },
                body: ""
            }).fail(function(ex) {
                docker_debug("stop failed:", id, ex);
            }).done(function(resp) {
                docker_debug("stopped:", id, resp);
            }).always(function() {
                not_waiting(id);
            });
        };
        this.restart = function restart(id) {
            waiting(id);
            docker_debug("restarting:", id);
            return http.post("/v1.10/containers/" + encodeURIComponent(id) + "/restart").fail(function(ex) {
                docker_debug("restart failed:", id, ex);
            }).done(function(resp) {
                docker_debug("restarted:", id, resp);
            }).always(function() {
                not_waiting(id);
            });
        };
        this.create = function create(name, options) {
            docker_debug("creating:", name);
            return http.request({
                method: "POST",
                path: "/v1.10/containers/create",
                params: {
                    name: name
                },
                headers: {
                    "Content-Type": "application/json"
                },
                body: JSON.stringify(options || {})
            }).fail(function(ex) {
                docker_debug("create failed:", name, ex);
            }).done(function(resp) {
                docker_debug("created:", name, resp);
            }).then(JSON.parse);
        };
        this.search = function search(term) {
            docker_debug("searching:", term);
            return http.get("/v1.10/images/search", {
                term: term
            }).fail(function(ex) {
                docker_debug("search failed:", term, ex);
            }).done(function(resp) {
                docker_debug("searched:", term, resp);
            });
        };
        this.commit = function create(id, repotag, options, run_config) {
            var args = {
                container: id,
                repo: repotag
            };
            $.extend(args, options);
            waiting(id);
            docker_debug("committing:", id, repotag, options, run_config);
            return http.request({
                method: "POST",
                path: "/v1.10/commit",
                params: args,
                headers: {
                    "Content-Type": "application/json"
                },
                body: JSON.stringify(run_config || {})
            }).fail(function(ex) {
                docker_debug("commit failed:", repotag, ex);
            }).done(function(resp) {
                docker_debug("committed:", repotag);
            }).always(function() {
                not_waiting(id);
            }).then(JSON.parse);
        };
        this.rm = function rm(id) {
            waiting(id);
            docker_debug("deleting:", id);
            return http.request({
                method: "DELETE",
                path: "/v1.10/containers/" + encodeURIComponent(id),
                body: ""
            }).fail(function(ex) {
                docker_debug("delete failed:", id, ex);
            }).done(function(resp) {
                docker_debug("deleted:", id, resp);
                remove_container(id);
            }).always(function() {
                not_waiting(id);
            });
        };
        this.rmi = function rmi(id) {
            waiting(id);
            docker_debug("deleting:", id);
            return http.request({
                method: "DELETE",
                path: "/v1.10/images/" + encodeURIComponent(id),
                body: ""
            }).fail(function(ex) {
                docker_debug("delete failed:", id, ex);
            }).done(function(resp) {
                docker_debug("deleted:", id, resp);
                remove_image(id);
            }).always(function() {
                not_waiting(id);
            });
        };
        function change_cgroup(directory, cgroup, filename, value) {
            var path = "/sys/fs/cgroup/" + directory + "/" + cgroup + "/" + filename;
            var command = "echo '" + value.toFixed(0) + "' > " + path;
            docker_debug("changing cgroup:", command);
            cockpit.spawn([ "sh", "-c", command ]).fail(function(ex) {
                console.warn(ex);
            });
        }
        this.change_memory_limit = function change_memory_limit(id, value) {
            if (value === undefined || value <= 0) value = -1;
            return change_cgroup("memory", this.containers[id].CGroup, "memory.limit_in_bytes", value);
        };
        this.change_swap_limit = function change_swap_limit(id, value) {
            if (value === undefined || value <= 0) value = -1;
            return change_cgroup("memory", this.containers[id].CGroup, "memory.memsw.limit_in_bytes", value);
        };
        this.change_cpu_priority = function change_cpu_priority(id, value) {
            if (value === undefined || value <= 0) value = 1024;
            return change_cgroup("cpuacct", this.containers[id].CGroup, "cpu.shares", value);
        };
        this.setup_cgroups_plot = function setup_cgroups_plot(element, sample_index, colors) {
            function is_container(cgroup) {
                return !!container_from_cgroup(cgroup);
            }
            return shell.setup_multi_plot(element, monitor, sample_index, colors, is_container);
        };
        this.machine_info = function machine_info() {
            return shell.util.machine_info();
        };
        this.info = function info() {
            return http.get("/v1.10/info").fail(function(ex) {
                docker_debug("info failed:", ex);
            }).done(function(resp) {
                docker_debug("info:", resp);
            });
        };
        this.close = function close() {
            $(monitor).off("NewSample", handle_new_samples);
            monitor = null;
            if (dbus_client) dbus_client.release();
            dbus_client = null;
            connected = null;
        };
        this.connect = function connect() {
            if (!connected) perform_connect();
            return connected.promise();
        };
        this.maybe_reconnect = function maybe_reconnect() {
            if (got_failure) {
                this.close();
                perform_connect();
            }
            return connected.promise();
        };
    }
})(jQuery, cockpit, shell, modules);